{$N+}
{$A-}
unit Allocatr ;
{
        Program Name : Allocatr
        Package Name : LB
        Purpose      : Allocator class
        Institution  : Conroy & Conroy Co.
        Date Written : 1998
        Written By   : Alan Conroy
        Version      : 1.0D

        Copyright (C) 1998-2003 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        * 15-Nov-2002    EAC         Added debug interface.         *
        * 16-Nov-2002    EAC         Allow viewing of AT in debug.  *
        *  2-Dec-2002    EAC         Handle strings in objects.     *
        * 14-May-2003    EAC         Fix problem with uninitialized *
        *                            Lowest_Free_Space.             *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit contains the TAllocator class, which is used to manage
        storage.  It also contains TAT, a TAllocator which manages storage via
        an allocation table.
}

interface

{$I EDEFINES.INC}

uses Common, { TBase_Object }
     _DebugIn, { PSDebug_Interface }
     _Stores, { TStore_Address }
     TypeDefs ; { pByte_Array }

type TAllocator = object( TBase_Object )
               public { API... }
                   Allow_Cache : boolean ;

                   function Get_Offset : TStore_Address ; virtual ;
                   function Get_Resolution : integer ; virtual ;
                   function Get_Size : longint ; virtual ;
                   function Get_Store : PSStore ; virtual ;

                   procedure Set_Offset( Value : TStore_Address ) ; virtual ;
                   procedure Set_Resolution( Value : integer ) ; virtual ;
                   procedure Set_Size( Value : longint ) ; virtual ;
                   procedure Set_Store( Value : PSStore ) ; virtual ;

               public { API... }
                   procedure Flush ;  virtual ; { Write AT changes to store }
                   function Allocate( Size : TStore_Address ) : TStore_Address ;
                       virtual ;
                   function Allocate_At( Pos, Size : TStore_Address ) : boolean ;
                       virtual ;
                   procedure Deallocate( PTR : TStore_Address ;
                       Size : TStore_Address ) ; virtual ;
                   function Debugger : PSDebug_Interface ; virtual ;
                   procedure Load_Table ; virtual ;
                   function MaxSpace : longint ; virtual ;
		   function SpaceAvail : longint ; virtual ;
           end ; { TAllocator }

     TAT = object( TAllocator )
               public { Constructors and destructors... }
                   constructor Init( _Resolution : integer ) ;
                   destructor Done ; virtual ;

               private { Instance data... }
                   AT : pByte_Array ; { Current AT segment }
                   Byte_Mapping : longint ;
                   In_Flush : boolean ;
                   Segments : PPointer_Array ;
                   Segment : integer ; { Which segment AT points to }
                   _Offset : TStore_Address ; { AT offset in store }
                   _Size : longint ; { AT size, in bytes }
                   _Store : PSStore ; { Store where AT is saved }
                   Lowest_Free_Space : TStore_Address ;
                   Modified : boolean ; { True if unsaved modifications }
                   {TODO: Make this a range so we only write out the changed portion}
                   Resolution : integer ; { How many bytes each bit represents }
{$IFDEF DELPHI}
                   Segment_Size : integer ;
{$ELSE}
                   Segment_Size : word ; { Size of AT segment, in bytes }
{$ENDIF}
                   { Size of currently loaded segment }

                   Segment_Mapping : longint ; { How much one segment maps }

               private { Internal utility routines... }
                   procedure Get_Segment( _Segment : longint ) ;

               public { API... }
                   function Get_Modified : boolean ; virtual ;
                   function Get_Offset : TStore_Address ; virtual ;
                   function Get_Resolution : integer ; virtual ;
                   function Get_Size : longint ; virtual ;
                   function Get_Store : PSStore ; virtual ;

                   procedure Set_Modified( Value : boolean ) ; virtual ;
                   procedure Set_Offset( Value : TStore_Address ) ; virtual ;
                   procedure Set_Resolution( Value : integer ) ; virtual ;
                   procedure Set_Size( Value : longint ) ; virtual ;
                   procedure Set_Store( Value : PSStore ) ; virtual ;

               public { API... }
                   procedure Flush ; virtual ; { Write AT changes to store }
                   function Allocate( Size : TStore_Address ) : TStore_Address ;
                       virtual ;
                   function Allocate_At( Pos, Size : TStore_Address ) : boolean ;
                       virtual ;
                   procedure Deallocate( PTR : TStore_Address ;
                       Size : TStore_Address ) ; virtual ;
                   function Debugger : PSDebug_Interface ; virtual ;
                   procedure Load_Table ; virtual ;
                   function MaxSpace : longint ; virtual ;
		   function SpaceAvail : longint ; virtual ;
           end ; { TAT }
     PAT = ^TAT ;

implementation

uses CommonUt, { Boolean_To_String }
     Compatib, { ReallocMem }
     DebugInt, { TText_Debugger }
     ErrMastr, { Log_Error }
     Num1s, { Num1 }
     Standard, { Bit_Values }
     _UEHDefs, { TUEC }
     UStrings ; { Boolean_To_String }

{ TAllocator methods... }

{ API... }

function TAllocator.Get_Offset : TStore_Address ;

begin
    Result := 0 ; { We are abstract }
end ;


function TAllocator.Get_Resolution : integer ;

begin
    Result := 0 ; { We are abstract }
end ;


function TAllocator.Get_Size : longint ;

begin
    Result := 0 ; { We are abstract }
end ;


function TAllocator.Get_Store : PSStore ;

begin
    Result := nil ; { We are abstract }
end ;


procedure TAllocator.Set_Offset( Value : TStore_Address ) ;

begin
end ;


procedure TAllocator.Set_Resolution( Value : integer ) ;

begin
end ;


procedure TAllocator.Set_Size( Value : longint ) ;

begin
end ;


procedure TAllocator.Set_Store( Value : PSStore ) ;

begin
end ;


procedure TAllocator.Flush ;

begin
end ;


function TAllocator.Allocate( Size : TStore_Address ) : TStore_Address ;

begin
    Allocate := 0 ;
end ;


function TAllocator.Allocate_At( Pos, Size : TStore_Address ) : boolean ;

begin
    Allocate_At := False ;
end ;


procedure TAllocator.Deallocate( PTR : TStore_Address ; Size : TStore_Address ) ;

begin
end ;


function TAllocator.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;


procedure TAllocator.Load_Table ;

begin
end ;


function TAllocator.MaxSpace : longint ;

begin
    MaxSpace := 0 ; { We are abstract }
end ;


function TAllocator.SpaceAvail : longint ;

begin
    SpaceAvail := 0 ; { We are abstract }
end ;



{ TByte_Array_Debugger... }

type PByte_Array_Debugger = ^TByte_Array_Debugger ;
     TByte_Array_Debugger = object( TSText_Debugger )
                                private
                                    _Instance : PByte_Array ;
                                    _Size : integer ;
                                    S : string[ 255 ] ;

                                public
                                    constructor Init( I : PByte_Array ;
					Size : integer ) ;

                                    function Count : longint ; virtual ;

                                    function Child( Ordinal : longint ) : PSDebug_Interface ;
                                        virtual ;
                            end ;

constructor TByte_Array_Debugger.Init( I : PByte_Array ; Size : integer ) ;

begin
    inherited Init ;

    _Instance := I ;
    _Size := Size ;
end ;


function TByte_Array_Debugger.Count : longint ;

begin
    Count := _Size ;
end ;


function TByte_Array_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    S := Num1( Ordinal ) + ' : ' + Num1( _Instance^[ Ordinal ] ) ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ;



type PAT_Debugger = ^TAT_Debugger ;
     TAT_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PAT ;
                              S : string[ 255 ] ;

                          public
                              constructor Init( I : PAT ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                      end ;

constructor TAT_Debugger.Init( I : PAT ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TAT_Debugger.Count : longint ;

begin
    Count := 8 ;
end ;


function TAT_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;
    R1 : PByte_Array_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : begin
                R.Free ;
                R1 := new( PByte_Array_Debugger, Init( @_Instance^.AT, _Instance.Segment_Size ) ) ;
                R := PSText_Debugger( R1 ) ;
                S := 'AT' ;
            end ;
        1 : S := 'Modified = ' + Boolean_To_String( _Instance^.Modified ) ;
        2 : S := 'Offset = ' + Num1( _Instance^.Get_Offset ) ;
        3 : S := 'Segment = ' + Num1( _Instance^.Segment ) ;
        4 : S := 'Segment_Size = ' + Num1( _Instance^.Segment_Size ) ;
        5 : S := 'Size = ' + Num1( _Instance^.Get_Size ) ;
        6 : S := 'Lowest_Free_Space = ' + Num1( _Instance^.Lowest_Free_Space ) ;
        7 : S := 'Segment_Mapping = ' + Num1( _Instance^.Segment_Mapping ) ;
    end ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ; { TAT_Debugger.Child }



{ TAT methods... }

{ Constructors and destructors... }

constructor TAT.Init( _Resolution : integer ) ;

begin
    inherited Init ;

    In_Flush := False ;
    _Offset := 0 ;
    _Size := 0 ;
    _Store := nil ;
    Modified := False ;
    Resolution := _Resolution ;
    AT := nil ;
    Segments := nil ;
    Segment := -1 ; { No segment loaded yet }
    Lowest_Free_Space := 0 ;
    Byte_Mapping := 8 * Resolution ; { How much one AT byte maps }
    Segment_Mapping := 32768 * Byte_Mapping ;
    { How many data bytes one full AT segment maps }
end ;


destructor TAT.Done ;

begin
    Flush ; { Update store }
    Set_Size( 0 ) ; { Clear cache }

    inherited Done ;
end ;


{ Internal utility routines... }

procedure TAT.Get_Segment( _Segment : longint ) ; { Retrieve a segment }

begin
    if( ( not Allow_Cache ) and ( _Segment = Segment ) ) then
    begin
	exit ; { Don't write cache unless necessary }
    end ;
    if( Modified ) then { Need to write out cached AT segment }
    begin
	Flush ;
    end ;
    if( ( _Segment < 0 ) or ( _Segment = Segment ) ) then
    begin
	exit ; { Just wanted to flush AT }
    end ;

    { Now load the new segment... }
    if( _Segment * 32768 > _Size ) then
    begin
        Segment := -1 ; { Tried to get non-existent segment }
        Segment_Size := 0 ;
        exit ;
    end ;
    Segment := _Segment ;
    AT := Segments^[ Segment ] ;
    if( ( Segment + 1 ) * 32768 > _Size ) then
    begin
        Segment_Size := _Size mod 32768 ;
    end else
    begin
        Segment_Size := 32768 ;
    end ;
end ; { TAT.Get_Segment }


{ API... }

function TAT.Get_Modified : boolean ;

begin
    Get_Modified := Modified ;
end ;


function TAT.Get_Offset : TStore_Address ;

begin
    Get_Offset := _Offset ;
end ;


function TAT.Get_Resolution : integer ;

begin
    Get_Resolution := Resolution ;
end ;


function TAT.Get_Size : longint ;

begin
    Get_Size := _Size ;
end ;


function TAT.Get_Store : PSStore ;

begin
    Get_Store := _Store ;
end ;


procedure TAT.Set_Modified( Value : boolean ) ;

begin
    Modified := Value ;
end ;


procedure TAT.Set_Offset( Value : TStore_Address ) ;

begin
    if( _Offset <> Value ) then
    begin
        _Offset := Value ;
    end ;
end ;


procedure TAT.Set_Resolution( Value : integer ) ;

begin
    if( Value <> Resolution ) then
    begin
        Resolution := Value ;
    end ;
end ;


procedure TAT.Set_Size( Value : longint ) ;

var Old_Segments, New_Segments : longint ;
    P : pointer ;

begin
    if( _Size <> Value ) then
    begin
        Modified := True ;
        Old_Segments := ( _Size + 32767 ) div 32768 ; { Number of current segments }
        New_Segments := ( Value + 32767 ) div 32768 ; { Number of new segments }
        if( Value < _Size ) then { Shrinking the table }
        begin
            while( Old_Segments > New_Segments ) do
            begin
                freemem( Segments^[ Old_Segments - 1 ], 32768 ) ;
                dec( Old_Segments ) ;
            end ;
            P := Segments ;
            ReallocMem( P, Old_Segments * sizeof( pointer ), New_Segments * sizeof( pointer ) ) ;
            Segments := P ;
        end else
        begin
            P := Segments ;
            ReallocMem( P, Old_Segments * sizeof( pointer ), New_Segments * sizeof( pointer ) ) ;
            Segments := P ;
            while( Old_Segments < New_Segments ) do
            begin
                getmem( Segments^[ Old_Segments ], 32768 ) ;
                if( Segments^[ Old_Segments ] = nil ) then { Allocation failed }
                begin
                    _Size := Old_Segments * 32768 ;
                    exit ;
                end ;
                fillchar( Segments^[ Old_Segments ]^, 32768, 0 ) ;
                inc( Old_Segments ) ;
            end ;
        end ;
        _Size := Value ;
    end ;
end ; { TAT.Set_Size }


procedure TAT.Set_Store( Value : PSStore ) ;

begin
    if( _Store <> Value ) then
    begin
        _Store := Value ;
    end ;
end ;


procedure TAT.Flush ; { Write AT changes to store }

var Dummy : TRecord_Size ;
    Loop : integer ;
    M, P : TStore_Address ;
    UEC : TUEC ;

begin
    if( ( not In_Flush ) and ( _Store <> nil ) and Modified ) then
    begin
        In_Flush := True ;
        M := _Store^.Extended_Size ;
        if( _Offset + _Size > M ) then { Table size plus offset exceeds store size }
        begin
            _Store^.Extend( _Offset + _Size - M ) ;
        end ;
        P := _Offset ;
        for Loop := 0 to ( _Size div 32768 ) - 1 do
        begin
            Dummy := _Store^.Write_Data( Segments^[ Loop ]^, P, 32768, UEC ) ;
            if( Dummy <> 32768 ) then
            begin
                halt ; { Cannot flush AT - file is corrupt }
            end ;
            P := P + 32768 ;
        end ;
        M := _Size mod 32768 ;
        Dummy := _Store^.Write_Data( Segments^[ _Size div 32768 ]^, P, M, UEC ) ;
        if( Dummy <> M ) then
        begin
            halt ; { Cannot flush AT - file is corrupt }
        end ;

        Modified := False ;
        In_Flush := False ;
    end ;
end ; { TAT.Flush }


function TAT.Allocate( Size : TStore_Address ) : TStore_Address ;
{ Allocate Size bytes and return pointer.  Return 0 if error. }

var AT_Byte, AT_Bit : longint ; { Current byte and bit within AT }
    Counter : longint ;
    Offset : TStore_Address ;
    Offset_Segment, Offset_Byte, Offset_Bit : longint ;

    procedure Set_Start ;

    begin
        Offset := Segment * ( Segment_Mapping div Byte_Mapping ) + AT_Byte ; { Allocation table byte offset }
        Offset := Offset * Byte_Mapping + AT_Bit * Resolution ; { Remember start }
        Offset_Segment := Segment ;
        Offset_Byte := AT_Byte ;
        Offset_Bit := AT_Bit ;
    end ;

begin
    Allocate := 0 ; { Return zero (0) if error }
    if( Size < 1 ) then { Cannot allocate nothing, or less than nothing }
    begin
        exit ;
    end ;
    Size := ( Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    { Find a place where we can allocate the requested space... }
    Offset := 0 ; { So we know if allocation failed }
    Get_Segment( Lowest_Free_Space div Segment_Mapping ) ;
    { Start at the beginning }

    AT_Byte := ( Lowest_Free_Space - ( Segment * Segment_Mapping ) ) div Byte_Mapping ;
    if( AT_Byte >= Segment_Size ) then
    begin
        AT_Byte := Segment_Size - 1 ;
    end ;
    AT_Bit := 0 ;
    Counter := 0 ; { Number of bytes available in current segment }
    while( Counter < Size ) do { Find necessary free space }
    begin
        if( AT^[ AT_Byte ] = 255 ) then { No room in entire byte }
        begin
            Counter := 0 ;
            AT_Bit := 7 ; { Skip rest of bits in this byte }
        end else
        begin
            if(
                ( AT_Bit = 0 )
                and
                ( Size - Counter >= Byte_Mapping )
                and
                ( AT^[ AT_Byte ] = 0 )
              ) then
            begin { Quick check for entire byte }
                if( Counter = 0 ) then
                begin
                    Set_Start ;
                end ;
                Counter := Counter + Byte_Mapping ;
                AT_Bit := 7 ;
                if( Counter >= Size ) then
                begin
                    break ;
                end ;
            end else
            begin { Check bit by bit }
                if( ( AT^[ AT_Byte ] and Bit_Values[ AT_Bit ] ) = 0 ) then
                begin
                    if( Counter = 0 ) then { First free block }
                    begin
                        Set_Start ;
                    end ;
                    Counter := Counter + Resolution ; { Count up free bytes }
                    if( Counter >= Size ) then
                    begin
                        break ;
                    end ;
                end else { Ran out of free blocks before we found sufficient space }
                begin
                    Counter := 0 ; { Start over }
                end ;
            end ;
        end ;
        inc( AT_Bit ) ; { Check next bit of current byte }
        if( AT_Bit > 7 ) then { Ran out of bits on this byte }
        begin
            AT_Bit := 0 ; { Start at first bit of next byte }
            inc( AT_Byte ) ; { Move to next byte }
            if( AT_Byte = Segment_Size ) then { Past AT segment }
            begin
                AT_Byte := 0 ; { Start at first byte of next segment }
                Get_Segment( Segment + 1 ) ; { Move to next segment }
            end ;
            if( AT_Byte >= Segment_Size ) then { AT was not big enough }
            begin
                exit ;
            end ; { if( AT_Byte > Segment_Size ) }
        end ; { if( AT_Bit > 7 ) }
    end ; { while( Counter > _Size ) }

    { Now mark space as allocated in AT... }
    Get_Segment( Offset_Segment ) ;
    AT_Byte := Offset_Byte ;
    AT_Bit := Offset_Bit ;
    while( Counter > 0 ) do
    begin
        Modified := True ;
        if( ( AT_Bit = 0 ) and ( Counter >= Byte_Mapping ) ) then
        begin { Quick check for entire byte }
            AT^[ AT_Byte ] := 255 ;
            Counter := Counter - Byte_Mapping ;
            AT_Bit := 7 ;
            if( Counter >= Size ) then
            begin
                break ;
            end ;
        end else
        begin
            AT^[ AT_Byte ] := AT^[ AT_Byte ] or Bit_Values[ AT_Bit ] ; { Mark as used }
            Counter := Counter - Resolution ;
        end ;
        inc( AT_Bit ) ; { Check next bit of current byte }
        if( AT_Bit > 7 ) then { Ran out of bits on this byte } if( Counter > 0 ) then
        { ...and not yet finished }
        begin
            AT_Bit := 0 ; { Start at first bit... }
            inc( AT_Byte ) ; {...of next byte}
            if( AT_Byte = Segment_Size ) then { Past segment end }
            begin
                AT_Byte := 0 ; { Start at first byte of next segment }
                Get_Segment( Segment + 1 ) ; { Move to next segment }
            end ; { if( AT_Byte = Byte_Mapping ) }
        end ; { if( AT_Bit > 7 ) }
    end ; { while Counter > 0 }
    Modified := True ;
    Allocate := Offset ; { Return offset of allocation }
    if( Offset <> 0 ) then
    begin
        if( Offset = Lowest_Free_Space ) then
        begin
            Lowest_Free_Space := Lowest_Free_Space + Size ;
        end ;
    end ;
end ; { TAT.Allocate }


function TAT.Allocate_At( Pos, Size : TStore_Address ) : boolean ;

var AT_Byte, AT_Bit : longint ; { Current byte and bit within AT }
    Counter : longint ;

begin
    Allocate_At := False ; { Return false if error }
    if( Size < 1 ) then { Cannot allocate nothing }
    begin
        exit ;
    end ;
    Size := ( Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    Get_Segment( Pos div Segment_Mapping ) ;
    AT_Byte := ( Pos - ( Segment * Segment_Mapping ) ) div Byte_Mapping ;
    if( AT_Byte >= Segment_Size ) then
    begin
        exit ;
    end ;
    AT_Bit := ( Pos - ( ( Segment * Segment_Mapping ) + ( AT_Byte * Byte_Mapping ) ) ) div Resolution ;
    Counter := 0 ; { Number of bytes available in current segment }
    while( Counter < Size ) do { Check for necessary free space }
    begin
        if(
            ( AT_Bit = 0 )
            and
            ( Size - Counter >= Byte_Mapping )
            and
            ( AT^[ AT_Byte ] = 0 )
          ) then
        begin { Quick check for entire byte }
            Counter := Counter + Byte_Mapping ;
            AT_Bit := 7 ;
            if( Counter >= Size ) then
            begin
                break ;
            end ;
        end else
        begin { Check bit by bit }
            if( ( AT^[ AT_Byte ] and Bit_Values[ AT_Bit ] ) = 0 ) then
            begin
                Counter := Counter + Resolution ; { Count up free bytes }
                if( Counter >= Size ) then
                begin
                    break ;
                end ;
            end else { Insufficient space at location }
            begin
                exit ;
            end ;
        end ;
        inc( AT_Bit ) ; { Check next bit of current byte }
        if( AT_Bit > 7 ) then { Ran out of bits on this byte }
        begin
            AT_Bit := 0 ; { Start at first bit of next byte }
            inc( AT_Byte ) ; { Move to next byte }
            if( AT_Byte = Byte_Mapping ) then { Past AT segment }
            begin
                if( ( Segment + 1 ) * Byte_Mapping < _Size ) then
                { if more of the Allocation Table to process }
                begin
                    AT_Byte := 0 ; { Start at first byte of next segment }
                    Get_Segment( Segment + 1 ) ; { Move to next segment }
                end ;
            end ;
            if( AT_Byte >= Segment_Size ) then { AT was not big enough }
            begin
                exit ;
            end ; { if( AT_Byte > Segment_Size ) }
        end ; { if( AT_Bit > 7 ) }
    end ; { while( Counter > _Size ) }

    { Now mark space as allocated in AT... }
    Get_Segment( Pos div Segment_Mapping ) ;
    AT_Byte := ( Pos - ( Segment * Segment_Mapping ) ) div Byte_Mapping ;
    AT_Bit := ( Pos - ( ( Segment * Segment_Mapping ) + ( AT_Byte * Byte_Mapping ) ) ) div Resolution ;
    while( Counter > 0 ) do
    begin
        Modified := True ;
        if( ( AT_Bit = 0 ) and ( Counter >= Byte_Mapping ) ) then
        begin { Quick check for entire byte }
            AT^[ AT_Byte ] := 255 ;
            Counter := Counter - Byte_Mapping ;
            AT_Bit := 7 ;
            if( Counter >= Size ) then
            begin
                break ;
            end ;
        end else
        begin
            AT^[ AT_Byte ] := AT^[ AT_Byte ] or Bit_Values[ AT_Bit ] ; { Mark as used }
            Counter := Counter - Resolution ;
        end ;
        inc( AT_Bit ) ; { Check next bit of current byte }
        if( AT_Bit > 7 ) then { Ran out of bits on this byte } if( Counter > 0 ) then
        { ...and not yet finished }
        begin
            AT_Bit := 0 ; { Start at first bit... }
            inc( AT_Byte ) ; {...of next byte}
            if( AT_Byte = Byte_Mapping ) then { Past AT segment end }
            begin
                AT_Byte := 0 ; { Start at first byte of next segment }
                Get_Segment( Segment + 1 ) ; { Move to next segment }
            end ; { if( AT_Byte = Byte_Mapping ) }
        end ; { if( AT_Bit > 7 ) }
    end ; { while Counter > 0 }
    Modified := True ;
    Allocate_At := True ; { Success }
    if( Pos = Lowest_Free_Space ) then
    begin
        Lowest_Free_Space := Lowest_Free_Space + Size ;
    end ;
end ; { TAT.Allocate_At }


procedure TAT.Deallocate( PTR : TStore_Address ; Size : TStore_Address ) ;

var AT_Byte, AT_Bit : longint ;

begin
    if( Size = 0 ) then
    begin
        exit ;
    end ;
    Size := ( Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Deallocate minimum multiple bytes at a time }

    if( Lowest_Free_Space > PTR ) then
    begin
        Lowest_Free_Space := PTR ;
    end ;

    Get_Segment( PTR div Segment_Mapping ) ;
    PTR := PTR - Segment * Segment_Mapping ;
    AT_Byte := PTR div Byte_Mapping ;
    PTR := PTR - AT_Byte * Byte_Mapping ;
    AT_Bit := PTR div Resolution ;
    while( Size > 0 ) do
    begin
        if( ( AT_Bit = 0 ) and ( Size >= Byte_Mapping ) ) then
        begin
            AT_Bit := 7 ;
            AT^[ AT_Byte ] := 0 ;
            Size := Size - Byte_Mapping ;
        end else
        begin
            AT^[ AT_Byte ] := AT^[ AT_Byte ] and ( not Bit_Values[ AT_Bit ] ) ;
            Size := Size - Resolution ;
        end ;

        if( Size > 0 ) then
        begin
            inc( AT_Bit ) ;
            if( AT_Bit > 7 ) then
            begin
                AT_Bit := 0 ; { Move to first bit... }
                inc( AT_Byte ) ; { ...of next byte }
                if( AT_Byte = Segment_Size ) then
                begin
                    Modified := True ;
                    Flush ;
                    AT_Byte := 0 ; { Start at first byte of next segment }
                    Get_Segment( Segment + 1 ) ;
                    if( Segment_Size = 0 ) then
                    begin
                        Size := 0 ; { Done if at end of AT }
                    end ;
                end ;
            end ; { if AT_Bit > 7 }
        end ; { if _Size > 0 }
    end ; { while _Size > 0 }
    Modified := True ;
    Flush ;
end ; { TAT.Deallocate }


function TAT.Debugger : PSDebug_Interface ;

var R : PAT_Debugger ;

begin
    system.new( R, Init( @Self ) ) ;
    Debugger := R ;
end ;


procedure TAT.Load_Table ; { Load from store }

var Dummy : TRecord_Size ;
    Loop : integer ;
    M, P : TStore_Address ;
    UEC : TUEC ;

begin
    if( ( _Store <> nil ) and ( _Size > 0 ) ) then
    begin
{$IFDEF DELPHI}
        P := _Offset ;
{$ELSE}
        P := trunc( _Offset ) ;
{$ENDIF}
        if( _Size > 32768 ) then
        begin
            for Loop := 0 to ( _Size div 32768 ) - 1 do
            begin
                Dummy := _Store^.Read_Data( Segments^[ Loop ]^, P, 32768, UEC ) ;
                if( Dummy <> 32768 ) then
                begin
                    Log_Error( UEC, 'Allocation Table load failure', SEL_Fatal ) ;
                end ;
                P := P + 32768 ;
            end ;
        end ;
        M := _Size mod 32768 ;
        Dummy := _Store^.Read_Data( Segments^[ _Size div 32768 ]^, P, M, UEC ) ;
        if( Dummy <> M ) then
        begin
            Log_Error( UEC, 'Allocation Table load failure', SEL_Fatal ) ;
        end ;

        Modified := False ;
    end ;
end ;


function TAT.MaxSpace : longint ; { Return largest contiguous space }

var AT_Byte, AT_Bit : integer ; { Current byte and bit within AT }
    Counter, Max : longint ;

begin
    Get_Segment( Lowest_Free_Space div Segment_Mapping ) ;
    { Start at the beginning }

    AT_Byte := ( Lowest_Free_Space - ( Segment * Segment_Mapping ) ) div Byte_Mapping ;
    if( AT_Byte >= Segment_Size ) then
    begin
        AT_Byte := Segment_Size - 1 ;
    end ;
    AT_Bit := 0 ;
    Counter := 0 ; { Number of bytes available in current segment }
    Max := 0 ; { Maximum space found }
    while( AT_Bit <= 7 ) do { Find free space }
    begin
	if( ( AT^[ AT_Byte ] and Bit_Values[ AT_Bit ] ) = 0 ) then
        begin
	    Counter := Counter + Resolution ; { Count up free bytes }
        end else
        begin { Ran out of free blocks }
            if( Counter > Max ) then
            begin
                Max := Counter ;
            end ;
            Counter := 0 ; { Start over }
        end ;
	inc( AT_Bit ) ; { Check next bit of current byte }
	if( AT_Bit > 7 ) then { Ran out of bits on this byte }
	begin
	    AT_Bit := 0 ; { Start at first bit of next byte }
	    inc( AT_Byte ) ; { Move to next byte }
	    if AT_Byte = Byte_Mapping then { Past AT segment }
	    begin
		AT_Byte := 0 ; { Start at first byte of next segment }
		Get_Segment( Segment + 1 ) ; { Move to next segment }
	    end ; { if AT_Byte > Byte_Mapping then }
	    if( AT_Byte >= Segment_Size ) then { AT was not big enough }
	    begin
		AT_Bit := 8 ; { Terminate loop }
		if( Counter > Max ) then
                begin
		    Max := Counter ;
                end ;
	    end ; { if AT_Byte >= AT_Segment_Size }
	end ; { if AT_Bit > 7 }
    end ; { while AT_Bit <= 7 }
    MaxSpace := Max ;
end ; { TAT.MaxSpace }


function TAT.SpaceAvail : longint ;
{ Return available space within table }

var AT_Byte, AT_Bit : integer ; { Current byte and bit within AT }
    Counter : longint ;

begin
    Get_Segment( Lowest_Free_Space div Segment_Mapping ) ;
    { Start at the beginning }

    AT_Byte := ( Lowest_Free_Space - ( Segment * Segment_Mapping ) ) div Byte_Mapping ;
    if( AT_Byte >= Segment_Size ) then
    begin
        AT_Byte := Segment_Size - 1 ;
    end ;
    AT_Bit := 0 ;
    Counter := 0 ; { Number of bytes available in current segment }
    while( AT_Bit <= 7 ) do { Find free space }
    begin
	if( ( AT^[ AT_Byte ] and Bit_Values[ AT_Bit ] ) = 0 ) then
        begin
            Counter := Counter + Resolution ; { Count up free bytes }
        end ;
	inc( AT_Bit ) ; { Check next bit of current byte }
	if( AT_Bit > 7 ) then { Ran out of bits on this byte }
	begin
	    AT_Bit := 0 ; { Start at first bit of next byte }
	    inc( AT_Byte ) ; { Move to next byte }
	    if( AT_Byte = Segment_Size ) then { Past AT segment }
	    begin
		AT_Byte := 0 ; { Start at first byte of next segment }
		Get_Segment( Segment + 1 ) ; { Move to next segment }
	    end ; { if( AT_Byte = Byte_Mapping ) }
	    if( AT_Byte >= Segment_Size ) then { AT was not big enough }
            begin
	        AT_Bit := 8 ; { Terminate loop }
            end ;
	end ;
    end ;
    SpaceAvail := Counter ;
end ; { TAT.SpaceAvail }



end.
