{
        Program Name : Database
        Package Name : LB
        Purpose      : Ancestral database objects
        Institution  : Conroy & Conroy Co.
        Date Written : 26-Jun-1996
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This module defines the base database objects.
}

unit Database ;

{$I EDEFINES.INC}

interface

uses Collect, { TCollection }
     Common, { TBase_Object }
     _UEHDefs ; { TUEC }

const DatabaseErr_Facility = 49 ;
      DatabaseErr_Success = 0 ;
      DatabaseErr_Not_Supported = 1 ; { Operation not supported }
      DatabaseErr_Invalid_Table = 2 ; { Reference to table that is not in the database }
      DatabaseErr_Duplicate_Table = 3 ; { Tried to add a table to the same database more than once }
      DatabaseErr_Table_Not_Connected = 4 ; { Tried to perform a database operation on an unconnected table }
      DatabaseErr_Record_Not_Found = 5 ; { Record was not found }
      DatabaseErr_Field_Not_Found = 6 ; { Field was not found }
      DatabaseErr_Expression_Syntax_Error = 7 ; { Selection expression error }
      DatabaseErr_Key_Not_Found = 8 ; { Key was not found }

      { Database Field Types... }
const DBFT_Undefined = 0 ;
      DBFT_String = 1 ;
      DBFT_Integer = 2 ; { 4-byte 2's complement integer }
      DBFT_Float = 3 ; { 8-byte (double) IEEE FP }
      DBFT_Boolean = 4 ; { 1-byte }
      DBFT_Binary = 5 ;

type TSelect_Operation = ( SO_Equal, SO_Not_Equal, SO_LT, SO_LTE, SO_GT, SO_GTE, SO_Invalid ) ;

type TDatabase_Field = record
                           Name : string[ 255 ] ; { Field name }
                           Offset : longint ; { Offset byte of field in record }
                           Len : longint ; { 0 = variable length }
                           Data_Type : longint ; { See TDBFT_* }
                           Upper_Name : string[ 255 ] ; { Uppercase version of Name }
                           Version : longint ; { Version of TDatabase_Field (10 = V1.0) }
                       end ;
     PDatabase_Field = ^TDatabase_Field ;

     TDatabase_Fields = object( TCollection )
                            public { Constructors and destructors... }
                                destructor Done ; virtual ;

                            public { API... }
                                procedure Add( Rec : PDatabase_Field ) ;
                                procedure Add_Field( Name : string ;
                                    Offset, Len, Data_Type : longint ) ;
                        end ;
     PDatabase_Fields = ^TDatabase_Fields ;

     TDatabase_Key = record
                         Allow_Duplicates : boolean ;
                         Name : string[ 255 ] ; { Name of key }
                         Fields : PDatabase_Fields ;
                     end ;
     PDatabase_Key = ^TDatabase_Key ;

     TDatabase_Keys = object( TCollection )
			  public { Constructors and destructors... }
                              destructor Done ; virtual ;

                          public { API... }
                              procedure Add( Key : PDatabase_Key ) ;
                      end ;
     PDatabase_Keys = ^TDatabase_Keys ;

     TSelection_Record = record
                             Field : longint ; { Field number (0 = key) }
                             Start : longint ; { Starting position within field }
                             Len : longint ; { Size of selection area }
                             Operation : TSelect_Operation ; { Type of operation }
                             Comparison : string ; { Data to compare with }
                             Version : longint ; { Version of TSelection_Record (10 = V1.0) }
                         end ;
     PSelection_Record = ^TSelection_Record ;

type PDatabase = ^TDatabase ;
     PDataBase_Table = ^TDatabase_Table ;
     PDatabase_Record = ^TDatabase_Record ;

     TDatabase_Cursor = object( TBase_Object )
                            public { Instance data... }
                                Database : PDatabase_Table ;
                                Context : PDatabase_Record ;
                                Key : longint ;

                            public { Constructor/destructor }
                                constructor Init ;

                            public { API... }
                                function First : PDatabase_Record ; virtual ;
                                function Previous : PDatabase_Record ; virtual ;
                                function Next : PDatabase_Record ; virtual ;
                                function Last : PDatabase_Record ; virtual ;
                        end ;
     PDatabase_Cursor = ^ TDatabase_Cursor ;

     TDatabase_Record = object( TBase_Object ) { A record within a table }
			    private { Instance data... }
                                _Database : PDatabase_Table ;

                            public { Constructor... }
                                constructor Init ;

                            public { API... }
                                function Key_Count : longint ; virtual ;
                                { Returns number of keys defined. }

                                function Get_Key( Index : longint ) : string ;
                                    virtual ;
                                { Returns the value of the key for this
                                  record for the specified index. }

                                function Get_Database : PDatabase_Table ;
                                    virtual ;
                                { Returns database/table that this record
                                  belongs in. }

                                function Get_Serialization : string ; virtual ;
                                { Returns a serialized version of the record. }

                                function Key_Match( Index : longint ;
                                    const S : string ) : boolean ; virtual ;
                                { Returns true if the passed string matches the
                                  record's key. }

                                function Match( Index : integer ;
                                    Select_Spec : PCollection ) : boolean ;
                                    virtual ;
                                { Returns True if the record matches the passed
                                  selection specification. }

                                procedure Set_Database( DB : PDatabase_Table ) ;
                                    virtual ;
                                { Returns the database/table that this record
                                  belongs in. }

                                procedure Set_From_Serialization( const Value : string ) ;
                                    virtual ;
                                { Sets the record from a serialized string. }

                                function Size : longint ; virtual ;
                                { Returns size of record, in bytes. }

                                procedure Update ; virtual ;
                                { Commits changes to the record to the
                                  database. }
                        end ;

     TLock_Type = ( LOCK_Read, LOCK_Write, LOCK_All ) ;

     TDatabase_Table = object( TBase_Object )
                           private
                               DB_Error : TUEC ; { Indicates status of last operation }
                               _Database : PDatabase ; { Database to which we belong (nil in view mode) }
                               View_Table : PDatabase_Table ; { Table that we are a view of (nil if not in view mode) }
                               List : TCollection ; { List of records in view (view mode only) }
                               _Name : string[ 255 ] ; { Name of table }
                               Fields : PDatabase_Fields ; { Record definition for this table (list of fields) }
                               Keys : PDatabase_Keys ; { List of indexes for this table }

                           public { Constructors and destructors... }
                               constructor Init ;
                               destructor Done ; virtual ;

                           public { Descendant support... }
                               procedure Set_Error( F, C : longint ) ; virtual ;
                               { Sets current error code. }

                           public { API... }
                               function Add( Rec : PDatabase_Record ) : PDatabase_Record ;
                                   virtual ;
                               { Adds the record to the database.  If a record
                                 with the matching key is found and duplicates
                                 are not allowed, it returns the existing
                                 record.  Otherwise, it returns the passed
                                 record. }

                               function Duplicates_Allowed( Index : longint ) : boolean ;
                                   virtual ;
                               { Returns True if the specified key allows
                                 duplicates. }

                               function Begin_Transaction : longint ; virtual ;
                               { Begins a transaction and returns an ID.  If no
                                 transaction is begun (such as when transaction
                                 processing is not supported by the database,
                                 the function returns 0. }

                               procedure Cancel_Transaction( ID : longint ) ;
                                   virtual ;
                               { Ends the specified transaction, cancelling all
                                 pending changes.  ID is the value returned by
                                 Begin_Transaction. }

                               function Count : longint ; virtual ;
                               { Returns count of records in table (-1 if
                                 operation is not supported). }

                               procedure Delete(  Index : longint ;
                                   const Key : string ) ; virtual ;
                               { Delete the specified record }

                               procedure End_Transaction( ID : longint ) ;
                                   virtual ;
                               { Ends the specified transaction, committing all
                                 pending changes.  ID is the value returned by
                                 Begin_Transaction.  After this call, the
                                 transaction reference by ID can no longer be
                                 referenced. }

                               function Get( Index : longint ;
                                   const Key : string ) : PDatabase_Record ;
                                   virtual ;
                               { Get the record with the specified key.  Returns
                                 nil if error. }

                               function Get_Database : PDatabase ;
                               { Returns database to which this table belongs.
                                 If the table is not part of another database,
                                 this returns nil. }

                               procedure Get_Last_Error( var UEC : TUEC ) ;
                                   virtual ;
                               { Sets UEC to last error code. }

                               function Get_Layout : PDatabase_Fields ;
                                   virtual ;
                               { Returns the record layout. }

                               function Get_Name : string ; virtual ;
                               { Returns table name }

                               procedure Lock( Locks : TLock_Type ) ; virtual ;
                               { Locks current record }

                               function Make_Cursor( Key : longint ) : PDatabase_Cursor ;
                                   virtual ;
                               { Creates a cursor for this table using the
                                 specified key. }

                               function Next( Index : longint ;
                                   Rec : PDatabase_Record ) : PDatabase_Record ;
                                   virtual ;
                               { Returns next record after specified record.  If
                                 nil is passed, the first record is returned.
                                 If there is no next record, nil is returned. }

                               function Previous( Index : longint ;
                                   Rec : PDatabase_Record ) : PDatabase_Record ;
                                   virtual ;
                               { Returns record prior to the specified record.
                                 If nil is passed, the last record is
                                 returned.  If there is no previous record, nil
                                 is returned. }

                               procedure Remove_From_View( Rec : PDatabase_Record ) ;
                                   virtual ;
                               { Deletes the record from the view, but not from
                                 the database. }

                               function Select( Index : integer ;
                                   const Expression : string ) : PDatabase_Table ;
                               { Performs a selection using the speciried index,
                                 returning a view containing all records
                                 matching the passed expression. }

                               procedure Set_Database( DB : PDatabase ) ;
                               { Makes this table part of the specified table. }

                               procedure Set_Keys( K : PDatabase_Keys ) ;
                                   virtual ;
                               { Defines the key structure }

                               procedure Set_Layout( Value : PDatabase_Fields ) ;
                                   virtual ;
                               { Defines the record layout. }

                               procedure Set_Name( Value : string ) ; virtual ;
                               { Sets name of the table/view }

                               procedure Set_View( DB : PDatabase_Table ) ;
                                   virtual ;
                               { Sets us up as a view into the specified
                                 database.  If nil is passed, we are no longer a
                                 view. }

                               procedure Unlock( Locks : TLock_Type ) ;
                                   virtual ;
                               { Unlocks current record }

                               procedure Zero ; virtual ;
                               { Deletes entire database }
	               end ; { TDatabase_Table }

     TDatabase = object( TBase_Object ) { A collection of tables }
		     private
                         DB_Error : TUEC ; { Indicates status of last operation }
                         Table_List : TCollection ; { Collection of tables }
                         _Name : string[ 255 ] ; { Database name }

                     public { Constructors and destructors... }
                         constructor Init ;
                         destructor Done ; virtual ;

                     public { Descendant support... }
			 function Relational : boolean ; virtual ;
			 { Returns true if we are a relational database. }

                         procedure Set_Error( F, C : longint ) ; virtual ;
                         { Sets current error code. }

                     public { API... }
                         procedure Add( Table : PDatabase_Table ;
                             Rec : PDatabase_Record ) ; virtual ;
                         { Adds the specified record to the specified table. }

                         procedure Add_Table( Table : PDatabase_Table ) ;
                             virtual ;
                         { Adds the specified table to the database. }

                         function Begin_Transaction : longint ; virtual ;
                         { Begins a transaction and returns an ID.  If no
                           transaction is begun (such as when transaction
                           processing is not supported by the database, the
                           function returns 0. }

                         procedure Cancel_Transaction( ID : longint ) ;
                             virtual ;
                         { Ends the specified transaction, cancelling all
                           pending changes.  ID is the value returned by
                           Begin_Transaction. }

                         procedure Delete( Table : PDatabase_Table ;
                              Index : longint ; const Key : string ) ; virtual ;
                         { Delete the specified record from the specified
                           table. }

                         procedure End_Transaction( ID : longint ) ; virtual ;
                         { Ends the specified transaction, committing all
                           pending changes.  ID is the value returned by
                           Begin_Transaction.  After this call, the transaction
                           reference by ID can no longer be referenced. }

                         function Get( Table : PDatabase_Table ;
                              Index : longint ;
                              const Key : string ) : PDatabase_Record ;
                              virtual ;
                         { Get the record with the specified key in the
                           specified table.  Returns nil if error. }

                         procedure Get_Last_Error( var UEC : TUEC ) ; virtual ;

                         function Get_Name : string ; virtual ;
                         { Returns database name }

                         procedure Lock( Locks : TLock_Type ) ; virtual ;
                         { Locks current record }

                         procedure Remove_Table( Table : PDatabase_Table ) ;
                             virtual ;
                         { Removes the specified table to the database. }

                         procedure Set_Name( Value : string ) ; virtual ;
                         { Sets name of the database }

                         procedure Unlock( Locks : TLock_Type ) ; virtual ;
                         { Unlocks current record }

                         procedure Zero ; virtual ; { Deletes entire database }
		 end ; { TDatabase }

type PISAM_DataBase = ^TISAM_Database ;
     TISAM_DataBase = object( TDatabase_Table )
                          public { API... }
			     function Get_GE( Index : longint ;
                                 const Key : string ) : PDatabase_Record ;
                                 virtual ;
			     { Get the first record whose key is greater than
			       or equal to the specified key. }

			     function Get_LE( Index : longint ;
                                 const Key : string ) : PDatabase_Record ;
				 virtual ;
			     { Get the last record whose key is less than
			       or equal to the specified key. }

                             function Key_Count : integer ; virtual ;

                             function Make_Cursor( Key : longint ) : PDatabase_Cursor ;
                                 virtual ;
		      end ;

     TISAM_Database_Cursor = object( TDatabase_Cursor )
                                 public { API... }
                                     function First_GE( const _Key : string ) : PDatabase_Record ;
                                         virtual ;
                             end ;
     PISAM_Database_Cursor = ^TISAM_Database_Cursor ;

implementation

uses { Borland... }
     SysUtils, { inttostr }

     { C&C... }
     _App, { Application_Manager }
     Parse ; { Parse_Command }

{ TDatabase_Keys methods... }

{ Constructors and destructors... }

destructor TDatabase_Keys.Done ;

var Dummy : integer ;

begin
    for Dummy := 0 to Count - 1 do
    begin
        dispose( PDatabase_Key( At( Dummy ) ) ) ;
    end ;

    inherited Done ;
end ;


{ API... }

procedure TDatabase_Keys.Add( Key : PDatabase_Key ) ;

begin
    Insert( Key ) ;
end ;


{ TDatabase_Fields methods... }

{ Constructors and destructors... }

destructor TDatabase_Fields.Done ;

var Loop : longint ;

begin
    for Loop := 0 to Count - 1 do
    begin
        dispose( PDatabase_Field( At( Loop ) ) ) ;
    end ;

    inherited Done ;
end ;


{ API... }

procedure TDatabase_Fields.Add( Rec : PDatabase_Field ) ;

begin
    Insert( Rec ) ;
end ;


procedure TDatabase_Fields.Add_Field( Name : string ;
    Offset, Len, Data_Type : longint ) ;

var Rec : PDatabase_Field ;

begin
    new( Rec ) ;
    Rec^.Name := Name ;
    Rec^.Offset := Offset ;
    Rec^.Len := Len ;
    Rec^.Data_Type := Data_Type ;
    Rec^.Upper_Name := uppercase( Name ) ;
    Rec^.Version := 10 ;
    Add( Rec ) ;
end ;



{ TDatabase_Cursor methods... }

{ Constructors/destructors... }

constructor TDatabase_Cursor.Init ;

begin
    inherited Init ;

    Database := nil ;
    Context := nil ;
    Key := 0 ;
end ;


{ API... }

function TDatabase_Cursor.First : PDatabase_Record ;

begin
    Context := Database^.Next( Key, nil ) ;
    First := Context ;
end ;


function TDatabase_Cursor.Previous : PDatabase_Record ;

begin
    Context := Database^.Previous( Key, Context ) ;
    Previous := Context ;
end ;


function TDatabase_Cursor.Next : PDatabase_Record ;

begin
    Context := Database^.Next( Key, Context ) ;
    Next := Context ;
end ;


function TDatabase_Cursor.Last : PDatabase_Record ;

begin
    Context := Database^.Previous( key, nil ) ;
    Last := Context ;
end ;



{ TISAM_Database_Record methods... }

function TISAM_Database_Cursor.First_GE( const _Key : string ) : PDatabase_Record ;

begin
    Context := PISAM_Database( Database )^.Get_GE( Key, _Key ) ;
    First_GE := Context ;
end ;



{ TDatabase_Record methods... }

{ Constructor... }

constructor TDatabase_Record.Init ;

begin
    inherited Init ;

    _Database := nil ;
end ;


{ API... }

function TDatabase_Record.Key_Count : longint ;

begin
    Key_Count := 0 ;
end ;


function TDatabase_Record.Get_Key( Index : longint ) : string ;
{ Returns the value of the key for this record. }

begin
    Get_Key := '' ; { Uninitialized record }
end ;


function TDatabase_Record.Get_Database : PDatabase_Table ;
{ Returns current database that this record belongs in. }

begin
    Get_Database := _Database ;
end ;


function TDatabase_Record.Get_Serialization : string ;
{ Returns a serialized version of the record. }

begin
    Get_Serialization := '' ;
end ;


function TDatabase_Record.Key_Match( Index : longint ;
    const S : string ) : boolean ;
{ Returns true if the passed string matches the record's key. }

begin
    Key_Match := ( S = Get_Key( Index ) ) ;
end ;


function TDatabase_Record.Match( Index : integer ;
    Select_Spec : PCollection ) : boolean ;

begin
    Match := False ; { Assume no match }
end ;


procedure TDatabase_Record.Set_Database( DB : PDatabase_Table ) ;
{ Returns current database that this record belongs in. }

begin
    _Database := DB ;
end ;


function TDatabase_Record.Size : longint ;
{ Returns size of record, in bytes. }

begin
    Size := 0 ;
end ;


procedure TDatabase_Record.Set_From_Serialization( const Value : string ) ;
{ Sets the record from a serialized string. }

begin
    Application_Manager^.Fatal_Exception( 'Abstract method called: TDatabase_Record.Set_From_Serialization' ) ;
    halt ;
end ;


procedure TDatabase_Record.Update ;
{ Commits changes to the record to the database. }

begin
    Application_Manager^.Fatal_Exception( 'Abstract method called: TDatabase_Record.Update' ) ;
    halt ;
end ;



{ TDatabase_Table methods... }

constructor TDatabase_Table.Init ; { Initialize database }

begin
    inherited Init ;

    fillchar( DB_Error, sizeof( DB_Error ), 0 ) ; { No errors so far }
    _Database := nil ; { Not part of another database }
    _Name := '' ;
    Fields := nil ;
    Keys := nil ;
    
    { Set-up so we are not a view... }
    View_Table := nil ;
    List.Init( 4, 4 ) ;
end ;


destructor TDatabase_Table.Done ;

begin
    Zero ;
    List.Done ;
    
    inherited Done ;
end ;


{ Internal support only... }

function TDatabase_Table.Next( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Next := nil ;
end ;


function TDatabase_Table.Previous( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;

begin
    Previous := nil ;
end ;


{ Descendant support... }

procedure TDatabase_Table.Set_Error( F, C : longint ) ;

begin
    DB_Error.Facility := F ;
    DB_Error.Code := C ;
end ;


{ API... }

function TDatabase_Table.Add( Rec : PDatabase_Record ) : PDatabase_Record ;
{ Adds the specified record to the database. }

begin
    Add := nil ;
end ;


function TDatabase_Table.Duplicates_Allowed( Index : longint ) : boolean ;

begin
    if( ( Keys = nil ) or ( Index >= Keys.Count ) ) then
    begin
        Duplicates_Allowed := False ;
    end else
    begin
        Duplicates_Allowed := PDatabase_Key( Keys.At( Index ) ).Allow_Duplicates ;
    end ;
end ;


function TDatabase_Table.Begin_Transaction : longint ;

begin
    if( View_Table <> nil ) then
    begin
        Begin_Transaction := View_Table^.Begin_Transaction ;
    end else
    if( _Database <> nil ) then
    begin
        Begin_Transaction := _Database^.Begin_Transaction ;
    end else
    begin
        Begin_Transaction := 0 ; { Transactions not supported by default }
    end ;
end ;


procedure TDatabase_Table.Cancel_Transaction( ID : longint ) ;

begin
    if( View_Table <> nil ) then
    begin
        View_Table^.Cancel_Transaction( ID ) ;
    end else
    if( _Database <> nil ) then
    begin
        _Database^.Cancel_Transaction( ID ) ;
    end else
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Table_Not_Connected ) ;
    end ;
end ;


function TDatabase_Table.Count : longint ;

begin
    Set_Error( 0, 0 ) ;
    if( View_Table <> nil ) then
    begin
        Count := List.Count ;
    end else
    begin
        Count := -1 ; { Not supported }
    end ;
end ;


procedure TDatabase_Table.Delete( Index : longint ; const Key : string ) ;

begin
    if( View_Table <> nil ) then
    begin
        View_Table^.Delete( Index, Key ) ;
        View_Table^.Get_Last_Error( DB_Error ) ;
    end else
    begin
        Application_Manager^.Fatal_Exception( 'Abstract method called: TDatabase_Table.Delete' ) ;
        halt ;
    end ;
end ;


procedure TDatabase_Table.End_Transaction( ID : longint ) ;

begin
    if( View_Table <> nil ) then
    begin
        View_Table^.End_Transaction( ID ) ;
    end else
    if( _Database <> nil ) then
    begin
        _Database^.End_Transaction( ID ) ;
    end else
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Table_Not_Connected ) ;
    end ;
end ;


function TDatabase_Table.Get( Index : longint ;
    const Key : string ) : PDatabase_Record ;

begin
    Get := nil ;
    Application_Manager^.Fatal_Exception( 'Abstract method called: TDatabase_Table.Get' ) ;
    halt ;
end ;


function TDatabase_Table.Get_Database : PDatabase ;

begin
    Get_Database := _Database ;
end ;


procedure TDatabase_Table.Get_Last_Error( var UEC : TUEC ) ;
{ Return last error code }

begin
    UEC := DB_Error ;
end ;


function TDatabase_Table.Get_Layout : PDatabase_Fields ;

begin
    Get_LayOut := Fields ;
end ;


function TDatabase_Table.Get_Name : string ;

begin
    Get_Name := _Name ;
end ;


procedure TDatabase_Table.Lock( Locks : TLock_Type ) ;
{ Locks current record.  TDatabase_Table is a single-user database, so this does
  nothing. }

begin
end ;


function TDatabase_Table.Make_Cursor( Key : longint ) : PDatabase_Cursor ;

var Res : PDatabase_Cursor ;

begin
    if( Key = 0 ) then
    begin
        Res := new( PDatabase_Cursor, Init ) ;
        Res^.Database := @self ;
        Res^.Key := Key ;
        Make_Cursor := Res ;
    end else
    begin
        Make_Cursor := nil ;
        Set_Error( DatabaseErr_Facility, DatabaseErr_Key_Not_Found ) ;
    end ;
end ;


procedure TDatabase_Table.Remove_From_View( Rec : PDatabase_Record ) ;

var Dummy : longint ;

begin
    if( View_Table <> nil ) then
    begin
        Dummy := List.IndexOf( Rec ) ;
        if( Dummy = -1 ) then
        begin
            Set_Error( DatabaseErr_Facility, DatabaseErr_Record_Not_Found ) ;
        end else
        begin
            List.AtDelete( Dummy ) ;
            Set_Error( 0, 0 ) ;
        end ;
    end else
    begin
        { Not a view - do nothing }
    end ;
end ;


function TDatabase_Table.Select( Index : integer ;
    const Expression : string ) : PDatabase_Table ;

var {Cursor : PISAM_Database_Cursor ;}
    Cursor : PDatabase_Cursor ;
    Dummy : longint ;
    Field : longint ;
    Field_Rec : PDatabase_Field ;
    Found : longint ;
    L : longint ;
    Res : PDatabase_Table ;
    Op : TSelect_Operation ;
    Rec : PDatabase_Record ;
    S, Work : string ;
    Select_Spec : TCollection ;
    Sel_Rec : PSelection_Record ;
    I : longint ;
    B : longbool ;
    R : double ;
    Switch : string ;

    function Get_Operation( const S : string ) : TSelect_Operation ;

    begin
        if( S = '=' ) then
        begin
            Get_Operation := SO_Equal ;
        end else
        if( ( S = '<>' ) or ( S = '#' ) ) then
        begin
            Get_Operation := SO_Not_Equal ;
        end else
        if( S = '<' ) then
        begin
            Get_Operation := SO_LT ;
        end else
        if( S = '<=' ) then
        begin
            Get_Operation := SO_LTE ;
        end else
        if( S = '>' ) then
        begin
            Get_Operation := SO_GT ;
        end else
        if( S = '>=' ) then
        begin
            Get_Operation := SO_GTE ;
        end else
        begin
            Get_Operation := SO_Invalid ;
        end ;
    end ;


    procedure Erase_Spec ;

    var Dummy : longint ;
        Sel_Rec : PSelection_Record ;

    begin
        for Dummy := 0 to Select_Spec.Count - 1 do
        begin
            Sel_Rec := PSelection_Record( Select_Spec.At( Dummy ) ) ;
            freemem( Sel_Rec, sizeof( Sel_Rec^ ) ) ;
        end ;
        Select_Spec.Done ;
    end ;

begin { TDatabase_Table.Select }
    Res := new( PDatabase_Table, Init ) ;
    Res^.Set_View( @self ) ;
    if( Fields <> nil ) then
    begin
        Select_Spec.Init( 4, 4 ) ;

        { Parse selection expression... }
        Work := Expression ;
        while( length( Work ) <> 0 ) do
        begin
            { Get field name... }
            Found := -1 ;
            Field_Rec := nil ;
            S := uppercase( Parse_Parameter( ' ', Work ) ) ;
            if( S = 'KEY' ) then
            begin
                Found := 0 ;
            end else
            begin
                for Field := 0 to Fields.Count - 1 do
                begin
                    Field_Rec := PDatabase_Field( Fields.At( Field ) ) ;
                    if( Field_Rec^.Upper_Name = S ) then
                    begin
                        Found := Field + 1 ;
                        break ;
                    end ;
                end ; { for Field := 0 to Fields.Count - 1 }
            end ;
            if( Found = -1 ) then
            begin
                Select := Res ;
                Erase_Spec ;
		Set_Error( DatabaseErr_Facility, DatabaseErr_Field_Not_Found ) ;
                exit ;
            end ;

            { Get comparison operator... }
            S := Parse_Parameter( ' ', Work ) ;
            Op := Get_Operation( S ) ;
            if( Op = SO_Invalid ) then
            begin
                Select := Res ;
                Erase_Spec ;
		Set_Error( DatabaseErr_Facility, DatabaseErr_Expression_Syntax_Error ) ;
                exit ;
            end ;

            { Setup selection record... }
            Sel_Rec := new( PSelection_Record ) ;
            Sel_Rec^.Field := Found ;
            Sel_Rec^.Start := 0 ;
	    if( Found > 0 ) then
            begin
                Sel_Rec^.Len := Field_Rec^.Len ;
            end else
            begin
                Sel_Rec^.Len := -1 ; { Key field defaults to entire length }
            end ;
            Sel_Rec^.Operation := Op ;
            Sel_Rec^.Version := 10 ;

            { Get comparison value... }
            S := Parse_Parameter( ' ', Work ) ;
            if( Found = 0 ) then
            begin
                if( Parse_Switch( 'N|UMBER', '', S, Switch ) <> 0 ) then
		begin
                    if( length( Switch ) > 0 ) then
                    begin
                        Select := Res ;
                        Erase_Spec ;
                        Set_Error( DatabaseErr_Facility, DatabaseErr_Expression_Syntax_Error ) ;
                        exit ;
                    end ;
                    Dummy := strtoint( S ) ;
                    S := '    ' ;
                    move( Dummy, S[ 1 ], 4 ) ;
                end ; { if( Parse_Switch( 'N|UMBER', '', S, Switch ) <> 0 )  }
            end else
            begin
                case Field_Rec^.Data_Type of
                    DBFT_Integer:
                        begin
                            val( S, I, Dummy ) ;
                            L := Field_Rec^.Len ;
                            if( L > sizeof( I ) ) then
                            begin
                                L := sizeof( I ) ;
                            end ;
                            move( I, S[ 1 ], L ) ;
                        end ;
                    DBFT_Float:
                        begin
                            val( S, R, Dummy ) ;
                            L := Field_Rec^.Len ;
                            if( L > sizeof( R ) ) then
                            begin
                                L := sizeof( R ) ;
                            end ;
                            move( R, S[ 1 ], L ) ;
                        end ;
                    DBFT_Boolean:
                        begin
                            S := uppercase( S ) ;
                            if( S = copy( 'TRUE', 1, length( S ) ) ) then
                            begin
                                B := True ;
                            end else
                            begin
                                B := False ;
                            end ;
                            L := Field_Rec^.Len ;
                            if( L > sizeof( B ) ) then
                            begin
                                L := sizeof( B ) ;
                            end ;
                            move( B, S[ 1 ], L ) ;
                            Dummy := 0 ;
                        end ;
                    else Dummy := 0 ;
                end ;
                if( Dummy <> 0 ) then
                begin
                    Select := Res ;
                    Erase_Spec ;
                    Set_Error( DatabaseErr_Facility, DatabaseErr_Expression_Syntax_Error ) ;
                    exit ;
                end ;
            end ; { if( Found > 0 ) }
            Sel_Rec^.Comparison := S ;
            Select_Spec.Insert( Sel_Rec ) ;
        end ; { while( length( Work ) <> 0 ) }

{        Cursor := PISAM_Database_Cursor( Make_Cursor( Index ) ) ;}
        Cursor := PDatabase_Cursor( Make_Cursor( Index ) ) ;
        if( Select_Spec.Count > 0 ) then
        begin
            { Now do the selection... }
            Sel_Rec := PSelection_Record( Select_Spec.At( 0 ) ) ;
            Rec := nil ;
            if( Sel_Rec^.Field = 0 ) then { Doing a selection primarily on key }
            begin
(*
                case Sel_Rec^.Operation of
                    SO_GT, SO_GTE: Rec := Cursor.First_GE( Sel_Rec^.Comparison ) ;
                    else Rec := Cursor.First ;
                end ;
*)
            end else
            begin
                Rec := Cursor.First ;
            end ;
            while( Rec <> nil ) do
            begin
                if( Rec^.Match( Index, @Select_Spec ) ) then
                begin
                    Res^.List.Insert( Rec ) ;
                end ;
                Rec := Cursor.Next ;
            end ;
        end ; { if( Select_Spec.Count > 0 ) }

        { Clean up... }
        Erase_Spec ;
        dispose( Cursor, Done ) ;
        Set_Error( 0, 0 ) ;
    end ; { if( Fields <> nil ) }
    Select := Res ;
end ; { TDatabase_Table.Select }


procedure TDatabase_Table.Set_Database( DB : PDatabase ) ;

begin
    if( _Database <> nil ) then
    begin
        _Database^.Remove_Table( @self ) ;
    end ;
    _Database := DB ;
end ;


procedure TDatabase_Table.Set_Keys( K : PDatabase_Keys ) ;

begin
    Keys := K ;
end ;


procedure TDatabase_Table.Set_Layout( Value : PDatabase_Fields ) ;

begin
    Fields := Value ;
end ;


procedure TDatabase_Table.Set_Name( Value : string ) ;

begin
    _Name := Value ;
end ;


procedure TDatabase_Table.Set_View( DB : PDatabase_Table ) ;

begin
    if( ( _Database <> nil ) and ( View_Table = nil ) ) then
    begin
        Set_Database( nil ) ;
    end ;
    View_Table := DB ;
    Zero ;
end ;


procedure TDatabase_Table.Unlock( Locks : TLock_Type ) ;
{ Unlocks current record.  TDatabase_Table is a single-user database, so this
  does nothing. }

begin
end ;


procedure TDatabase_Table.Zero ; { Deletes entire table/view }

begin
    if( View_Table <> nil ) then
    begin
        List.Done ;
        List.Init( 4, 4 ) ;
    end else
    begin
        { Abstract method - do nothing }
    end ;
end ;



{ TISAM_Database methods... }

{ API... }

function TISAM_Database.Get_GE( Index : longint ;
    const Key : string ) : PDatabase_Record ;
{ Get the first record whose key is greater than or equal to the specified
  key. }

begin
    Get_GE := nil ;
    Application_Manager^.Fatal_Exception( 'Abstract method called: TISAM_Database.Get_GE' ) ;
    halt ;
end ;


function TISAM_Database.Get_LE( Index : longint ;
    const Key : string ) : PDatabase_Record;
{ Get the last record whose key is less than or equal to the specified key. }

begin
    Get_LE := nil ;
    Application_Manager^.Fatal_Exception( 'Abstract method called: TISAM_Database.Get_LE' ) ;
    halt ;
end ;


function TISAM_Database.Key_Count : integer ;

begin
    if( Keys = nil ) then
    begin
        Key_Count := 1 ;
    end else
    begin
        Key_Count := Keys.Count + 1 ;
    end ;
end ;


function TISAM_Database.Make_Cursor( Key : longint ) : PDatabase_Cursor ;

var Res : PISAM_Database_Cursor ;

begin
    if( Key <= Key_Count ) then
    begin
        Res := new( PISAM_Database_Cursor, Init ) ;
        Res^.Database := @self ;
        Res^.Key := Key ;
        Make_Cursor := Res ;
    end else
    begin
        Make_Cursor := nil ;
        Set_Error( DatabaseErr_Facility, DatabaseErr_Key_Not_Found ) ;
    end ;
end ;


{ TDatabase methods... }

{ Constructors and destructors... }

constructor TDatabase.Init ;

begin
    inherited Init ;

    Table_List.Init( 4, 4 ) ;
    _Name := '' ;
end ;


destructor TDatabase.Done ;

var Dummy : longint ;

begin
    for Dummy := 0 to Table_List.Count - 1 do
    begin
        dispose( PDatabase_Table( Table_List.At( Dummy ) ), Done ) ;
    end ;
    Table_List.Done ;

    inherited Done ;
end ;


{ Descendant support... }

function TDatabase.Relational : boolean ;

begin
    Relational := True ; { Most databases are, these days }
end ;


procedure TDatabase.Set_Error( F, C : longint ) ;

begin
    DB_Error.Facility := F ;
    DB_Error.Code := C ;
end ;


{ API... }

procedure TDatabase.Add( Table : PDatabase_Table ; Rec : PDatabase_Record ) ;

begin
    if( ( Table = nil ) or ( Table^.Get_Database <> @self ) ) then
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Invalid_Table ) ;
        exit ;
    end ;
    Table^.Add( Rec ) ;
end ;


procedure TDatabase.Add_Table( Table : PDatabase_Table ) ;

begin
    if( Table_List.IndexOf( Table ) <> -1 ) then
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Duplicate_Table ) ;
        exit ;
    end ;
    Table_List.Insert( Table ) ;
    Table^.Set_Database( @self ) ;
end ;


function TDatabase.Begin_Transaction : longint ;

begin
    Begin_Transaction := 0 ; { Not supported }
end ;


procedure TDatabase.Cancel_Transaction( ID : longint ) ;

begin
    Set_Error( DatabaseErr_Facility, DatabaseErr_Not_Supported ) ;
end ;


procedure TDatabase.Delete( Table : PDatabase_Table ; Index : longint ;
    const Key : string ) ;

begin
    if( ( Table = nil ) or ( Table^.Get_Database <> @self ) ) then
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Invalid_Table ) ;
        exit ;
    end ;
    Table^.Delete( Index, Key ) ;
    Table^.Get_Last_Error( DB_Error ) ;
end ;


procedure TDatabase.End_Transaction( ID : longint ) ;

begin
    Set_Error( DatabaseErr_Facility, DatabaseErr_Not_Supported ) ;
end ;


function TDatabase.Get( Table : PDatabase_Table ;
     Index : longint ; const Key : string ) : PDatabase_Record ;

begin
    if( ( Table = nil ) or ( Table^.Get_Database <> @self ) ) then
    begin
        Get := nil ;
        Set_Error( DatabaseErr_Facility, DatabaseErr_Invalid_Table ) ;
        exit ;
    end ;
    Get := Table^.Get( Index, Key ) ;
end ;


procedure TDatabase.Get_Last_Error( var UEC : TUEC ) ;

begin
    UEC := DB_Error ;
end ;


function TDatabase.Get_Name : string ;

begin
    Get_Name := _Name ;
end ;


procedure TDatabase.Lock( Locks : TLock_Type ) ;

begin
end ;


procedure TDatabase.Remove_Table( Table : PDatabase_Table ) ;

var Dummy : longint ;

begin
    Dummy := Table_List.IndexOf( Table ) ;
    if( Dummy = -1 ) then
    begin
        Set_Error( DatabaseErr_Facility, DatabaseErr_Invalid_Table ) ;
    end else
    begin
        Set_Error( 0, 0 ) ;
        Table_List.AtDelete( Dummy ) ;
        Table^.Set_Database( nil ) ;
    end ;
end ;


procedure TDatabase.Set_Name( Value : string ) ;

begin
    _Name := Value ;
end ;


procedure TDatabase.Unlock( Locks : TLock_Type ) ;

begin
end ;


procedure TDatabase.Zero ;

var Dummy : longint ;

begin
    for Dummy := 0 to Table_List.Count - 1 do
    begin
        PDatabase_Table( Table_List.At( Dummy ) )^.Zero ;
    end ;
end ;


end.
