{$N+}
{
        Program Name : ErrMastr
        Package Name : LB
        Purpose      : Master error control
        Institution  : Conroy & Conroy Co.
        Date Written : 10-Jan-1999
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 1999-2001 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  16-Nov-2002   EAC         Fix BP compilation error.      *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides a single place for all application error
	handling.

          USE_UEH should be defined for applications which are going to use the
        UEH DLL, but the DLL itself should not define USE_UEH.  The DLL should
        define UEH instead.
}

unit ErrMastr ;

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF BP}
     Objects, { TCollection }
     Strings, { strpas }
{$ENDIF}

     { C&C... }
     _UEHDefs, { tUEC }
     UEHDefs ; { TStandard_Error_Interface }

      { Master Error Operation.. }
const TMEO_Silent = 0 ;
      TMEO_Ignore = 1 ;
      TMEO_Show = 2 ;

      { Standard Error Levels... }
const SEL_Success = 0 ;
      SEL_Notification = 1 ;
      SEL_Warning = 2 ;
      SEL_Error = 3 ;
      SEL_Fatal = 4 ;

{$IFDEF DELPHI}
function Error_Code( Facility, Code : longint ) : tUEC ;
{$ELSE}
function Error_Code( Facility, Code : longint ) : double ;
{$ENDIF}
function Get_Error_Text( UEC : tUEC ; Level : longint ) : PChar ;
function Last_Error : PChar ;
procedure Log_Error( UEC : tUEC ; const T : string ; Level : longint ) ;
function Register_Error_Interface( EI : PSError_Interface ) : longbool ;
procedure Set_Master_Error_Operation( MEO : longint ) ;
function Facility_Count : longint ;
function Facility_From_Index( Index : longint ) : longint ;
function Facility_Name_From_Index( Index : longint ) : PChar ;
function Unregister_Error_Interface( Facility : longint ) : longbool ;
function Get_Error_Text_Ex( UEC : tUEC ; Level : longint ;
    var Size, Typ : longint ) : PChar ;
function Severity( UEC : TUEC ) : longint ;
function Version : longint ;


implementation

uses { C&C... }
     _App, { Application_Manager }

{$IFDEF DELPHI}
     { Borland... }
     Dialogs, { ShowMessage }
     SysUtils, { inttostr }
     Windows, { FormatMessage }

     { C&C... }
     Collect { TCollection }
{$ELSE}
  {$IFDEF WINDOWS}
     WinDOS { GetDate }
  {$ELSE}
     DOS { GetEnv }
  {$ENDIF}
{$ENDIF}
     ;

{$IFDEF USE_UEH}
var UEH32_Handle : integer = 0 ;
    UEH_Error_Code : function ( Facility, Code : longint ) : tUEC ; stdcall ;
    UEH_Get_Error_Text : function ( UEC : tUEC ; Level : longint ) : PChar ;
        stdcall ;
    UEH_Get_Error_Text_Ex : function ( UEC : tUEC ; Level : longint ;
        var Size, Typ : longint ) : PChar ; stdcall ;
    UEH_Last_Error : function  : PChar ; stdcall ;
    UEH_Log_Error : procedure ( UEC : tUEC ; T : PChar ; Level : longint ) ;
        stdcall ;
    UEH_Register_Error_Interface : function ( EI : PSError_Interface ) : longbool ;
        stdcall ;
    UEH_Set_Master_Error_Operation : procedure ( MEO : longint ) ; stdcall ;
    UEH_Facility_Count : function  : longint ; stdcall ;
    UEH_Facility_From_Index : function ( Index : longint ) : longint ; stdcall ;
    UEH_Facility_Name_From_Index : function ( Index : longint ) : PChar ;
        stdcall ;
    UEH_Unregister_Error_Interface : function ( Facility : longint ) : longbool ;
        stdcall ;
    UEH_Version : function : longint ; stdcall ;
    UEH_Severity : function( UEC : TUEC ) : longint ; stdcall ;
{$ENDIF}

{$IFDEF DELPHI}
type tDOS_EI = class( TStandard_Error_Interface )
                   public
                       function Get_Facility : longint ; override ;
                       function Get_Name : PChar ; override ;
                       function Translate_Error( Code : longint ) : PChar ;
                           override ;
                       procedure Done ; override ;
                       function Facility_Version : longint ; override ;
               end ;
{$ENDIF}
type tSDOS_EI = object( TSStandard_Error_Interface )
                    public
                        function Get_Facility : longint ; virtual ;
                        function Get_Name : string ; virtual ;
                        function Translate_Error( Code : longint ) : string ;
                            virtual ;
                        function Facility_Version : longint ; virtual ;
                end ;
                
{$IFDEF DELPHI}
type tWindows_EI = class( TStandard_Error_Interface )
                       public
                           function Get_Facility : longint ; override ;
                           function Get_Name : PChar ; override ;
                           function Translate_Error( Code : longint ) : PChar ;
                               override ;
                           procedure Done ; override ;
                           function Facility_Version : longint ; override ;
                   end ;
{$ENDIF}
type tSWindows_EI = object( TSStandard_Error_Interface )
                        public
                            function Get_Facility : longint ; virtual ;
                            function Get_Name : string ; virtual ;
                            function Translate_Error( Code : longint ) : string ;
                                virtual ;
                            function Facility_Version : longint ; virtual ;
                    end ;

{$IFNDEF DELPHI}
type pSDOS_EI = ^tSDOS_EI ;
     pSWindows_EI = ^tSWindows_EI ;
{$ENDIF}


function Date : string ;

var Day_String, Year_String : string[ 4 ] ;
    Year, Month, Day : Word ;
{$IFNDEF DELPHI}
    DOW : word ;
{$ENDIF}

Const Months : Array[ 1..12 ] Of string[ 3 ] =
      ( 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' ) ;

begin
{$IFDEF DELPHI}
    DecodeDate( SysUtils.Date, Year, Month, Day ) ;
{$ELSE}
    Getdate( Year, Month, Day, Dow ) ;
{$ENDIF}
    Str( Year, Year_String ) ;
    Str( Day, Day_String ) ; { Convert day and year to strings }
    if( Length( Day_String ) = 1 ) then
    begin
        Day_String := '0' + Day_String ;
    end ;
    Date := Day_String + '-' + Months[ Month ] + '-' + Year_String ;
    { Now return value }
end ;


function Time : string ;

var Am_Pm : string[ 2 ] ;
    Hour, Minute, Second, Hundreths : Word ;
    Hour_String, Minute_String : string[ 2 ] ;

begin
{$IFDEF DELPHI}
    DecodeTime( SysUtils.Time, Hour, Minute, Second, Hundreths ) ;
{$ELSE}
    Gettime( Hour, Minute, Second, Hundreths ) ;
{$ENDIF}
    Am_Pm := 'AM' ;
    if( Hour > 11 ) then
    begin
        Am_Pm := 'PM' ;
        if( Hour > 12 ) then
        begin
            Hour := Hour - 12 ;
        end ;
    end ;
    if( Hour = 0 ) then
    begin
        Hour := 12 ;
    end ;
    Str( Hour, Hour_String ) ;
    Str( Minute, Minute_String ) ; { Convert hour and minute to strings }
    if( Length( Minute_String ) = 1 ) then
    begin
        Minute_String := '0' + Minute_String ;
    end ;
    Time := Hour_String + ':' + Minute_String + ' ' + Am_Pm ; { Now return value }
end ;


procedure Clear_IOResult ;

{$IFNDEF DELPHI}
var Dummy : integer ;
{$ENDIF}

begin
{$IFNDEF DELPHI}
    Dummy :=
{$ENDIF}
    IOResult ;
end ;


{ Copied from Standard.  Must have here because we cannot reference Standard in
  this unit or else it initializes before we do - which is bad. }
procedure CC_Log( const X : string ) ; { Log a message }

{$IFDEF DELPHI}
    function GetEnv( const X : string ) : string ;

    var Y : array[ 0..127 ] of char ;

    begin
        if( GetEnvironmentVariable( PChar( X ), Y, 127 ) = 0 ) then
        begin
            GetEnv := '' ; { Not defined }
        end else
        begin
            GetEnv := strpas( Y ) ;
        end ;
    end ;
{$ELSE}
  {$IFDEF WINDOWS}
    function GetENV( const X : string ) : string ;

    Type Env_Array = array[ 1..255 ] of Byte ;
         Env_Pointer = ^Env_Array ;

    var A : integer ;
        Ep : Env_Pointer ;
        Env : string ;
        Z : array[ 0..255 ] of char ;

    begin
        StrPCopy( Z, X ) ;
        Ep := Env_Pointer( Getenvvar( Z ) ) ;
        Env := '' ;
        if( Ep <> nil ) then {If variable is defined}
        begin
            A := 1 ;
            while( Ep^[ A ] <> 0 ) do {Until end of variable}
            begin
                Env := Env + chr( Ep^[ A ] ) ;
                inc( A ) ;
            end ;
        end ;
        GetENV := Env ;
    end ;
  {$ENDIF}

    function FileExists( const S : string ) : boolean ;

    var F : file ;
	I : integer ;

    begin
	assign( F, S ) ;
	{$I-}
	reset( F ) ;
	{$I+}
	I := IOResult ;
	FileExists := ( I = 0 ) ;
	if( I = 0 ) then
	begin
	    {$I-}
	    close( F ) ;
	    {$I+}
	end ;
	I := IOResult ;
    end ;
{$ENDIF}

var F : text ;
    S : string ;

begin
{$IFDEF DELPHI}
  try
{$ENDIF}
    S := GetEnv( 'CCLOG' ) ;
    if( length( S ) > 0 ) then { Symbol is defined }
    begin
        assign( F, S ) ;
        if( FileExists( S ) ) then
        begin
            {$I-}
            append( F ) ;
            {$I+}
        end else
        begin
            {$I-}
            rewrite( F ) ;
            {$I+}
        end ;
        if( IOResult <> 0 ) then
        begin
            exit ; { Error opening file }
        end ;
        if( Application_Manager <> nil ) then
        begin
{$IFDEF DELPHI}
            try
{$ENDIF}
                S := Application_Manager^.Get_Name ;
                if( length( S ) > 0 ) then
                begin
                    S := S + ' ' ;
                end ;
                if( length( Application_Manager^.Get_Version ) > 0 ) then
                begin
                    S := S + Application_Manager^.Get_Version ;
                end ;
{$IFDEF DELPHI}
            except
            end ;
{$ENDIF}
        end else
        begin
            S := '' ;
        end ;
        {$I-}
        writeln( F, S, ' on ', Date, ' at ', Time, ': ', X ) ;
        {$I+}
        Clear_IOResult ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
    end ; { if( length( S ) > 0 ) }
{$IFDEF DELPHI}
  except
  end ;
{$ENDIF}
end ; { CC_Log }



const DOS_Name : PChar = 'MSDOS' ;

var DOS_Error_Text : array[ 0..127 ] of char ;

function DOS_ERT( Code : integer ) : string ;

var Err_String : string ;

begin
    case Code of
	0 : Err_String := 'Success' ;
        1 : Err_String := 'Invalid function number' ;
	2 : Err_String := 'File not found' ;
	3 : Err_String := 'Path not found' ;
	4 : Err_String := 'Too many open files' ;
	5 : Err_String := 'File access denied' ;
	6 : Err_String := 'Invalid file handle' ;
	7 : Err_String := 'Memory blocks destroyed' ;
	8 : Err_String := 'Insufficient memory' ;
	9 : Err_String := 'Invalid memory block address' ;
	10 : Err_String := 'Invalid environment' ;
	11 : Err_String := 'Invalid format' ;
	12 : Err_String := 'Invalid file access code' ;
	13 : Err_String := 'Invalid data' ;
	14 : Err_String := 'Unknown unit' ;
	15 : Err_String := 'Invalid drive number' ;
	16 : Err_String := 'Cannot remove current directory' ;
	17 : Err_String := 'Cannot rename across drives' ;
	18 : Err_String := 'No more files' ;
	19 : Err_String := 'Disk is write-protected' ;
	20 : Err_String := 'Unknown unit' ;
	21 : Err_String := 'Drive not ready' ;
	22 : Err_String := 'Unknown command' ;
	23 : Err_String := 'CRC error in data' ;
	24 : Err_String := 'Bad request structure length' ;
	25 : Err_String := 'Disk seek error' ;
	26 : Err_String := 'Unknown media type' ;
	27 : Err_String := 'Sector not found' ;
	28 : Err_String := 'Printer out of paper' ;
	29 : Err_String := 'Device write fault' ;
	30 : Err_String := 'Device read fault' ;
	31 : Err_String := 'General failure' ;
	32 : Err_String := 'Sharing violation' ;
	33 : Err_String := 'Lock violation' ;
	34 : Err_String := 'Invalid disk change' ;
	35 : Err_String := 'FCB unavailable' ;
	36 : Err_String := 'Sharing buffer exceeded' ;
	38 : Err_String := 'Unable to complete file operation' ;
	50 : Err_String := 'Unsupported network request' ;
	51 : Err_String := 'Remote machine not listening' ;
	52 : Err_String := 'Duplicate name on network' ;
	53 : Err_String := 'Network name not found' ;
	54 : Err_String := 'Network busy' ;
	55 : Err_String := 'Device no longer exists on network' ;
	56 : Err_String := 'NetBIOS command limit exceeded' ;
	57 : Err_String := 'Error in network adapter hardware' ;
	58 : Err_String := 'Incorrect response from network' ;
	59 : Err_String := 'Unexpected network error' ;
	60 : Err_String := 'Remote adapter incompatible' ;
	61 : Err_String := 'Print queue full' ;
	62 : Err_String := 'Queue not full' ;
	63 : Err_String := 'Not enough room for print file' ;
	64 : Err_String := 'Network name deleted' ;
	65 : Err_String := 'Access denied' ;
	66 : Err_String := 'Incorrect network device type' ;
	67 : Err_String := 'Network name not found' ;
	68 : Err_String := 'Network name limit exceeded' ;
	69 : Err_String := 'netBIOS session limit exceeded' ;
	70 : Err_String := 'Temporary pause' ;
	71 : Err_String := 'Network request not accepted' ;
	72 : Err_String := 'Print or disk redirection paused' ;
	80 : Err_String := 'File already exists' ;
	81 : Err_String := 'Duplicate FCB' ;
	82 : Err_String := 'Cannot make directory' ;
	83 : Err_String := 'Fail on Int 24H (critical error)' ;
	84 : Err_String := 'Out of structures' ;
	85 : Err_String := 'Already assigned' ;
	86 : Err_String := 'Invalid password' ;
	87 : Err_String := 'Invalid parameter' ;
	88 : Err_String := 'Net write fault' ;
	89 : Err_String := 'No process slots available' ;
	90 : Err_String := 'Not frozen' ;
	91 : Err_String := 'Timer service table overflow' ;
	92 : Err_String := 'Timer service table duplicate' ;
	93 : Err_String := 'There were no items to operate upon' ;
	95 : Err_String := 'Interrupted system call' ;
        100 : Err_String := 'Too many semaphores' ;
        101 : Err_String := 'Exclusive semaphore already owned' ;
        102 : Err_String := 'Semaphore is set' ;
        103 : Err_String := 'Too many semaphore requests' ;
        104 : Err_String := 'Invalid at interrupt time' ;
        105 : Err_String := 'WaitSem found that Semaphore owner died' ;
        106 : Err_String := 'Too many processes have this semaphore' ;
	107 : Err_String := 'Insert disk B into drive A' ;
	108 : Err_String := 'Drive locked by another process' ;
	109 : Err_String := 'Write on pipe with no reader' ;
	110 : Err_String := 'Open/create failed' ;
	111 : Err_String := 'Buffer passed to system call' ;
	112 : Err_String := 'Not enough space on the disk' ;
	113 : Err_String := 'Cannot allocate another search' ;
	114 : Err_String := 'Target handle in DOSDUPHANDLE' ;
	115 : Err_String := 'Bad user virtual address' ;
	116 : Err_String := 'VIOKBD Request' ;
	117 : Err_String := 'Invalid category for DEVIOCTL' ;
	118 : Err_String := 'Invalid value for Verify switch' ;
	119 : Err_String := 'Bad driver level for DOSDEVIOCTL ' ;
	120 : Err_String := 'Call not implemented' ;
	121 : Err_String := 'Semaphore timeout' ;
	122 : Err_String := 'Insufficient buffer' ;
	123 : Err_String := 'Illegal character or malformed name' ;
	124 : Err_String := 'Invalid level' ;
	125 : Err_String := 'No volume label' ;
	126 : Err_String := 'Module not found' ;
	127 : Err_String := 'Process not found' ;
	128 : Err_String := 'CWait found no children' ;
	129 : Err_String := 'CWait child not complete' ;
	130 : Err_String := 'Invalid handle operation' ;
	131 : Err_String := 'Negative seek' ;
	132 : Err_String := 'Seek error on device' ;
	133 : Err_String := 'Is join target' ;
	134 : Err_String := 'Is joined' ;
	135 : Err_String := 'Is substituted' ;
	136 : Err_String := 'Not joined' ;
	137 : Err_String := 'Not substituted' ;
	138 : Err_String := 'Join to join' ;
	139 : Err_String := 'Substitute to Substitute' ;
	140 : Err_String := 'Join to Substitute' ;
	141 : Err_String := 'Substitute to join' ;
	142 : Err_String := 'Drive busy' ;
	143 : Err_String := 'Same drive' ;
	144 : Err_String := 'Directory not root' ;
	145 : Err_String := 'Directory not empty' ;
	146 : Err_String := 'Is Substitute path' ;
	147 : Err_String := 'Is join path' ;
	148 : Err_String := 'Path busy' ;
	149 : Err_String := 'Is Substitute target' ;
	150 : Err_String := 'System trace error' ;
	151 : Err_String := 'DOSMUXSEMWAIT error' ;
	152 : Err_String := 'Too many MUXWAITERS ' ;
	153 : Err_String := 'Invalid list format' ;
	154 : Err_String := 'Label too long' ;
	155 : Err_String := 'Too many TCBs' ;
	156 : Err_String := 'Signal refused' ;
	157 : Err_String := 'Discarded' ;
	158 : Err_String := 'Not locked' ;
	159 : Err_String := 'Bad ThreadID address' ;
	160 : Err_String := 'Bad arguments' ;
	161 : Err_String := 'Bad pathname' ;
        162 : Err_String := 'Signal pending' ;
	163 : Err_String := 'Uncetain media' ;
	164 : Err_String := 'Max threads reached' ;
	165 : Err_String := 'Monitors not supported' ;
	180 : Err_String := 'Invalid segment number' ;
	181 : Err_String := 'Invalid call gate' ;
	182 : Err_String := 'Invalid ordinal' ;
	183 : Err_String := 'Already exists' ;
	184 : Err_String := 'No child process' ;
	185 : Err_String := 'Child alive nowait' ;
	186 : Err_String := 'Invalid flag number' ;
	187 : Err_String := 'Semaphore not found' ;
	188 : Err_String := 'Invalid starting code segment' ;
	189 : Err_String := 'Invalid stack segment' ;
	190 : Err_String := 'Invalid module type' ;
	191 : Err_String := 'Invalid EXE signature' ;
	192 : Err_String := 'EXE marked invalid' ;
	193 : Err_String := 'Bad EXE format' ;
	194 : Err_String := 'Iterated data exceeds 64K' ;
	195 : Err_String := 'Invalid MINALLOCSIZE' ;
	196 : Err_String := 'DYNLINK from invalid ring' ;
	197 : Err_String := 'IOPL not enabled' ;
	198 : Err_String := 'Invalid SEGDPL' ;
	199 : Err_String := 'AUTODATASEG exceeds 64K' ;
        200 : Err_String := 'Ring2 segment must be movable' ;
        201 : Err_String := 'Relocation chain needs SEGLIM' ;
        202 : Err_String := 'Infinite loop in relocation chain' ;
        203 : Err_String := 'Environment variable not found' ;
        204 : Err_String := 'Not current country' ;
        205 : Err_String := 'No signal sent' ;
        206 : Err_String := 'Filename exceeded range' ;
        207 : Err_String := 'Ring2 stack in use' ;
        208 : Err_String := 'Meta expansion too long' ;
        209 : Err_String := 'Invalid signal number' ;
        210 : Err_String := 'Thread 1 inactive' ;
        211 : Err_String := 'Information not available' ;
        212 : Err_String := 'Locked' ;
        213 : Err_String := 'Bad Dynalink' ;
        214 : Err_String := 'Too many modules' ;
        215 : Err_String := 'Nesting not allowed' ;
	216 : Err_String := 'General Protection fault' ;

        { RAS error codes... }
        600 : Err_String := 'An operation is pending.' ;
        601 : Err_String := 'The port handle is invalid.' ;
        602 : Err_String := 'The port is already open.' ;
        603 : Err_String := 'Caller''s buffer is too small.' ;
        604 : Err_String := 'Wrong information specified.' ;
        605 : Err_String := 'Cannot set port information.' ;
        606 : Err_String := 'The port is not connected.' ;
        607 : Err_String := 'The event is invalid.' ;
        608 : Err_String := 'The device does not exist.' ;
        609 : Err_String := 'The device type does not exist.' ;
        610 : Err_String := 'The buffer is invalid.' ;
        611 : Err_String := 'The route is not available.' ;
        612 : Err_String := 'The route is not allocated.' ;
        613 : Err_String := 'Invalid compression specified.' ;
        614 : Err_String := 'Out of buffers.' ;
        615 : Err_String := 'The port was not found.' ;
        616 : Err_String := 'An asynchronous request is pending.' ;
        617 : Err_String := 'The port or device is already disconnecting.' ;
        618 : Err_String := 'The port is not open.' ;
        619 : Err_String := 'The port is disconnected.' ;
        620 : Err_String := 'There are no endpoints.' ;
        621 : Err_String := 'Cannot open the phone book file.' ;
        622 : Err_String := 'Cannot load the phone book file.' ;
        623 : Err_String := 'Cannot find the phone book entry.' ;
        624 : Err_String := 'Cannot write the phone book file.' ;
        625 : Err_String := 'Invalid information found in the phone book.' ;
        626 : Err_String := 'Cannot load a string.' ;
        627 : Err_String := 'Cannot find key.' ;
        628 : Err_String := 'The port was disconnected.' ;
        629 : Err_String := 'The port was disconnected by the remote machine.' ;
        630 : Err_String := 'The port was disconnected due to hardware failure.' ;
        631 : Err_String := 'The port was disconnected by the user.' ;
        632 : Err_String := 'The structure size is incorrect.' ;
        633 : Err_String := 'The port is already in use or is not configured for Remote Access dialout.' ;
        634 : Err_String := 'Cannot register your computer on the remote network.' ;
        635 : Err_String := 'Unknown error.' ;
        636 : Err_String := 'The wrong device is attached to the port.' ;
        637 : Err_String := 'The string could not be converted.' ;
        638 : Err_String := 'The request has timed out.' ;
        639 : Err_String := 'No asynchronous net available.' ;
        640 : Err_String := 'A NetBIOS error has occurred.' ;
        641 : Err_String := 'The server cannot allocate NetBIOS resources needed to support the client.' ;
        642 : Err_String := 'One of your NetBIOS names is already registered on the remote network.' ;
        643 : Err_String := 'A network adapter at the server failed.' ;
        644 : Err_String := 'You will not receive network message popups.' ;
        645 : Err_String := 'Internal authentication error.' ;
        646 : Err_String := 'The account is not permitted to log on at this time of day.' ;
        647 : Err_String := 'The account is disabled.' ;
        648 : Err_String := 'The password has expired.' ;
        649 : Err_String := 'The account does not have Remote Access permission.' ;
        650 : Err_String := 'The Remote Access server is not responding.' ;
        651 : Err_String := 'Your modem (or other connecting device) has reported an error.' ;
        652 : Err_String := 'Unrecognized response from the device.' ;
        653 : Err_String := 'A macro required by the device was not found in the device .INF file section.' ;
        654 : Err_String := 'A command or response in the device .INF file section refers to an undefined macro' ;
        655 : Err_String := 'The <message> macro was not found in the device .INF file section.' ;
        656 : Err_String := 'The <defaultoff> macro in the device .INF file section contains an undefined macro' ;
        657 : Err_String := 'The device .INF file could not be opened.' ;
        658 : Err_String := 'The device name in the device .INF or media .INI file is too long.' ;
        659 : Err_String := 'The media .INI file refers to an unknown device name.' ;
        660 : Err_String := 'The device .INF file contains no responses for the command.' ;
        661 : Err_String := 'The device .INF file is missing a command.' ;
        662 : Err_String := 'Attempted to set a macro not listed in device .INF file section.' ;
        663 : Err_String := 'The media .INI file refers to an unknown device type.' ;
        664 : Err_String := 'Cannot allocate memory.' ;
        665 : Err_String := 'The port is not configured for Remote Access.' ;
        666 : Err_String := 'Your modem (or other connecting device) is not functioning.' ;
        667 : Err_String := 'Cannot read the media .INI file.' ;
        668 : Err_String := 'The connection dropped.' ;
        669 : Err_String := 'The usage parameter in the media .INI file is invalid.' ;
        670 : Err_String := 'Cannot read the section name from the media .INI file.' ;
        671 : Err_String := 'Cannot read the device type from the media .INI file.' ;
        672 : Err_String := 'Cannot read the device name from the media .INI file.' ;
        673 : Err_String := 'Cannot read the usage from the media .INI file.' ;
        674 : Err_String := 'Cannot read the maximum connection BPS rate from the media .INI file.' ;
        675 : Err_String := 'Cannot read the maximum carrier BPS rate from the media .INI file.' ;
        676 : Err_String := 'The line is busy.' ;
        677 : Err_String := 'A person answered instead of a modem.' ;
        678 : Err_String := 'There is no answer.' ;
        679 : Err_String := 'Cannot detect carrier.' ;
        680 : Err_String := 'There is no dial tone.' ;
        681 : Err_String := 'General error reported by device.' ;
        682 : Err_String := 'ERROR WRITING SECTIONNAME' ;
        683 : Err_String := 'ERROR WRITING DEVICETYPE' ;
        684 : Err_String := 'ERROR WRITING DEVICENAME' ;
        685 : Err_String := 'ERROR WRITING MAXCONNECTBPS' ;
        686 : Err_String := 'ERROR WRITING MAXCARRIERBPS' ;
        687 : Err_String := 'ERROR WRITING USAGE' ;
        688 : Err_String := 'ERROR WRITING DEFAULTOFF' ;
        689 : Err_String := 'ERROR READING DEFAULTOFF' ;
        690 : Err_String := 'ERROR EMPTY INI FILE' ;
        691 : Err_String := 'Access denied because username and/or password is invalid on the domain.' ;
        692 : Err_String := 'Hardware failure in port or attached device.' ;
        693 : Err_String := 'ERROR NOT BINARY MACRO' ;
        694 : Err_String := 'ERROR DCB NOT FOUND' ;
        695 : Err_String := 'ERROR STATE MACHINES NOT STARTED' ;
        696 : Err_String := 'ERROR STATE MACHINES ALREADY STARTED' ;
        697 : Err_String := 'ERROR PARTIAL RESPONSE LOOPING' ;
        698 : Err_String := 'A response keyname in the device .INF file is not in the expected format.' ;
        699 : Err_String := 'The device response caused buffer overflow.' ;
        700 : Err_String := 'The expanded command in the device .INF file is too long.' ;
        701 : Err_String := 'The device moved to a BPS rate not supported by the COM driver.' ;
        702 : Err_String := 'Device response received when none expected.' ;
        703 : Err_String := 'ERROR INTERACTIVE MODE' ;
        704 : Err_String := 'ERROR BAD CALLBACK NUMBER' ;
        705 : Err_String := 'ERROR INVALID AUTH STATE' ;
        706 : Err_String := 'ERROR WRITING INITBPS' ;
        707 : Err_String := 'X.25 diagnostic indication.' ;
        708 : Err_String := 'The account has expired.' ;
        709 : Err_String := 'Error changing password on domain.' ;
        710 : Err_String := 'Serial overrun errors were detected while communicating with your modem.' ;
        711 : Err_String := 'RasMan initialization failure. Check the event log.' ;
        712 : Err_String := 'Biplex port is initializing. Wait a few seconds and redial.' ;
        713 : Err_String := 'No active ISDN lines are available.' ;
        714 : Err_String := 'Not enough ISDN channels are available to make the call.' ;
        715 : Err_String := 'Too many errors occurred because of poor phone line quality.' ;
        716 : Err_String := 'The Remote Access IP configuration is unusable.' ;
        717 : Err_String := 'No IP addresses are available in the static pool of Remote Access IP addresses.' ;
        718 : Err_String := 'PPP timeout.' ;
        719 : Err_String := 'PPP terminated by remote machine.' ;
        720 : Err_String := 'No PPP control protocols configured.' ;
        721 : Err_String := 'Remote PPP peer is not responding.' ;
        722 : Err_String := 'The PPP packet is invalid.' ;
        723 : Err_String := 'The phone number, including prefix and suffix, is too long.' ;
        724 : Err_String := 'The IPX protocol cannot dial-out on the port because the computer is an IPX router.' ;
        725 : Err_String := 'The IPX protocol cannot dial-in on the port because the IPX router is not installed..' ;
        726 : Err_String := 'The IPX protocol cannot be used for dial-out on more than one port at a time.' ;
        727 : Err_String := 'Cannot access TCPCFG.DLL.' ;
        728 : Err_String := 'Cannot find an IP adapter bound to Remote Access.' ;
        729 : Err_String := 'SLIP cannot be used unless the IP protocol is installed.' ;
        730 : Err_String := 'Computer registration is not complete.' ;
        731 : Err_String := 'The protocol is not configured.' ;
        732 : Err_String := 'The PPP negotiation is not converging.' ;
        733 : Err_String := 'The PPP control protocol for this network protocol is not available on the server.' ;
        734 : Err_String := 'The PPP link control protocol terminated..' ;
        735 : Err_String := 'The requested address was rejected by the server.' ;
        736 : Err_String := 'The remote computer terminated the control protocol.' ;
        737 : Err_String := 'Loopback detected.' ;
        738 : Err_String := 'The server did not assign an address.' ;
        739 : Err_String := 'The remote server cannot use the Windows NT encrypted password.' ;
        740 : Err_String := 'The TAPI devices configured for Remote Access ' +
		'failed to initialize or were not installed correctly.' ;
        741 : Err_String := 'The local computer does not support encryption.' ;
        742 : Err_String := 'The remote server does not support encryption.' ;
        743 : Err_String := 'The remote server requires encryption.' ;
        744 : Err_String := 'Cannot use the IPX net number assigned by the remote server. Check the event log.' ;
        752 : Err_String := 'A syntax error was encountered while processing a script.' ;
        else
        Begin
            Str( Code, Err_String ) ;
            Err_String := 'Unused error number ' + Err_String ;
        end ;
    end ; { Case Code of }
    DOS_ERT := Err_String ;
end ; { DOS_ERT }


{$IFDEF DELPHI}
{ TDOS_EI methods... }

function tDOS_EI.Get_Facility : longint ;

begin
    Get_Facility := 1 ;
end ;


function tDOS_EI.Get_Name : PChar ;

begin
    Get_Name := DOS_Name ;
end ;


function tDOS_EI.Translate_Error( Code : longint ) : PChar ;

var Err_String : string ;

begin
    Err_String := DOS_ERT( Code ) + #0 ;
    move( Err_String[ 1 ], DOS_Error_Text[ 0 ], length( Err_String ) ) ;
    Translate_Error := @DOS_Error_Text ;
end ; { tDOS_EI.Translate_Error }


procedure tDOS_EI.Done ;

begin
{ Destroyed in this unit. }
end ;


function tDOS_EI.Facility_Version : longint ;

begin
    Facility_Version := 21 ; { DOS 2.1 }
end ;
{$ENDIF}



{ TSDOS_EI methods... }

function tSDOS_EI.Get_Facility : longint ;

begin
    Get_Facility := 1 ;
end ;


function tSDOS_EI.Get_Name : string ;

begin
    Get_Name := strpas( DOS_Name ) ;
end ;


function tSDOS_EI.Translate_Error( Code : longint ) : string ;

var Err_String : string ;

begin
    Err_String := DOS_ERT( Code ) ;
    Translate_Error := Err_String ;
end ; { tSDOS_EI.Translate_Error }


function tSDOS_EI.Facility_Version : longint ;

begin
    Facility_Version := 21 ; { DOS 2.1 }
end ;



const Windows_Name : PChar = 'MS Windows' ;

var Windows_Error_Text : array[ 0..127 ] of char ;

{$IFDEF DELPHI}
{ tWindows_EI methods... }

function tWindows_EI.Get_Facility : longint ;

begin
    Get_Facility := 0 ;
end ;


function tWindows_EI.Get_Name : PChar ;

begin
    Get_Name := Windows_Name ;
end ;


function tWindows_EI.Translate_Error( Code : longint ) : PChar ;

begin
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_IGNORE_INSERTS,
        nil, Code, 0, @Windows_Error_Text, 128, nil ) ;
    Windows_Error_Text[ 127 ] := #0 ;
    Translate_Error := @Windows_Error_Text ;
end ;


procedure tWindows_EI.Done ;

begin
{ Destroyed in this unit. }
end ;


function tWindows_EI.Facility_Version : longint ;

begin
    Facility_Version := 40 ; { MS Windows 95/NT and later }
end ;
{$ENDIF}



{ tSWindows_EI methods... }

function tSWindows_EI.Get_Facility : longint ;

begin
    Get_Facility := 0 ;
end ;


function tSWindows_EI.Get_Name : string ;

begin
    Get_Name := strpas( Windows_Name ) ;
end ;


function tSWindows_EI.Translate_Error( Code : longint ) : string ;

var S : string ;

begin
    str( Code, S ) ;
    S := 'Unknown error code ' + S ;
    Translate_Error := S ;
end ;


function tSWindows_EI.Facility_Version : longint ;

begin
    Facility_Version := 40 ; { MS Windows 95/NT and later }
end ;



{ Internal unit data... }

var Interfaces : TCollection ;
    Facility_IDs : TCollection ;
    Last_Error_Text : array[ 0..254 ] of char ;
    _Last_Error : tUEC ;
    _Last_Error_Level : longint ;
    Master_Error_Operation : integer ;
{$IFDEF DELPHI}
    State : integer = 0 ;
{$ELSE}
    State : integer ;
{$ENDIF}

{ Internal unit utility methods... }

{ Unit API... }

{$IFDEF DELPHI}
function Error_Code( Facility, Code : longint ) : tUEC ;
{$ELSE}
function Error_Code( Facility, Code : longint ) : double ;
{$ENDIF}

var U : tUEC ;

begin
    U.Facility := Facility ;
    U.Code := Code ;
{$IFDEF DELPHI}
    Error_Code := U ;
{$ELSE}
    Error_Code := double( U ) ;
{$ENDIF}
end ;


function Get_Error_Text( UEC : tUEC ; Level : longint ) : PChar ;

var Dummy : integer ;
    EI : PSError_Interface ;
    Num1, S, Txt : string ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Get_Error_Text ) ) then
    begin
        Result := UEH_Get_Error_Text( UEC, Level ) ;
        exit ;
    end ;
{$ENDIF}
    if( State = 0 ) then { Not in a valid state }
    begin
        Dummy := -1 ;
    end else
    begin
        Dummy := Facility_IDs.IndexOf( pointer( UEC.Facility ) ) ;
    end ;
    if( Dummy = -1 ) then
    begin
        str( UEC.Facility, Num1 ) ;
        str( UEC.Code, S ) ;
        Txt := 'Facility ' + num1 + ' - ' + 'Error ' + S ;
    end else
    begin
        EI := Interfaces.At( Dummy ) ;
{$IFDEF DELPHI}
        Txt := strpas( TError_Interface( EI ).Get_Name ) + ' - ' +
            strpas( TError_Interface( EI ).Translate_Error( UEC.Code ) ) ;
        if( TError_Interface( EI ).Version >= 12 ) then { Supports severity }
        begin
            if( Level = -1 ) then
            begin
                Level := TError_Interface( EI ).Severity( UEC.Code ) ;
            end ;
        end ;
{$ELSE}
        Txt := EI^.Get_Name + ' - ' + EI^.Translate_Error( UEC.Code ) ;
        if( EI^.Version >= 12 ) then { Supports severity }
        begin
            if( Level = -1 ) then
            begin
                Level := EI^.Severity( UEC.Code ) ;
            end ;
        end ;
{$ENDIF}
    end ;
    case Level of
        SEL_Success : S := 'Success' ;
        SEL_Notification : S := 'Notification ' ;
        SEL_Warning : S := 'Warning ' ;
        SEL_Error : S := 'Error ' ;
        SEL_Fatal : S := 'Fatal error ' ;
        else S := '' ;
    end ;
    if( length( S ) > 0 ) then
    begin
        str( UEC.Code, Num1 ) ;
        S := S + num1 + ': ' ;
    end ;
    S := S + Txt + #0 ;
    move( S[ 1 ], Last_Error_Text[ 0 ], length( S ) ) ;
    Get_Error_Text := @Last_Error_Text ;
end ; { Get_Error_Text }


function Get_Error_Text_Ex( UEC : tUEC ; Level : longint ;
    var Size, Typ : longint ) : PChar ;

var Dummy : integer ;
    EI : PSError_Interface ;
    P : PChar ;
    Num1, S, Txt : string ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Get_Error_Text_Ex ) ) then
    begin
        Result := UEH_Get_Error_Text_Ex( UEC, Level, Size, Typ ) ;
        exit ;
    end ;
{$ENDIF}
    if( State = 0 ) then { Not in a valid state }
    begin
        Dummy := -1 ;
    end else
    begin
        Dummy := Facility_IDs.IndexOf( pointer( UEC.Facility ) ) ;
    end ;
    if( Dummy > -1 ) then
    begin
        EI := Interfaces.At( Dummy ) ;
{$IFDEF DELPHI}
        if( TError_Interface( EI ).Version >= 12 ) then { Supports extended error translation }
        begin
            Txt := strpas( TError_Interface( EI ).Get_Name ) + ' - ' +
                strpas( TError_Interface( EI ).Translate_Error_Ex( UEC.Code, Size, Typ ) ) ;
            if( Level = -1 ) then
            begin
                Level := TError_Interface( EI ).Severity( UEC.Code ) ;
            end ;
{$ELSE}
        if( EI^.Version >= 12 ) then { Supports extended error translation }
        begin
            Txt := EI^.Get_Name + ' - ' + EI^.Translate_Error_Ex( UEC.Code, Size, Typ ) ;
            if( Level = -1 ) then
            begin
                Level := EI^.Severity( UEC.Code ) ;
            end ;
{$ENDIF}
            case Level of
                SEL_Success : S := 'Success' ;
                SEL_Notification : S := 'Notification ' ;
                SEL_Warning : S := 'Warning ' ;
                SEL_Error : S := 'Error ' ;
                SEL_Fatal : S := 'Fatal error ' ;
                else S := '' ;
            end ;
            if( length( S ) > 0 ) then
            begin
                str( UEC.Code, Num1 ) ;
                S := S + num1 + ': ' ;
            end ;
            S := S + Txt ;
            Size := length( S ) ;
            S := S + #0 ;
            move( S[ 1 ], Last_Error_Text[ 0 ], length( S ) ) ;
            Get_Error_Text_Ex := @Last_Error_Text ;
            exit ;
        end ; { if( EI.Version >= 12 ) }
    end ; { if( Dummy > -1 ) }

    P := Get_Error_Text( UEC, Level ) ;
    if( P = nil ) then
    begin
        Size := 0 ;
    end else
    begin
        Size := strlen( P ) ;
    end ;
    Typ := 0 ; { 7-bit ASCII }
    Get_Error_Text_Ex := P ;
end ; { Get_Error_Text_Ex }


function Last_Error : PChar ;

begin
    Last_Error := Get_Error_Text( _Last_Error, _Last_Error_Level ) ;
end ;


procedure Log_Error( UEC : tUEC ; const T : string ; Level : longint ) ;

var Txt : string ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Log_Error ) ) then
    begin
        CC_Log( T ) ; { Log a message }
        UEH_Log_Error( UEC, PChar( T ), Level ) ;
        exit ;
    end ;
{$ENDIF}
    if( Master_Error_Operation = TMEO_Ignore ) then
    begin
	exit ; { Ignore errors }
    end ;
    _Last_Error := UEC ;
    _Last_Error_Level := Level ;
    if( Master_Error_Operation = TMEO_Show ) then
    begin
        Txt := strpas( Get_Error_Text( UEC, Level ) ) ;
        CC_Log( Txt + T ) ; { Log a message }
{$IFDEF DELPHI}
        if( Level = SEL_Fatal ) then
        begin
            raise Exception.Create( Txt + T ) ;
        end ;
	ShowMessage( Txt + T ) ;
{$ELSE}
	{$I-}
        writeln( Txt, T ) ;
        {$I+}
        Clear_IOResult ;
        if( Level = SEL_Fatal ) then
        begin
            halt ;
        end ;
{$ENDIF}
    end ;
end ; { Log_Error }


{ Registers the error interface, returning True if succesful; false otherwise. }
function Register_Error_Interface( EI : PSError_Interface ) : longbool ;

var Dummy, Index : integer ;
    Old_EI : PSError_Interface ;
    Facility_Version, Old_Facility_Version, Version, Old_Version : longint ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Register_Error_Interface ) ) then
    begin
        Result := UEH_Register_Error_Interface( EI ) ;
        exit ;
    end ;
{$ENDIF}
    if( EI = nil ) then
    begin
        Register_Error_Interface := False ;
        exit ;
    end ;
    if( State = 0 ) then { Not in a valid state }
    begin
        Register_Error_Interface := False ;
        exit ;
    end ;
    Register_Error_Interface := True ;
{$IFDEF DELPHI}
    Dummy := TError_Interface( EI ).Get_Facility ;
    Version := TError_Interface( EI ).Version ;
    Facility_Version := TError_Interface( EI ).Facility_Version ;
{$ELSE}
    Dummy := EI^.Get_Facility ;
    Version := EI^.Version ;
    Facility_Version := EI^.Facility_Version ;
{$ENDIF}
    Index := Facility_IDs.IndexOf( pointer( Dummy ) ) ;
    if( Index <> -1 ) then { This facility was already registered }
    begin
        Old_EI := PSError_Interface( Interfaces.At( Index ) ) ;
{$IFDEF DELPHI}
	Old_Version := TError_Interface( Old_EI ).Version ;
	Old_Facility_Version := TError_Interface( Old_EI ).Facility_Version ;
{$ELSE}
	Old_Version := Old_EI^.Version ;
	Old_Facility_Version := Old_EI^.Facility_Version ;
{$ENDIF}
        if(
            ( Old_Facility_Version < Facility_Version )
            or
            (
              ( Old_Facility_Version = Facility_Version )
              and
              ( Old_Version < Version )
            )
          ) then
        begin
            { We get here if the previous registered facility was an older
              version of the facility, or the same facility version, but
              supporting an older UEH specification. }
{$IFDEF DELPHI}
            Old_EI^.Done ; { Get rid of the old one }
{$ELSE}
            Old_EI^.Done ; { Get rid of the old one }
{$ENDIF}
            Unregister_Error_Interface( Dummy ) ; { Remove it from our list }
        end else
        begin
            Register_Error_Interface := False ;
            exit ;
        end ;
    end ;
    Facility_IDs.Insert( pointer( Dummy ) ) ;
    Interfaces.Insert( EI ) ;
end ; { Register_Error_Interface }


procedure Set_Master_Error_Operation( MEO : longint ) ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Set_Master_Error_Operation ) ) then
    begin
        UEH_Set_Master_Error_Operation( MEO ) ;
        exit ;
    end ;
{$ENDIF}
    Master_Error_Operation := MEO ;
end ;


function Facility_Count : longint ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Facility_Count ) ) then
    begin
        Result := UEH_Facility_Count ;
        exit ;
    end ;
{$ENDIF}
    if( State = 0 ) then { Not in a valid state }
    begin
        Facility_Count := 0 ;
    end else
    begin
        Facility_Count := Facility_IDs.Count ;
    end ;
end ;


function Facility_From_Index( Index : longint ) : longint ;

begin
    if( ( Index < 0 ) or ( Index >= Facility_Count ) ) then
    begin
        Facility_From_Index := -1 ;
        exit ;
    end ;
{$IFDEF USE_UEH}
    if( assigned( UEH_Facility_From_Index ) ) then
    begin
        Result := UEH_Facility_From_Index( Index ) ;
        exit ;
    end ;
{$ENDIF}
    if( ( State = 0 ) or ( Facility_IDs.Count <= Index ) or ( Index < 0 ) ) then { Not in a valid state, or out of range }
    begin
        Facility_From_Index := -1 ;
    end else
    begin
        Facility_From_Index := longint( Facility_IDs.At( Index ) ) ;
    end ;
end ;


{$IFNDEF DELPHI}
var Temp_Facility_Name_From_Index : string[ 255 ] ;
{$ENDIF}

function Facility_Name_From_Index( Index : longint ) : PChar ;

var EI : PSError_Interface ;

begin
    if( ( Index < 0 ) or ( Index >= Facility_Count ) ) then
    begin
        Facility_Name_From_Index := nil ;
        exit ;
    end ;
{$IFDEF USE_UEH}
    if( assigned( UEH_Facility_Name_From_Index ) ) then
    begin
        Result := UEH_Facility_Name_From_Index( Index ) ;
        exit ;
    end ;
{$ENDIF}
    if( ( State = 0 ) or ( Facility_IDs.Count <= Index ) or ( Index < 0 ) ) then { Not in a valid state, or out of range }
    begin
        Facility_Name_From_Index := nil ;
    end else
    begin
        EI := Interfaces.At( Index ) ;
{$IFDEF DELPHI}
        Facility_Name_From_Index := TError_Interface( EI ).Get_Name ;
{$ELSE}
        Temp_Facility_Name_From_Index := EI^.Get_Name ;
        Facility_Name_From_Index := @Temp_Facility_Name_From_Index[ 1 ] ;
{$ENDIF}
    end ;
end ;


function Unregister_Error_Interface( Facility : longint ) : longbool ;

var Dummy : integer ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Unregister_Error_Interface ) ) then
    begin
        Result := UEH_Unregister_Error_Interface( Facility ) ;
        exit ;
    end ;
{$ENDIF}
    { The following check is here because, for some unknown reason on Windows
      2000, the DLL finalization code is called prior to the shut down of other
      DLLs which reference this one.  Our finalization clears the valid state
      and releases memory, then other DLLs may call this one to unregister their
      error interfaces, which would result in an error if we did our normal
      processing.  So we ignore calls when our state is invalid. }
    if( State = 0 ) then { Not in a valid state }
    begin
        Unregister_Error_Interface := False ;
        exit ;
    end ;
    Unregister_Error_Interface := True ;
    Dummy := Facility_IDs.IndexOf( pointer( Facility ) ) ;
    if( Dummy = -1 ) then
    begin
        Unregister_Error_Interface := False ;
        exit ;
    end ;
    Facility_IDs.AtDelete( Dummy ) ;
    Interfaces.AtDelete( Dummy ) ;
end ; { Unregister_Error_Interface }


function Version : longint ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Version ) ) then
    begin
        Result := UEH_Version ;
        exit ;
    end ;
{$ENDIF}
    Version := 12 ; { Compatible with UEH specification 1.2 }
end ;


function Severity( UEC : TUEC ) : longint ;

var Dummy : integer ;
    EI : PSError_Interface ;

begin
{$IFDEF USE_UEH}
    if( assigned( UEH_Severity ) ) then
    begin
        Result := UEH_Severity( UEC ) ;
        exit ;
    end ;
{$ENDIF}
    if( State = 0 ) then { Not in a valid state }
    begin
        Severity := UEH_Error ;
        exit ;
    end ;
    Dummy := Facility_IDs.IndexOf( pointer( UEC.Facility ) ) ;
    if( Dummy = -1 ) then
    begin
        Severity := UEH_Error ;
        exit ;
    end ;
    EI := Interfaces.At( Dummy ) ;
{$IFDEF DELPHI}
    if( TError_Interface( EI ).Version >= 12 ) then { Supports severity }
    begin
        Severity := TError_Interface( EI ).Severity( UEC.Code ) ;
        exit ;
    end ;
{$ELSE}
    if( EI^.Version >= 12 ) then { Supports severity }
    begin
        Severity := EI^.Severity( UEC.Code ) ;
        exit ;
    end ;
{$ENDIF}
    Severity := UEH_Error ;
end ; { Severity }


{ Unit initialization/finalization code... }

{$IFDEF DELPHI}
var DOS_EI : TDOS_EI ;
    Windows_EI : TWindows_EI ;
{$ELSE}
var DOS_EI : pSDOS_EI ;
    Windows_EI : pSWindows_EI ;
{$ENDIF}

procedure Initialize_Unit ;

begin
    CC_Log( 'Entering ErrMastr.Initialize_Unit' ) ;
    { Setup... }
    _Last_Error_Level := 0 ;
    Master_Error_Operation := TMEO_Silent ;
{$IFNDEF DELPHI}
    State := 0 ;
{$ENDIF}
    if( State = 1 ) then { Already initialized }
    begin
        exit ;
    end ;

    { Setup unit variables... }
    Interfaces.Init( 0, 4 ) ;
    Facility_IDs.Init( 0, 4 ) ;
    _Last_Error.Facility := 0 ;
    _Last_Error.Code := 0 ;

{$IFDEF USE_UEH}
    UEH32_Handle := LoadLibrary( 'UEH32.DLL' ) ;
    if( UEH32_Handle <> 0 ) then { Found library }
    begin
        UEH_Error_Code := GetProcAddress( UEH32_Handle, 'UEH_Error_Code' ) ;
        UEH_Get_Error_Text := GetProcAddress( UEH32_Handle, 'UEH_Get_Error_Text' ) ;
        UEH_Get_Error_Text_Ex := GetProcAddress( UEH32_Handle, 'UEH_Get_Error_Text_Ex' ) ;
        UEH_Last_Error := GetProcAddress( UEH32_Handle, 'UEH_Last_Error' ) ;
        UEH_Log_Error := GetProcAddress( UEH32_Handle, 'UEH_Log_Error' ) ;
        UEH_Register_Error_Interface := GetProcAddress( UEH32_Handle, 'UEH_Register_Error_Interface' ) ;
        UEH_Set_Master_Error_Operation := GetProcAddress( UEH32_Handle, 'UEH_Set_Master_Error_Operation' ) ;
        UEH_Facility_Count := GetProcAddress( UEH32_Handle, 'UEH_Facility_Count' ) ;
        UEH_Facility_From_Index := GetProcAddress( UEH32_Handle, 'UEH_Facility_From_Index' ) ;
        UEH_Facility_Name_From_Index := GetProcAddress( UEH32_Handle, 'UEH_Facility_Name_From_Index' ) ;
        UEH_Unregister_Error_Interface := GetProcAddress( UEH32_Handle, 'UEH_Unregister_Error_Interface' ) ;
        UEH_Version := GetProcAddress( UEH32_Handle, 'UEH_Version' ) ;
        UEH_Severity := GetProcAddress( UEH32_Handle, 'UEH_Severity' ) ;
        exit ;
    end ;
{$ENDIF}

    { Create default error interfaces for DOS and Windows.. }
    State := 1 ;
{$IFDEF DELPHI}
    DOS_EI := TDOS_EI.Create ;
    Windows_EI := TWindows_EI.Create ;
    Register_Error_Interface( PSError_Interface( DOS_EI ) ) ;
    Register_Error_Interface( PSError_Interface( Windows_EI ) ) ;
{$ELSE}
    new( DOS_EI, Init ) ;
    new( Windows_EI, Init ) ;
    Register_Error_Interface( DOS_EI ) ;
    Register_Error_Interface( Windows_EI ) ;
{$ENDIF}
    CC_Log( 'Exiting ErrMastr.Initialize_Unit' ) ;
end ; { Initialize_Unit }


procedure Terminate_Unit ;

{$IFNDEF USE_UEH}
  {$IFNDEF UEH}
var Dummy : integer ;
  {$ENDIF}
{$ENDIF}

begin
    CC_Log( 'Entering ErrMastr.Terminate_Unit' ) ;
{$IFDEF USE_UEH}
    if( UEH32_Handle <> 0 ) then
    begin
        FreeLibrary( UEH32_Handle ) ;
    end ;
{$ELSE}
  {$IFDEF DELPHI}
    Unregister_Error_Interface( DOS_EI.Get_Facility ) ;
    Unregister_Error_Interface( Windows_EI.Get_Facility ) ;
    DOS_EI.Free ;
    DOS_EI := nil ;
    Windows_EI.Free ;
    Windows_EI := nil ;
    State := 0 ;
  {$ELSE}
    Unregister_Error_Interface( DOS_EI^.Get_Facility ) ;
    Unregister_Error_Interface( Windows_EI^.Get_Facility ) ;
    DOS_EI^.Done ;
    Windows_EI^.Done ;
  {$ENDIF}
  {$IFNDEF UEH}
    for Dummy := 0 to Interfaces.Count - 1 do
    begin
        if( Interfaces.At( Dummy ) <> nil ) then
        begin
    {$IFDEF DELPHI}
            TError_Interface( Interfaces.At( Dummy ) ).Done ;
    {$ELSE}
            PSError_Interface( Interfaces.At( Dummy ) )^.Done ;
    {$ENDIF}
        end ;
    end ;
  {$ENDIF}
    Interfaces.Done ;
    Facility_IDs.Done ;
{$ENDIF}
    CC_Log( 'Exiting ErrMastr.Terminate_Unit' ) ;
end ; { Terminate_Unit }

{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}

    Initialize_Unit ;

{$IFDEF DELPHI}
finalization
    Terminate_Unit ;
{$ENDIF}
end.




