{
        Program Name : FileDB
        Package Name : LB
        Purpose      : File database
        Institution  : Conroy & Conroy Co.
        Date Written : 06-Jul-1996
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This module defines a file database.  That is, a database
	implemented via the DOS file system (not a database of files).
	Each file is considered a record within the database.  The record
	size is the file size (plus n bytes).  The key is the path and
	filename.  The key is considered the first n bytes of the record,
	where n is passed to the constructor.

	  Note that file databases cannot have duplicate keys.
}

unit FileDB ;

interface

{$I EDEFINES.INC}

uses
{$IFDEF DELPHI}
     Windows,
{$ELSE}
     WinDOS,
{$ENDIF}
     Database,
     _Types ; { Small_String }

const DBErr_Success = 0 ; { Last operation was successful }
      DBErr_DatabaseNotOpen = -1 ; { Database is not open }
      DBErr_RecordNotFound = -2 ; { Specified key doesn't exist in database }
      DBErr_DuplicateKey = -3 ; { Attempt to create record with existing key }
      DBErr_InvalidRecordSize = -4 ; { Attempt to create a record too small }
      DBErr_InvalidKey = -5 ; { Attempt to use a null key }
      DBErr_RecordOverflow = -6 ; { Attempt to write past end of record }
      DBErr_NoCurrentRecord = -7 ; { Attempted operation on current record when none }
      DBErr_KeyNotChangable = -8 ; { Attempted to change key part of record }

type TComparison = ( COMPARE_EQ, COMPARE_NE, COMPARE_GE, COMPARE_GT, COMPARE_LE, COMPARE_LT ) ;

type PFile_DataBase = ^TFile_Database ;
     TFile_DataBase = object( TISAM_Database )
			private { Instance data }
			     Current_File : file ;
			     Current_Record : Small_String ;
                             Key_Size : integer ; { Size of filename key }
			     Cache_Pointer : PChar ; { Pointer to cached record }
			     Cache_Size : integer ; { Size of cached record }
			     Root : Small_String ; { Root directory of database }

                        public { Constructors and destructors }
			     constructor Create( Dir : string ; Size : integer ) ;

			     constructor Init( Dir : string ; Size : integer ) ;
			     { Initialize database }

{$IFDEF DELPHI}
			     destructor Destroy ; virtual ;
			     { Destruct object }

{$ELSE}
			     destructor Destroy ;
			     { Destruct object }

{$ENDIF}
			     destructor Done ; virtual ;
			     { Destruct object }

			private { Internal utility functions }
			    procedure Reset_Pointer ;

			public { Public API... }
			     procedure Delete( Index : longint ;
                                 const Key : string ) ; virtual ;
			     { Delete the specified record }

			     function Find_Key( Index : longint ;
                                 const Key : string ;
                                 Typ : TComparison ) : string ;
			     { Find key relative to current key }

			     procedure Flush ; virtual ;
			     { Flushes current record buffer, if any. }

			     function Get( Index : longint ;
                                 const Key : string ) : PDatabase_Record ;
                                 virtual ;
			     { Get data from specified offset in specified
			       record, filling specified buffer of specified
			       size. }

			     function Get_GE( Index : longint ;
                                 const Key : string ) : PDatabase_Record ; virtual ;
			     { Get first record whose key is greater than or
			       equal to the specified key. }

			     function Get_LE( Index : longint ;
                                 const Key : string ) : PDatabase_Record ; virtual ;
			     { Get last record whose key is less than or equal
			       to the specified key. }

			     function Next( Index : longint ;
                                   Rec : PDatabase_Record ) : PDatabase_Record ; virtual ;
			     { Get next record from current one, filling
			       specified buffer of specified size. }

			     function Previous( Index : longint ;
                                   Rec : PDatabase_Record  ) : PDatabase_Record ; virtual ;
			     { Get prior record from current one, filling
			       specified buffer of specified size. }

			     procedure Lock( Locks : TLock_Type ) ; virtual ;
			     { Locks current record }

			     procedure Put( Key : string ; Len, Size : integer ;
				var Buffer ) ; virtual ;
			     { Create a new record with the specified key and
			       length.  Write Size bytes from specified buffer
			       to start of new record. }

			     function Record_Size( Index : longint ;
                                 const Key : string ) : longint ; virtual ;
			     { Returns size of specified record, in bytes. }

			     function Record_Pointer( Index : longint ;
                                 const Key : string ) : pointer ; virtual ;
			     { Returns a pointer to the specified record.  If
			       the record does not exist, nil is returned. }

			     procedure Unlock( Locks : TLock_Type ) ; virtual ;
			     { Unlocks current record }

			     procedure Update( Index : longint ;
                                 const Key : string ;
 				 Offset, Size : integer ; var Buffer ) ;
				 virtual ;
			     { Updates specified record at specified offset for
			       specified bytes from specified buffer. }

			     procedure Zero ; virtual ;
			     { Deletes entire database }
	      end ;

implementation

uses Compatib,
     Files, { TFind_Record }
     _UEHDefs ; { TUEC }

{ TFile_Database methods... }

{ Internal utility functions... }

procedure TFile_Database.Reset_Pointer ;

begin
    if( Cache_Pointer <> nil ) then
    begin
	freemem( Cache_Pointer, Cache_Size ) ;
    end ;
    Cache_Pointer := nil ;
end ;


{ Constructors and destructors... }

constructor TFile_Database.Create( Dir : string ; Size : integer ) ;

begin
    inherited Init ;
    
    Init( Dir, Size ) ;
end ;


constructor TFile_Database.Init( Dir : string ; Size : integer ) ;
{ Initialize database }

begin
    inherited Init ;

    if( Dir[ length( Dir ) ] <> '\' ) then
    begin
	Dir := Dir + '\' ;
    end ;
    Root := Dir ;
    Key_Size := Size ;
    if( Size < 1 ) then
    begin
	writeln( 'Invalid key size' ) ;
	halt ;
    end ;
    Cache_Pointer := nil ;
end ;


destructor TFile_Database.Destroy ;

begin
    Done ;
end ;


destructor TFile_Database.Done ;
{ Destruct object }

begin
    Flush ;
end ;


{ API... }

procedure TFile_Database.Delete( Index : longint ; const Key : string ) ; { Delete the specified record }

var Dummy : integer ;
    K : string ;

begin
    Flush ; { Make sure it is closed (also sets DB_Error to 0) }

    { Delete the file... }
    assign( Current_File, Root + Key ) ;
    {$I-}
    erase( Current_File ) ;
    {$I+}
    Dummy := IOResult ;
    Set_Error( 0, Dummy ) ;
    if( ( Dummy = 2 ) or ( Dummy = 3 ) ) then
    begin
	Set_Error( 0, DBErr_RecordNotFound ) ;
	exit ;
    end ;

    { Delete the path... }
    Dummy := length( Key ) ;
    while( ( Dummy > 0 ) and ( Key[ Dummy ] <> '\' ) ) do
    begin
	dec( Dummy ) ;
    end ;
    K := copy( Key, 1, Dummy - 1 ) ; { Strip filename }
    Dummy := 0 ;
    while( ( Dummy = 0 ) and ( length( K ) > 0 ) ) do
    begin
	{$I-}
	Rmdir( Root + K ) ;
	{$I+}
	Dummy := IOResult ;
	if( Dummy = 0 ) then { If successfully deleted the directory }
	begin
	    Dummy := length( K ) ;
	    while( ( K[ Dummy ] <> '\' ) and ( Dummy > 0 ) ) do
	    begin
		dec( Dummy ) ; { Find last directory in key }
	    end ;
	    K := copy( K, 1, Dummy - 1 ) ; { Remove last directory }
	    Dummy := 0 ;
	end ;
    end ;
end ; { TFile_Database.Delete }


{ Find key with specified relation to current key }
function TFile_Database.Find_Key( Index : longint ; const Key : string ;
    Typ : TComparison ) : string ;

var Best_Match : string ;
    Dummy : integer ;
    Find_Record : TFind_Record ;
    S : string ;

begin
    { Find end of last directory in key... }
    Dummy := length( Key ) ;
    while( ( Dummy > 0 ) and ( Key[ Dummy ] <> '\' ) ) do
    begin
	dec( Dummy ) ;
    end ;

    { Attempt to find a file GE Key in the current directory... }
    Best_Match := '' ;
    Initialize_Find_Record( Find_Record ) ;
    S := Lookup( Root + copy( Key, 1, Dummy ) + '...\*.*', Find_Record ) ;
    while( length( S ) > 0 ) do
    begin
	if(
	    ( copy( S, length( S ) - 1, 2 ) = '\.' )
	    or
	    ( copy( S, length( S ) - 2, 3 ) = '\..' ) ) then
	begin
	    S := Lookup( '', Find_Record  ) ;
	    continue ;
	end ;
	case Typ of
	    COMPARE_GE:
		if(
		    ( S >= Root + Key )
		    and
		    (
		      ( S < Best_Match )
		      or
		      ( length( Best_Match ) = 0 )
		    )
		  ) then
		begin
		    Best_Match := S ;
		end ;
	    COMPARE_GT:
		if(
		    ( S > Root + Key )
		    and
		    (
		      ( S < Best_Match )
		      or
		      ( length( Best_Match ) = 0 )
		    )
		  ) then
		begin
		    Best_Match := S ;
		end ;
	    COMPARE_LE:
		if(
		    ( S <= Root + Key )
		    and
		    (
		      ( S > Best_Match )
		      or
		      ( length( Best_Match ) = 0 )
		    )
		  ) then
		begin
		    Best_Match := S ;
		end ;
	    COMPARE_LT:
		if(
		    ( S < Root + Key )
		    and
		    (
		      ( S > Best_Match )
		      or
		      ( length( Best_Match ) = 0 )
		    )
		  ) then
		begin
		    Best_Match := S ;
		end ;
	end ;
	S := Lookup( '', Find_Record ) ; { Find next }
    end ;

    { If not found, start backing up through directories... }
    while( ( length( Best_Match ) = 0 ) and ( Dummy > 0 ) ) do
    begin
	dec( Dummy ) ;
        while( ( Dummy > 0 ) and ( Key[ Dummy ] <> '\' ) ) do
	begin
	    dec( Dummy ) ;
	end ;
	S := Lookup( Root + copy( Key, 1, Dummy ) + '...\*.*', Find_Record ) ;
	while( length( S ) > 0 ) do
	begin
	    if(
		( copy( S, length( S ) - 1, 2 ) = '\.' )
		or
		( copy( S, length( S ) - 2, 3 ) = '\..' ) ) then
	    begin
		S := Lookup( '', Find_Record ) ;
		continue ;
	    end ;
	    case Typ of
		COMPARE_GE:
		    if(
			( S >= Root + Key )
			and
			(
			  ( S < Best_Match )
			  or
			  ( length( Best_Match ) = 0 )
			)
		      ) then
		    begin
			Best_Match := S ;
		    end ;
		COMPARE_GT:
		    if(
			( S > Root + Key )
			and
			(
			  ( S < Best_Match )
			  or
			  ( length( Best_Match ) = 0 )
			)
		      ) then
		    begin
			Best_Match := S ;
		    end ;
		COMPARE_LE:
		    if(
			( S <= Root + Key )
			and
			(
			  ( S > Best_Match )
			  or
			  ( length( Best_Match ) = 0 )
			)
		      ) then
		    begin
			Best_Match := S ;
		    end ;
		COMPARE_LT:
		    if(
			( S < Root + Key )
			and
			(
			  ( S > Best_Match )
			  or
			  ( length( Best_Match ) = 0 )
			)
		      ) then
		    begin
			Best_Match := S ;
		    end ;
	    end ;
	    S := Lookup( '', Find_Record ) ; { Find next }
	end ;
    end ;
    Uninitialize_Find_Record( Find_Record ) ;
    Find_Key := copy( Best_Match, length( Root ) + 1, 254 ) ;
end ; { TFile_Database.Find_Key }


procedure TFile_Database.Flush ; { Flushes current record buffer, if any. }

begin
    if( length( Current_Record ) > 0 ) then
    begin
	if( Cache_Pointer <> nil ) then
	begin
	    seek( Current_File, 0 ) ;
	    blockwrite( Current_File, Cache_Pointer[ Key_Size ], Cache_Size - Key_Size ) ;
	end ;
	close( Current_File ) ;
	Current_Record := '' ;
	Reset_Pointer ;
    end ;
    Set_Error( 0, DBErr_Success ) ; { No errors possible }
end ;


function TFile_Database.Get( Index : longint ;
    const Key : string ) : PDatabase_Record ;
{ Get data from specified offset in specified record, filling specified buffer
  of specified size. }

var Dummy : integer ;
(*
    P : PChar ;
    S : string ;
    Start : integer ; { Starting position in buffer }
*)

begin
    Get := nil ;
    if( Key <> Current_Record ) then { Locate record... }
    begin
	Flush ;
	assign( Current_File, Root + Key ) ;
	{$I-}
	reset( Current_File, 1 ) ;
	{$I+}
	Dummy := IOResult ;
        Set_Error( 0, Dummy ) ;
	if( ( Dummy = 2 ) or ( Dummy = 3 ) ) then
	begin
	    Set_Error( 0, DBErr_RecordNotFound ) ;
	    exit ;
	end ;
	Current_Record := Key ;
    end ;
    Reset_Pointer ;

(*
    { Set up to copy record data to buffer... }
    if( Size = 0 ) then { Does not want data }
    begin
	exit ;
    end ;

    Start := 0 ;

    { Handle key field }
    if( Offset < Key_Size ) then { Wants "key" field }
    begin
	Dummy := Size ;
	if( Dummy > Key_Size ) then
	begin
	    Dummy := Key_Size ;
	end ;
	Dummy := Dummy - Offset ;
	S := Current_Record ;
	Size := Size - Dummy ;
	if( Dummy > length( S ) ) then
	begin
	    Dummy := length( S ) ;
	end ;
	move( S[ 1 ], Buffer, Dummy ) ;
	Start := Key_Size - Offset ;
    end ;

    { Move data to buffer... }
    if( Size > filesize( Current_File ) ) then
    begin
	Size := filesize( Current_File ) ;
    end ;
    seek( Current_File, Offset - Key_Size ) ;
    P := PChar( @Buffer ) ;
    {$I-}
    blockread( Current_File, P[ Start ], Size ) ;
    {$I+}
    DB_Error := IOResult ;
*)
end ; { TFile_Database.Get }


function TFile_Database.Get_GE( Index : longint ; const Key : string ) : PDatabase_Record ;
{ Get first record whose key is greater than or equal to the specified key. }

var Best_Match : string ; { Best match found }

begin
    { Assume success }
    Set_Error( 0, 0 ) ;
    Get_GE := nil ;

    Best_Match := Find_Key( Index, Key, COMPARE_GE ) ;

    { Return result, etc. }
    if( length( Best_Match ) = 0 ) then
    begin
	Set_Error( 0, DBErr_RecordNotFound ) ;
    end else
    begin
	Get_GE := Get( Index, Best_Match ) ;
    end ;
end ; { TFile_Database.Get_GE }


function TFile_Database.Get_LE( Index : longint ;
    const Key : string ) : PDatabase_Record ;
{ Get last record whose key is less than or equal to the specified key. }

var Best_Match : string ;

begin
    { Assume success }
    Set_Error( 0, 0 ) ;
    Get_LE := nil ;

    Best_Match := Find_Key( Index, Key, COMPARE_LE ) ;

    { Return result, etc. }
    if( length( Best_Match ) = 0 ) then
    begin
	Set_Error( 0, DBErr_RecordNotFound ) ;
    end else
    begin
	Get_LE := Get( Index, Best_Match ) ;
    end ;
end ;


function TFile_Database.Next( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;
{ Get next record from current one, filling specified buffer of specified size. }

var Best_Match : string ;

begin
    Next := nil ;
    if( length( Current_Record ) = 0 ) then
    begin
	Set_Error( 0, DBErr_NoCurrentRecord ) ;
	exit ;
    end ;

    { Assume success }
    Set_Error( 0, 0 ) ;

    Best_Match := Find_Key( Index, Current_Record, COMPARE_GT ) ;

    { Return result, etc. }
    if( length( Best_Match ) = 0 ) then
    begin
	Set_Error( 0, DBErr_RecordNotFound ) ;
    end else
    begin
	Next := Get( Index, Best_Match ) ;
    end ;
end ;


function TFile_Database.Previous( Index : longint ;
    Rec : PDatabase_Record ) : PDatabase_Record ;
{ Get prior record from current one, filling specified buffer of specified
  size. }

var Best_Match : string ;

begin
    Previous := nil ;
    if( length( Current_Record ) = 0 ) then
    begin
	Set_Error( 0, DBErr_NoCurrentRecord ) ;
	exit ;
    end ;

    { Assume success }
    Set_Error( 0, 0 ) ;

    Best_Match := Find_Key( Index, Current_Record, COMPARE_LT ) ;

    { Return result, etc. }
    if( length( Best_Match ) = 0 ) then
    begin
	Set_Error( 0, DBErr_RecordNotFound ) ;
    end else
    begin
	Previous := Get( Index, Best_Match ) ;
    end ;
end ;


procedure TFile_DataBase.Lock( Locks : TLock_Type ) ;
{ Locks current record. }

begin
    if( length( Current_Record ) = 0 ) then
    begin
	Set_Error( 0, DBErr_NoCurrentRecord ) ;
	exit ;
    end ;
end ;


procedure TFile_Database.Put( Key : string ; Len, Size : integer ;
    var Buffer ) ;
{ Create a new record with the specified key and length.  Write Size bytes from
  specified buffer to start of new record, past the key. }

var Dummy : integer ;
    Buf : array[ 0..256 ] of char ;

begin
    { Verify parameters }
    if( Len < 1 + Key_Size ) then
    begin
	Set_Error( 0, DBErr_InvalidRecordSize ) ;
	exit ;
    end ;
    if( length( Key ) = 0 ) then
    begin
	Set_Error( 0, DBErr_InvalidKey ) ;
	exit ;
    end ;
    if( Size > Len - Key_Size ) then
    begin
	Set_Error( 0, DBErr_RecordOverflow ) ;
	exit ;
    end ;

    Flush ; { Clear existing record and reset status }

    { Make sure no duplicates exist... }
    assign( Current_File, Root + Key ) ;
    {$I-}
    reset( Current_File, 1 ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy = 0 ) then
    begin
	close( Current_File ) ;
	Set_Error( 0, DBErr_DuplicateKey ) ;
	exit ;
    end ;

    { Make sure path exists... }
    Dummy := pos( '\', Root + Key ) ;
    while( Dummy <= length( Root + Key ) ) do
    begin
	if( copy( Root + Key, Dummy, 1 ) = '\' ) then
	begin
	    {$I-}
	    MKDir( copy( Root + Key, 1, Dummy - 1 ) ) ;
	    {$I+}
	    Set_Error( 0, IOResult ) ;
	end ;
	inc( Dummy ) ;
    end ;

    { Create new record... }
    {$I-}
    rewrite( Current_File, 1 ) ;
    {$I+}
    Dummy := IOResult ;
    Set_Error( 0, Dummy ) ;
    if( Dummy <> 0 ) then
    begin
	exit ;
    end ;
    Fillchar( Buf, sizeof( Buf ), 0 ) ;
    Len := Len - Key_Size ; { Subtract key portion }
    while( Len > 0 ) do
    begin
	if( Len > sizeof( Buf ) ) then
	begin
	    blockwrite( Current_File, Buf, sizeof( Buf ) ) ;
	    Len := Len - sizeof( Buf ) ;
	end else
	begin
	    blockwrite( Current_File, Buf, Len ) ;
	    Len := 0 ;
	end ;
    end ;
    Seek( Current_File, 0 ) ;
    blockwrite( Current_File, Buffer, Size ) ;
    Current_Record := Key ;
    Reset_Pointer ;
end ; { TFile_Database.Put }


function TFile_Database.Record_Size( Index : longint ;
    const Key : string ) : longint ;
{ Returns size of specified record, in bytes.  If Key is null, returns size of
  current record. }

var K : string ;
    UEC : TUEC ;

begin
    Set_Error( 0, 0 ) ;
    Record_Size := 0 ;

    if( length( Key ) = 0 ) then
    begin
	if( length( Current_Record ) = 0 ) then
	begin
	    Record_Size := 0 ;
	    Set_Error( 0, DBErr_NoCurrentRecord ) ;
	    exit ;
	end ;
	K := Current_Record ;
    end else
    begin
        K := Key ;
    end ;
    if( K <> Current_Record ) then { Need to get specified record }
    begin
	Get( Index, K ) ;
        Get_last_Error( UEC ) ;
	if( UEC.Code <> DBErr_Success ) then
	begin
	    exit ;
	end ;
    end ;
    Record_Size := Key_Size + filesize( Current_File ) ;
end ;


function TFile_Database.Record_Pointer( Index : longint ;
    const Key : string ) : pointer ;
{ Returns a pointer to the specified record.  If the record does not exist,
  nil is returned. }

var UEC : TUEC ;

begin
    Set_Error( 0, 0 ) ;
    Record_Pointer := nil ;
    
    { Handle already having it }
    if( Cache_Pointer <> nil ) then
    begin
	Record_Pointer := Cache_Pointer ;
	exit ;
    end ;

    Cache_Size := Record_Size( Index, Key ) ;
    Get_Last_Error( UEC ) ;
    if( UEC.Code = DBErr_Success ) then
    begin
	getmem( Cache_Pointer, Cache_Size ) ;
	Get( Index,  Key ) ;
        Get_Last_Error( UEC ) ;
	if( UEC.Code <> DBErr_Success ) then
	begin
	    Cache_Pointer := nil ;
	end ;
    end ;
end ;


procedure TFile_Database.Unlock( Locks : TLock_Type ) ;
{ Unlocks current record. }

begin
    if( length( Current_Record ) = 0 ) then
    begin
	Set_Error( 0, DBErr_NoCurrentRecord ) ;
	exit ;
    end ;
end ;


procedure TFile_Database.Update( Index : longint ; const Key : string ;
    Offset, Size : integer ; var Buffer) ;
{ Updates the specified record at specified offset for specified bytes from
  specified buffer.  This updates both the file and the cache buffer, if any. }

var UEC : TUEC ;

begin
    { Check for key change }
    if( Offset < Key_Size ) then
    begin
	Set_Error( 0, DBErr_KeyNotChangable ) ;
	exit ;
    end ;

    { Do the update... }
    Get( Index, Key ) ;
    Get_Last_Error( UEC ) ;
    if( UEC.Code = DBErr_Success ) then
    begin
	Offset := Offset - Key_Size ;
	seek( Current_File, Offset ) ;
	blockwrite( Current_File, PChar( @Buffer )[ Offset ], Size ) ;
	if( Cache_Pointer <> nil ) then
	begin
	    move( PChar( @Buffer )[ Offset ], Cache_Pointer[ Offset ], Size ) ;
	end ;
    end ;
end ; { TFile_Database.Update }


procedure TFile_Database.Zero ; { Deletes entire database }

var Attr : word ;
    Deleted : integer ; { Number of files/directories deleted }
    Find_Record : TFind_Record ;
    S : string ;

begin
    if( length( Root ) = 0 ) then { Must be open first }
    begin
	Set_Error( 0, DBErr_DatabaseNotOpen ) ;
	exit ;
    end ;

    { Close open file }
    if( length( Current_Record ) > 0 ) then
    begin
	close( Current_File ) ;
    end ;

    { Delete all files and directories in tree... }
    Initialize_Find_Record( Find_Record ) ;
    repeat
	Deleted := 0 ;
	S := Lookup( Root + '...\*.*', Find_Record ) ;
	while( length( S ) <> 0 ) do
	begin
	    if(
		( copy( S, length( S ) - 1, 2 ) <> '\.' )
		and
		( copy( S, length( S ) - 2, 3 ) <> '\..' )
	      ) then
	    begin
		assign( Current_File, S ) ;
		Attr := FileGetAttr( S ) ;
		if( ( Attr and faDirectory ) <> 0 ) then
		begin
		    {$I-}
		    rmdir( S ) ;
		    {$I+}
		    if( IOResult = 0 ) then
		    begin
			inc( Deleted ) ;
		    end ;
		end else
		begin
		    erase( Current_File ) ;
		    inc( Deleted ) ;
		end ;
	    end ;
	    S := Lookup( '', Find_Record ) ;
	end ; { while( length( S ) <> 0 ) }
    until Deleted = 0 ;
    Uninitialize_Find_Record( Find_Record ) ;
    Current_Record := '' ;
    Reset_Pointer ;
end ; { TFile_Database.Zero }



end.
