{
              Program Name : GetChar
              Package Name : LB
              Purpose      : Character I/O routines
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-1988
              Written By   : Alan Conroy
              Version      : 3.0

	      Copyright (C) 1988 by Alan Conroy.  Released to the Public Domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions handle special keyboard input requests.
}

unit Getchars ;

interface

{$I EDEFINES.INC}

uses Common, { TBase_Object }
     _Types ; { Small_String }

type TIO_Source = object( TBase_Object )
                      public { API... }
                          procedure Clear ; virtual ; { Clear any pending I/O }
                  end ;

type TInput_Source = object( TIO_Source )
                         private { Instance data... }
                             Input_Buffer : string ;

                         public { API... }
                             constructor Init ;

                             procedure Clear ; virtual ; { Clear any pending I/O }

                             { Forces input to the source. }
                             function Force_Input( S : string ) : boolean ;
                                 virtual ;

                             { Forces a special non-ASCII value. }
                             function Force_Special_Input( S : string ) : boolean ;
                                 virtual ;

                             { Returns a single "raw" character. }
                             function Get_Char( Option : integer ) : char ;
                                 virtual ;

                             { Returns a single "cooked" character. }
                             function Inkey : string ; virtual ;

                             { Returns True if any input is waiting. }
                             function Pending_Input : boolean ; virtual ;
                     end ;

type TDefault_Input = object( TInput_Source )
                         public { API... }
                             procedure Clear ; virtual ; { Clear any pending I/O }

                             { Forces input to the source. }
                             function Force_Input( S : string ) : boolean ;
                                 virtual ;

                             { Forces a special non-ASCII value. }
                             function Force_Special_Input( S : string ) : boolean ;
                                 virtual ;

                             { Returns a single "raw" character. }
                             function Get_Char( Option : integer ) : char ;
                                 virtual ;

                             { Returns a single "cooked" character. }
                             function Inkey : string ; virtual ;

                             { Returns True if any input is waiting. }
                             function Pending_Input : boolean ; virtual ;
                      end ;

function Input_Line( I : TInput_Source ; X : integer ) : Small_String ;

var Default_Input : TDefault_Input ;

{$I EDEFINES.INC}

implementation

uses
{$IFDEF DELPHI}
     Windows, { PeekMessage }
     Messages, { WM_Char }
     Forms, { Application }
{$ENDIF}
     ASCIIDef, { CR }
     O_S, { OS }
     Standard, { Clear_IOResult }
     TypeDefs ; { Max_String_Length }

{ TIO_Source methods... }

{ API... }

procedure TIO_Source.Clear ;

begin
    { This routine left intentionally blank... }
end ;


{ TInput_Source methods... }

constructor TInput_Source.Init ;

begin
    inherited Init ;

{$IFDEF DELPHI}
    Initialize( Input_Buffer ) ;
{$ELSE}
    Input_Buffer := '' ;
{$ENDIF}
end ;


{ API... }

procedure TInput_Source.Clear ;

begin
    inherited Clear ;
    Input_Buffer := '' ;
end ;


function TInput_Source.Force_Input( S : string ) : boolean ;

begin
    if( length( S ) + length( Input_Buffer ) > Max_String_Length ) then
    begin
        Force_Input := False ;
        exit ;
    end ;
    Force_Input := True ;
    Input_Buffer := Input_Buffer + S ;
end ;


function TInput_Source.Force_Special_Input( S : string ) : boolean ;

begin
    Force_Special_Input := False ; { We only understand ASCII }
end ;


function TInput_Source.Get_Char( Option : integer ) : char ;

begin
    if( length( Input_Buffer ) > 0 ) then
    begin
        Get_Char := Input_Buffer[ 1 ] ;
        Input_Buffer := copy( Input_Buffer, 2, length( Input_Buffer ) ) ;
    end else
    begin
        Get_Char := #0 ;
    end ;
end ;


function TInput_Source.Inkey : string ;

begin
    InKey := Get_Char( 0 ) ;
end ;


function TInput_Source.Pending_Input : boolean ;

begin
    Pending_Input := ( length( Input_Buffer ) > 0 ) ;
end ;



{ TDefault_Input methods... }

{ API... }


{ Places input into the buffer, as if the user typed it.  It returns true if
  success or false if the buffer is full.  }
function TDefault_Input.Force_Input( S : string ) : boolean ;

begin
{$IFDEF DELPHI}
    while( length( S ) > 0 ) do
    begin
	PostMessage( 0, WM_CHAR, ord( S[ 1 ] ), 0 ) ;
        S := copy( S, 2, length( S ) ) ;
    end ;
    Force_Input := True ;
{$ELSE}
    Force_Input := inherited Force_Input( S ) ;
{$ENDIF}
end ; { Force_Input }


function TDefault_Input.Force_Special_Input( S : string ) : boolean ;
{ This function allows forcing a single special character.  The parameter is
  the key name of the character to force to the input buffer. }

var SS : string ;

    procedure Convert ;

    begin
        SS := '' ;
        if( S = 'ALTBACK' ) then
        begin
            SS := NUL + char( 14 ) ;
            exit ;
        end ;
        if( S = 'A[' ) then
        begin
            SS := NUL + char( 26 ) ;
            exit ;
        end ;
        if( S = 'A]' ) then
        begin
            SS := NUL + char( 27 ) ;
            exit ;
        end ;
        if( S = 'A;' ) then
        begin
            SS := NUL + char( 39 ) ;
            exit ;
        end ;
        if( S = 'A"' ) then
        begin
            SS := NUL + char( 40 ) ;
            exit ;
        end ;
        if( S = 'A`' ) then
        begin
            SS := NUL + char( 41 ) ;
            exit ;
        end ;
        if( S = 'A\' ) then
        begin
            SS := NUL + char( 43 ) ;
            exit ;
        end ;
        if( S = 'A,' ) then
        begin
            SS := NUL + char( 51 ) ;
            exit ;
        end ;
        if( S = 'A.' ) then
        begin
            SS := NUL + char( 52 ) ;
            exit ;
        end ;
        if( S = 'A/' ) then
        begin
            SS := NUL + char( 53 ) ;
            exit ;
        end ;
        if( S = 'AKP-' ) then
        begin
            SS := NUL + char( 74 ) ;
            exit ;
        end ;
        if( S = 'AKP+' ) then
        begin
            SS := NUL + char( 78 ) ;
            exit ;
        end ;
        if( S = 'CKP-' ) then
        begin
            SS := NUL + char( 142 ) ;
            exit ;
        end ;
        if( S = 'CKP+' ) then
        begin
            SS := NUL + char( 144 ) ;
            exit ;
        end ;
        if( S = 'AKP/' ) then
        begin
            SS := NUL + char( 149 ) ;
            exit ;
        end ;
        if( S = 'ALTHOME' ) then
        begin
            SS := NUL + char( 151 ) ;
            exit ;
        end ;
        if( S = 'ALTUP' ) then
        begin
            SS := NUL + char( 152 ) ;
            exit ;
        end ;
        if( S = 'ALTPAGEUP' ) then
        begin
            SS := NUL + char( 153 ) ;
            exit ;
        end ;
        if( S = 'ALTLEFT' ) then
        begin
            SS := NUL + char( 155 ) ;
            exit ;
        end ;
        if( S = 'ALTRIGHT' ) then
        begin
            SS := NUL + char( 157 ) ;
            exit ;
        end ;
        if( S = 'ALTEND' ) then
        begin
            SS := NUL + char( 159 ) ;
            exit ;
        end ;
        if( S = 'ALTDOWN' ) then
        begin
            SS := NUL + char( 160 ) ;
            exit ;
        end ;
        if( S = 'ALTPAGEDOWN' ) then
        begin
            SS := NUL + char( 161 ) ;
            exit ;
        end ;
        if( S = 'ALTINSERT' ) then
        begin
            SS := NUL + char( 162 ) ;
            exit ;
        end ;
        if( S = 'ALTDELETE' ) then
        begin
            SS := NUL + char( 163 ) ;
            exit ;
        end ;
        if( S = 'AKP/' ) then
        begin
            SS := NUL + char( 164 ) ;
            exit ;
        end ;
        if( S = 'ALTENTER' ) then
        begin
            SS := NUL + char( 166 ) ;
            exit ;
        end ;
        if( S = 'CKP4' ) then
        begin
            SS := NUL + char( 115 ) ;
            exit ;
        end ;
        if( S = 'CKP6' ) then
        begin
            SS := NUL + char( 116 ) ;
            exit ;
        end ;
        if( S = 'CKP8' ) then
        begin
            SS := NUL + char( 141 ) ;
            exit ;
        end ;
        if( S = 'CKP2' ) then
        begin
            SS := NUL + char( 145 ) ;
            exit ;
        end ;
        if( S = 'CKP9' ) then
        begin
            SS := NUL + char( 132 ) ;
            exit ;
        end ;
        if( S = 'CF11' ) then
        begin
            SS := NUL + char( 137 ) ;
            exit ;
        end ;
        if( S = 'CF12' ) then
        begin
            SS := NUL + char( 138 ) ;
            exit ;
        end ;
        if( S = 'AF11' ) then
        begin
            SS := NUL + char( 139 ) ;
            exit ;
        end ;
        if( S = 'AF12' ) then
        begin
            SS := NUL + char( 140 ) ;
            exit ;
        end ;
        if( S = 'A8' ) then
        begin
            SS := NUL + char( 127 ) ;
            exit ;
        end ;
        if( S = 'A9' ) then
        begin
            SS := NUL + char( 128 ) ;
            exit ;
        end ;
        if( S = 'A0' ) then
        begin
            SS := NUL + char( 129 ) ;
            exit ;
        end ;
        if( S = 'A-' ) then
        begin
            SS := NUL + char( 130 ) ;
            exit ;
        end ;
        if( S = 'A=' ) then
        begin
            SS := NUL + char( 131 ) ;
            exit ;
        end ;
        if( S = 'AQ' ) then
        begin
            SS := NUL + NUL ;
            exit ;
        end ;
        if( S = 'AS' ) then
        begin
            SS := NUL + US ;
            exit ;
        end ;
        if( S = 'AA' ) then
        begin
            SS := NUL + RS ;
            exit ;
        end ;
        if( S = 'UP' ) then
        begin
            SS := NUL + 'H' ;
            exit ;
        end ;
        if( S = 'DOWN' ) then
        begin
            SS := NUL + 'P' ;
            exit ;
        end ;
        if( S = 'RIGHT' ) then
        begin
            SS := NUL + 'M' ;
            exit ;
        end ;
        if( S = 'LEFT' ) then
        begin
            SS := NUL + 'K' ;
            exit ;
        end ;
        if( S = 'DELETE' ) then
        begin
            SS := NUL + 'S' ;
            exit ;
        end ;
        if( S = 'E1' ) then { Shift F1, etc }
        begin
            SS := NUL + 'T' ;
            exit ;
        end ;
        if( S = 'E2' ) then
        begin
            SS := NUL + 'U' ;
            exit ;
        end ;
        if( S = 'E3' ) then
        begin
            SS := NUL + 'V' ;
            exit ;
        end ;
        if( S = 'E4' ) then
        begin
            SS := NUL + 'W' ;
            exit ;
        end ;
        if( S = 'E5' ) then
        begin
            SS := NUL + 'X' ;
            exit ;
        end ;
        if( S = 'E6' ) then
        begin
            SS := NUL + 'Y' ;
            exit ;
        end ;
        if( S = 'E7' ) then
        begin
            SS := NUL + 'Z' ;
            exit ;
        end ;
        if( S = 'E8' ) then
        begin
            SS := NUL + '[' ;
            exit ;
        end ;
        if( S = 'E9' ) then
        begin
            SS := NUL + '\' ;
            exit ;
        end ;
        if( S = 'E10' ) then
        begin
            SS := NUL + ']' ;
            exit ;
        end ;
        if( S = 'E11' ) then
        begin
            SS := NUL + #135 ;
            exit ;
        end ;
        if( S = 'E12' ) then
        begin
            SS := NUL + #136 ;
            exit ;
        end ;
        if( S = 'F1' ) then
        begin
            SS := NUL + ';' ;
            exit ;
        end ;
        if( S = 'F2' ) then
        begin
            SS := NUL + '<' ;
            exit ;
        end ;
        if( S = 'F3' ) then
        begin
            SS := NUL + '=' ;
            exit ;
        end ;
        if( S = 'F4' ) then
        begin
            SS := NUL + '>' ;
            exit ;
        end ;
        if( S = 'F5' ) then
        begin
            SS := NUL + '?' ;
            exit ;
        end ;
        if( S = 'F6' ) then
        begin
            SS := NUL + '@' ;
            exit ;
        end ;
        if( S = 'F7' ) then
        begin
            SS := NUL + 'A' ;
            exit ;
        end ;
        if( S = 'F8' ) then
        begin
            SS := NUL + 'B' ;
            exit ;
        end ;
        if( S = 'F9' ) then
        begin
            SS := NUL + 'C' ;
            exit ;
        end ;
        if( S = 'F10' ) then
        begin
            SS := NUL + 'D' ;
            exit ;
        end ;
        if( S = 'F11' ) then
        begin
            SS := NUL + chr( 133 ) ;
            exit ;
        end ;
        if( S = 'F12' ) then
        begin
            SS := NUL + chr( 134 ) ;
            exit ;
        end ;
        if( S = 'F13' ) then
        begin
            SS := NUL + 'V' ;
            exit ;
        end ;
        if( S = 'F14' ) then
        begin
            SS := NUL + 'W' ;
            exit ;
        end ;
        if( S = 'F15' ) then
        begin
            SS := NUL + 'X' ;
            exit ;
        end ;
        if( S = 'F16' ) then
        begin
            SS := NUL + 'Y' ;
            exit ;
        end ;
        if( S = 'F17' ) then
        begin
            SS := NUL + 'Z' ;
            exit ;
        end ;
        if( S = 'F18' ) then
        begin
            SS := NUL + '[' ;
            exit ;
        end ;
        if( S = 'F19' ) then
        begin
            SS := NUL + '\' ;
            exit ;
        end ;
        if( S = 'F20' ) then
        begin
            SS := NUL + ']' ;
            exit ;
        end ;
        if( S = 'AF1' ) then
        begin
            SS := NUL + 'h' ;
            exit ;
        end ;
        if( S = 'AF2' ) then
        begin
            SS := NUL + 'i' ;
            exit ;
        end ;
        if( S = 'AF3' ) then
        begin
            SS := NUL + 'j' ;
            exit ;
        end ;
        if( S = 'AF4' ) then
        begin
            SS := NUL + 'k' ;
            exit ;
        end ;
        if( S = 'AF5' ) then
        begin
            SS := NUL + 'l' ;
            exit ;
        end ;
        if( S = 'AF6' ) then
        begin
            SS := NUL + 'm' ;
            exit ;
        end ;
        if( S = 'AF7' ) then
        begin
            SS := NUL + 'n' ;
            exit ;
        end ;
        if( S = 'AF8' ) then
        begin
            SS := NUL + 'o' ;
            exit ;
        end ;
        if( S = 'AF9' ) then
        begin
            SS := NUL + 'p' ;
            exit ;
        end ;
        if( S = 'AF10' ) then
        begin
            SS := NUL + 'q' ;
            exit ;
        end ;
        if( S = 'AF11' ) then
        begin
            SS := NUL + #139 ;
            exit ;
        end ;
        if( S = 'AF12' ) then
        begin
            SS := NUL + #140 ;
            exit ;
        end ;
        if( S = 'CF1' ) then
        begin
            SS := NUL + '^' ;
            exit ;
        end ;
        if( S = 'CF2' ) then
        begin
            SS := NUL + '_' ;
            exit ;
        end ;
        if( S = 'CF3' ) then
        begin
            SS := NUL + '`' ;
            exit ;
        end ;
        if( S = 'CF4' ) then
        begin
            SS := NUL + 'a' ;
            exit ;
        end ;
        if( S = 'CF5' ) then
        begin
            SS := NUL + 'b' ;
            exit ;
        end ;
        if( S = 'CF6' ) then
        begin
            SS := NUL + 'c' ;
            exit ;
        end ;
        if( S = 'CF7' ) then
        begin
            SS := NUL + 'd' ;
            exit ;
        end ;
        if( S = 'CF8' ) then
        begin
            SS := NUL + 'e' ;
            exit ;
        end ;
        if( S = 'CF9' ) then
        begin
            SS := NUL + 'f' ;
            exit ;
        end ;
        if( S = 'CF10' ) then
        begin
            SS := NUL + 'g' ;
            exit ;
        end ;
        if( S = 'BACK' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( ( S = 'KP0' ) or ( S = 'INSERT' ) ) then
        begin
            SS := NUL + 'R' ;
            exit ;
        end ;
        if( S = 'KP9' ) then
        begin
            SS := NUL + 'I' ;
            exit ;
        end ;
        if( (S = 'HOME' ) or ( S = 'KP7' ) ) then
        begin
            SS := NUL + 'G' ;
            exit ;
        end ;
        if( S = 'CKP4' ) then
        begin
            SS := NUL + 's' ;
            exit ;
        end ;
        if( S = 'CKP6' ) then
        begin
            SS := NUL + 't' ;
            exit ;
        end ;
        if( ( S = 'KP1' ) or ( S = 'END' ) ) then
        begin
            SS := NUL + 'O' ;
            exit ;
        end ;
        if( S = 'KP3' ) then
        begin
            SS := NUL + 'Q' ;
            exit ;
        end ;
        if( S = 'CKP7' ) then
        begin
            SS := NUL + 'w' ;
            exit ;
        end ;
        if( S = 'CKP1' ) then
        begin
            SS := NUL + 'u' ;
            exit ;
        end ;
        if( S = 'CKP3' ) then
        begin
            SS := NUL + 'v' ;
            exit ;
        end ;
        if( S = 'A1' ) then
        begin
            SS := NUL + 'x' ;
            exit ;
        end ;
        if( S = 'A2' ) then
        begin
            SS := NUL + 'y' ;
            exit ;
        end ;
        if( S = 'A3' ) then
        begin
            SS := NUL + 'z' ;
            exit ;
        end ;
        if( S = 'A4' ) then
        begin
            SS := NUL + '{' ;
            exit ;
        end ;
        if( S = 'A5' ) then
        begin
            SS := NUL + '|' ;
            exit ;
        end ;
        if( S = 'A6' ) then
        begin
            SS := NUL + '}' ;
            exit ;
        end ;
        if( S = 'A7' ) then
        begin
            SS := NUL + '~' ;
            exit ;
        end ;
        if( S = 'AB' ) then
        begin
            SS := NUL + '0' ;
            exit ;
        end ;
        if( S = 'AC' ) then
        begin
            SS := NUL + '.' ;
            exit ;
        end ;
        if( S = 'AD' ) then
        begin
            SS := NUL + ' ' ;
            exit ;
        end ;
        if( S = 'AE' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AF' ) then
        begin
            SS := NUL + '!' ;
            exit ;
        end ;
        if( S = 'AG' ) then
        begin
            SS := NUL + '"' ;
            exit ;
        end ;
        if( S = 'AH' ) then
        begin
            SS := NUL + '#' ;
            exit ;
        end ;
        if( S = 'AI' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AJ' ) then
        begin
            SS := NUL + '$' ;
            exit ;
        end ;
        if( S = 'AK' ) then
        begin
            SS := NUL + '%' ;
            exit ;
        end ;
        if( S = 'AL' ) then
        begin
            SS := NUL + '&' ;
            exit ;
        end ;
        if( S = 'AM' ) then
        begin
            SS := NUL + '2' ;
            exit ;
        end ;
        if( S = 'AN' ) then
        begin
            SS := NUL + '1' ;
            exit ;
        end ;
        if( S = 'AO' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AP' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AQ' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AR' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AT' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AU' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AV' ) then
        begin
            SS := NUL + '/' ;
            exit ;
        end ;
        if( S = 'AW' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AX' ) then
        begin
            SS := NUL + '-' ;
            exit ;
        end ;
        if( S = 'AY' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AZ' ) then
        begin
            SS := NUL + ',' ;
            exit ;
        end ;

        if( S = 'UP' ) then
        begin
            SS := ESC + '[A' ;
            exit ;
        end ;
        if( S = 'DOWN' ) then
        begin
            SS := ESC + '[B' ;
            exit ;
        end ;
        if( S = 'RIGHT' ) then
        begin
            SS := ESC + '[C' ;
            exit ;
        end ;
        if( S = 'LEFT' ) then
        begin
            SS := ESC + '[D' ;
            exit ;
        end ;
        if( S = 'COMPOSE' ) then
        begin
            SS := ESC + '[10' ;
            exit ;
        end ;
        if( S = 'F9' ) then
        begin
            SS := ESC + '[00' ;
            exit ;
        end ;
        if( S = 'AF1' ) then
        begin
            SS := ESC + '[1~a' ;
            exit ;
        end ;
        if( S = 'AF2' ) then
        begin
            SS := ESC + '[2~a' ;
            exit ;
        end ;
        if( S = 'AF3' ) then
        begin
            SS := ESC + '[3~a' ;
            exit ;
        end ;
        if( S = 'F18' ) then
        begin
            SS := ESC + '[12a' ;
            exit ;
        end ;

        if( S = 'F19' ) then
        begin
            SS := ESC + '[13a' ;
            exit ;
        end ;
        if( S = 'F14' ) then
        begin
            SS := '[?6a' ;
            exit ;
        end ;
        if( S = 'F6' ) then
        begin
            SS := '[17a' ;
            exit ;
        end ;
        if( S = 'F17' ) then
        begin
            SS := '[31a' ;
            exit ;
        end ;
        if( S = 'F10' ) then
        begin
            SS := '[21a' ;
            exit ;
        end ;
        if( S = 'F20' ) then
        begin
            SS := '[34a' ;
            exit ;
        end ;
        if( S = 'F4' ) then
        begin
            SS := '[24a' ;
            exit ;
        end ;
        if( S = 'F7' ) then
        begin
            SS := '[18a' ;
            exit ;
        end ;
        if( S = 'F15' ) then
        begin
            SS := '[28a' ;
            exit ;
        end ;
        if( S = 'F8' ) then
        begin
            SS := '[19a' ;
            exit ;
        end ;
        if( S = 'F16' ) then
        begin
            SS := '[29a' ;
            exit ;
        end ;

        if( S = 'AF4' ) then
        begin
            SS := ESC + '[4' ;
            exit ;
        end ;
        if( S = 'AF5' ) then
        begin
            SS := ESC + '[5' ;
            exit ;
        end ;
        if( S = 'AF6' ) then
        begin
            SS := ESC + '[6' ;
            exit ;
        end ;

        if( S = 'PF1' ) then
        begin
            SS := ESC + 'OP' ;
            exit ;
        end ;
        if( S = 'PF2' ) then
        begin
            SS := ESC + 'OQ' ;
            exit ;
        end ;
        if( S = 'PF3' ) then
        begin
            SS := ESC + 'OR' ;
            exit ;
        end ;
        if( S = 'PF4' ) then
        begin
            SS := ESC + 'OS' ;
            exit ;
        end ;
        if( S = 'ENTER' ) then
        begin
            SS := ESC + 'OM' ;
            exit ;
        end ;
        if( S = 'KP0' ) then
        begin
            SS := ESC + 'Op' ;
            exit ;
        end ;
        if( S = 'KP1' ) then
        begin
            SS := ESC + 'Oq' ;
            exit ;
        end ;
        if( S = 'KP2' ) then
        begin
            SS := ESC + 'Or' ;
            exit ;
        end ;
        if( S = 'KP3' ) then
        begin
            SS := ESC + 'Os' ;
            exit ;
        end ;
        if( S = 'KP4' ) then
        begin
            SS := ESC + 'Ot' ;
            exit ;
        end ;
        if( S = 'KP5' ) then
        begin
            SS := ESC + 'Ou' ;
            exit ;
        end ;
        if( S = 'KP6' ) then
        begin
            SS := ESC + 'Ov' ;
            exit ;
        end ;
        if( S = 'KP7' ) then
        begin
            SS := ESC + 'Ow' ;
            exit ;
        end ;
        if( S = 'KP8' ) then
        begin
            SS := ESC + 'Ox' ;
            exit ;
        end ;
        if( S = 'KP9' ) then
        begin
            SS := ESC + 'Oy' ;
            exit ;
        end ;
        if( S = 'KP-' ) then
        begin
            SS := ESC + 'Om' ;
            exit ;
        end ;
        if( S = 'KP,' ) then
        begin
            SS := ESC + 'Ol' ;
            exit ;
        end ;
        if( ( S = 'KP.' ) or ( S = 'DELETE' ) ) then
        begin
            SS := ESC + 'On' ;
            exit ;
        end ;
    end ;

begin
    Convert ;
    if( length( Input_Buffer ) + length( SS ) > Max_String_Length ) then
    begin
        Force_Special_Input := False ;
        exit ;
    end ;
    Force_Special_Input := True ;
    Input_Buffer := Input_Buffer + SS ;
end ; { Force_Special_Input }


function TDefault_Input.Get_Char( Option : integer ) : char ;
{ This function returns a single character typed at the keyboard.  If Option is
  non-zero, control returns immediately whether or not a character was waiting
  (in which case the returned character is zero, if no input was waiting). }

var B : boolean ;
    C : Char ;

begin
    C := inherited Get_Char( Option ) ;
    if( C <> #0 ) then
    begin
        Get_Char := C ;
        exit ;
    end ;
    Get_Char := OS^.Get_Char( ( Option = 0 ), B ) ;
end ; { TDefault_Input.Get_Char }


procedure TDefault_Input.Clear ; { Clear input buffer }

begin
    inherited Clear ;

    OS^.Cancel_Typeahead ;
end ;


function TDefault_Input.Inkey : string ;

var Command : string[ 11 ] ; { Input or command line }
{$IFDEF CONSOLE}
    Dummy : DWORD ;
    Count : integer ;
    Handle : integer ;
    Input_Data : TInputRecord ;
    W : integer ;

label Loop ;
{$ENDIF}

begin
{$IFDEF CONSOLE}
Loop:
    Command := inherited Get_Char( 0 ) ;
    if( Command <> #0 ) then { Pending input }
    begin
        W := 0 ;
    end else
    begin
        Handle := GetStdHandle( STD_INPUT_HANDLE ) ;
        Command := '' ;
        GetNumberOfConsoleInputEvents( Handle, Dummy ) ;
        while( Dummy = 0 ) do
        begin
            GetNumberOfConsoleInputEvents( Handle, Dummy ) ;
        end ;

        { Get the data... }
        ReadConsoleInput( Handle, Input_Data, 1, Dummy ) ;
        if(
            ( Input_Data.EventType <> KEY_EVENT )
            or
            ( not Input_Data.Event.KeyEvent.bKeyDown )
          ) then
        begin
            goto Loop ; { Skip anything but key-down }
        end ;

        Command := Input_Data.Event.KeyEvent.AsciiChar ;
        W := Input_Data.Event.KeyEvent.wVirtualKeyCode ;

        { Handle repeats... }
        Count := Input_Data.Event.KeyEvent.wRepeatCount - 1 ;
        Dummy := length( Input_Buffer ) + Count ;
        setlength( Input_Buffer, Dummy ) ;
        while( Count > 0 ) do
        begin
            Input_Buffer[ Dummy ] := Command[ 1 ] ;
            dec( Dummy ) ;
            dec( Count ) ;
        end ;
    end ; { if( Command <> #0 ) }
    case W of
        VK_LBUTTON : Command := 'LBUTTON' ;
        VK_RBUTTON : Command := 'RBUTTON' ;
        VK_CANCEL : Command := 'CANCEL' ;
        VK_MBUTTON : Command := 'MBUTTON' ;
        VK_BACK	: Command := BS ;
        VK_TAB : Command := HT ;
        VK_CLEAR : Command := 'CLEAR' ;
        VK_RETURN : Command := CR ;
        VK_SHIFT : Command := 'SHIFT' ;
        VK_CONTROL : Command := 'CONTROL' ;
        VK_MENU	: Command := 'ALT' ;
        VK_PAUSE : Command := 'PAUSE' ;
        VK_CAPITAL : Command := 'CAPS' ;
        VK_ESCAPE : Command := ESC ;
        VK_SPACE : Command := ' ' ;
        VK_PRIOR : Command := 'PAGEUP' ;
        VK_NEXT	: Command := 'PAGEDOWN' ;
        VK_END : Command := 'END' ;
        VK_HOME	: Command := 'HOME' ;
        VK_LEFT	: Command := 'LEFT' ;
        VK_UP : Command := 'UP' ;
        VK_RIGHT : Command := 'RIGHT' ;
        VK_DOWN	: Command := 'DOWN' ;
        VK_SELECT : Command := 'SELECT' ;
        VK_PRINT : Command := 'PRINT' ;
        VK_EXECUTE : Command := 'EXECUTE' ;
        VK_SNAPSHOT : Command := 'PRINT' ;
        VK_INSERT : Command := 'INSERT' ;
        VK_DELETE : Command := 'DELETE' ;
        VK_HELP	: Command := 'HELP' ;
        VK_LWIN	: Command := 'LWIN' ;
        VK_RWIN	: Command := 'RWIN' ;
        VK_APPS	: Command := 'APPS' ;
        VK_NUMPAD0 : Command := 'KP0' ;
        VK_NUMPAD1 : Command := 'KP1' ;
        VK_NUMPAD2 : Command := 'KP2' ;
        VK_NUMPAD3 : Command := 'KP3' ;
        VK_NUMPAD4 : Command := 'KP4' ;
        VK_NUMPAD5 : Command := 'KP5' ;
        VK_NUMPAD6 : Command := 'KP6' ;
        VK_NUMPAD7 : Command := 'KP7' ;
        VK_NUMPAD8 : Command := 'KP8' ;
        VK_NUMPAD9 : Command := 'KP9' ;
        VK_MULTIPLY : Command := 'KP*' ;
        VK_ADD : Command := 'KP+' ;
        VK_SEPARATOR : Command := 'ENTER' ;
        VK_SUBTRACT : Command := 'KP-' ;
        VK_DECIMAL : Command := 'KP.' ;
        VK_DIVIDE : Command := 'KP/' ;
        VK_F1 : Command := 'F1' ;
        VK_F2 : Command := 'F2' ;
        VK_F3 : Command := 'F3' ;
        VK_F4 : Command := 'F4' ;
        VK_F5 : Command := 'F5' ;
        VK_F6 : Command := 'F6' ;
        VK_F7 : Command := 'F7' ;
        VK_F8 : Command := 'F8' ;
        VK_F9 : Command := 'F9' ;
        VK_F10 : Command := 'F10' ;
        VK_F11 : Command := 'F11' ;
        VK_F12 : Command := 'F12' ;
        VK_F13 : Command := 'F13' ;
        VK_F14 : Command := 'F14' ;
        VK_F15 : Command := 'F15' ;
        VK_F16 : Command := 'F16' ;
        VK_F17 : Command := 'F17' ;
        VK_F18 : Command := 'F18' ;
        VK_F19 : Command := 'F19' ;
        VK_F20 : Command := 'F20' ;
        VK_F21 : Command := 'F21' ;
        VK_F22 : Command := 'F22' ;
        VK_F23 : Command := 'F23' ;
        VK_F24 : Command := 'F24' ;
        VK_NUMLOCK : Command := 'NUMLOCK' ;
        VK_SCROLL : Command := 'SCROLL' ;
        VK_PROCESSKEY : Command := 'PROCESS' ;
        VK_ATTN	: Command := 'ATTN' ;
        VK_CRSEL : Command := 'CRSEL' ;
        VK_EXSEL : Command := 'EXSEL' ;
        VK_EREOF : Command := 'EREOF' ;
        VK_PLAY	: Command := 'PLAY' ;
        VK_ZOOM	: Command := 'ZOOM' ;
        VK_PA1 : Command := 'PA1' ;
        VK_OEM_CLEAR : Command := 'CLEAR' ;
    end ;
    if( length( Command ) = 0 ) then
    begin
        goto Loop ;
    end ;
    case Input_Data.Event.KeyEvent.dwControlKeyState of
        LEFT_ALT_PRESSED, RIGHT_ALT_PRESSED : if( ( length( Command ) > 1 ) and ( copy( Command, 1, 2 ) <> 'KP' ) ) then
                                              begin
                                                  Command := 'ALT' + Command ;
                                              end else
                                              begin
                                                  Command := 'A' + Command ;
                                              end ;
        LEFT_CTRL_PRESSED, RIGHT_CTRL_PRESSED : if( length( Command ) > 1 ) then
                                                begin
                                                    Command := 'C' + Command ;
                                                end ;
    end ;
    InKey := Command ;
{$ELSE}
    Command[ 1 ] := Get_Char( 0 ) ;
    if( ( Command[ 1 ] <> NUL ) and ( Command[ 1 ] <> ESC ) ) then
    { Normal character }
    begin
        Inkey := Command[ 1 ] ;
        exit ;
    end ;

    if( Command[ 1 ] = NUL ) then { IBM character }
    begin
        Command[ 2 ] := Get_Char( 1 ) ;
        if( Command[ 2 ] = NUL ) then
	begin
	    Inkey := 'AQ' ;
	end ;
        if( Command[ 2 ] = US ) then
	begin
	    Inkey := 'AS' ;
	end ;
        if( Command[ 2 ] = RS ) then
	begin
	    Inkey := 'AA' ;
	end ;
        case Command[ 2 ] of
            #14 : Inkey := 'ALTBACK' ;
            #26 : Inkey := 'A[' ;
            #27 : Inkey := 'A]' ;
            #39 : Inkey := 'A;' ;
            #40 : Inkey := 'A"' ;
            #41 : InKey := 'A`' ;
            #43 : Inkey := 'A\' ;
            #51 : Inkey := 'A,' ;
            #52 : Inkey := 'A.' ;
            #53 : Inkey := 'A/' ;
            #74 : Inkey := 'AKP-' ;
            #78 : Inkey := 'AKP+' ;
            #142 : Inkey := 'CKP-' ;
            #144 : Inkey := 'CKP+' ;
            #149 : Inkey := 'AKP/' ;
            #151 : Inkey := 'ALTHOME' ;
            #152 : Inkey := 'ALTUP' ;
            #153 : Inkey := 'ALTPAGEUP' ;
            #155 : Inkey := 'ALTLEFT' ;
            #157 : Inkey := 'ALTRIGHT' ;
            #159 : Inkey := 'ALTEND' ;
            #160 : Inkey := 'ALTDOWN' ;
            #161 : Inkey := 'ALTPAGEDOWN' ;
            #162 : Inkey := 'ALTINSERT' ;
            #163 : Inkey := 'ALTDELETE' ;
            #164 : Inkey := 'AKP/' ;
            #166 : Inkey := 'ALTENTER' ;

            #127 : Inkey := 'A8' ;
            #128 : Inkey := 'A9' ;
            #129 : Inkey := 'A0' ;
            #130 : Inkey := 'A-' ;
            #131 : Inkey := 'A=' ;
            #132 : Inkey := 'CKP9' ;
            #133 : Inkey := 'F11' ;
            #134 : Inkey := 'F12' ;
            #135 : Inkey := 'E11' ;
            #136 : Inkey := 'E12' ;
            #137 : Inkey := 'CF11' ;
            #138 : Inkey := 'CF12' ;
            #139 : Inkey := 'AF11' ;
            #140 : Inkey := 'AF12' ;
            #141 : Inkey := 'CKP8' ;
            #145 : Inkey := 'CKP2' ;
            'H' : Inkey := 'UP' ;
            'P' : Inkey := 'DOWN' ;
            'M' : Inkey := 'RIGHT' ;
            'K' : Inkey := 'LEFT' ;
	    'S' : Inkey := 'DELETE' ;
            ';' : Inkey := 'F1' ;
            '<' : Inkey := 'F2' ;
            '=' : Inkey := 'F3' ;
            '>' : Inkey := 'F4' ;
            '?' : Inkey := 'F5' ;
            '@' : Inkey := 'F6' ;
            'A' : Inkey := 'F7' ;
            'B' : Inkey := 'F8' ;
            'C' : Inkey := 'F9' ;
            'D' : Inkey := 'F10' ;
            'T' : Inkey := 'E1' ; { Shift F1, etc }
            'U' : Inkey := 'E2' ;
            'V' : Inkey := 'E3' ;
            'W' : Inkey := 'E4' ;
            'X' : Inkey := 'E5' ;
            'Y' : Inkey := 'E6' ;
            'Z' : Inkey := 'E7' ;
            '[' : Inkey := 'E8' ;
            '\' : Inkey := 'E9' ;
            ']' : Inkey := 'E10' ;
            'h' : Inkey := 'AF1' ; { Alt F1, etc }
            'i' : Inkey := 'AF2' ;
            'j' : Inkey := 'AF3' ;
            'k' : Inkey := 'AF4' ;
            'l' : Inkey := 'AF5' ;
            'm' : Inkey := 'AF6' ;
            'n' : Inkey := 'AF7' ;
            'o' : Inkey := 'AF8' ;
            'p' : Inkey := 'AF9' ;
            'q' : Inkey := 'AF10' ;
            '^' : Inkey := 'CF1' ; { Control F1 is CF1, etc }
            '_' : Inkey := 'CF2' ;
            '`' : Inkey := 'CF3' ;
            'a' : Inkey := 'CF4' ;
            'b' : Inkey := 'CF5' ;
            'c' : Inkey := 'CF6' ;
            'd' : Inkey := 'CF7' ;
            'e' : Inkey := 'CF8' ;
            'f' : Inkey := 'CF9' ;
            'g' : Inkey := 'CF10' ;
            '' : Inkey := 'BACK' ; { Back tab }
            'R' : Inkey := 'INSERT' ;
            'I' : Inkey := 'KP9' ; { PAGEUP }
            'G' : Inkey := 'HOME' ;
            's' : Inkey := 'CKP4' ;
            't' : Inkey := 'CKP6' ;
            'O' : Inkey := 'END' ;
            'Q' : Inkey := 'KP3' ; { PAGEDOWN }
            'w' : Inkey := 'CKP7' ;
            'u' : Inkey := 'CKP1' ;
            'v' : Inkey := 'CKP3' ;
            'x' : Inkey := 'A1' ;
            'y' : Inkey := 'A2' ;
            'z' : Inkey := 'A3' ;
            '{' : Inkey := 'A4' ;
            '|' : Inkey := 'A5' ;
            '}' : Inkey := 'A6' ;
            '~' : Inkey := 'A7' ;
            '0' : Inkey := 'AB' ;
            '.' : Inkey := 'AC' ;
            ' ' : Inkey := 'AD' ;
            '' : Inkey := 'AE' ;
            '!' : Inkey := 'AF' ;
            '"' : Inkey := 'AG' ;
            '#' : Inkey := 'AH' ;
            '' : Inkey := 'AI' ;
            '$' : Inkey := 'AJ' ;
            '%' : Inkey := 'AK' ;
            '&' : Inkey := 'AL' ;
            '2' : Inkey := 'AM' ;
            '1' : Inkey := 'AN' ;
            '' : Inkey := 'AO' ;
            '' : Inkey := 'AP' ;
            '' : Inkey := 'AQ' ;
            '' : Inkey := 'AR' ;
            '' : Inkey := 'AT' ;
            '' : Inkey := 'AU' ;
            '/' : Inkey := 'AV' ;
            '' : Inkey := 'AW' ;
            '-' : Inkey := 'AX' ;
            '' : Inkey := 'AY' ;
            ',' : Inkey := 'AZ' ;
        end ;
        exit ;
    end ;

    { Escape character... }
    Command[ 1 ] := Get_Char( 1 ) ;
    if( Command[ 1 ] = NUL ) then { Just an escape }
    begin
        Inkey := ESC ;
        exit ;
    end ;

    { Special character... }
    Command[ 2 ] := Get_Char( 0 ) ; { Get third character }
    if( Command[ 1 ] = chr( 91 ) ) {"["} Then
    begin
        case Command[ 2 ] Of
            'A' : Inkey := 'UP' ;
            'B' : Inkey := 'DOWN' ;
            'C' : Inkey := 'RIGHT' ;
            'D' : Inkey := 'LEFT' ;
            '1'..'3' : begin
                           Command[ 3 ] := Get_Char( 1 ) ;
                           Command[ 4 ] := Get_Char( 1 ) ;
                           case Command[ 3 ] of
                               '~' : Inkey := 'E' + Command[ 2 ] ;
                               '0' : if( Command[ 2 ] = '1' ) then
				     begin
					 Inkey := 'COMPOSE' ;
				     end else
				     begin
                                         Inkey := 'F9' ;
				     end ;
                               '2' : Inkey := 'F18' ;
                               '3' : Inkey := 'F19' ;
                               '6' : Inkey := 'F14' ;
                               '7' : Inkey := 'F6' ;
                               '1' : if( Command[ 2 ] = '3' ) then Inkey := 'F17'
                                         Else Inkey := 'F10' ;
                               '4' : if( Command[ 2 ] = '3' ) then Inkey := 'F20'
                                         Else Inkey := 'F4' ;
                               '8' : if( Command[ 2 ] = '1' ) then Inkey := 'F7'
                                         Else Inkey := 'F15' ;
                               '9' : if( Command[ 2 ] = '1' ) then Inkey := 'F8'
                                         Else Inkey := 'F16' ;
                           end ;
                           exit ;
                       end ;
            '4' : Inkey := 'E4' ;
            '5' : Inkey := 'E5' ;
            '6' : Inkey := 'E6' ;
        end ;
        exit ;
    end ;
    if( Command[ 1 ] = 'O' ) then
    begin
        case Command[ 2 ] of
            'P' : Inkey := 'PF1' ;
            'Q' : Inkey := 'PF2' ;
            'R' : Inkey := 'PF3' ;
            'S' : Inkey := 'PF4' ;
            'M' : Inkey := 'ENTER' ;
            'p' : Inkey := 'KP0' ;
            'q' : Inkey := 'KP1' ;
            'r' : Inkey := 'KP2' ;
            's' : Inkey := 'KP3' ;
            't' : Inkey := 'KP4' ;
            'u' : Inkey := 'KP5' ;
            'v' : Inkey := 'KP6' ;
            'w' : Inkey := 'KP7' ;
            'x' : Inkey := 'KP8' ;
            'y' : Inkey := 'KP9' ;
            'm' : Inkey := 'KP-' ;
            'l' : Inkey := 'KP,' ;
            'n' : Inkey := 'KP.' ;
        end ;
    end ;
{$ENDIF}
end ; { TDefault_Input.InKey }


function TDefault_Input.Pending_Input : boolean ;

var B : boolean ;
    C : Char ;

begin
    B := inherited Pending_Input ; { Handle any buffered input }
    if( B ) then
    begin
        Pending_Input := True ;
        exit ;
    end ;
    C := OS^.Get_Char( False, B ) ;
    if( B ) then { There was some input }
    begin
        Input_Buffer := Input_Buffer + C ;
        Pending_Input := True ;
        exit ;
    end ;
    Pending_Input := False ;
end ; { TDefault_Input.Pending_Input }



function Input_Line( I : TInput_Source ; X : Integer ) : Small_String ;
{ This function returns a string of input after taking it with or without echo.
  Ending delimeters are returned as well. }

    procedure _Write( const S : string ) ;

    begin
        {$I-}
        write( S ) ;
        {$I+}
        Clear_IOResult ;
    end ;


    procedure _Writeln( const S : string ) ;

    begin
        {$I-}
        writeln( S ) ;
        {$I+}
        Clear_IOResult ;
    end ;

var Accumulate : Small_String ;
    Count : Integer ;
    Reg : Char ;
    Reg_Ax : Integer ;

label Loop ;

begin
    Count := 1 ;
Loop:
    Reg := I.Get_Char( 0 ) ;
    Reg_Ax := Ord( Reg ) And 127 ;
    Accumulate[ Count ] := Chr( Reg_Ax ) ;
    Count := Count + 1 ;
    if( ( Reg_Ax = 10 ) or ( Reg_Ax = 13 ) or ( Reg_Ax = 12 ) or ( Reg_Ax = 27 ) ) then
    { LF, CR, FF, ESC are delimiters }
    begin
        if( X <> 0 ) then
	begin
	    if( Reg_Ax = 27 ) then
	    begin
		_Write( '$' ) ;
	    end else
	    begin
		_Write( Reg ) ;
	    end ;
	end ;
        if( Reg_Ax = 13 ) then
        begin
            Accumulate[ Count ] := LF ;
            Count := Count + 1 ;
            if( X <> 0 ) Then
	    begin
		_Write( Chr( 10 ) ) ;
	    end ;
        end ;
        if( Reg_Ax = 10 ) then
        begin
            Accumulate[ Count ] := CR ;
            Count := Count + 1 ;
            if( X <> 0 ) then
	    begin
		_Write( CR ) ;
	    end ;
        end ;
        Accumulate[ 0 ] := Chr( Count - 1 ) ;
        Input_Line := Accumulate ;
        exit ;
    end ;
    if( Reg_Ax = 21 ) then { Control-U }
    begin
        Count := 1 ;
        if( X <> 0 ) then
	begin
	    _Writeln( '^U' ) ;
	end ;
        goto Loop ;
    end ;
    if( Reg_Ax = 127 ) then { Delete }
    begin
        Count := Count - 2 ;
        if( Count < 1 ) then
	begin
	    Count := 1 ;
	end ;
        if( X <> 0 ) then
	begin
	    _Write( ' ' ) ;
	end ;
        goto Loop ;
    end ;
    if( Count = 254 ) then
    begin
        _Write( '' ) ; { End of input buffer }
        Count := Count - 1 ;
        goto Loop ;
    end ;
    if( X <> 0 ) then
    begin
	_Write( Reg ) ;
    end ;
    goto Loop ;
end ; { Input_Line }


begin
    Default_Input.Init ;
end.


