{
        Program Name : KeyMap
        Package Name : LB
        Purpose      : Simple keymapper
        Institution  : Conroy & Conroy Co.
        Date Written : 5-August-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TKey_Map provides a simple key-to-action-name mapping for Windows
        applications.
}

unit KeyMap ;

interface

uses // Borland...
     Classes ; // TStringList

type TWindows_Key = class
                        public // Constructors and destructors...
                            constructor Create( _Code : integer ;
                                _Shift : TShiftState ) ;

                        public
                            Code : integer ; // VK code
                            Shift : TShiftState ;

                            function As_String : string ;
                    end ;

type TKey_Map = class( TStringList )
                    public // Constructors and destructors...
                        constructor Create ;
                        destructor Destroy ; override ;

                    private // Instance data...
                        Captions : TStringList ;

                    protected // Property handlers...
                        function Get_Caption( Index : integer ) : string ;
                        function Get_Tag( Index : integer ) : integer ;
                        procedure Set_Caption( Index : integer ;
                            const Value : string ) ;
                        procedure Set_Tag( Index, Value : integer ) ;

                    public // API...
                        procedure Add_Key( const Name, _Caption : string ;
                            Key : TWindows_Key ; Tag : integer = 0 ) ;
                        property Caption[ Index : integer ] : string
                            read Get_Caption
                            write Set_Caption ;
                        function Keys( Index : integer ) : TWindows_Key ;
                        property Tag[ Index : integer ] : integer
                            read Get_Tag
                            write Set_Tag ;
                end ;

function String_To_Key( S : string ) : TWindows_Key ;

implementation

uses // Borland...
     SysUtils, // uppercase

     // C&C...
     VCL_Std ; // VK_To_String

// Unit API...

function String_To_Key( S : string ) : TWindows_Key ;

var Code : integer ;
    Loop : integer ;
    Shift : TShiftState ;

begin
    // Setup...
    Code := 0 ;
    Shift := [] ;
    S := uppercase( S ) ;

    // Parse
    for Loop := 0 to 2 do // Check all combinations of Shift/Control/Alt
    begin
        if( copy( S, 1, 5 ) = 'CTRL+' ) then
        begin
            Shift := Shift + [ssCtrl] ;
            S := copy( S, 6, length( S ) ) ;
        end ;
        if( copy( S, 1, 4 ) = 'ALT+' ) then
        begin
            Shift := Shift + [ssAlt] ;
            S := copy( S, 5, length( S ) ) ;
        end ;
        if( copy( S, 1, 6 ) = 'SHIFT+' ) then
        begin
            Shift := Shift + [ssShift] ;
            S := copy( S, 7, length( S ) ) ;
        end ;
    end ;
    if( S = 'CONTROL' ) then
    begin
        Shift := Shift + [ssCtrl] ;
    end else
    if( S = 'SHIFT' ) then
    begin
        Shift := Shift + [ssShift] ;
    end else
    if( S = 'ALT' ) then
    begin
        Shift := Shift + [ssAlt] ;
    end else
    if( length( S ) > 0 ) then
    begin
        Code := String_To_VK( S ) ;
    end ;

    Result := TWindows_Key.Create( Code, Shift ) ;
end ; // String_To_Key


// TWindows_Key methods...

// Constructors and destructors...

constructor TWindows_Key.Create( _Code : integer ; _Shift : TShiftState ) ;

begin
    inherited Create ;

    Code := _Code ;
    Shift := _Shift ;
end ;


function TWindows_Key.As_String : string ;

begin
    if( Code = 0 ) then // No key code
    begin
        if( Shift = [ssShift] ) then
        begin
            Result := 'SHIFT' ;
            exit ;
        end ;
        if( Shift = [ssAlt] ) then
        begin
            Result := 'ALT' ;
            exit ;
        end ;
        if( Shift = [ssCtrl] ) then
        begin
            Result := 'CONTROL' ;
            exit ;
        end ;
    end ;
    Result := '' ;
    if( ssCtrl in Shift ) then
    begin
        Result := Result + 'Ctrl+' ;
    end ;
    if( ssShift in Shift ) then
    begin
        Result := Result + 'Shift+' ;
    end ;
    if( ssAlt in Shift ) then
    begin
        Result := Result + 'Alt+' ;
    end ;
    if( Code = 0 ) then
    begin
        Result := copy( Result, 1, length( Result ) - 1 ) ; // Remove trailing "+"
    end else
    begin
        Result := Result + VK_To_String( Code ) ;
    end ;
end ; // TWindows_Key.As_String



// TKey_Map methods...

// Constructors and destructors...

constructor TKey_Map.Create ;

begin
    inherited Create ;

    Captions := TStringList.Create ;
end ;


destructor TKey_Map.Destroy ;

begin
    Captions.Free ;
    Captions := nil ;

    inherited Destroy ;
end ;


// Property handlers...

function TKey_Map.Get_Caption( Index : integer ) : string ;

begin
    Result := Captions[ Index ] ;
end ;


function TKey_Map.Get_Tag( Index : integer ) : integer ;

begin
    Result := integer( Captions.Objects[ Index ] ) ;
end ;


procedure TKey_Map.Set_Caption( Index : integer ; const Value : string ) ;

begin
    Captions[ Index ] := Value ;
end ;


procedure TKey_Map.Set_Tag( Index, Value : integer ) ;

begin
    Captions.Objects[ Index ] := pointer( Value ) ;
end ;


// API...

procedure TKey_Map.Add_Key( const Name, _Caption : string ;
    Key : TWindows_Key ; Tag : integer = 0 ) ;

begin
    AddObject( Name, Key ) ;
    Captions.AddObject( _Caption, pointer( Tag ) ) ;
end ;


function TKey_Map.Keys( Index : integer ) : TWindows_Key ;

begin
    Result := TWindows_Key( Objects[ Index ] ) ;
end ;


end.

