{
        Program Name : SCU
        Package Name : SCU
        Purpose      : SCU definition
        Institution  : Conroy & Conroy Co.
        Date Written : 16-November-2001
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2001 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************
}

unit SCUInc ;

interface

uses // Other...
     _DebugIn, // PDebug_Interface
     _UEHDefs, // TUEC
     Loggers ; // PSLogger

{$I EDEFINES.INC}

const SCUErr_Facility = 22 ;
      SCUErr_Success = 0 ;
      SCUErr_Value_Not_Found = 1 ;
      SCUErr_Invalid_Size = 2 ;
      SCUErr_Exists = 3 ; { Key or value already exists }
      SCUErr_Alarm_Failure = 4 ; {Operation caused an alarm which user used to abort operation }

{$IFDEF DELPHI}
type T_SCU = class
		public { API... }
		    { Instance support... }
		    function Initialize : TUEC ; virtual ; stdcall ; abstract ;
		    procedure Terminate ; virtual ; stdcall ; abstract ;

                    function Init( Name : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Version : longint ; virtual ; stdcall ; abstract ;

		    function Write( Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Read( Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Read_Size( Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Read_Type( Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Exists( Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
		    function Key_Exists( Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;

		    function Lookup( Path : PChar ; DT : longint ) : PChar ;
                        virtual ; stdcall ; abstract ;

		    function Delete( Path : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;
		    function Create( Path : PChar ; DT : longint ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Create_And_Write( Path : PChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Rename( Old, New : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;

		    function Set_Value( User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Get_Value( User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Value_Size( User, Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Value_Type( User, Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Value_Exists( User, Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
		    function Delete_Value( User, Path : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;
		    function Create_Value( User, Path : PChar ;
                        DT : longint ) : TUEC ; virtual ; stdcall ; abstract ;
                    function Verify : TUEC ; virtual ; stdcall ; abstract ;
                    function Debugger : PDebug_Interface ;
                        virtual ; stdcall ; abstract ;

                    function Read_Value( User, Path : PChar ;
                        var DT, Size : longint ; var Buffer : PChar) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    { Reads the specified value.  If result indicates a success,
                      DT is the type of the data, Size is the actual data size,
                      and Buffer is a pointer to a buffer containing the data.
                      DT, Size, and Buffer are ignored on call, and set on
                      return.

                      This function serves as a single call to: 1) determine if
                      the value exists, 2) determine what type it is, and
                      3) read the value's data. }

                    function Init_And_Extend( Name : PChar ;
                        Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes. }

                    function Delete_With_Callback( Path : PChar ;
                        Logger : PSLogger ) : TUEC ;
                        virtual ; stdcall ; abstract ;

                { V1.2 interface: }
                public { Contextual version of instance support... }
                    function Init_Context( Name : PChar ; Flags : longint ) : longint ;
                        virtual ; stdcall ; abstract ;
                    procedure Terminate_Context( Context : longint ) ;
                        virtual ; stdcall ; abstract ;
                    function Init_And_Extend_Context( Name : PChar ;
                        Size : longint ; Flags : longint ) : longint ;
                        virtual ; stdcall ; abstract ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes.  Flags and 1 = turn on caching }

                public { Contextual version of Low-level API... }
		    function Write_Context( Context : longint ; Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Read_Context( Context : longint ; Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Read_Size_Context( Context : longint ; Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Read_Type_Context( Context : longint ; Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Exists_Context( Context : longint ; Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
		    function Key_Exists_Context( Context : longint ; Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;

		    function Lookup_Context( Context : longint ; Path : PChar ; DT : longint ) : PChar ;
                        virtual ; stdcall ; abstract ;

		    function Delete_Context( Context : longint ; Path : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;
		    function Create_Context( Context : longint ; Path : PChar ; DT : longint ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Create_And_Write_Context( Context : longint ; Path : PChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Rename_Context( Context : longint ; Old, New : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;

                public { Contextual version of Standard API... }
		    function Set_Value_Context( Context : longint ; User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Get_Value_Context( Context : longint ; User, Path : PChar ; Data : pointer ;
			Size : longint ) : TUEC ; virtual ; stdcall ; abstract ;
		    function Value_Size_Context( Context : longint ; User, Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Value_Type_Context( Context : longint ; User, Path : PChar ) : longint ;
                        virtual ; stdcall ; abstract ;
		    function Value_Exists_Context( Context : longint ; User, Path : PChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
		    function Delete_Value_Context( Context : longint ; User, Path : PChar ) : TUEC ;
                        virtual ; stdcall ; abstract ;
		    function Create_Value_Context( Context : longint ; User, Path : PChar ;
                        DT : longint ) : TUEC ; virtual ; stdcall ; abstract ;
                    function Verify_Context( Context : longint ) : TUEC ;
                        virtual ; stdcall ; abstract ;
                    function Debugger_Context( Context : longint ) : PDebug_Interface ;
                        virtual ; stdcall ; abstract ;

                    function Read_Value_Context( Context : longint ; User, Path : PChar ;
                        var DT, Size : longint ; var Buffer : PChar) : TUEC ;
                        virtual ; stdcall ; abstract ;

                    function Delete_With_Callback_Context( Context : longint ;
                        Path : PChar ; Logger : PSLogger ) : TUEC ;
                        virtual ; stdcall ; abstract ;
            end ;
{$ENDIF}

const SCUDT_Undefined = 0 ;
      SCUDT_Key = 1 ;
      SCUDT_String = 2 ;
      SCUDT_Integer = 3 ; { 4-bit signed }
      SCUDT_Boolean = 4 ;
      SCUDT_Float = 5 ; { Double-precision }
      SCUDT_Binary = 6 ;
      SCUDT_All = -1 ; { All types }

implementation

end.
