{$N+}
{
        Program Name : Standard
        Package Name : LB
        Purpose      : Standard Pascal routines
        Institution  : Conroy & Conroy Co.
        Date Written : 28-Nov-88
        Written By   : Alan Conroy
        Version      : 2.1I

        Copyright (c) 1988-2007 by Alan Conroy.  Released to the Public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        * 21-Mar-2002   EAC            Made Edit work more like *
        *                              BASIC-Plus.              *
        *                                                       *
        *  1-Nov-2002   EAC            Moved Boolean_To_String  *
        *                              to ustrings unit.        *
        *                              Added Format_Time,       *
        *                              Time_To_String, and      *
        *                              Parse_Time.              *
        *  6-Nov-2002   EAC            Moved time functions to  *
        *                              dates unit.              *
        *                                                       *
        *  1-Apr-2005   EAC            Refactoring.             *
        *                                                       *
        *  9-Aug-2006   EAC            Fix insert in            *
        *                              TInteger_List64.         *
        *                                                       *
        * 17-Jan-2007   EAC            Added Tag to             *
        *                              TInteger_List64.         *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit contains standard Pascal support routines for Conroy &
        Conroy Co.
}

{$I EDEFINES.INC}

{$IFDEF DELPHI}
{$DEFINE WINDOWS}
{$ENDIF}

unit Standard ;

interface

uses
{$IFDEF DELPHI}
     { Borland... }
     Classes,
{$ENDIF}
     { C&C... }
     _DebugIn, { PSDebug_Interface }
     TypeDefs ;

var Bit_Values : array[ 0..31 ] of longint ;
{$IFDEF DELPHI}
    XBit_Values : array[ 0..63 ] of int64 ;
{$ENDIF}

{ Standard function declarations... }
procedure CC_Log( const X : string ) ;
procedure Clear_IOResult ;
function Command_Line : string ;
function Compare_Memory( var X, Y ; Size : longint ) : boolean ; { Compare memory }
procedure Error ; { Error handler }
function Ert( Err : Integer ) : string ; { Return error text }

{ Return position of extension in passed filename.  Returns 0 if no extension. }
function Extension_Pos( const X : string ) : integer ;

{ Return position of extension in passed filename.  Returns length + 1 if no
  extension. }
function Implicit_Extension_Pos( const X : string ) : integer ;

function Get_Serial_Number : longint ;
function OS_ID : string ; { Returns name and version of current operating system }
function PID : longint ; { Current process ID }
{$IFNDEF Windows}
function MSDOS_Error : Integer ; { Return DOS extended error code }
{$ENDIF}
{$IFDEF DELPHI}
function MaxAvail : longint ;
{$ENDIF}
function Bits_For_Value( Value : longint ) : integer ;
{$IFDEF DELPHI}
function XBits_For_Value( Value : int64 ) : integer ;
{$ENDIF}

var Exitsave : Pointer ;

{$IFDEF DEBUG}
procedure Test_Unit ;
{$ENDIF}

{$IFDEF DELPHI}
type TInteger64_List = class( TList )
                           protected { Property handlers }
                               function Get_Index( Index : integer ) : int64 ;
                                   virtual ;
                               procedure Set_Index( Index : integer ;
                                   Value : int64 ) ; virtual ;

                           public { API... }
                               Tag : integer ;

                               function Add( Value : int64 ) : integer ;
                                   virtual ;
                               function Count : integer ;
                               procedure Delete( Index : integer ) ;
                               procedure Remove( Value : int64 ) ;
                               procedure Increment( Index : integer ) ;
                               function IndexOf( Value : int64 ) : integer ;
                               procedure Insert( Index : integer ;
                                   Value : int64 ) ;
                               function Index_With_Max : integer ;
                               procedure Sort( Mask : int64 ) ;

                               property Items[ Index : integer ] : int64
                                   read Get_Index
                                   write Set_Index ; default ;
                       end ;
{$ENDIF}

implementation

uses { Borland... }
{$IFDEF DELPHI}
     Windows,
     SysUtils,

     { Conroy & Conroy Co. }
     ASCIIDef,

{$ELSE}
     Strings, { strpas }

     { Conroy & Conroy Co. }
     TSRS,
{$ENDIF}
     _AEL, { TApplication_Event_Logger }
     _App, { Application_Manager }
     CommonUt, { Edit }
     Compatib, { FileExists }
     Dates, { Time }
     ErrMastr, { Error_Code }
     Num1S, { Num1 }
     O_S, { OS }
     _UEHDefs, { TUEC }
     UEHDefs ; { tError_Interface }

{$IFDEF DELPHI}
function MaxAvail : longint ;

var Status : TMemoryStatus ;
    Temp : cardinal ;

begin
    GlobalMemoryStatus( Status ) ;
    Temp := Status.dwAvailVirtual ;
    if( Temp > $7FFFFFFF ) then
    begin
        Temp := $7FFFFFFF ;
    end ;
    MaxAvail := Temp ;
end ;
{$ENDIF}


function Bits_For_Value( Value : longint ) : integer ;

var Loop : integer ;

begin
    if( Value < 0 ) then
    begin
        Value := -Value ;
    end ;

    for Loop := 0 to 31 do
    begin
        if( Value <= Bit_Values[ Loop ] ) then
        begin
            Bits_For_Value := Loop + 1 ;
            exit ;
        end ;
    end ;
    Bits_For_Value := 32 ;
end ;


{$IFDEF DELPHI}
function XBits_For_Value( Value : int64 ) : integer ;

var Loop : integer ;

begin
    if( Value < 0 ) then
    begin
        Value := -Value ;
    end ;

    for Loop := 0 to 63 do
    begin
        if( Value <= XBit_Values[ Loop ] ) then
        begin
            XBits_For_Value := Loop + 1 ;
            exit ;
        end ;
    end ;
    XBits_For_Value := 32 ;
end ;
{$ENDIF}


{$IFDEF Debug}
{var Allocation_Count : integer ;}


procedure Start_Memory_Test ;

begin
    {TODO}
end ;


procedure End_Memory_Test ;

begin
    {TODO}
end ;
{$ENDIF}


var Pascal_Name : array[ 0..127 ] of char ;

var Pascal_Error_Text : array[ 0..127 ] of char ;

function Pascal_ERT( Code : integer ) : string ;

var Err_String : string ;
{$IFDEF DELPHI}
    R : TUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : TUEC ;

begin
    case Code of
        { Runtime errors... }
        0 : Err_String := 'Success' ;
        100 : Err_String := 'Disk read error' ;
        101 : Err_String := 'Disk write error' ;
        102 : Err_String := 'File not assigned' ;
        103 : Err_String := 'File not open' ;
        104 : Err_String := 'File not open for input' ;
        105 : Err_String := 'File not open for output' ;
        106 : Err_String := 'Invalid numeric format' ;
        150..161 : begin
                        R := Error_Code( 1, Code - 131 ) ;
                        UEC := tUEC( R ) ;
                        Err_String := strpas( Get_Error_Text( UEC, -1 ) ) ;
                   end ;
        162 : Err_String := 'Hardware failure' ;

        { Fatal errors... }
        200 : Err_String := 'Division by zero' ;
        201 : Err_String := 'Range check error' ;
        202 : Err_String := 'Stack overflow error' ;
        203 : Err_String := 'Heap overflow error' ;
        204 : Err_String := 'Invalid pointer operation' ;
        205 : Err_String := 'Floating point overflow' ;
        206 : Err_String := 'Floating point underflow' ;
        207 : Err_String := 'Invalid floating point operation' ;
        208 : Err_String := 'Overlay manager not installed' ;
        209 : Err_String := 'Overlay file read error' ;
        210 : Err_String := 'Object not initialized' ;
        211 : Err_String := 'Call to abstract method' ;
        212 : Err_String := 'Stream registration error' ;
        213 : Err_String := 'Collection index out of range' ;
        214 : Err_String := 'Collection overflow error' ;
        215 : Err_String := 'Arithmetic overflow' ;
        216 : Err_String := 'Access violation' ;
        217 : Err_String := 'Control-C' ;
        218 : Err_String := 'Privileged instruction' ;
        219 : Err_String := 'Invalid typecast' ;
        220 : Err_String := 'Invalid variant typecast' ;
        221 : Err_String := 'Invalid variant operation' ;
        222 : Err_String := 'No variant method call dispatcher' ;
        223 : Err_String := 'Cannot create variant array' ;
        224 : Err_String := 'Variant does not contain array' ;
        225 : Err_String := 'Variant array bounds error' ;
        226 : Err_String := 'TLS initialization error' ;
        227 : Err_String := 'Assertion failed' ;
        228 : Err_String := 'Interface Cast Error' ;
        229 : Err_String := 'Safecall error' ;
        230 : Err_String := 'Unhandled exception' ;
        231 : Err_String := 'Too many nested exceptions' ;
        232 : Err_String := 'Fatal signal raised on a non-Delphi thread' ;
        else
        begin
            R := Error_Code( 1, Code ) ;
            UEC := tUEC( R ) ;
            Err_String := strpas( Get_Error_Text( UEC, -1 ) ) ;
        end ;
    end ;
    Pascal_ERT := Err_String ;
end ; { Pascal_ERT }


{$IFDEF DELPHI}
type tPascal_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; override ;
                                function Get_Name : PChar ; override ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ;
                                procedure Done ; override ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                  end ;

function tPascal_EI.Get_Facility : longint ;

begin
    Get_Facility := 2 ;
end ;


function tPascal_EI.Get_Name : PChar ;

var S : string ;

begin
    S := 'Delphi RT' ;
    move( S[ 1 ], Pascal_Name[ 0 ], length( S ) ) ;
    Get_Name := @Pascal_Name ;
end ;


function tPascal_EI.Translate_Error( Code : longint ) : PChar ;

var Err_String : string ;

begin
    Err_String := Pascal_ERT( Code ) + #0 ;
    move( Err_String[ 1 ], Pascal_Error_Text[ 0 ], length( Err_String ) ) ;
    Translate_Error := @Pascal_Error_Text ;
end ; { tPascal_EI.Translate_Error }


procedure tPascal_EI.Done ;

begin
    Free ;
end ;


function tPascal_EI.Facility_Version : longint ;

begin
    Facility_Version := 100 ; { Delphi (Pascal V10.0) }
end ;


function tPascal_EI.Version : longint ;

begin
    Version := 10 ; { Complliant with UEH V1.0 specification }
end ;
{$ENDIF}



type tSPascal_EI = object( TSStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : string ; virtual ;
                                function Translate_Error( Code : longint ) : string ;
                                    virtual ;
                                function Facility_Version : longint ; virtual ;
                                function Version : longint ; virtual ;
                  end ;
     PSPascal_EI = ^tSPascal_EI ;

function tSPascal_EI.Get_Facility : longint ;

begin
    Get_Facility := 2 ;
end ;


function tSPascal_EI.Get_Name : string ;

begin
    Get_Name := 'Pascal RT' ;
end ;


function tSPascal_EI.Translate_Error( Code : longint ) : string ;

var Err_String : string ;

begin
    Err_String := Pascal_ERT( Code ) ;
    Translate_Error := Err_String ;
end ; { tSPascal_EI.Translate_Error }


function tSPascal_EI.Facility_Version : longint ;

begin
    Facility_Version := 100 ; { Delphi (Pascal V10.0) }
end ;


function tSPascal_EI.Version : longint ;

begin
    Version := 10 ; { Compliant with UEH V1.0 specification }
end ;



{ Unit API... }

procedure Clear_IOResult ;

{$IFNDEF Delphi}
var Dummy : integer ;
{$ENDIF}

begin
{$IFNDEF Delphi}
    Dummy :=
{$ENDIF}
    IOResult ;
end ;


function Command_Line : string ;

begin
    Command_Line := OS^.Command_Line ;
end ;


{$IFNDEF Windows}
function MSDOS_Error : Integer ;

var UEC : TUEC ;

begin
    OS^.Last_Error( UEC ) ;
    MSDOS_Error := UEC.Code ;
end ;
{$ENDIF}


function Ert( Err : Integer ) : string ; { Return Pascal error text }

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
    R := Error_Code( 2, Err ) ;
    UEC := tUEC( R ) ;
    Ert := strpas( Get_Error_Text( UEC, -1 ) ) ;
end ;


procedure Error ; { Standard error handler procedure }

begin
    Exitproc := Exitsave ; { Restore old vector }
    if( Erroraddr = nil ) then
    begin
        Exit ; { Normal exit }
    end ;
    Erroraddr := nil ; { We are reporting the error }
    Writeln( Ert( Exitcode ) )
end ;


var Have_CCLog : boolean ;
    CCLog : string ;

procedure CC_Log( const X : string ) ; { Log a message }

var F : text ;
    S : string ;

begin
{$IFDEF DELPHI}
  try
{$ENDIF}
    if( not Have_CCLog ) then
    begin
        Have_CCLog := True ;
        CCLog := OS^.Get_Symbol( 'CCLOG' ) ;
    end ;
    if( length( CCLog ) > 0 ) then { Symbol is defined }
    begin
        assign( F, CCLog ) ;
        if( FileExists( CCLog ) ) then
        begin
            {$I-}
            append( F ) ;
            {$I+}
        end else
        begin
            {$I-}
            rewrite( F ) ;
            {$I+}
        end ;
        if( IOResult <> 0 ) then
        begin                                                                                        
            exit ; { Error opening file }
        end ;

        S := Application_Manager^.Get_Name ;
        if( length( S ) > 0 ) then
        begin
            S := S + ' ' ;
        end ;
        if( length( Application_Manager^.Get_Version ) > 0 ) then
        begin
            S := S + 'V' + Application_Manager^.Get_Version ;
        end ;
        {$I-}
        writeln( F, S, ' on ', Date, ' at ', Time, ': ', X ) ;
        {$I+}
        Clear_IOResult ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
    end ; { if( length( S ) > 0 ) }
{$IFDEF DELPHI}
  except { Don't let exception here end program }
  end ;
{$ENDIF}
end ; { CC_Log }


function Compare_Memory( var X, Y ; Size : longint ) : boolean ; { Compare memory }

begin
    Compare_Memory := True ; { Assume success }
    while( Size > 0 ) do
    begin
        dec( Size ) ;
        if( tByte_Array( X )[ Size ] <> tByte_Array( Y )[ Size ] ) then
        begin
            Compare_Memory := False ;
            exit ;
        end ;
    end ;
end ;


function Get_Serial_Number : longint ;

begin
    if( Application_Manager <> nil ) then
    begin
        Get_Serial_Number := Application_Manager^.Get_Serial_Number( 0 ) ;
    end else
    begin
        Get_Serial_Number := 0 ;
    end ;
end ;


function Extension_Pos( const X : string ) : integer ;
{ Return position of dot in filename extension, 0 if none. }

var Dummy : integer ;

begin
    Dummy := length( X ) ; { Start at end of filename }
    while( ( Dummy > 0 ) and ( X[ Dummy ] <> '.' ) and ( X[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ; { Work our way backwards until found or done }
    end ;
    if( ( Dummy > 0 ) and ( X[ Dummy ] <> '.' ) ) then { Not found }
    begin
        Dummy := 0 ;
    end ;
    Extension_Pos := Dummy ; { Return result }
end ;


function Implicit_Extension_Pos( const X : string ) : integer ;

var Dummy : integer ;

begin
    Dummy := Extension_Pos( X ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := length( X ) + 1 ;
    end ;
    Implicit_Extension_Pos := Dummy ;
end ;


function OS_ID : string ; { Returns name and version of current operating system }

begin
    OS_ID := OS^.Name + ' V' + OS^.Version ;
end ;


function PID : longint ; { Current process ID }

begin
    PID := OS^.PID ;
end ;


{ Unit functions }
procedure Load_Bit_Values ;

var D : integer ;

begin
    Bit_Values[ 0 ] := 1 ; { 2^0 }
{$IFDEF DELPHI}
    XBit_Values[ 0 ] := 1 ;
{$ENDIF}
    for D := 1 to 31 do
    begin
        Bit_Values[ D ] := Bit_Values[ D - 1 ] shl 1 ;
{$IFDEF DELPHI}
        XBit_Values[ D ] := XBit_Values[ D - 1 ] + XBit_Values[ D - 1 ] ;
{$ENDIF}
    end ;
{$IFDEF DELPHI}
    for D := 32 to 63 do
    begin
        XBit_Values[ D ] := XBit_Values[ D - 1 ] + XBit_Values[ D - 1 ] ;
    end ;
{$ENDIF}
end ;


type TStandard_Application_Event_Logger = object( TSApplication_Event_Logger )
                                              public { Constructors and destructors... }
                                                  constructor Init ;
                                                  destructor Done ; virtual ;

                                              public { API... }
                                                  Old_Event_Logger : PSApplication_Event_Logger ;

                                                  procedure Log_Event( const S : string ) ;
                                                      virtual ;
                                                  procedure Trace_Enter( const S : string ) ;
                                                      virtual ;
                                                  procedure Trace_Exit( const S : string ) ;
                                                      virtual ;
                                          end ;
     PStandard_Application_Event_Logger = ^TStandard_Application_Event_Logger ;

{ TStandard_Application_Event_Logger methods... }

{ Constructors and destructors... }

constructor TStandard_Application_Event_Logger.Init ;

begin
    inherited Init ;

    Old_Event_Logger := nil ;
end ;


destructor TStandard_Application_Event_Logger.Done ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Detach ;
        Old_Event_Logger := nil ;
    end ;

    inherited Done ;
end ;


{ API... }

procedure TStandard_Application_Event_Logger.Log_Event( const S : string ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Log_Event( S ) ;
    end ;
    CC_Log( S ) ;
end ;


procedure TStandard_Application_Event_Logger.Trace_Enter( const S : string ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Trace_Enter( S ) ;
    end ;
    CC_Log( 'Entering ' + S ) ;
end ;


procedure TStandard_Application_Event_Logger.Trace_Exit( const S : string ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Trace_Exit( S ) ;
    end ;
    CC_Log( 'Exiting ' + S ) ;
end ;


{$IFDEF DELPHI}
{ TInteger64_List methods... }

{ Property handlers... }

function TInteger64_List.Get_Index( Index : integer ) : int64 ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    Res._Low := integer( inherited Items[ Index * 2 ] ) ;
    Res._High := integer( inherited Items[ Index * 2 + 1 ] ) ;
    Result := Res._Full ;
end ;


procedure TInteger64_List.Set_Index( Index : integer ; Value : int64 ) ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer) ;
          end ;

begin
    Res._Full := Value ;
    inherited Items[ Index * 2 ] := pointer( Res._Low ) ;
    inherited Items[ Index * 2 + 1 ] := pointer( Res._High ) ;
end ;


function TInteger64_List.Add( Value : int64 ) : integer ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    if( Count = Capacity ) then { Need to expand }
    begin
        if( Count < 1024 ) then
        begin
            Capacity := Capacity * 4 ;
        end else
        begin
            Capacity := Capacity + 1024 ;
        end ;
    end ;
    Res._Full := Value ;
    inherited Add( pointer( Res._Low ) ) ;
    inherited Add( pointer( Res._High ) ) ;
    Result := Count - 1 ;
end ;


function TInteger64_List.Count : integer ;

begin
    Result := inherited Count ;
    if( Result > 0 ) then
    begin
        Result := Result div 2 ;
    end ;
end ;


procedure TInteger64_List.Delete( Index : integer ) ;

begin
    inherited Delete( Index * 2 ) ;
    inherited Delete( Index * 2 ) ;
end ;


procedure TInteger64_List.Remove( Value : int64 ) ;

var Dummy : integer ;

begin
    Dummy := IndexOf( Value ) ;
    if( Dummy <> -1 ) then
    begin
        Delete( Dummy ) ;
    end ;
end ;


procedure TInteger64_List.Increment( Index : integer ) ;

begin
    Items[ Index ] := Items[ Index ] + 1 ;
end ;


function TInteger64_List.IndexOf( Value : int64 ) : integer ;

var Loop : integer ;

begin
    IndexOf := -1 ;
    for Loop := 0 to Count - 1 do
    begin
        if( Items[ Loop ] = Value ) then
        begin
            IndexOf := Loop ;
            exit ;
        end ;
    end ;
end ;


procedure TInteger64_List.Insert( Index : integer ; Value : int64 ) ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    Res._Full := Value ;
    Index := Index * 2 ; // Convert from our 64-bit index to the 32-bit index
    inherited Insert( Index, pointer( Res._High ) ) ; // Insert high
    inherited Insert( Index, pointer( Res._Low ) ) ; // Insert low before High
end ;


function TInteger64_List.Index_With_Max : integer ;

var Loop : integer ;
    Max : int64 ;

begin
    Result := -1 ;
    Max := -( $7FFFFFFFFFFFFFFF ) ;
    for Loop := 0 to Count - 1 do
    begin
        if( Items[ Loop ] > Max ) then
        begin
            Max := Items[ Loop ] ;
            Result := Loop ;
        end ;
    end ;
end ;


procedure TInteger64_List.Sort( Mask : int64 ) ;

var Loop, Loop1 : integer ;
    I, I1, M, M1 : int64 ;

begin
    for Loop := 0 to Count - 1 do
    begin
        I := Items[ Loop ] ;
        M := I and Mask ;
        for Loop1 := Loop to Count - 1 do
        begin
            I1 := Items[ Loop1 ] ;
            M1 := I1 and Mask ;
            if( M > M1 ) then
            begin
                Items[ Loop ] := I1 ;
                Items[ Loop1 ] := I ;
                I := I1 ;
                M := M1 ;
            end ;
        end ;
    end ;
end ;
{$ENDIF}


{$IFDEF DELPHI}
var PEI : TPascal_EI ;
{$ELSE}
var PEI : pSPascal_EI ;
{$ENDIF}
    AEL : PStandard_Application_Event_Logger ;

procedure Unit_Initialize ;

begin
    { Log our entry... }
    AEL := nil ;
    if( Application_Manager <> nil ) then
    begin
        AEL := new( PStandard_Application_Event_Logger, Init ) ;
        AEL^.Old_Event_Logger := Application_Manager^.Get_Event_Logger ;
        AEL^.Old_Event_Logger^.Attach ;
        Application_Manager^.Set_Event_Logger( AEL ) ;
    end ;

    Have_CCLog := False ;
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Enter( 'Standard.Unit_Initialize' ) ;
    end ;

    { Initialize data... }
    Load_Bit_Values ;
{$IFDEF DELPHI}
    PEI := TPascal_EI.Create ;
{$ELSE}
    new( PEI, Init ) ;
{$ENDIF}
    if( not Register_Error_Interface( PSError_Interface( PEI ) ) ) then
    begin
{$IFDEF DELPHI}
        PEI.Free ;
{$ELSE}
        dispose( PEI, Done ) ;
{$ENDIF}
        PEI := nil ;
    end ;
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Exit( 'Standard.Unit_Initialize' ) ;
    end ;
end ; { Unit_Initialize }


procedure Unit_Finalization ;

begin
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Enter( 'Standard.Unit_Finalization' ) ;
    end ;
    if( PEI <> nil ) then
    begin
{$IFDEF DELPHI}
        Unregister_Error_Interface( PEI.Get_Facility ) ;
        PEI.Free ;
{$ELSE}
        Unregister_Error_Interface( PEI^.Get_Facility ) ;
        dispose( PEI, Done ) ;
{$ENDIF}
        PEI := nil ;
    end ;
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Exit( 'Standard.Unit_Finalization' ) ;
    end ;
end ;


{$IFDEF DEBUG}
procedure Test_Unit ;

    procedure Test_Edit ;

    begin
        { Remove leading/trailing spaces from a string... }
        if( Edit( '', 8 + 128 ) <> '' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' ', 8 + 128 ) <> '' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a b', 8 + 128 ) <> 'a b' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' ab', 8 + 128 ) <> 'ab' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( '  ', 8 + 128 ) <> '' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'ab ', 8 + 128 ) <> 'ab' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'ab', 32 ) <> 'AB' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a"b"', 32 ) <> 'A"B"' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a"b"c', 32 ) <> 'A"B"C' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' a"b"c ', 32 + 8 ) <> 'A"B"C ' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' a"b"c ', 32 + 128 ) <> ' A"B"C' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a"b"', 32 + 256 ) <> 'A"b"' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a"b"c', 32 + 256 ) <> 'A"b"C' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' a"b"c ', 32 + 8 + 256 ) <> 'A"b"C ' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( ' a"b"c ', 32 + 128 + 256 ) <> ' A"b"C' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a" ', 128 ) <> 'a"' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        if( Edit( 'a" ', 128 or 256 ) <> 'a" ' ) then
        begin
            raise Exception.Create( 'Edit failed' ) ;
        end ;
        {TODO: Need to test entire range of features }
    end ;

begin { Test_Unit }
    Test_Edit ;
end ;
{$ENDIF}

{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialize ;

{$IFDEF DELPHI}
finalization
    Unit_Finalization ;
{$ENDIF}
end.
