{$R-}
unit Symbols ;
{
        Program Name : Symbols
        Package Name : LB
        Purpose      : Symbol Table Manager
        Institution  : Conroy & Conroy Co.
        Date Written : 21-Jan-2002
        Written By   : Alan Conroy
        Version      : 1.0B

        Copyright (C) 2002-2004 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1-Nov-2002    EAC         Use TObject from common and    *
	*                            descend TSymbol_Deletion_Logger*
	*			     and TSymbol_Table from it.     *
        *                            Added TSymbol_Table.ERT.       *
        *                            Added PSymbol_Table.           *
        *  7-Dec-2004    EAC         Fix failure in Find_Next.      *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	Symbol table manager.
}

interface

uses Collect, { TCollection }
     Common, { TBase_Object }
     Compatib ; { SmallInt }

type String255 = string[ 255 ] ;

     PSymbol_Entry = ^TSymbol_Entry ;
     TSymbol_Entry = record
                        Name : ^String255 ; { Pointer to symbol name }
                        Address : int64 ; { Address of symbol }
                        Flags : longint ; { Flags }
			Data_Type : longint ; { Data type of symbol }
			Size : longint ; { Size of symbol, in bytes }
                        B_Link, L_Link, R_Link : PSymbol_Entry ;
                        { Back, left and right pointers to next symbols }
                        Data : pointer ; { Pointer to extended symbol data }
                    end ;

     TSymbol_Deletion_Logger = object( TBase_Object )
                                   public { API... }
                                       procedure Deletion_Notification( Symbol : PSymbol_Entry ) ;
                                           virtual ;
                               end ;
     PSymbol_Deletion_Logger = ^TSymbol_Deletion_Logger ;

     TSymbol_Table = object( TBase_Object )
                         private { Instance data... }
                             _Delete_Logger : PSymbol_Deletion_Logger ;
                             _Root : TCollection ; { Roots of symbol tables }

                         public { Constructors and destructors }
                             constructor Init( Delete_Logger : PSymbol_Deletion_Logger ) ;
                             destructor Done ; virtual ;

                         private { Internal utility routines... }
                             function _Add( Sym : string ; Addr : int64 ;
                                 Flg, D_T, Siz : longint ;
                                 Dat : pointer ; Table : PSymbol_Entry ) : integer ;
                             function _Find( Sym : string ; var Addr : int64 ;
                                 var Flg, D_T, Siz : integer ;
                                 var Dat : pointer ; Table : PSymbol_Entry ) : integer ;
                             function _Modify( Sym : string ; Addr : int64 ;
                                 Flg, D_T, Siz : longint ;
                                 Dat : pointer ; Table : PSymbol_Entry ) : integer ;
                             procedure _Delete( const Sym : string ;
                                 Table : PSymbol_Entry ) ;

                         public { API... }
                             Last_Symbol : PSymbol_Entry ;
                             Max_Symbol_Length : integer ;
                             Case_Sensitive : boolean ;

                             function Add( const Sym : string ; Addr : int64 ;
                                 Flg, D_T, Siz : longint ;
                                 Dat : pointer ) : integer ; virtual ;
                             function Find( const Sym : string ; var Addr : int64 ;
                                 var Flg, D_T, Siz : longint ;
                                 var Dat : pointer ) : integer ; virtual ;
                             function Find_Next( var Addr : int64 ;
                                 var Flg, D_T, Siz : longint ;
                                 var Dat : pointer ) : integer ; virtual ;
                             function Modify( Sym : string ; Addr : int64 ;
                                 Flg, D_T, Siz : longint ;
                                 Dat : pointer ) : integer ; virtual ;
                             procedure Delete( const Sym : string ) ; virtual ;
                             procedure Push_Level ;
                             procedure Pop_Level ;
                             procedure Clear ;
                             function ERT( X : integer ) : string ;
                     end ;
     PSymbol_Table = ^TSymbol_Table ;

function SERT( X : Integer ) : string ;

implementation

uses { C&C... }
     CommonUt, { Edit }
     XLate ; { Wildcard_Match }


function SERT( X : Integer ) : string ;
{ Return text of error for symbol table manager error number }

begin
    case X of
        0 : SErt := 'Success' ;
        1 : Sert := '?Invalid symbol name' ;
        2 : Sert := '?Reserved name' ;
        3 : Sert := '?Multiply defined symbol' ;
        4 : Sert := '?Undefined reference' ;
        5 : SErt := '?No more matches' ;
        else SERT := '?Unknown symbol error' ;
    end
end ;


{ TSymbol_Table methods... }

{ Constructors and destructors }

constructor TSymbol_Table.Init( Delete_Logger : PSymbol_Deletion_Logger ) ;

begin
    Max_Symbol_Length := 16 ;
    _Root.Init( 0, 16 ) ;
    _Delete_Logger := Delete_Logger ;
end ;


destructor TSymbol_Table.Done ;

begin
    Clear ;
    _Root.Done ;
end ;


{ Internal utility methods... }

function TSymbol_Table._Add( Sym : string ; Addr : int64 ; Flg, D_T, Siz : longint ;
    Dat : pointer ; Table : PSymbol_Entry ) : integer ;

label Loop, Create ;

var Current, Previous : PSymbol_Entry ;

begin
    _Add := 0 ;

    Sym := Copy( Sym, 1, Max_Symbol_Length ) ;
    if( not Case_Sensitive ) then
    begin
        Sym := Edit( Sym, 32 ) ;
    end ;

    Current := Table ; { Start at root }
    Previous := nil ;
Loop:
    if( Current <> nil ) then { Not yet at end }
    begin
        if( Current^.Name^ = Sym ) then { Found symbol in table }
        begin
            Result := 3 ; { Multiple definition }
            exit ;
        end ;
        Previous := Current ;
        if( Current^.Name^ > Sym ) then
        begin
            Current := Current^.L_Link ;
        end else
        begin
	    Current := Current^.R_Link ; { Follow link }
        end ;
        goto Loop ;
    end ;

Create:
    New( Current ) ; { Create pointer }
    if( Previous <> nil ) then
    begin
        if( Previous^.Name^ > Sym ) then
        begin
            Previous^.L_Link := Current ;
        end else
        begin
            Previous^.R_Link := Current ;
        end ;
    end ;
    Current^.L_Link := nil ;
    Current^.R_Link := nil ;
    Current^.B_Link := Previous ;
    Getmem( Current^.Name, Length( Sym ) + 1 ) ;
    Current^.Name^ := Sym ;
    Current^.Address := Addr ;
    Current^.Flags := Flg ;
    Current^.Data_Type := D_T ;
    Current^.Size := Siz ;
    Current^.Data := Dat ;
    Last_Symbol := Current ;
end ; { TSymbol_Table._Add }


function TSymbol_Table._Find( Sym : string ; var Addr : int64 ;
    var Flg, D_T, Siz : longint ;
    var Dat : pointer ; Table : PSymbol_Entry ) : integer ;

label Loop ;

var Current : PSymbol_Entry ;

begin
    _Find := 0 ; { Assume success }
    Sym := Copy( Sym, 1, Max_Symbol_Length ) ;
    if( not Case_Sensitive ) then
    begin
        Sym := Edit( Sym, 32 ) ;
    end ;

    if( Last_Symbol <> nil ) then { Cached symbol}
    begin
        if( Last_Symbol^.Name <> nil ) then
        begin
            if( WC_Match( Sym, Last_Symbol^.Name^ ) ) then
            begin
                Addr := Last_Symbol^.Address ;
                Flg := Last_Symbol^.Flags ;
                D_T := Last_Symbol^.Data_Type ;
                Siz := Last_Symbol^.Size ;
                Dat := Last_Symbol^.Data ;
                exit ;
            end ;
        end ;
    end ;

    Current := Table ; { Start at root }
Loop:
    if( Current <> nil ) then { Not yet at end }
    begin
        if( WC_Match( Sym, Current^.Name^ ) ) then { Found symbol in table }
        begin
            Addr := Current^.Address ;
            Flg := Current^.Flags ;
            D_T := Current^.Data_Type ;
            Siz := Current^.Size ;
            Dat := Current^.Data ;
            Last_Symbol := Current ;
            exit ;
        end ;
        if( Current^.Name^ > Sym ) then
        begin
            Current := Current^.L_Link ;
        end else
        begin
	    Current := Current^.R_Link ; { Follow link }
        end ;
        goto Loop ;
    end ;

    Result := 4 ;
end ; { TSymbol_Table._Find }


function TSymbol_Table._Modify( Sym : string ;
    Addr : int64 ; Flg, D_T, Siz : longint ;
    Dat : pointer ; Table : PSymbol_Entry ) : integer ;

label Loop ;

var A : int64 ;
    F, D, S, Dummy : longint ;
    Da : pointer ;

begin
    Dummy := _Find( Sym, A, F, D, S, Da, Table ) ;
    _Modify := Dummy ;
    if( Dummy <> 0 ) then
    begin
        exit ;
    end ;

    Last_Symbol^.Address := Addr ;
    Last_Symbol^.Flags := Flg ;
    Last_Symbol^.Data_Type := D_T ;
    Last_Symbol^.Size := Siz ;
    Last_Symbol^.Data := Dat ;
end ; { TSymbol_Table._Modify }


{ API... }

function TSymbol_Table.Add( const Sym : string ; Addr : int64 ;
    Flg, D_T, Siz : longint ; Dat : pointer ) : integer ;

var Current : PSymbol_Entry ;
    Table : integer ;

begin
    Table := _Root.Count - 1 ; { Use highest table }
    if( Table < 0 ) then
    begin
        _Root.Insert( nil ) ;
        Table := 0 ;
    end ;

    if( _Root.At( Table ) = nil ) then { No entries yet }
    begin
        New( Current ) ; { Create pointer }
        _Root.AtPut( Table, Current ) ;
        Current^.L_Link := nil ;
        Current^.R_Link := nil ;
        Current^.B_Link := nil ;
	Getmem( Current^.Name, Length( Sym ) + 1 ) ;
        Current^.Name^ := Sym ;
        Current^.Address := Addr ;
        Current^.Flags := Flg ;
        Current^.Data_Type := D_T ;
	Current^.Size := Siz ;
        Current^.Data := Dat ;
	Last_Symbol := Current ;
        Add := 0 ;
        exit ;
    end ;

    Add := _Add( Sym, Addr, Flg, D_T, Siz, Dat, PSymbol_Entry( _Root.At( Table ) ) ) ;
end ; { TSymbol_Table.Add }


function TSymbol_Table.Find( const Sym : string ; var Addr : int64 ;
    var Flg, D_T, Siz : longint ; var Dat : pointer ) : integer ;

label Symbol_Loop, Try_Next ;

var _Result : integer ;
    Table : integer ;

begin
    Table := _Root.Count ; { Use highest table (+1) }
    if( Table < 0 ) then
    begin
        _Root.Insert( nil ) ;
        Table := 0 ;
    end ;

Try_Next:
    if( Table > 0 ) then
    begin
	Table := Table - 1 ; { Try next table down }
    end ;

    if( _Root.At( Table ) = nil ) then { No entries yet }
    begin
        if( Table > 0 ) then
        begin
            goto Try_Next ;
        end ;
        Result := 4 ; { Reference to undeclared variable }
        Exit ;
    end ;

    _Result := _Find( Sym, Addr, Flg, D_T, Siz, Dat, PSymbol_Entry( _Root.At( Table ) ) ) ;
    if( ( _Result <> 0 ) and ( Table > 0 ) ) then
    begin
	goto Try_Next ;
    end ;
    Find := _Result ;
end ; { TSymbol_Table.Find }


function TSymbol_Table.Find_Next( var Addr : int64 ; var Flg, D_T, Siz : integer ;
    var Dat : pointer ) : integer ;

var Previous : PSymbol_Entry ;
    Saved : PSymbol_Entry ;

begin
    if( Last_Symbol = nil ) then
    begin
        Find_Next := Find( '*', Addr, Flg, D_T, Siz, Dat ) ;
    end else
    begin
        if( Last_Symbol^.R_Link <> nil ) then
        begin
            Last_Symbol := Last_Symbol^.R_Link ;
            while( Last_Symbol^.L_Link <> nil ) do
            begin
                Last_Symbol := Last_Symbol^.L_Link ;
            end ;
        end else
        begin
            Saved := Last_Symbol ;
            while( True ) do
            begin
                Previous := Last_Symbol^.B_Link ;
                if( Previous = nil ) then
                begin
                    Find_Next := 5 ; { Backed-up to root - we are done }
                    exit ;
                end ;
                if( Previous^.R_Link = Last_Symbol ) then
                begin
                    Last_Symbol := Previous ;
                end else
                begin
                    break ;
                end ;
            end ; { while( True ) }
            if( Saved = Last_Symbol ) then { No more symbols - we are done }
            begin
                Find_Next := 5 ;
                exit ;
            end ;
            Addr := Last_Symbol^.Address ;
            Flg := Last_Symbol^.Flags ;
            D_T := Last_Symbol^.Data_Type ;
            Siz := Last_Symbol^.Size ;
            Dat := Last_Symbol^.Data ;
        end ; { if( Last_Symbol^.R_Link <> nil ) }
        Find_Next := 0 ;
    end ; { if( Last_Symbol = nil ) }
end ; { TSymbol_Table.Find_Next }


function TSymbol_Table.Modify( Sym : string ; Addr : int64 ; Flg, D_T, Siz : longint ;
    Dat : pointer ) : integer ;

label Symbol_Loop, Try_Next ;

var _Result : integer ;
    Table : integer ;

begin
    Table := _Root.Count ; { Use highest table (+1) }
    if( Table < 0 ) then
    begin
        _Root.Insert( nil ) ;
        Table := 0 ;
    end ;

Try_Next:
    if( Table > 0 ) then
    begin
	Table := Table - 1 ; { Try next table down }
    end ;

    if( _Root.At( Table ) = nil ) then { No entries yet }
    begin
        if( Table > 0 ) then
        begin
            goto Try_Next ;
        end ;
        Result := 4 ; { Reference to undeclared symbol }
        Exit ;
    end ;

    _Result := _Modify( Sym, Addr, Flg, D_T, Siz, Dat, PSymbol_Entry( _Root.At( Table ) ) ) ;
    if( ( _Result <> 0 ) and ( Table > 0 ) ) then
    begin
	goto Try_Next ;
    end ;
    Modify := _Result ;
end ; { TSymbol_Table.Modify }


procedure TSymbol_Table.Delete( const Sym : string ) ;

label Symbol_Loop, Try_Next ;

var Table : integer ;

begin
    Table := _Root.Count ; { Use highest table (+1) }
    if( Table < 0 ) then
    begin
        _Root.Insert( nil ) ;
        Table := 0 ;
    end ;

Try_Next:
    if( Table > 0 ) then
    begin
	Table := Table - 1 ; { Try next table down }
    end ;

    if( _Root.At( Table ) = nil ) then { No entries yet }
    begin
        if( Table > 0 ) then
        begin
            goto Try_Next ;
        end ;
        Exit ; { Undefined symbol }
    end ;

    _Delete( Sym, PSymbol_Entry( _Root.At( Table ) ) ) ;
end ; { TSymbol_Table.Delete }


procedure TSymbol_Table._Delete( const Sym : string ; Table : PSymbol_Entry ) ;

var {Current, Previous,} Tmp : PSymbol_Entry ;
    Addr : int64 ;
    Flg, D_T, Siz : longint ;
    Dat : pointer ;

begin
    if( _Find( Sym, Addr, Flg, D_T, Siz, Dat, Table ) = 0 ) then
    begin
        if( ( Last_Symbol^.R_Link <> nil ) and ( Last_Symbol^.L_Link = nil ) ) then
        {Right branch but no left branch}
        begin
            Last_Symbol^.R_Link^.B_Link := Last_Symbol^.B_Link ;
            if( Table <> Last_Symbol ) then
            begin
                if( Last_Symbol^.B_Link^.L_Link = Last_Symbol )
                    then
                        Last_Symbol^.B_Link^.L_Link := Last_Symbol^.R_Link
                    else
                        Last_Symbol^.B_Link^.R_Link := Last_Symbol^.R_Link ;
            end ;
            if( _Delete_Logger <> nil ) then { Notify user of symbol deletion }
            begin
                _Delete_Logger.Deletion_Notification( Last_Symbol ) ;
            end ;
            freemem( Last_Symbol^.Name, Length( Last_Symbol^.Name^ ) + 1 ) ;
            dispose( Last_Symbol ) ;
            Last_Symbol := nil ;
            exit ;
        end ;
        if( ( Last_Symbol^.L_Link = nil ) and ( Last_Symbol^.R_Link = nil ) ) then {Terminal node}
        begin
            if( Table <> Last_Symbol ) then
            begin
                if( Last_Symbol^.B_Link^.L_Link = Last_Symbol )
                    then
                        Last_Symbol^.B_Link^.L_Link := nil
                    else
                        Last_Symbol^.B_Link^.R_Link := nil ;
            end ;
            if( _Delete_Logger <> nil ) then { Notify user of symbol deletion }
            begin
                _Delete_Logger.Deletion_Notification( Last_Symbol ) ;
            end ;
            freemem( Last_Symbol^.Name, Length( Last_Symbol^.Name^ ) + 1 ) ;
            dispose( Last_Symbol ) ;
            Last_Symbol := nil ;
            exit ;
        end ;
        {Bring up left branch}
        Last_Symbol^.L_Link^.B_Link := Last_Symbol^.B_Link ;
        if( Table <> Last_Symbol ) then
        begin
            if Last_Symbol^.B_Link^.L_Link = Last_Symbol
                then
                    Last_Symbol^.B_Link^.L_Link := Last_Symbol^.L_Link
                else
                    Last_Symbol^.B_Link^.R_Link := Last_Symbol^.L_Link ;
        end ;
        if( Last_Symbol^.R_Link <> nil ) then {A right branch to handle}
        begin
            Tmp := Last_Symbol^.L_Link ; {Start at replacement node}
            while( Tmp^.R_Link <> nil ) do
            begin
                Tmp := Tmp^.R_Link ; {Go all the way right}
            end ;
            Tmp^.R_Link := Last_Symbol^.R_Link ;
            Tmp^.R_Link^.B_Link := Tmp ;
        end ;
        if( _Delete_Logger <> nil ) then { Notify user of symbol deletion }
        begin
            _Delete_Logger.Deletion_Notification( Last_Symbol ) ;
        end ;
        freemem( Last_Symbol^.Name, Length( Last_Symbol^.Name^ ) + 1 ) ;
        dispose( Last_Symbol ) ;
        Last_Symbol := nil ;
    end ; { if }
end ; { TSymbol_Table.Delete }


procedure TSymbol_Table.Push_Level ;

begin
    _Root.Insert( nil ) ;
end ;


procedure TSymbol_Table.Pop_Level ;

var Current, Previous : PSymbol_Entry ;

begin
    if( _Root.Count > 0 ) then
    begin
        { Delete all of the symbols in this table... }
        Current := PSymbol_Entry( _Root.At( _Root.Count - 1 ) ) ;
        while( Current <> nil ) do
        begin
            while( Current^.L_Link <> nil ) do
            begin
                Current := Current^.L_Link ; { Start at far left node of tree }
            end ;
            Previous := nil ;
            while(
                   ( Current <> nil )
                   and
                   (
                     ( Current^.R_Link = nil )
                     or
                     ( Previous = Current^.R_Link )
                   )
                 ) do { No right branch from here }
            begin
                Previous := Current ;
                Current := Current^.B_Link ; { Back up }
                if( _Delete_Logger <> nil ) then { Notify user of symbol deletion }
                begin
                    _Delete_Logger.Deletion_Notification( Previous ) ;
                end ;
                if( Last_Symbol = Previous ) then
                begin
                    Last_Symbol := nil ;
                end ;
                freemem( Previous^.Name, Length( Previous^.Name^ ) + 1 ) ;
                dispose( Previous ) ;
                if( Current <> nil ) then
                begin
                    if( Previous = Current^.L_Link ) then
                    begin
                        Previous := nil ;
                        continue ; { Backing up from left branch - check this node }
                    end ;
                end ; { if( Current <> nil ) }
            end ; { while() }
            if( Current <> nil ) then
            begin
                Current := Current^.R_Link ;
            end ;
        end ; { while( Current <> nil ) }
    end ; { if( _Root.Count > 0 ) }

    _Root.AtDelete( _Root.Count - 1 ) ;
end ; { TSymbol_Table.Pop_Level }


procedure TSymbol_Table.Clear ;

begin
    while( _Root.Count > 0 ) do
    begin
        Pop_Level ;
    end ;
end ;


function TSymbol_Table.ERT( X : integer ) : string ;

begin
    ERT := SERT( X ) ;
end ;



{ TSymbol_Deletion_Logger methods... }

{ API... }

procedure TSymbol_Deletion_Logger.Deletion_Notification( Symbol : PSymbol_Entry ) ;

begin
    { This routine left intentionally blank }
end ;



end.
