{
        Program Name : TDialogs
        Package Name : LB
        Purpose      : Base Delphi dialog class
        Institution  : Conroy & Conroy Co.
        Date Written : 2-July-2004
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2004 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Standard base for dialogs.  This supports an initializer on the
        Show_Modal to allow the form to be pre-loaded with data.
}

unit TDialogs ;

interface

uses // Borland...
     BehavMan, // TBehavior_Manager
     Buttons, // TSpeedButton
     Classes, // TShiftState
     Forms, // TForm
     StdCtrls, // TButton

     // C&C...
     B_Dialog_Button_Nav, // PForm_Button_Navigation_Behavior
     B_Dialog_Default_Key, // PForm_Default_Key_Behavior
     VCL_Std ; // Adjust_Button_Panel

type TDialog = class( TForm )
                   private // Instance data...
                       _Button_Navigation : TForm_Button_Navigation_Behavior ;
                       _Behavior_Manager : TBehavior_Manager ;
                       _Default_Key_Processing : TForm_Default_Key_Behavior ;
                       _Initialized : boolean ;

                   public // Constructors and destructors...
                       destructor Destroy ; override ;

                   protected // Property handlers...
                       function Get_Behavior_Manager : TBehavior_Manager ;
                       function Get_Button_Navigation : boolean ;
                       function Get_Default_Key_Processing : boolean ;
                       procedure Set_Button_Navigation( Value : boolean ) ;
                       procedure Set_Default_Key_Processing( Value : boolean ) ;

                   protected // Callbacks...
                       procedure CB_Button_Panel_Resize( Sender : TObject ) ;

                   public // API...
                       property Behavior_Manager : TBehavior_Manager
                           read Get_Behavior_Manager ;

                       property Button_Navigation : boolean
                           read Get_Button_Navigation
                           write Set_Button_Navigation ;

                       property Default_Key_Processing : boolean
                           read Get_Default_Key_Processing
                           write Set_Default_Key_Processing ;

                       { Returns an initializer string for the current dialog
                         state/contents. }
                       function Get_Initializer : string ; virtual ;

                       { Show the form modally and return result.  Initializer
                         contains the dialog initialization specification. }
                       function Show_Modal( const Initializer : string ) : integer ;
                           virtual ;

                       { Called when Show_Modal's initializer indicates that the
                         dialog state should be updated based on the values set
                         thus far. }
                       procedure Update_State ; virtual ;

                       procedure FormKeyDown( Sender : TObject ; var Key : Word ;
                           Shift : TShiftState ) ; virtual ;
               end ;

implementation

uses // Borland...
     Windows, // VK_DOWN
     ExtCtrls, // TPanel
     ComCtrls, // TListView
     Controls, // TControl
     SysUtils, // uppercase

     // C&C...
     Events, // TEvent
     HTML, // THTML_Parser
     UStrings, // Escape_String
     WinEvent ; // Create_Windows_Key_Down_Event

// TDialog methods...

// Constructors and destructors...

destructor TDialog.Destroy ;

begin
    if( _Behavior_Manager <> nil ) then
    begin
        _Behavior_Manager.Free ;
        _Behavior_Manager := nil ;
    end ;

    inherited Destroy ;
end ;


// Property handlers...

function TDialog.Get_Behavior_Manager : TBehavior_Manager ;

begin
    if( _Behavior_Manager = nil ) then
    begin
        _Behavior_Manager := TBehavior_Manager.Create ;
    end ;
    Result := _Behavior_Manager ;
end ;


function TDialog.Get_Button_Navigation : boolean ;

begin
    Result := ( _Button_Navigation <> nil ) ;
end ;


function TDialog.Get_Default_Key_Processing : boolean ;

begin
    Result := ( _Default_Key_Processing <> nil ) ;
end ;


procedure TDialog.Set_Button_Navigation( Value : boolean ) ;

begin
    if( Button_Navigation <> Value ) then // A change
    begin
        if( Value ) then
        begin
            _Button_Navigation := TForm_Button_Navigation_Behavior.Create ;
            Behavior_Manager.Add( _Button_Navigation ) ;
        end else
        begin
            Behavior_Manager.Remove( _Button_Navigation ) ;
            _Button_Navigation.Free ;
        end ;
    end ;
    if( Value ) then
    begin
        OnKeyDown := FormKeyDown ;
        KeyPreview := True ;
    end ;
end ;


procedure TDialog.Set_Default_Key_Processing( Value : boolean ) ;

begin
    if( Default_Key_Processing <> Value ) then
    begin
        if( Value ) then
        begin
            _Default_Key_Processing := TForm_Default_Key_Behavior.Create ;
            Behavior_Manager.Add( _Default_Key_Processing ) ;
        end else
        begin
            Behavior_Manager.Remove( _Default_Key_Processing ) ;
            _Default_Key_Processing.Free ;
        end ;
    end ;
    if( Value ) then
    begin
        OnKeyDown := FormKeyDown ;
        KeyPreview := True ;
    end ;
end ;


// Callbacks...

procedure TDialog.CB_Button_Panel_Resize( Sender : TObject ) ;

begin
    Adjust_Button_Panel( TWinControl( Sender ) ) ;
end ;


// API...

function TDialog.Get_Initializer : string ;

    function Process_Children( C : TWinControl ) : string ; forward ;


    function Control_Description( Child : TControl ) : string ;

    var Typ : string ;

    begin
        Typ := '' ;
        if( ( Child is TButton ) or ( Child is TSpeedButton ) ) then
        begin
            Typ := '</command>' ;
            Result := '<command' ;
        end else
        if( Child is TListView ) then
        begin
            Typ := '</input>' ;
            Result := '<input' ;
        end else
        begin
            Typ := '</input>' ;
            Result := '<input' ;
        end ;
        Result := Result + ' name=' + Child.Name ;
        if( Child.Enabled ) then
        begin
            Result := Result + ' enabled' ;
        end else
        begin
            Result := Result + ' noenabled' ;
        end ;
        if( Child.Visible ) then
        begin
            Result := Result + ' visible' ;
        end else
        begin
            Result := Result + ' novisible' ;
        end ;
        if( Control_Has_Children( Child ) ) then
        begin
            Result := Result + '>' + Process_Children( TWinControl( Child ) ) + Typ ;
        end ;
    end ;


    function Process_Children( C : TWinControl ) : string ;

    var Child : TControl ;
        Loop : integer ;

    begin
        Result := '' ;
        for Loop := 0 to C.ControlCount - 1 do
        begin
            Child := C.Controls[ Loop ] ;
            Result := Result + Control_Description( Child ) ;
        end ;
    end ;

begin
    Result := Control_Description( self ) ;
end ; // TDialog.Get_Initializer


function TDialog.Show_Modal( const Initializer : string ) : integer ;

    function Find_Child( C : TWinControl ; const Name : string ) : TControl ;

    var Loop : integer ;

    begin
        Result := nil ;
        for Loop := 0 to C.ControlCount - 1 do
        begin
            if( C.Controls[ Loop ].Name = Name ) then
            begin
                Result := C.Controls[ Loop ] ;
                exit ;
            end ;
        end ;
    end ;


    function Standard_Processing( C : TWinControl ;
        Tag : THTML_Tag ) : TControl ;

    var Loop : integer ;

    begin
        // Setup...
        Result := nil ; // Assume failure

        // Find the child
        for Loop := 0 to Tag.Count - 1 do
        begin
            if( uppercase( Tag.Attribute_Name( Loop ) ) = 'NAME' ) then
            begin
                Result := Find_Child( C, Tag.Attribute_Value( Loop ) ) ;
                break ;
            end ;
        end ;

        // Process standard attributes
        if( Result <> nil ) then // If we found the control
        begin
            for Loop := 0 to Tag.Count - 1 do
            begin
                if( uppercase( Tag.Attribute_Name( Loop ) ) = 'ENABLED' ) then
                begin
                    Result.Enabled := True ;
                end else
                if( uppercase( Tag.Attribute_Name( Loop ) ) = 'NOENABLED' ) then
                begin
                    Result.Enabled := False ;
                end else
                if( uppercase( Tag.Attribute_Name( Loop ) ) = 'VISIBLE' ) then
                begin
                    Result.Visible := True ;
                end else
                if( uppercase( Tag.Attribute_Name( Loop ) ) = 'NOVISIBLE' ) then
                begin
                    Result.Visible := False ;
                end ;
            end ;
        end ;
    end ;


    procedure Parse_Input( C : TWinControl ; Master_Tag : THTML_Tag ; const Image : string ) ;
        forward ;

    procedure Parse_Command( C : TWinControl ; Master_Tag : THTML_Tag ; const Image : string ) ;

    var Control : TControl ;
        Parser : THTML_Parser ;
        S : string ;
        Tag : THTML_Tag ;

    begin
        Control := Standard_Processing( C, Master_Tag ) ;
        if( Control = nil ) then
        begin
            exit ;
        end ;
        Parser := THTML_Parser.Create ;
        Parser.Source := Image ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Token ;
            if( copy( S, 1, 1 ) = '<' ) then
            begin
                Tag := THTML_Tag.Create( S ) ;
                if( uppercase( Tag.Name ) = 'COMMAND' ) then
                begin
                    Parse_Command( TWinControl( Control ), Tag, Parser.Get_Section( 'command' ) ) ;
                end else
                if( uppercase( Tag.Name ) = 'INPUT' ) then
                begin
                    Parse_Input( TWinControl( Control ), Tag, Parser.Get_Section( 'input' ) ) ;
                end else
                begin
                    Parser.Get_Section( S ) ;
                end ;
                Tag.Free ;
            end ;
        end ;
        Parser.Free ;
    end ;


    procedure Parse_Input( C : TWinControl ; Master_Tag : THTML_Tag ; const Image : string ) ;

    var Control : TControl ;
        Parser : THTML_Parser ;
        S : string ;
        Tag : THTML_Tag ;

    begin
        Control := Standard_Processing( C, Master_Tag ) ;
        if( Control = nil ) then
        begin
            exit ;
        end ;
        Parser := THTML_Parser.Create ;
        Parser.Source := Image ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Token ;
            if( copy( S, 1, 1 ) = '<' ) then
            begin
                Tag := THTML_Tag.Create( S ) ;
                if( uppercase( Tag.Name ) = 'COMMAND' ) then
                begin
                    Parse_Command( TWinControl( Control ), Tag, Parser.Get_Section( 'command' ) ) ;
                end else
                if( uppercase( Tag.Name ) = 'INPUT' ) then
                begin
                    Parse_Input( TWinControl( Control ), Tag, Parser.Get_Section( 'input' ) ) ;
                end else
                begin
                    Parser.Get_Section( S ) ;
                end ;
                Tag.Free ;
            end ;
        end ;
        Parser.Free ;
    end ;

var Control : TControl ;
    Parser : THTML_Parser ;
    S : string ;
    Tag : THTML_Tag ;

begin // TDialog.Show_Modal
    // Do general setup on first show...
    if( not _Initialized ) then
    begin
        _Initialized := True ;
        Control := Find_Child( self, 'Button_Panel' ) ;
        if( ( Control <> nil ) and ( Control is TPanel ) ) then
        begin
            TPanel( Control ).OnResize := CB_Button_Panel_Resize ;
        end ;
    end ;

    // Process initializer commands...
    if( length( Initializer ) <> 0 ) then
    begin
        Parser := THTML_Parser.Create ;
        Parser.Source := Initializer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Token ;
            if( copy( S, 1, 1 ) = '<' ) then
            begin
                Tag := THTML_Tag.Create( S ) ;
                if( uppercase( Tag.Name ) = 'COMMAND' ) then
                begin
                    Parse_Command( self, Tag, Parser.Get_Section( 'command' ) ) ;
                end else
                if( uppercase( Tag.Name ) = 'INPUT' ) then
                begin
                    Parse_Input( self, Tag, Parser.Get_Section( 'input' ) ) ;
                end else
                begin
                    Parser.Get_Section( S ) ;
                end ;
                Tag.Free ;
            end ;
        end ;
        Parser.Free ;
        Result := mrOK ;
        exit ;
    end ;
    Result := ShowModal ;
end ; // TDialog.Show_Modal


procedure TDialog.Update_State ;

begin
    // This routine intentionally left blank...
end ;



procedure TDialog.FormKeyDown( Sender : TObject ; var Key : Word ;
    Shift : TShiftState ) ;

var Event : TEvent ;

begin
    if( _Behavior_Manager <> nil ) then
    begin
        Event := Create_Windows_Key_Down_Event( self, Key, Shift ) ;
        try
            if( not Behavior_Manager.Allow_Normal_Processing( Event ) ) then
            begin
                Key := 0 ;
                exit ;
            end ;
            Behavior_Manager.Execute( Event ) ;
        finally
            Event.Free ;
        end ;
    end ;
end ;


end.
