{$I EDEFINES.INC}
{$IFDEF DELPHI}
{$H-}
{$V-}
{$ENDIF}
{
          Program Name : UEDITOR
          Package Name : LB
          Purpose      : Text Editing Engine (tEditor)
          Institution  : Conroy & Conroy
          Original Vers: 22-Mar-89
          Date Written : 21-Jul-1990
          Written By   : Alan Conroy
          Version      : 2.2C

          Copyright (C) 1989-2002 by Alan Conroy.  Released to the public domain.

          TO THE GLORY OF GOD THROUGH JESUS CHRIST

	*********************************************************
	*                                                       *
	*        M O D I F I C A T I O N   H I S T O R Y        *
	*                                                       *
	*     DATE      BY             REASON                   *
	*                                                       *
	*  1-Nov-2002   EAC            Code cleanup.            *
	* 13-Nov-2002   EAC            Fix problem caused by    *
        *                              last change.             *
        *  2-Dec-2002   EAC            Initialize strings.      *
	*                                                       *
	*********************************************************

	*********************************************************
	*                                                       *
	*           P R O G R A M   P U R P O S E               *
	*                                                       *
	*********************************************************

	  Conceptual equivalent to VAX TPU engine.
}

unit UEditor ;

interface

uses TypeDefs,
{$IFDEF DELPHI}
     Classes, Controls,
{$ENDIF}
     SMU,
     _Types ; { Small_String }

const eds_SubWindow = 1 ; {Sub window within window}
      eds_ScrollOnRedraw = 2 ; {Scroll on redraw of adjustment}
      eds_CursorBound = 4 ; {Bind cursor to line length}
      eds_Beep = 8 ; {Beep on warning/error}
      eds_AutoIndent = 16 ; {Automatically indent new lines}
      eds_NoMessages = 32 ; {No error/warning messages}
      eds_NoTabs = 64 ; {No special tab processing}
      eds_NoCursorWrap = 128 ; { No wrapping of cursor }
      eds_Case_Sensitive = 256 ; { Case-sensitive searches }

      edn_Memory = 1 ; {Not enough memory}
      edn_Top = 2 ; {Backup past BOB}
      edn_Bottom = 3 ; {Advance beyond EOB}
      edn_Length = 4 ; {Line length exceeds 255 characters}
      edn_NotFound = 5 ; {Text not found}
      edn_NoSelection = 6 ; {Nothing selected}
      edn_Selection = 7 ; {Selection in progress}
      edn_Position = 8 ; {Changed cursor position}
      edn_Modify = 9 ; {Buffer modification flag has changed}
      edn_LineCount = 10 ; { Line count change }
      edn_Scroll = 11 ; { Position within buffer change }
      edn_Frozen = 12 ; { Selection is frozen }

Type Anchor_Type = record {Anchor type}
		       Line, End_Line : Longint ;
		       Column, P_Column, End_Column, End_P_Column : Byte ;
		   end ;
     Line_Pointer = ^Line_Record ;
     Line_Record = record
			Last_Line, Next_Line : Line_Pointer ;
			Text : Small_String ;
		   end ;
     tNotify = procedure( X : integer ) ;
{$IFDEF DELPHI}
     tEditor_Notify_Event = procedure( X : integer ) of object ;
{$ENDIF}

     Translation_Type = array[ 0..255 ] of string[ 5 ] ;

     { Only used within this unit, but in interface due to reference in
       tEditor. }
     tLine_Buffer = object
			Start : Line_Pointer ;
			Ending : Line_Pointer ;
			Last_Line : longint ;

			constructor Init ;
			destructor Done ;
			procedure Append( S : string ) ;
			procedure Zero ;
		    end ;
     pLine_Buffer = ^tLine_Buffer ;

type pEditor = ^tEditor ;
     tEditor = object
                   public
                    Anchor : Anchor_Type ;
                    Any_Changes : boolean ; {True if ANY changes made to file}
                    Border : integer ; {Border around window}
                    Changes : Boolean ; {True if changes made to the current line}
                    Current_Line : Line_Pointer ; {Pointer to current line in buffer}
                    Cursor_Column : integer ; { Current column of cursor }
                    Cursor_Row : integer ; {Current row of cursor}
                    { Note that row is the row we sit on and column is the
                      column we sit on (for wide carets) or the one to the right
                      of the caret (for narrow carets). }
                    End_Of_Buffer_String : Small_String ; {What to show as end-of-buffer}
                    End_Of_Line_Character : char ; {What to show as end-of-long-line}
                    Error_Flag : boolean ; {TRUE if error currently displayed}
                    File_Root : pLine_Buffer ;
                    Forwards : boolean ; { Moving in a forward direction }
                    Last_Screen_Line : longint ; {Previous Screen_Line as of last Redraw}
                    Logical : Small_String ;
                    Logical_Column, Physical_Column : integer ; {Editing pointers}
                    Logical_Text, Physical_Text : string ; { Current line text }
                    Lower_Third_Row : integer ;
                    Max_Row : integer ;
                    Middle_Row : integer ;
                    Upper_Third_Row : integer ;
                    Options : longint ; {Operational options}
                    Quiet_Mode : boolean ;
                    Screen_Column : integer ; {First column visible on screen}
                    Screen_Line : longint ; {Line in file corresponding to line 1 of screen}
                    Screen_Window : PWindow ;
                    Selection : boolean ; { True if a selection }
                    Selection_Is_Frozen : boolean ; { True if selection is frozen }
                    Tab_Size : byte ; { Size of tab}
                    Text_Height : integer ;
                    Translation : Translation_Type ;
                    Window_Height : integer ; { Height of window }
                    Window_Width : integer ; { Width of window }
                    Notification_Routine : tNotify ; { Notification routine in caller}
{$IFDEF DELPHI}
                    Notification_Routine1 : tEditor_Notify_Event ;
{$ENDIF}

                  private

                    function Offset_Position( var S : string; Offset : integer ) : integer ;
                    procedure Notify( X : integer ) ;
                    Procedure Flag_Changes ;
                    function Tab_Position( X : integer ) : integer ;
                    function Get_Line_Pointer( Line : longint ) : Line_Pointer ;
                    Procedure Handle_Selection( Row, Column : integer ;
                        var _Text : String ) ;
                    Procedure Adjust ;
                    procedure Adjust_EOB ;
                    Procedure Delete_CRLF( Explicit : Boolean ) ;
                    Procedure New_Line ;
                    procedure Redraw_Line ;

                  public
                    constructor Init ;
                    destructor Done ;
                    procedure Initialize ;

                    procedure Advance ; {Advance to beginning of next line}
                    function At_EOB : boolean ;
                    function At_BOB : boolean ;
                    function At_EOL : boolean ;
                    function At_BOL : boolean ;
                    Procedure Bottom ;
                    procedure Calculate_Rows( Height : integer ) ;
                    Procedure Change_Case ;
                    procedure Clear( Draw : boolean ) ;
                    procedure Convert_Offset_To_Position( Offset : longint; var Line, Column : longint ) ;
                    function Convert_Position_To_Offset( Line, Column : longint ) : longint ;
                    procedure Copy_Text( Paste_Buffer : pLine_Buffer ) ;
                    procedure Cursor_Up ;
                    procedure Cursor_Down ; { Move cursor down one line }
                    procedure Cursor_Left ; { Move cursor left one character }
                    procedure Cursor_Right ; { Move cursor one character to the right }
                    procedure Delete_Char_Back ;
                    procedure Delete_Char( Explicit : Boolean ) ;
                    procedure Delete_Current_Line ;
                    procedure Delete_Lines( Start, Count : integer ) ;
                    procedure Delete_Text ;
                    procedure Delete_Word ;
                    function Do_Search( Search_Text : string ) : boolean ; {Search for text}
                    procedure Do_Update ;
                    procedure EOL ; {Move to end of current line}
                    function Fetch_Line( Line_Number : longint ) : string ;
                    procedure Format( X : String ) ;
                    procedure Freeze_Selection ;
                    procedure Get_Selection_Range( var First_Line, Final_Line : longint ;
                        var First_Column, Final_Column : integer ) ;
                    procedure Get_Text( T : PChar ) ;
                    procedure Half_Page ;
                    procedure Insert_Blank_Line ;
                    procedure Insert_Chars( Character : String ) ; {Character(s) insertion}
                    procedure Lowercase_Word ; { Change current word to lowercase }
                    procedure Move_Word ; {Move to beginning of next word}
                    procedure Page ;
                    procedure Page_Down ;
                    procedure Page_Up ;
                    procedure Paste( Paste_Buffer : pLine_Buffer ) ;
                    procedure Position_Caret ;
                    procedure Position_Cursor( Line, Column : longint ) ;
                    procedure Redraw( Start : integer ) ; {Draw/redraw screen}
                    procedure Resize ;
                    function Screen_To_Text_Column( Line, Column : longint ) : longint ;
                    procedure Scroll_Columns( X : integer ) ;
                    procedure Search( Search_Text : string ; Start_Line,
                        Start_Column : longint ;
                        var Result_Line, Result_Column : longint ) ; {Search}
                    procedure Select ;
                    procedure Select_Text( Start_Line, Start_Column, End_Line,
                        End_Column : integer ) ;
                    procedure Set_Buffer( X : pLine_Buffer ) ;
                    procedure Set_Cursor_Column( X : integer ) ;
                    procedure Set_Cursor_Row( X : integer ) ;
                    procedure Set_Line( X : integer ; Y : string ) ;
                    function Something_Selected : boolean ;
                    function Text_Size : longint ; { Size of all text, in bytes }
                    function Text_To_Screen_Column( Line, Column : longint ) : longint ;
                    procedure Top ;
                    procedure Unselect ;
                    procedure Uppercase_Word ; { Convert current word to uppercase }
                    procedure Word_Backward ;
                    procedure Word_Forward ;
                    procedure Word_Range( var Starting, Ending : integer ) ;
{$IFDEF Test}
                    procedure Test_Class ;
{$ENDIF}
	      end ;

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

const Line_Overhead = 9 ; { Overhead for each line (1 byte length + 2 4-byte pointers) }

implementation

{$IFDEF DELPHI}
uses Windows, Graphics,
{$ELSE}
  {$IFNDEF WINDOWS}
uses Dos,
  {$ELSE}
uses WinTypes, WinProcs,
  {$ENDIF}
{$ENDIF}
     ASCIIDef,
     CommonUt, { Edit }
     Compatib ;


constructor tLine_Buffer.Init ;

begin
    Start := nil ;
    Ending := nil ;
    Last_Line := 0 ;
end ;


destructor tLine_Buffer.Done ;

begin
    Zero ;
end ;


procedure tLine_Buffer.Zero ; {Zero buffer's storage}

var T_Pointer : Line_Pointer ;

begin
    while( Start <> nil ) do {Reclaim memory}
    begin
	T_Pointer := Start^.Next_Line ;
	Freemem( Start, Length( Start^.Text ) + Line_Overhead ) ;
	Start := T_Pointer ;
    end ;
    Ending := nil ;
    Last_Line := 0 ;
end ;


procedure tLine_Buffer.Append( S : string ) ;

var Previous_Line, This_Line : Line_Pointer ;

begin
    if( maxavail < length( S ) + Line_Overhead ) then
    begin
	exit ;
    end ;
    Previous_Line := nil ;
    if( Start = nil ) then
    begin
	getmem( Start, length( S ) + Line_Overhead ) ;
	This_Line := Start ;
    end else
    begin
	Previous_Line := Ending ;
	getmem( Previous_Line^.Next_Line, length( S ) + Line_Overhead ) ;
	This_Line := Previous_Line^.Next_Line ;
    end ;
    This_Line^.Text := S ;
    This_Line^.Next_Line := nil ;
    This_Line^.Last_Line := Previous_Line ;
    Ending := This_Line ;
    Last_Line := Last_Line + 1 ; {One more line in the buffer}
end ; { tLine_Buffer.Append }



{ tEditor methods... }

function tEditor.At_EOL : boolean ;

begin
    At_EOL := ( Physical_Column > length( Physical_Text ) ) ;
end ;


function tEditor.At_BOL : boolean ;

begin
    At_BOL := ( Physical_Column = 1 ) ;
end ;


function tEditor.At_EOB : boolean ;

begin
    At_EOB := ( Current_Line = nil ) ;
end ;


function tEditor.At_BOB : boolean ;

begin
    if( File_Root^.Start = nil ) then { Nothing in buffer }
    begin
	At_BOB := true ;
	exit ;
    end ;
    if( Current_Line = nil ) then { EOB }
    begin
	At_BOB := False ;
	exit ;
    end ;
    At_BOB := ( Current_Line^.Last_Line = nil ) ;
end ;


function tEditor.Offset_Position( var S : string ; Offset : integer ) : integer ;
{ Convert screen Offset to string position }

var X : integer ;

begin
    X := 1 ;
    while(
	     (
	         ( Get_Text_Extent( Screen_Window, S, X ) and 32767 ) < Offset
	     )
	     and
	     ( X < length( S ) )
	 ) do
    begin
	X := X + 1 ;
    end ;
    Offset_Position := X ;
end ;


procedure tEditor.Scroll_Columns( X : integer ) ;

begin
    Screen_Column := Screen_Column + X ;
    if( Screen_Column < 1 ) then
    begin
	Screen_Column := 1 ;
    end ;
{$IFNDEF WINDOWS}
    if( Screen_Column + Window_Width > 257 ) then
    begin
	Screen_Column := 257 - Window_Width ;
    end ;
{$ENDIF}
    Last_Screen_Line := 0 ; {Force redraw}
    Redraw( 1 ) ;
end ;


procedure tEditor.Set_Cursor_Row( X : integer ) ;

begin
    Cursor_Row := X ;
    Notify( EDN_Position ) ;
end ;


procedure tEditor.Set_Cursor_Column( X : integer ) ;

begin
    Cursor_Column := X ;
    Notify( EDN_Position ) ;
end ;


function tEditor.Fetch_Line( Line_Number : longint ) : string ;
{Return text of requested line}

var TPointer : Line_Pointer ;

begin
    if ( Line_Number > File_Root^.Last_Line ) or ( Line_Number < -1 ) or ( Line_Number = 0 ) then
    {Line does not exist}
    begin
	Fetch_Line := '' ;
	exit ;
    end ;
    if ( Line_Number = Screen_Line + Cursor_Row - 1 ) or ( Line_Number = -1 ) then
    {Current line}
    begin
	Fetch_Line := Physical_Text ;
    end else
    begin
	TPointer := File_Root^.Start ; {Start with line 1}
	while( Line_Number > 1 ) do
	begin
	    dec( Line_Number ) ;
	    TPointer := TPointer^.Next_Line ; {Find requested line}
	end ;
	Fetch_Line := TPointer^.Text ;
    end ;
end ; { tEditor.Fetch_Line }


procedure tEditor.Convert_Offset_To_Position( Offset : longint ;
    var Line, Column : longint ) ;
{ Return line and column number for the passed character offset. }

var Count, Last_Count, Row : longint ;
    TPointer : Line_Pointer ;

begin
    Last_Count := 0 ;
    Row := 1 ;
    TPointer := File_Root^.Start ;
    Count := length( TPointer^.Text ) + 2 ;
    while Count < Offset do
    begin
	Inc( Row ) ;
	TPointer := TPointer^.Next_Line ;
	Last_Count := Count ;
	Count := Count + length( TPointer^.Text ) + 2 ;
    end ;
    Line := Row ;
    Column := Offset - Last_Count ;
end ;


function tEditor.Convert_Position_To_Offset( Line, Column : longint ) : longint ;
{Return offset of character at passed line and column.  Line -1 means the
 current line, column -1 means the current column.}

var Count, Row : longint ;
    TPointer : Line_Pointer ;

begin
    if Line = -1 then
    begin
	Line := Screen_Line + Cursor_Row - 1 ;
    end ;
    Count := 0 ;
    Row := 1 ;
    TPointer := File_Root^.Start ;
    while Row < Line do
    begin
	Count := Count + length( TPointer^.Text ) + 2 ;
	inc( Row ) ;
	TPointer := TPointer^.Next_Line ;
    end ;
    if( Column = -1 ) then
    begin
	Column := Physical_Column ;
    end ;
    if( Column = 0 ) then
    begin
	Column := 1 ;
    end ;
    Convert_Position_To_Offset := Count + Column - 1 ;
end ; { tEditor.Convert_Position_To_Offset }


procedure tEditor.Notify( X : integer ) ;

begin
    if( @Notification_Routine <> nil ) then
    begin
	Notification_Routine( X ) ;
    end ;
{$IFDEF DELPHI}
     if( assigned( Notification_Routine1 ) ) then
     begin
         Notification_Routine1( X ) ;
     end ;
{$ENDIF}

end ;


procedure tEditor.Calculate_Rows( Height : integer ) ;
{Calculate middle, lower third, and upper third rows.}

var Dummy : integer ;
    Text_Width : integer ;

begin
    Text_Height := Get_Text_Extent( Screen_Window,
	'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890',
	62 ) shr 16 ;
    Text_Width := Get_Text_Extent( Screen_Window, 'W', 1 ) and 32767 ;
    Get_Window_Size( Screen_Window, Dummy, Window_Width ) ;
    Window_Height := Height ;
    Max_Row := Window_Height div Text_Height ; { Rows fully visible }
    SMU.Resize( Screen_Window, Max_Row, Window_Width div Text_Width ) ;
    Middle_Row := Max_Row div 2 ;
    if( Middle_Row < 1 ) then
    begin
	Middle_Row := 1 ; { Minimum of 1 }
    end ;
    if( ( Options and eds_SubWindow ) <> 0 ) then { Sub window within window }
    begin
	Lower_Third_Row := Max_Row div 3 ;
	Upper_Third_Row := Max_Row - Lower_Third_Row ;
	if( Lower_Third_Row > Middle_Row )
	    then
		Lower_Third_Row := Middle_Row ;
	if( Lower_Third_Row < 1 )
	    then
		Lower_Third_Row := 1 ;
	if( Upper_Third_Row < Middle_Row )
	    then
		Upper_Third_Row := Middle_Row ;
	if( Upper_Third_Row > Max_Row )
	    then
		Upper_Third_Row := Max_Row ;
    end else
    begin
	Lower_Third_Row := 1 ;
	Upper_Third_Row := Max_Row ;
    end ;
end ;


procedure tEditor.Resize ; { Resize window }

var Height, Width : integer ;

begin
    Get_Window_Size( Screen_Window, Height, Width ) ;
    Calculate_Rows( Height ) ;
end ;


procedure tEditor.Flag_Changes ; { Flag any changes to line }

begin
    Changes := True ;
    Logical_Column := Cursor_Column ;
    Notify( edn_Modify ) ;
end ;


procedure tEditor.Clear( Draw : boolean ) ;

begin
    File_Root^.Zero ;
    Changes := false ;
    Current_Line := nil ;
    Physical_Column := 1 ;
    Set_Cursor_Column( 1 ) ;
    Set_Cursor_Row( 1 ) ;
    Physical_Text := '' ;
    Logical_Column := 1 ;
    Screen_Line := 1 ;
    if( Draw ) then
    begin
	Clear_Window( Screen_Window ) ;
	Redraw( 1 ) ;
    end ;
end ;


function tEditor.Tab_Position( X : integer ) : integer ;
{Given position X, return position of next tab}

var Dummy : integer ;

begin
    Dummy := ( X + Tab_Size - 1 ) ;
    Tab_Position := ( Dummy div Tab_Size ) * Tab_Size + 1 ;
end ;


procedure tEditor.Format( X : string ) ;
{This procedure translates physical text to logical text}

Label Finished ;

Var A : Integer ; {Loop variable}
    C : Char ;
    Dummy : Integer ; {Dummy variable}

begin
    Fillchar( Logical, 256, 32 ) ;
    Logical := '' ;
    for A := 1 to Length( X ) do {for each character in physical string}
    begin
	C := X[ A ] ; {Get current character}
	if( ( ( Options and eds_NoTabs ) = 0 ) and ( C = HT ) ) then {Tab}
	begin
	    Dummy := Tab_Position( length( Logical ) + 1 ) - 1 ;
	    {New location}
	    while( Length( Logical ) < Dummy ) do
	    begin
		if( length( Logical ) = 255 ) then
		begin
		    goto Finished ;
		end ;
		Logical := Logical + ' ' ; {Space to next tab position}
	    end ;
	end else
	begin
	    if( length( Logical ) + length( Translation[ ord( C ) ] ) > 255 ) then
	    begin
		goto Finished ;
	    end ;
	    Logical := Logical + Translation[ Ord( C ) ] ;
	end
    end ;
Finished:
    if( Length( Logical ) - Screen_Column + 1 > Window_Width ) then
    begin
	if( End_Of_Line_Character <> #0 ) then {Have an EOL character}
	begin
	    Logical[
		Offset_Position( Logical, Window_Width + Screen_Column - 1 ) ] :=
		End_of_Line_Character ;
	end ;
    end ;
    if( Screen_Column > 1 ) then
    begin
	if( End_Of_Line_Character <> #0 ) then {Have an EOL character}
	begin
	    Logical[ Offset_Position( Logical, Screen_Column ) ] :=
		End_of_Line_Character ;
	end ;
    end ;
end ; { tEditor.Format }


procedure tEditor.Do_Update ; {Update memory with edited line - if needed}

Var T_Pointer : Line_Pointer ;

begin
    if Changes then {Changes made}
    begin
	Any_Changes := True ;
	Changes := False ;
	if Current_Line = nil then Exit ; {No update if at EOB}
	if Maxavail < 256 then {Ran out of memory}
	begin
	    Notify( edn_Memory ) ;
	    exit ;
	end ;
	Getmem( T_Pointer, Length( Physical_Text ) + Line_Overhead ) ;
	{Allocate new line}
	T_Pointer^.Last_Line := Current_Line^.Last_Line ; {Link in new line}
	T_Pointer^.Next_Line := Current_Line^.Next_Line ;
	if T_Pointer^.Last_Line <> nil then {Link previous line to new one}
	    T_Pointer^.Last_Line^.Next_Line := T_Pointer ;
	if T_Pointer^.Next_Line <> nil then {Link next line to new one}
	    T_Pointer^.Next_Line^.Last_Line := T_Pointer ;
	T_Pointer^.Text := Physical_Text ; {Store text for line}
	if File_Root^.Ending = Current_Line then File_Root^.Ending := T_Pointer ;
	if File_Root^.Start = Current_Line then File_Root^.Start := T_Pointer ;
	Freemem( Current_Line, Length( Current_Line^.Text ) + Line_Overhead ) ;
	{Delete old line}
	Current_Line := T_Pointer ;
    end ;
end ;


function tEditor.Get_Line_Pointer( Line : longint ) : Line_Pointer ;
{Returns pointer to specified line}

var T_Pointer : Line_Pointer ;

begin
    T_Pointer := File_Root^.Start ;
    while( Line > 1 ) do
    begin
	if( T_Pointer = nil ) then
	begin
	    Get_Line_Pointer := nil ;
	    exit ;
	end ;
	T_Pointer := T_Pointer^.Next_Line ;
	Line := Line - 1 ;
    end ;
    Get_Line_Pointer := T_Pointer ;
end ;


procedure tEditor.Position_Caret ;

var Actual_Column : integer ;

begin
    if( Cursor_Column + Border - Screen_Column + 1 > Window_Width ) then
    begin
	Actual_Column := Window_Width - Border + Screen_Column - 1 ;
    end else
    begin
	Actual_Column := Cursor_Column ;
    end ;
    Move_to( Screen_Window, Cursor_Row, Actual_Column - Screen_Column + 1 ) ;
end ;


procedure TEditor.Get_Selection_Range( var First_Line, Final_Line : longint ;
    var First_Column, Final_Column : integer ) ;

var A : longint ; { Used for reversing an inverted range }

begin
    if( Selection_Is_Frozen ) then
    begin
        First_Line := Anchor.Line ;
        First_Column := Anchor.P_Column ;
        Final_Line := Anchor.End_Line ;
        Final_Column := Anchor.End_P_Column - 1 ;
    end else
    begin
        First_Line := Anchor.Line ;
        First_Column := Anchor.P_Column ;
        Final_Line := Screen_Line + Cursor_Row - 1 ;
        Final_Column := Physical_Column - 1 ;
    end ;

    { Handle inverted range }
    if(
        ( Final_Line < First_Line )
        or
        (
          ( Final_Line = First_Line )
          and
          ( Final_Column < First_Column )
        )
      ) then
    begin
        A := Final_Line ;
        Final_Line := First_Line ;
        First_Line := A ;
        A := Final_Column ;
        Final_Column := First_Column - 1 ;
        First_Column := A + 1 ;
    end ;
end ;


procedure tEditor.Redraw( Start : integer ) ;
{ Draw/Redraw screen, starting with line Start }

type Video_State = ( Normal, Reverse ) ;

var A : Integer ;
    After_Start, Before_End : boolean ;
    Col : integer ; { Current column in line processing }
    Image : string ; { Current output string }
    Text_Pointer : Line_Pointer ;
    First_Line, Final_Line : longint ;
    First_Column, Final_Column : integer ;
    Forced_Redraw : Boolean ;
    New_State, State : Video_State ;
    Rect : SMU.TRect ;
    This_First_Column, This_Final_Column : integer ;

begin
    Do_Update ;
    if( Quiet_Mode ) then
    begin
	exit ;
    end ;
{$IFDEF WINDOWS}
    Clear_Cursor( 0, 0 ) ;
{$ENDIF}
    Forced_Redraw := True ; { Default to Redraw rather than scroll }
    if( ( Options and eds_ScrollOnRedraw ) = 0 ) then
    begin
	Last_Screen_Line := 0 ;
    end ;
    if( not Selection ) then
    begin
	First_Line := 0 ;
	Final_Line := 0 ;
	if( Last_Screen_Line > 0 ) then { No forced Redraw }
	begin
	    if( Abs( Screen_Line - Last_Screen_Line ) <= Max_Row ) then
	    { Difference of one screen or less }
	    begin
		if( Last_Screen_Line <> Screen_Line ) then { Movement needed }
		begin
		    Forced_Redraw := False ;
		end ;
	    end ;
	end ;
    end else { Selection under way }
    begin { Something is selected }
        Get_Selection_Range( First_Line, Final_Line, First_Column, Final_Column ) ;
    end ;

    A := Start ; { First line to redraw }
    if( not Forced_Redraw ) then { Scroll screen instead of Redrawing it }
    begin
	A := 1 ; { Start at line 1 in this case }
	if( Last_Screen_Line > Screen_Line ) then { Scroll down needed }
	begin
	    while( Last_Screen_Line > Screen_Line ) do
	    begin
		Rect.Top := Cursor_Row ;
		Rect.Bottom := Max_Row ;
		Rect.Left := 0 ;
		Rect.Right := Screen_Window^.Columns ;
		Scroll_Window( Screen_Window, 0, 1, Rect ) ;
		Last_Screen_Line := Last_Screen_Line - 1 ;
	    end ;
	end else { Scroll up needed }
	begin
	    while( Last_Screen_Line < Screen_Line ) do
	    begin
		Rect.Top := Cursor_Row ;
		Rect.Bottom := Max_Row ;
		Rect.Left := 0 ;
		Rect.Right := Screen_Window^.Columns ;
		Scroll_Window( Screen_Window, 0, -1, Rect ) ;
		Last_Screen_Line := Last_Screen_Line + 1 ;
	    end ;
	end ;
    end ;

    { Redraw the screen... }
    This_First_Column := 0 ;
    This_Final_Column := 0 ;
    Text_Pointer := Get_Line_Pointer( Screen_Line + A - 1 ) ;
    Select_Window( Screen_Window ) ;
    Move_To( Screen_Window, A, 2 - Screen_Column ) ;
    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Normal video and home cursor }
    if( Text_Pointer = nil ) then { Nothing in buffer }
    begin
	Erase_EOL( Screen_Window ) ; { Blank line first }
	Text_Out( Screen_Window, End_Of_Buffer_String,
	    length( End_Of_Buffer_String ) ) ; { Show end of buffer }
    end ;
    while( A <= Max_Row ) do { Until end of screen or file }
    begin
        if( Text_Pointer = nil ) then
	begin
	    Logical := '' ;
	end else
	begin
	    if(
                ( Screen_Line + A - 1 >= First_Line )
                and
		( Screen_Line + A - 1 <= Final_Line )
                and
                ( Selection )
              ) then
            begin
                { Convert from physical columns to logical columns... }
                Format( copy( Text_Pointer^.Text, 1, First_Column - 1 ) ) ;
                This_First_Column := length( Logical ) ;
                Format( copy( Text_Pointer^.Text, 1, Final_Column ) ) ;
                This_Final_Column := length( Logical ) ;
            end ;
	    Format( Text_Pointer^.Text ) ; { Get entire line }
	end ;
        State := Normal ;
        Col := 1 ; { Current column being processed }
        Image := '' ;
        while( Col <= length( Logical ) ) do { Until end of line }
        begin
            New_State := Normal ;
            After_Start := false ;
            Before_End := false ;
            if( Selection ) then { Selection under way }
            begin
	        if ( Screen_Line + A - 1 > First_Line ) or
                    ( ( Screen_Line + A - 1 = First_Line ) and
                    ( Col >= This_First_Column ) ) then
		begin
		    After_Start := true ;
		end ;
		if ( Screen_Line + A - 1 < Final_Line ) or
                    ( ( Screen_Line + A - 1 = Final_Line ) and
                    ( Col <= This_Final_Column ) ) then
		begin
		    Before_End := true ;
		end ;
            end ;
            if( After_Start and Before_End ) then { Within selection }
	    begin
		New_State := Reverse ;
	    end ;
            if( New_State <> State ) then { State change }
            begin
                Text_Out( Screen_Window, Image, length( Image ) ) ;
                Image := '' ;
                State := New_State ;
                if State = Normal then
		begin
		    Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
		    { Normal video here }
		end else
		begin
		    Set_Window( Screen_Window, 0, 0, 0, -1 ) ; { Reverse video }
		end ;
            end ; { if New_State <> State }
	    Image := Image + Logical[ Col ] ;
            inc( Col ) ; { Move to next character }
        end ; { while Col <= length( Logical ) }
        Text_Out( Screen_Window, Image, length( Image ) ) ;
        { Show remainder of line }
	Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Normal video }
        Erase_EOL( Screen_Window ) ; { Blank rest of line }
	if( A < Max_Row ) then
	begin
	    Move_To( Screen_Window, A + 1, 2 - Screen_Column ) ;
	end ;
	if( Text_Pointer <> nil ) then
	begin
	    Text_Pointer := Text_Pointer^.Next_Line ;
	    if(
		( Text_Pointer = nil ) { if at the end of the buffer }
		and
		( A < Max_Row ) { not at end of screen }
	      ) then
	    begin
		Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Normal video }
		Text_Out( Screen_Window, End_Of_Buffer_String,
		    length( End_Of_Buffer_String ) ) ; { Show end of buffer }
		Erase_EOL( Screen_Window ) ; { Blank line first }
	    end ;
	end ;
        A := A + 1 ;
    end ; { while ( A <= Max_Row ) }

    Update_Window( Screen_Window ) ;
    Position_Caret ;
    Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
{$IFDEF WINDOWS}
    Set_Cursor( Cursor_Row, Cursor_Column - Screen_Column + 1 ) ;
{$ENDIF}
end ; { TEditor.Redraw }


procedure tEditor.Handle_Selection( Row, Column : integer ;
    var _Text : String ) ; { Reverse video for selected text }

var C, CC : Byte ; { Dummy variable }
    Character : string ;
    First_Line, Final_Line : longint ;
    First_Column, Final_Column : integer ;

begin
    if Selection then { Selection in progress } 
    begin
        Get_Selection_Range( First_Line, Final_Line, First_Column, Final_Column ) ;

	if ( Screen_Line + Row - 1 >= First_Line ) and
	    ( Screen_Line + Row - 1 <= Final_Line ) then { Within select range }
	begin
	    Move_To( Screen_Window, Row, 2 - Screen_Column ) ;
	    if( First_Line = Final_Line ) then { Selection within a line }
	    begin
		Character := Copy( _Text, 1, First_Column - 1 ) ;
		Format( Character ) ;
		C := Length( Logical ) ; { end of first segment }
		Character := Copy( _Text, 1, Final_Column ) ;
		Format( Character ) ;
		CC := length( Logical ) ; { end of second segment }
		Format( _Text ) ;
		Text_Out( Screen_Window, Logical, C ) ;
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ; { Reverse video }
		Text_Out( Screen_Window, Copy( Logical, C + 1, CC - C ),
		    CC - C  ) ;
		Set_Window( Screen_Window, 0, 0, 0, -32 ) ;
		Text_Out( Screen_Window, Copy( Logical, CC + 1, 255 ),
		    length( Logical ) - CC ) ;
	    end else
	    if Screen_Line + Row - 1 = First_Line then
	    begin
		Character := Copy( _Text, 1, First_Column - 1 ) ;
		Format( Character ) ;
		C := Length( Logical ) ;
		Format( _Text ) ;
		Logical := copy( Logical, Screen_Column, 255 ) ;
		Text_Out( Screen_Window, Logical, C ) ;
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ; { Reverse video }
		Text_Out( Screen_Window, Copy( Logical, C + 1, 255 ),
		    length( Logical ) - C ) ;
	    end else
	    if( Screen_Line + Row - 1 = Final_Line ) then
	    begin
		Character := Copy( _Text, 1, Final_Column ) ;
		Format( Character ) ;
		C := Length( Logical ) ;
		Format( _Text ) ;
		Delete( Logical, 1, Screen_Column - 1 ) ;
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ; { Reverse video }
		Text_Out( Screen_Window, Logical, C ) ;
		Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Normal video }
		Text_Out( Screen_Window, Copy( Logical, C + 1, 255 ),
		    length( Logical ) - C ) ;
	    end else
	    begin
		Format( _Text ) ;
		Set_Window( Screen_Window, 0, 0, 0, -1 ) ; { Reverse video }
		Text_Out( Screen_Window, Logical, length( Logical ) ) ;
	    end ;
	    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Back to normal }
	    Update_Window( Screen_Window ) ;
	end else
        begin
            Format( _Text ) ;
            Move_To( Screen_Window, Row, 2 - Screen_Column ) ;
	    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Back to normal }
            Text_Out( Screen_Window, Logical, length( Logical ) ) ;
	    Update_Window( Screen_Window ) ;
        end ;
    end ; { if Selection }
end ; { TEditor.Handle_Selection }


procedure tEditor.Adjust ; { Adjust cursor position based on tabs, EOL, etc. }

Var L, P : Integer ; { Work variables (Character, logical, physical pointers) }
    Character : string ;

begin
    Set_Cursor_Column( Logical_Column ) ;
    {Return to proper column.  Logical_Column only changes when a vertical
     cursor movement or editing takes place.}
    if( Cursor_Column > Length( Logical_Text ) ) then {Past end of line}
    begin
	if( ( Options and eds_CursorBound ) = 0 ) then
	begin
	    Physical_Column := length( Physical_Text ) +
		Cursor_Column - length( Logical_Text ) ;
	end else {Don't exceed line length}
	begin
	    Set_Cursor_Column( Length( Logical_Text ) + 1 ) ;
	    Physical_Column := length( Physical_Text ) + 1 ;
	end ;
	Exit ;
    end ;
    L := 0 ; {Start at first character position}
    P := 1 ;
    while( L < Cursor_Column - 1 ) do
    begin
	Character := copy( Physical_Text, 1, P ) ; {Get next character}
	Format( Character ) ;
	L := length( Logical ) ; {Get next character}
	P := P + 1 ; {Look at next character}
    end ;
    Set_Cursor_Column( L + 1 ) ;
    Physical_Column := P ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ; { tEditor.Adjust }


procedure tEditor.Adjust_EOB ; { Make sure EOB is as close to bottom as possible }

var Scrolled : boolean ;

begin
    Scrolled := False ;
    while( ( Screen_Line > 1 ) and ( Screen_Line + Max_Row - 2 > File_Root^.Last_Line ) ) do
    begin
        Scrolled := True ;
	Screen_Line := Screen_Line - 1 ;
	Set_Cursor_Row( Cursor_Row + 1 ) ;
    end ;
    if( Scrolled ) then
    begin
	Notify( edn_Scroll ) ;
    end ;
end ;


procedure tEditor.Cursor_Down ; { Move cursor down one line }

var A : Integer ; { Loop variable }
    T_Pointer : Line_Pointer ; {Temporary line pointer}
    Rect : SMU.TRect ;

begin
    if( Current_Line = nil ) then {Bottom of buffer}
    begin
	Notify( edn_Bottom ) ;
	Exit ;
    end ;
    Do_Update ; {Update current line}
    Set_Cursor_Row( Cursor_Row + 1 ) ;
    Handle_Selection( Cursor_Row - 1, Window_Width + Screen_Column, Physical_Text ) ;
    Current_Line := Current_Line^.Next_Line ; {Forward}
    if( Current_Line = nil ) then {At EOB}
    begin
	Physical_Text := '' ;
	Fillchar( Logical[ 1 ], 255, 32 ) ; {Blank the line first}
	Logical := End_Of_Buffer_String ;
	Set_Cursor_Column( 1 ) ;
	Physical_Column := 1 ;
    end else
    begin
	Physical_Text := Current_Line^.Text ;
	Format( Physical_Text ) ;
	Logical_Text := Logical ;
	Adjust ;
    end ;
    if( Cursor_Row > Upper_Third_Row ) then {Scrolling time} 
	if( Screen_Line + Max_Row - 2 < File_Root^.Last_Line ) then
    begin
	Screen_Line := Screen_Line + 1 ;
	Notify( edn_Scroll ) ;
	Last_Screen_Line := Screen_Line ;
	Set_Cursor_Row( Upper_Third_Row ) ;
	Move_To( Screen_Window, Max_Row, 2 - Screen_Column ) ;
        { Go to bottom of screen}

	Rect.Top := 0 ;
	Rect.Bottom := Max_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, -1, Rect ) ;
	{Scroll up one line}

	T_Pointer := Current_Line ;
	for A := Upper_Third_Row to Max_Row - 1 do
	begin
	    if( T_Pointer <> nil ) then
	    begin
		T_Pointer := T_Pointer^.Next_Line ; {Find line}
	    end ;
	end ;
	if T_Pointer = nil then
	begin
	    Text_Out( Screen_Window, End_Of_Buffer_String,
		length( End_Of_Buffer_String ) ) ;
	    Erase_EOL( Screen_Window ) ; { Blank rest of line }
	end else
	begin
	    Format( T_Pointer^.Text ) ;
	    Text_Out( Screen_Window, Logical, length( Logical ) ) ;
	    {Display newly appearing line}
	    Handle_Selection( Max_Row, 1, T_Pointer^.Text ) ;
	end ;
	Update_Window( Screen_Window ) ;
    end ;
    Position_Caret ; {Return to cursor}
end ;


procedure tEditor.Cursor_Left ; { Move cursor left one character }

Var Character : string ;
    Dummy : Byte ; {Saved cursor column}
    Real_Length : byte ;

begin
    if( Physical_Column = 1 ) then { Going past start of line }
    begin
        if( ( Options and eds_NoCursorWrap ) <> 0 ) then
        begin
            exit ; { No cursor wrapping }
        end ;
	if( Cursor_Row + Screen_Line < 3 ) then {Top of buffer}
	begin
	    Cursor_Up ; {Show error}
	end else
	begin
	    Cursor_Up ;
	    Set_Cursor_Column( Length( Logical_Text ) + 1 ) ;
	    Physical_Column := Length( Physical_Text ) + 1 ;
	    Logical_Column := Cursor_Column ;
	end ;
	Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
	Exit ;
    end ;

    Physical_Column := Physical_Column - 1 ;
    { Physical movement is one character }

    if(
        ( ( Options and eds_CursorBound ) = 0 )
        and
        ( Physical_Column > length( Physical_Text ) )
      ) then
    begin
        Set_Cursor_Column( Cursor_Column - 1 ) ;
        Logical_Column := Cursor_Column ;
        exit ;
    end ;

    Dummy := Cursor_Column ;
    Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
    Format( Character ) ;
    Real_Length := length( Logical ) ;
    if( Real_Length <= Window_Width + Screen_Column - 1 ) then
    begin
	Set_Cursor_Column( Real_Length + 1 ) ;
    end ;
    if( Dummy > Window_Width + Screen_Column - 1 ) then
	if( Cursor_Column <= Window_Width + Screen_Column - 1 ) then
    {Cursor became visible}
    begin
	Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
	Format( Character ) ;
	Set_Cursor_Column( length( Logical ) + 1 ) ;
    end ;
    Logical_Column := Cursor_Column ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ;


procedure tEditor.Cursor_Right ; {Move cursor one character to the right}

var Character : string ;

begin
    if Current_Line = nil then {if at EOB}
    begin
	Cursor_Down ; {Display error}
	Exit ;
    end ;
    if Physical_Column > Length( Physical_Text ) then { Going past end of line }
    begin
        if( ( Options and eds_NoCursorWrap ) = 0 ) then
        begin
            if( Current_Line <> nil ) then { not at EOB }
            begin
                Do_Update ;
                Set_Cursor_Column( 1 ) ;
                Physical_Column := 1 ;
                Logical_Column := 1 ;
            end ;
            Cursor_Down ;
        end else
        begin
            if( ( Options and eds_CursorBound ) = 0 ) then
            begin
                inc( Physical_Column ) ;
                Set_Cursor_Column( Cursor_Column + 1 ) ;
                Logical_Column := Cursor_Column ;
            end else
            begin
                exit ; { Cursor is bound }
            end ;
        end ;
        exit ;
    end ;
    Physical_Column := Physical_Column + 1 ;
    { Physical movement is one character }
    Character := copy( Physical_Text, 1, Physical_Column - 1 ) ;
    Format( Character ) ;
    Set_Cursor_Column( length( Logical ) + 1 ) ;
    Logical_Column := Cursor_Column ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ; { tEditor.Cursor_Right }


procedure tEditor.Cursor_Up ;

var A : Integer ; {Loop variable}
    T_Pointer : Line_Pointer ; {Temp line pointer}
    Rect : SMU.TRect ;

begin
    if( Cursor_Row + Screen_Line < 3 ) then {Top of buffer}
    begin
	Notify( edn_Top ) ;
	Exit ;
    end ;
    Do_Update ; { Update current line }
    Set_Cursor_Row( Cursor_Row - 1 ) ;
    Handle_Selection( Cursor_Row + 1, 1, Physical_Text ) ;
    if Current_Line = nil {Sitting at EOB}
	then
	    Current_Line := File_Root^.Ending
	else
	    Current_Line := Current_Line^.Last_Line ; {Backup}
    Physical_Text := Current_Line^.Text ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Adjust ;
    if( Cursor_Row < Lower_Third_Row ) then {Scrolling time} if Screen_Line > 1 then
    begin
	Screen_Line := Screen_Line - 1 ;
	Notify( edn_Scroll ) ;
	Last_Screen_Line := Screen_Line ;
	Set_Cursor_Row( Lower_Third_Row ) ;

	Rect.Top := 0 ;
	Rect.Bottom := Max_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, 1, Rect ) ;
	{ Insert a blank line at top (scroll down) }

	T_Pointer := Current_Line ;
	for A := 1 to Lower_Third_Row - 1 do
	begin
	    T_Pointer := T_Pointer^.Last_Line ; { Find line }
	end ;
	Format( T_Pointer^.Text ) ;
	Move_To( Screen_Window, 1, 2 - Screen_Column ) ;
	Text_Out( Screen_Window, Logical, length( Logical ) ) ;
	{ Write newly appearing line }
	Update_Window( Screen_Window ) ;
	Handle_Selection( 1, Window_Width + Screen_Column, T_Pointer^.Text ) ;
    end ;
    Position_Caret ; { Return to cursor }
end ;


procedure tEditor.Delete_CRLF( Explicit : Boolean ) ;
{ Delete CRLF between last line and this line.  WARNING: Should only be called
  from Delete_Char.  Use Delete_Char to delete lines. }

Var A, B, Dummy : integer ; {Dummy variables}
    L_Pointer, T_Pointer : Line_Pointer ;
    Rect : SMU.TRect ;

begin
    T_Pointer := Current_Line ;
    if(
        ( Screen_Line = 1 )
        or
        (
          ( Cursor_Row > Middle_Row )
          and
	  ( Screen_Line + Max_Row - 2 < File_Root^.Last_Line )
        )
      ) then
    begin  { If deletion of line should cause a scroll-up }
	for Dummy := Cursor_Row to Max_Row do
	begin
	    if( T_Pointer <> nil ) then
            begin
                T_Pointer := T_Pointer^.Next_Line ;
            end ;
	end ;

	{ Scroll up one line }
	Set_Cursor_Row( Cursor_Row - 1 ) ; { Backup one line }
	Rect.Top := Cursor_Row  ;
	Rect.Bottom := Max_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, -1, Rect ) ;

	Move_To( Screen_Window, Max_Row, 1 ) ;
	A := Max_Row ; {for selection routine}
	B := 1 ;
    end else
    begin
	if T_Pointer <> nil then
	begin
	    A := 1 ;
	end else
	begin
	    A := 2 ;
	    T_Pointer := File_Root^.Ending ;
	end ;

	{Scroll down}
	Rect.Top := 0 ;
	Rect.Bottom := Cursor_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, 1, Rect ) ;
	Screen_Line := Screen_Line - 1 ;
	Notify( edn_Scroll ) ;
	Last_Screen_Line := Screen_Line ;
	for Dummy := Cursor_Row Downto A do if T_Pointer <> nil then
	begin
	    T_Pointer := T_Pointer^.Last_Line ; { Get newly appearing line }
	end ;
	Move_To( Screen_Window, 1, 1 ) ;
	A := 1 ; { for selection routine }
	B := Window_Width + Screen_Column ;
    end ;
    if( T_Pointer = nil ) then { Screen is not full }
    begin
	Fillchar( Logical, 255, 32 ) ; { Blank it }
	if( File_Root^.Last_Line > Max_Row - 1 ) then
	{ if last screen line (minus 1) is EOB }
	begin
	    Logical := End_Of_Buffer_String ; { Show EOB }
	end ;
    end else
    begin
	Format( T_Pointer^.Text ) ; { else format newly appearing line }
    end ;
    Move_To( Screen_Window, A, 2 - Screen_Column ) ;
    Text_Out( Screen_Window, Logical, length( Logical ) ) ;
    {Display newly appearing line}
    Update_Window( Screen_Window ) ;
    Position_Caret ;
    File_Root^.Last_Line := File_Root^.Last_Line - 1 ;
    Notify( edn_LineCount ) ;
    L_Pointer := Current_Line ;
    if( L_Pointer = nil ) then
    begin
	L_Pointer := File_Root^.Ending ;
    end ;
    if( File_Root^.Ending = L_Pointer ) then {Deleted last line}
    begin
	File_Root^.Ending := L_Pointer^.Last_Line ;
    end ;
    if( File_Root^.Start = L_Pointer ) then {Deleted first line}
    begin
	File_Root^.Start := L_Pointer^.Last_Line ;
    end ;
    if( L_Pointer^.Last_Line <> nil ) then
    begin
	L_Pointer^.Last_Line^.Next_Line := L_Pointer^.Next_Line ;
    end ;
    if( L_Pointer^.Next_Line <> nil ) then
    begin
	L_Pointer^.Next_Line^.Last_Line := L_Pointer^.Last_Line ;
    end ;
    {Relink previous and next text lines}
    T_Pointer := Current_Line ;
    if Current_Line <> nil then Current_Line := Current_Line^.Last_Line ;
    {Backup one text line}
    Handle_Selection( A, B, T_Pointer^.Text ) ;
    if T_Pointer <> nil then
    begin
	Freemem( T_Pointer, Length( T_Pointer^.Text ) + Line_Overhead ) ;
	{Delete line}
    end ;
end ;


procedure tEditor.New_Line ; {Insert a new line/create two lines from one}

var Character : String ;
    Dummy : Integer ;
    T_Pointer : Line_Pointer ;
    Rect : SMU.TRect ;

begin
    if( Maxavail < 256 ) then
    begin
	Notify( edn_Memory ) ; {Error}
	exit ;
    end ;
    Character := Copy( Physical_Text, Physical_Column, 255 ) ;
    { Save last part of line }
    
    if( Physical_Column <= length( Physical_Text ) ) then
    begin
        SetLength( Physical_Text, Physical_Column - 1 ) ;
    end ;
    Do_Update ; {Truncate the current line}
    Dummy := -32 ; {Normal video}
    if( Selection and ( not Selection_Is_Frozen ) ) then if Anchor.Line < Screen_Line + Cursor_Row then
    begin
	Dummy := -1 ;
	{ Reverse video if after beginning of selection }
    end ;
    Set_Window( Screen_Window, 0, 0, 0, Dummy ) ; {Set to proper video mode}
    Format( Physical_Text ) ;
    Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
    Text_Out( Screen_Window, Logical, length( Logical ) ) ;
    { Display changed/new line }
    Erase_EOL( Screen_Window ) ; { Blank the line }
    Set_Window( Screen_Window, 0, 0, 0, -32 ) ; { Normal video }
    Physical_Text := Character ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Set_Cursor_Column( 1 ) ;
    Physical_Column := 1 ;
    Logical_Column := 1 ;
    File_Root^.Last_Line := File_Root^.Last_Line + 1 ;
    Notify( edn_LineCount ) ;
    if Current_Line = nil then {At EOB}
    begin
	Physical_Text := '' ; {Blank line}
	Getmem( T_Pointer, Line_Overhead ) ; {Allocate memory for blank line}
	T_Pointer^.Text := '' ;
	T_Pointer^.Next_Line := nil ;
	T_Pointer^.Last_Line := File_Root^.Ending ;
	if T_Pointer^.Last_Line <> nil then
	begin
	    T_Pointer^.Last_Line^.Next_Line := T_Pointer ;
	end ;
	File_Root^.Ending := T_Pointer ;
	if( File_Root^.Start = nil ) then
	begin
	    File_Root^.Start := File_Root^.Ending ;
	end ;
	Logical := End_Of_Buffer_String ;
	Logical_Text := '' ;
    end else
    begin
	Getmem( T_Pointer, Length( Physical_Text ) + Line_Overhead ) ;
	T_Pointer^.Text := Physical_Text ;
	T_Pointer^.Last_Line := Current_Line ;
	T_Pointer^.Next_Line := Current_Line^.Next_Line ;
	Current_Line^.Next_Line := T_Pointer ;
	if( T_Pointer^.Next_Line <> nil ) then
	begin
	    T_Pointer^.Next_Line^.Last_Line := T_Pointer ; {Insert new line}
	end ;
	if( File_Root^.Ending = Current_Line ) then
	begin
	    File_Root^.Ending := T_Pointer ;
	end ;
	Current_Line := T_Pointer ; {At new line}
    end ;

    if( ( Cursor_Row < Middle_Row ) or
	( Screen_Line + max_Row - 1 > File_Root^.Last_Line ) ) then
    begin
	Set_Cursor_Row( Cursor_Row + 1 ) ; {Scroll down}
        Rect.Top := Cursor_Row ;
	Rect.Bottom := Max_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, 1, Rect ) ;
    end else
    begin
	Screen_Line := Screen_Line + 1 ; {Scroll up}
	Notify( edn_Scroll ) ;
	Last_Screen_Line := Screen_Line ;
	Rect.Top := 0 ;
	Rect.Bottom := Cursor_Row ;
	Rect.Left := 0 ;
	Rect.Right := Screen_Window^.Columns ;
	Scroll_Window( Screen_Window, 0, -1, Rect ) ;
    end ;
    Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
    Text_Out( Screen_Window, Logical, length( Logical ) ) ;
    Update_Window( Screen_Window ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ;


procedure tEditor.Delete_Char_Back ;
{Delete character, taking indent into consideration}

var Character : string ;
    Dummy : integer ;
    Found_NonSpace : integer ;
    T_Pointer : Line_Pointer ;

    function Find_New_Column : boolean ;

    var Dummy : integer ;

    begin
	Find_New_Column := False ;
	Format( Character ) ;
	Found_NonSpace := 0 ;
	for Dummy := 1 to Cursor_Column - 1 do
	begin
	    if( Logical[ Dummy ] <> ' ' ) then
	    begin
		if( Found_NonSpace = 0 ) then
		begin
		    Found_NonSpace := Dummy ;
		end ;
	    end ;
	end ;
	if( Found_NonSpace > 0 ) then {Found a line with earlier indent}
	begin
	    Cursor_Column := Found_NonSpace ;
	    Flag_Changes ;

	    {Find last position in current line prior to cursor}
	    Dummy := 1 ;
	    Character := copy( Physical_Text, 1, Dummy ) ;
	    Format( Character ) ;
	    while( ( Get_Text_Extent( Screen_Window, Logical, length( Logical ) ) and 32767 ) <
		Cursor_Column ) do
	    begin
		Dummy := Dummy + 1 ;
		Character := copy( Physical_Text, 1, Dummy ) ;
		Format( Character ) ;
	    end ;
	    Dummy := Dummy - 1 ;
	    Character := copy( Physical_Text, 1, Dummy ) ;
	    Format( Character ) ;

	    {Pad up to cursor position}
	    while( ( Get_Text_Extent( Screen_Window, Logical, length( Logical ) ) and 32767 ) <
		Cursor_Column - 1 ) do {Need to add spaces}
	    begin
		Physical_Column := Physical_Column + 1 ;
		Physical_Text := copy( Physical_Text, 1, Dummy ) + ' ' +
		    copy( Physical_Text, Dummy + 1, 255 ) ;
		Dummy := Dummy + 1 ;
		Character := copy( Physical_Text, 1, Dummy ) ;
		Format( Character ) ;
	    end ;
	    Physical_Text := copy( Physical_Text, 1, Dummy ) +
		copy( Physical_Text, Physical_Column, 255 ) ;
	    Redraw_Line ;
	    Find_New_Column := True ;
	    exit ;
	end ;
    end ;

begin
    if( ( Physical_Column = 1 ) or {Deleting CRLF}
	( ( Options and eds_AutoIndent ) = 0 ) ) then {No auto-indent}
    begin
	Delete_Char( True ) ; {Normal delete}
	exit ;
    end ;

    if ( Options and eds_CursorBound ) = 0 then { Unbound cursor }
    begin
	if Physical_Column > length( Physical_Text ) + 1 then
	{if past end of line, jump to last character}
	begin
	    Set_Cursor_Column( length( Logical_Text ) + 1 ) ;
            Physical_Column := length( Physical_Text ) + 1 ;
	    exit ;
	end ;
    end ;

    if( ( Physical_Text[ Physical_Column - 1 ] <> HT ) and
	( Physical_Text[ Physical_Column - 1 ] <> ' ' ) ) then
    {not deleting a space}
    begin
	Delete_Char( True ) ;
	exit ;
    end ;

    for Dummy := Physical_Column - 1 downto 1 do
    begin
	if( ( Physical_Text[ Dummy ] <> HT ) and
	    ( Physical_Text[ Dummy ] <> ' ' ) ) then
	{Non-white space to the left of the cursor}
	begin
	    Delete_Char( True ) ;
	    exit ;
	end ;
        if( Physical_Column > length( Physical_Text ) + 1 ) then
        begin
            dec( Physical_Column ) ;
        end else
        begin
            Delete_Char( True ) ;
        end ;
    end ;

    {Find a previous line with an earlier indent}
    T_Pointer := Current_Line^.Last_Line ;
    while( T_Pointer <> nil ) do
    begin
	Character := T_Pointer^.Text ;
	if( Find_New_Column ) then
	begin
	    exit ;
	end ;
	T_Pointer := T_Pointer^.Last_Line ; {Try previous line}
    end ;

    {No previous line, try to find a subsequent line with an earlier indent}
    T_Pointer := Current_Line^.Next_Line ;
    while( T_Pointer <> nil ) do
    begin
	Character := T_Pointer^.Text ;
	if( Find_New_Column ) then
	begin
	    exit ;
	end ;
	T_Pointer := T_Pointer^.Next_Line ; {Try next line}
    end ;

    {All the way back}
    Physical_Text := copy( Physical_Text, Physical_Column, 255 ) ;
    Physical_Column := 1 ;
    Logical_Column := 1 ;
    Cursor_Column := 1 ;
    Redraw( Cursor_Row ) ;
end ;


procedure tEditor.Delete_Char( Explicit : Boolean ) ;
{ Delete character to left of cursor position }

Var Character : String ;
    Do_Left_Arrow : Boolean ;
    New_Column : integer ;
    Last_Line : Line_Pointer ;

begin
    Flag_Changes ;
    if( Physical_Column = 1 ) then {Deleting CRLF}
    begin
	if( Current_Line = nil ) then {At EOB}
	begin
	    if( File_Root^.Ending = nil ) then
	    begin
		Do_Left_Arrow := True ;
	    end else
	    begin
		if Length( File_Root^.Ending^.Text ) > 0 then
		begin
		    Do_Left_Arrow := True ;
		end else
		begin
		    Do_Left_Arrow := False ;
		end ;
	    end ;
	    if Do_Left_Arrow then {Nothing in file or last line not blank}
	    begin
		Cursor_Left ; {Equivalent of left arrow}
		Exit ;
	    end ;
	    Logical := '' ;
	    Fillchar( Logical[ 1 ], 255, 32 ) ;
	    Physical_Text := '' ;
	    Last_Line := File_Root^.Ending ;
	end else
	begin
	    Last_Line := Current_Line^.Last_Line ;
	end ;
	if Last_Line = nil then {At first line}
	begin
	    Notify( edn_Top ) ;
	    Exit ;
	end ;
	if( Length( Physical_Text ) + Length( Last_Line^.Text ) > 255 ) then
	begin
	    Notify( edn_Length ) ;
	    Exit ;
	end ;
	New_Column := Length( Physical_Text ) ; {Length of current line}
	if( Current_Line <> nil ) then
	begin
	    Format( Last_Line^.Text ) ;
	end ;
	Set_Cursor_Column( Length( Logical ) + 1 ) ; {Position cursor}
	if( Current_Line <> nil ) then
	begin
	    Physical_Text := Current_Line^.Last_Line^.Text + Physical_Text ;
	end ; { Concatenate current line and next line }
	Format( Physical_Text ) ;
	Logical_Text := Logical ;
	Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
        { Position to current line }
	Delete_CRLF( Explicit ) ;
	if( Current_Line = nil ) then
	begin
	    Logical_Text := End_Of_Buffer_String ;
	end ;
	Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
	Text_Out( Screen_Window, Logical_Text, length( Logical_Text ) ) ;
	{ Display newly appearing line }
	Erase_EOL( Screen_Window ) ;
	Update_Window( Screen_Window ) ;
	Physical_Column := Length( Physical_Text ) - New_Column + 1 ;
	Logical_Column := Cursor_Column ;
	if( Cursor_Column > Window_Width + Screen_Column - 1 ) then
	begin
	    Set_Cursor_Column( Window_Width + Screen_Column - 1 ) ;
	end ;
	Adjust ;
	Exit ;
    end ; { Delete CRLF }

    { Deleting something other than CRLF }
    if ( Options and eds_CursorBound ) = 0 then {Unbound cursor}
    begin
	if Physical_Column > length( Physical_Text ) + 1 then
	begin
	    Cursor_Left ;
	    exit ;
	end ;
    end ;

    Physical_Text := Copy( Physical_Text, 1, Physical_Column - 2 ) +
	Copy( Physical_Text, Physical_Column, Length( Physical_Text ) ) ;
    {Delete character}
    Physical_Column := Physical_Column - 1 ;
    Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
    Format( Character ) ;
    New_Column := length( Logical ) + 1 ;
    Set_Cursor_Column( Length( Logical ) + 1 ) ;
    Logical_Column := Cursor_Column ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    if( Cursor_Column <= Window_Width + Screen_Column - 1 ) then
    begin
	Position_Caret ;
	Text_Out( Screen_Window, Copy( Logical, New_Column, 255 ),
	    length( Logical ) - New_Column + 1 ) ; {Show changed line}
	Erase_EOL( Screen_Window ) ;
	Update_Window( Screen_Window ) ;
    end ;
    Position_Caret ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ;


procedure tEditor.Insert_Chars( Character : String ) ; {Character(s) insertion}

var Dummy : integer ;
    Command0 : string ;

begin
    Flag_Changes ;
    if( Character = CR ) then {Carriage Return}
    begin
	if ( ( Options and eds_AutoIndent ) = eds_AutoIndent ) and
	    ( Character = CR ) then {Auto-indent is active}
	begin
	    Dummy := 1 ;
	    if ( Physical_Text[ Dummy ] = ' ' ) or 
		( Physical_Text[ Dummy ] = HT ) then
	    {if line begins with space or tab}
	    begin
		while ( ( Physical_Text[ Dummy ] = ' ' ) or
		    ( ( ( Options and eds_NoTabs ) = 0 ) and ( Physical_Text[ Dummy ] = Chr( 9 ) ) ) )
		    and ( Dummy <= Length( Physical_Text ) ) do Inc( Dummy ) ;
		{Find first non-space/non-tab}
	    end ;
	    Command0 := Copy( Physical_Text, 1, Dummy - 1 ) ;
	    {Get space/tab portion}
	end else
	begin
	    Command0 := '' ;
	end ;
	New_Line ; {Break into two lines}
	if ( Options and eds_AutoIndent ) = eds_AutoIndent then
	{do auto-indent, if needed}
	begin
	    if Length( Command0 ) > 0 then
	    begin
		Insert_Chars( Command0 ) ;
	    end ;
	end ;
	exit ;
    end ;

    {Non-CR insertion}
    if Current_Line = nil then
    begin
	New_Line ; {Insert line at EOB}
	Cursor_Up ; {Stay at new line}
	Flag_Changes ;
    end ;

    { Handle situation when cursor is past end of line... }
    Format( Physical_Text ) ;
    while( length( Logical ) < Cursor_Column - 1 ) do
    begin
        Physical_Text := Physical_Text + ' ' ;
        Format( Physical_Text ) ;
    end ;

    { Check for maximum line length of 255 characters }
    if Length( Physical_Text ) + Length( Character ) > 255 then
    begin
	Notify( edn_Length ) ;
	Exit ;
    end ;

    { Insert the character(s)... }
    Physical_Text := Copy( Physical_Text, 1, Physical_Column - 1 ) + Character +
	Copy( Physical_Text, Physical_Column, 255 ) ;
    Physical_Column := Physical_Column + Length( Character ) ;
    Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
    Format( Character ) ;
    Set_Cursor_Column( Length( Logical ) + 1 ) ;
    Logical_Column := Cursor_Column ;
    Format( Physical_Text ) ;
    Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
    Text_Out( Screen_Window, Logical, length( Logical ) ) ;
    { Output updated line }
    Logical_Text := Logical ;
    Update_Window( Screen_Window ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ; { tEditor.Insert_Chars }


function tEditor.Screen_To_Text_Column( Line, Column : longint ) : longint ;
{ Translate from screen line/column position to physical text position }

var Dummy : integer ;
    Found : boolean ;
    This_Line : Line_Pointer ;
    This_Text : string ;

begin
    Do_Update ;

    { Bounds checks... }
    if( Column < 1 ) then
    begin
        Column := 1 ;
    end ;
    if( Line < 1 ) then
    begin
        Line := 1 ;
    end ;
    if( Line > File_Root^.Last_Line ) then
    begin
        Line := File_Root^.Last_Line + 1 ;
    end ;

    { Get line text... }
    This_Line := Get_Line_Pointer( Line ) ;
    if( This_Line = nil ) then
    begin
        This_Text := '' ;
    end else
    begin
        This_Text := This_Line^.Text ;
    end ;

    { Now translate into actual physical position... }
    Screen_To_Text_Column := 0 ;
    Found := False ;
    for Dummy := 1 to length( This_Text ) do
    begin
        Format( copy( This_Text, 1, Dummy ) ) ;
        if( Column <= length( Logical ) ) then
        begin
            Screen_To_Text_Column := Dummy ;
            Found := True ;
            break ;
        end ;
    end ;
    if( not Found ) then { Past end of line }
    begin
        Screen_To_Text_Column := Column - length( Logical ) + length( This_Text ) + 1 ;
    end ;
end ; { tEditor.Screen_To_Text_Column }


procedure tEditor.Search( Search_Text : string ;
    Start_Line, Start_Column : longint ;
    var Result_Line, Result_Column : longint ) ;

var A : integer ;
    This_Line : longint ;
    T_Pointer : Line_Pointer ;

begin
    Do_Update ; { Update current line }

    { Get line to start search at }
    if( Start_Line = -1 ) then { Current line }
    begin
	T_Pointer := Current_Line ;
	Start_Line := Screen_Line + Cursor_Row - 1 ;
    end else
    begin
	T_Pointer := Get_Line_Pointer( Start_Line ) ;
    end ;
    if( T_Pointer = nil ) then { Start line is out of range }
    begin
	Result_Line := 0 ;
	Result_Column := 0 ;
	exit ;
    end ;

    { do search... }
    if( Start_Column > length( T_Pointer^.Text ) ) then
    begin
        A := 0 ;
    end else
    begin
        { Search for text in first line }
        if( ( Options and eds_Case_Sensitive ) <> 0 ) then
        begin
            A := Pos( Search_Text,
                Copy( T_Pointer^.Text, Start_Column,
                length( T_Pointer^.Text ) ) ) ;
        end else
        begin
            A := Pos( Search_Text,
                Copy( Edit( T_Pointer^.Text, 32 ), Start_Column,
                length( T_Pointer^.Text ) ) ) ;
        end ;
    end ;
    if( A > 0 ) then { Found }
    begin
	Result_Line := Start_Line ;
	Result_Column := A + Start_Column - 1 ;
	exit ;
    end ;

    T_Pointer := T_Pointer^.Next_Line ;
    This_Line := Start_Line + 1 ;
    while( T_Pointer <> nil ) do { Until end of buffer }
    begin
        if( ( Options and eds_Case_Sensitive ) <> 0 ) then
        begin
	    A := Pos( Search_Text, T_Pointer^.Text ) ;
	end else
        begin
	    A := Pos( Search_Text, CommonUt.Edit( T_Pointer^.Text, 32 ) ) ;
        end ;
        if( A > 0 ) then { Found it }
	begin
	    Result_Line := This_Line ;
	    Result_Column := A ;
	    Exit ;
	end ;
	T_Pointer := T_Pointer^.Next_Line ;
	This_Line := This_Line + 1 ;
    end ;
    Result_Line := 0 ;
    Result_Column := 0 ; { Not found }
end ; { tEditor.Search }


procedure tEditor.Position_Cursor( Line, Column : longint ) ;
{ Position cursor and column to specified location }

var Character : string ;
    Do_Redraw : Boolean ; { TRUE if Redraw needs to be done }

begin
    Do_Update ;
    Do_Redraw := Selection ; { Must Redraw if selection }

    { Bounds checks... }
    if( Column < 1 ) then
    begin
        Column := 1 ;
    end ;
    if( Line < 1 ) then
    begin
        Line := 1 ;
    end ;
    if( Line > File_Root^.Last_Line ) then
    begin
        Line := File_Root^.Last_Line + 1 ;
    end ;

    { Get line text... }
    Current_Line := Get_Line_Pointer( Line ) ;
    if( Current_Line = nil ) then
    begin
        Physical_Text := '' ;
    end else
    begin
        Physical_Text := Current_Line^.Text ;
    end ;

    { Set the physical column to the requested column }
    Physical_Column := Column ;

    { Now set the position... }
    if( ( Options and eds_CursorBound ) <> 0 ) then { Bound cursor }
    begin
        Character := Copy( Physical_Text, 1, Column - 1 ) ;
        Format( Character ) ;
        Set_Cursor_Column( Length( Logical ) + 1 ) ;
    end else
    begin
        Set_Cursor_Column( Column ) ;
    end ;
    if( ( Options and eds_SubWindow ) <> 0 ) then
    begin
        if( Line >= Screen_Line + Upper_Third_Row ) then { Out of range }
        begin
            Screen_Line := Line - Middle_Row ;
            { Put us at mid line by default }
            if( Screen_Line + Max_Row - 2 > File_Root^.Last_Line ) then
            begin
                Screen_Line := File_Root^.Last_Line - Max_Row + 2 ;
                { Don't go too far }
            end ;
            if( Screen_Line < 1 ) then
            begin
                Screen_Line := 1 ; { ...in either direction}
            end ;
            Notify( edn_Scroll ) ;
            Do_Redraw := True ;
        end else
        begin
            if( Line < Screen_Line + Lower_Third_Row ) then
            begin
                Screen_Line := Line - Middle_Row ;
                if Screen_Line < 1 then
                begin
                    Screen_Line := 1 ;
                end ;
                Notify( edn_Scroll ) ;
                Do_Redraw := True ;
            end ;
        end ;
    end else
    begin
        if( Line >= Screen_Line + Max_Row ) then {Out of range}
        begin
            Screen_Line := Line - Max_Row ;
            if Screen_Line < 1 then
            begin
                Screen_Line := 1 ;
            end ;
            Notify( edn_Scroll ) ;
            Do_Redraw := True ;
        end else
        begin
            if( Line < Screen_Line ) then
            begin
                Screen_Line := Line ;
                if( Screen_Line < 1 ) then
                begin
                    Screen_Line := 1 ;
                end ;
                Notify( edn_Scroll ) ;
                Do_Redraw := True ;
            end ;
        end ;
    end ;
    
    Logical_Column := Cursor_Column ;
    Set_Cursor_Row( Trunc( Line - Screen_Line ) + 1 ) ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Position_Caret ;
    if( Do_Redraw ) then
    begin
	Redraw( 1 ) ;
    end ;
end ; { tEditor.Position_Cursor }


function tEditor.Do_Search( Search_Text : string ) : boolean ; { Search for text }

var Column : longint ;
    This_Line : Longint ;

begin
    Search( Search_Text, -1, Physical_Column + 1, This_Line, Column ) ;
    if( This_Line > 0 ) then { Found }
    begin
	Position_Cursor( This_Line, Column ) ;
	Do_Search := True ;
	Exit ;
    end ;
    Notify( edn_NotFound ) ;
    Do_Search := False ;
end ;


procedure tEditor.Advance ; { Advance to beginning of next line }

begin
    Cursor_Down ;
    Set_Cursor_Column( 1 ) ;
    Logical_Column := 1 ;
    Physical_Column := 1 ;
    if( Screen_Column > 1 ) then
    begin
	Screen_Column := 1 ;
	Last_Screen_Line := 0 ;
        Notify( edn_Scroll ) ;
	Redraw( 1 ) ;
    end else
    begin
	Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
    end ;
end ;


procedure tEditor.Lowercase_Word ; { Change current word to lowercase }

var Starting, Ending : integer ;

begin
    { Bounds check }
    if( Physical_Column > length( Physical_Text ) + 1 ) then
    begin
        exit ; { No current word }
    end ;

    { Find start of word }
    Starting := Physical_Column ;
    while(
           ( Starting > 0 )
           and
           ( Physical_Text[ Starting ] <> ' ' )
           and
           ( Physical_Text[ Starting ] <> HT )
         ) do
    begin
        dec( Starting ) ;
    end ;
    inc( Starting ) ;

    { Find end of word }
    Ending := Starting ;
    while(
           ( Ending <= length( Physical_Text ) + 1 )
           and
           ( Physical_Text[ Ending ] <> ' ' )
           and
           ( Physical_Text[ Ending ] <> HT )
         ) do
    begin
        inc( Ending ) ;
    end ;
    dec( Ending ) ;
    if( Starting >= Ending ) then { No current word }
    begin
        exit ;
    end ;

    { Modify word }
    Changes := True ;
    while( Starting < Ending ) do
    begin
        if( ( Physical_Text[ Starting ] >= 'A' ) and ( Physical_Text[ Starting ] <= 'Z' ) ) then
        begin
            Physical_Text[ Starting ] := chr( ord( Physical_Text[ Starting ] ) + 32 ) ;
        end ;
        inc( Starting ) ;
    end ;
    Redraw_Line ;
end ;


procedure tEditor.Move_Word ; { Move to beginning of next word }

begin
    if( Forwards ) then
    begin
        Word_Forward ;
    end else
    begin
        Word_Backward ;
    end ;
end ;


procedure tEditor.Copy_Text( Paste_Buffer : pLine_Buffer ) ;
{ Copy selected text into the paste buffer }

Var First_Line, Final_Line, Root_Line : Longint ;
    First_Column, Final_Column : integer ;
    L_Pointer, T_Pointer : Line_Pointer ;

begin
    if( not Selection ) then { Must have something selected in order to cut }
    begin
	Notify( edn_NoSelection ) ;
	Exit ;
    end ;

    Do_Update ; { Make sure everything is updated }

    { Determine start/end lines to cut }
    Get_Selection_Range( First_Line, Final_Line, First_Column, Final_Column ) ;

    if( First_Line = Final_Line ) then { Cut within a single line }
    begin
	Paste_Buffer^.Append( Copy( Physical_Text, First_Column,
	    Final_Column - First_Column + 1 ) ) ; {Get selected characters}
	Exit ;
    end ;

    {Multi-line cut operation}
    L_Pointer := Get_Line_Pointer( First_Line ) ; {Find first line to cut}
    if( L_Pointer = nil ) then
    begin
	Exit ; {Hit end of buffer}
    end ;
    if( Maxavail <= 255 + Line_Overhead ) then {Ran out of memory}
    begin
	Notify( edn_Memory ) ;
	exit ;
    end ;
    Paste_Buffer^.Append( Copy( L_Pointer^.Text, First_Column, 255 ) ) ;
    {Get portion of first line which is selected}

    Root_Line := First_Line + 1 ;
    T_Pointer := L_Pointer ;
    while( Root_Line < Final_Line ) do {Find last line to cut}
    begin
	if( T_Pointer <> nil ) then {not at buffer end}
	begin
	    T_Pointer := T_Pointer^.Next_Line ;
	    if( T_Pointer <> nil ) then
	    begin
		if( Maxavail <= 255 + Line_Overhead ) then {Ran out of memory}
		begin
		    Notify( edn_Memory ) ;
		    exit ;
		end ;
		Paste_Buffer^.Append( T_Pointer^.Text ) ;
	    end ;
	end ;
	Root_Line := Root_Line + 1 ;
    end ;
    if T_Pointer <> nil then {not at buffer end}
    begin
	T_Pointer := T_Pointer^.Next_Line ;
	if( T_Pointer <> nil ) then
	begin
	    if( Maxavail <= 255 + Line_Overhead ) then {Ran out of memory}
	    begin
		Notify( edn_Memory ) ;
		exit ;
	    end ;
	    Paste_Buffer^.Append( copy( T_Pointer^.Text, 1, Final_Column - 1 ) ) ;
	end else
	begin
	    Paste_Buffer^.Append( '' ) ;
	end ;
    end ;
end ;


procedure tEditor.Delete_Lines( Start, Count : integer ) ; { Delete lines }

var Current, Temp : Line_Pointer ;
    Previous : Line_Pointer ;

begin
    { Bounds checks }
    if( ( Count < 1 ) or ( Start + Count < 0 ) or ( Start > File_Root^.Last_Line ) ) then
    begin
        exit ;
    end ;

    { Delete the lines }
    Do_Update ;
    Current := Get_Line_Pointer( Start ) ;
    Previous := Current^.Last_Line ;
    while( ( Count > 0 ) and ( Current <> nil ) ) do
    begin
        Temp := Current ;
        Current := Current^.Next_Line ; { Move to next line }
	Freemem( Temp, Length( Temp^.Text ) + Line_Overhead ) ;
        dec( Count ) ; { One less line to delete }
        dec( File_Root^.Last_Line ) ; { One less line in buffer }
    end ;
    Notify( edn_LineCount ) ;

    { Relink the lines to either side of the deleted range... }
    if( Previous <> nil ) then
    begin
        Previous^.Next_Line := Current ;
    end ;
    if( Current <> nil ) then
    begin
        Current^.Last_Line := Previous ;
    end ;

    Adjust_EOB ;
end ; { tEditor.Delete_Lines }


procedure tEditor.Delete_Text ; { Delete selected text }

var Character : string ;
    Dummy : integer ; {Dummy variable}
    First_Line, Final_Line, Root_Line : Longint ;
    First_Column, Final_Column : integer ;
    L_Pointer, T_Pointer : Line_Pointer ;

begin
    if( not Selection ) then { Must have something selected in order to cut }
    begin
	Notify( edn_NoSelection ) ;
	exit ;
    end ;
    Selection := False ; {Delete operation ends selection}

    Do_Update ; {Make sure everything is updated}

    { Determine start/end lines to cut }
    Get_Selection_Range( First_Line, Final_Line, First_Column, Final_Column ) ;

    if( First_Line = Final_Line ) then { Cut within a single line }
    begin
        Physical_Column := First_Column ;

	Character := Copy( Physical_Text, 1, First_Column - 1 ) ;
	{Get portion of line before selected portion}
	Format( Character ) ;
	Set_Cursor_Column( Length( Logical ) + 1 ) ; {Resulting position}
	Physical_Text := Copy( Physical_Text, 1, First_Column - 1 ) +
	    Copy( Physical_Text, Final_Column + 1, 255 ) ;
	    {Extract cut text from line}
	Format( Physical_Text ) ;
	Logical_Text := Logical ;
	Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
	Text_out( Screen_Window, Logical, length( Logical ) ) ;
	{Show newly updated line}
	Erase_EOL( Screen_Window ) ;
	Update_Window( Screen_Window ) ;
	Position_Caret ;
	Flag_Changes ;
	Exit ;
    end ;

    {Multi-line cut operation}
    T_Pointer := Get_Line_Pointer( First_Line ) ;
    if( Current_Line = nil ) then {if at EOB}
    begin
	File_Root^.Ending := T_Pointer ;
    end ;
    Current_Line := T_Pointer ;
    T_Pointer := Current_Line^.Next_Line ;
    Root_Line := First_Line + 1 ;
    while( Root_Line < Final_Line ) do {Find last line to cut}
    begin
	if( T_Pointer <> nil ) then {not at buffer end}
	begin
	    L_Pointer := T_Pointer ;
	    T_Pointer := T_Pointer^.Next_Line ;
	    Freemem( L_Pointer, length( L_Pointer^.Text ) + Line_Overhead ) ;
	    File_Root^.Last_Line := File_Root^.Last_Line - 1 ;
	    {One less line in main buffer}
	end ;
	Root_Line := Root_Line + 1 ;
    end ;

    {At this point: Current_Line points to first line of selection
		    T_Pointer points to last line of selection}

    if( T_Pointer <> nil ) then {not at EOB}
    begin
	T_Pointer^.Last_Line := Current_Line ;
    end ;
    Current_Line^.Next_Line := T_Pointer ;

    {Now adjust first and last selected lines}
    Physical_Text := Current_Line^.Text ;
    SetLength( Physical_Text, First_Column - 1 ) ; {Adjust first line}
    Flag_Changes ;
    Do_Update ;
    Current_Line := Current_Line^.Next_Line ; {Go to next line}
    if Current_Line = nil then {At EOB}
    begin
	Physical_Text := '' ;
    end else
    begin
	Physical_Text := Copy( Current_Line^.Text, Final_Column, 255 ) ;
	{Adjust last line}
    end ;
    Flag_Changes ;
    Do_Update ;

    {Now get rid of extra line created by above process}
    Set_Cursor_Column( 1 ) ;
    Physical_Column := 1 ;
    Logical_Column := 1 ;
    Delete_Char( False ) ;
    Do_Update ;

    { Now show updated screen }
    Set_Cursor_Row( Trunc( First_Line - Screen_Line ) + 1 ) ;
    Adjust_EOB ;

    Last_Screen_Line := 0 ; {Force Redraw}
    if ( Cursor_Row < Lower_Third_Row ) or ( Cursor_Row > 127 ) then {Adjust screen}
    begin
	Dummy := Lower_Third_Row - Cursor_Row ; {Number of lines to adjust}
	Screen_Line := Screen_Line - Dummy ;
	Set_Cursor_Row( Cursor_Row + Dummy ) ;
	while Screen_Line < 1 do {First line can't be before line 1}
	begin
	    Screen_Line := Screen_Line + 1 ;
	    Set_Cursor_Row( Cursor_Row - 1 ) ;
	end ;
	Notify( edn_Scroll ) ;
    end ;
    Clear_Window( Screen_Window ) ; {Erase screen and home cursor}
    Position_Caret ;
    Redraw( 1 ) ; {Redraw screen}
end ;


procedure tEditor.EOL ; {Move to end of current line}

begin
    if( Current_Line = nil ) then
    begin
        exit ; {do nothing if at EOB}
    end ;
    if( Physical_Column > Length( Physical_Text ) ) then {At EOL}
    begin
	Cursor_Down ; {Goto EOL of next line}
    end ;
    Format( Physical_Text ) ;
    Logical_Column := length( Logical_Text ) + 1 ;
    Set_Cursor_Column( Logical_Column ) ;
    Physical_Column := length( Physical_Text ) + 1 ;
end ;


procedure tEditor.Page ; { Move one page }

begin
    if( Forwards ) then
    begin
        Page_Down ;
    end else
    begin
        Page_Up ;
    end ;
end ;


procedure tEditor.Page_Down ; { Advance one page }

var A, Dummy : integer ; {Dummy variables}

begin
    if( Current_Line = nil ) then Exit ; { Done if at EOB }
    Do_Update ;
    Physical_Text := '' ;

    { Move forward one screen, but not past EOB }
    A := 0 ;
    for Dummy := 1 to Max_Row - 1 do if( Current_Line <> nil ) then
    begin
	Current_Line := Current_Line^.Next_Line ;
	A := A + 1 ;
    end ;
    Set_Cursor_Row( Cursor_Row + A ) ;

    if( ( Cursor_Row > Middle_Row ) and ( File_Root^.Last_Line > Max_Row - 1 ) ) then
    { Need to scroll }
    begin
	Screen_Line := Screen_Line + Cursor_Row - Middle_Row ;
	Set_Cursor_Row( Middle_Row ) ;
	if( Screen_Line + Max_Row - 1 > File_Root^.Last_Line ) then
	{ Leave EOB at bottom }
	begin
	    Set_Cursor_Row( Cursor_Row + Trunc( Screen_Line + Max_Row -
		2 - File_Root^.Last_Line ) ) ;
	    Screen_Line := File_Root^.Last_Line - ( max_Row - 2 ) ;
	end ;
	Notify( edn_Scroll ) ;
	Redraw( 1 ) ;
    end ;

    if( Current_Line <> nil ) then
    begin
        Physical_Text := Current_Line^.Text ;
    end ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Adjust ;
end ; { tEditor.Page_Down }


procedure tEditor.Page_Up ; { Backup one page }

var A, Dummy : integer ; { Dummy variables }

begin
    if( ( Screen_Line = 1 ) and ( Cursor_Row = 1 ) ) then
    begin
        Exit ; { Done if at BOB }
    end ;
    Do_Update ;
    Physical_Text := '' ;

    if( Screen_Line + Cursor_Row - 1 <= Max_Row ) then { Already at the top }
    begin
        Screen_Line := 1 ;
        Cursor_Row := 1 ;
        Current_Line := File_Root^.Start ;
        Notify( edn_Scroll ) ;
        Redraw( 1 ) ;
    end else
    begin
        { Move bacward one screen, but not past BOB }
        if( Current_Line = nil ) then { At EOB }
        begin
            Current_Line := File_Root^.Ending ;
            A := 1 ;
        end else
        begin
            A := 0 ;
        end ;
        for Dummy := 1 to Max_Row - 1 do
        begin
            if( Current_Line^.Last_Line <> nil ) then
            begin
                Current_Line := Current_Line^.Last_Line ;
                A := A + 1 ;
            end ;
        end ;
        Set_Cursor_Row( Cursor_Row - A ) ;

        if( ( Cursor_Row < Middle_Row ) and ( Screen_Line > 1 ) ) then
        { Need to scroll }
        begin
            Screen_Line := Screen_Line + Cursor_Row - Middle_Row ;
            Set_Cursor_Row( Middle_Row ) ;
            if( Screen_Line < 1 ) then
            begin
                Set_Cursor_Row( Cursor_Row + Screen_Line ) ;
                Screen_Line := 1 ;
            end ;
            Notify( edn_Scroll ) ;
            Redraw( 1 ) ;
        end ;
    end ;

    Physical_Text := Current_Line^.Text ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Adjust ;
end ; { tEditor.Page_Up }


procedure tEditor.Get_Text( T : PChar ) ; { Return text of entire buffer }

var Current : Line_Pointer ;
    Dummy : longint ;

begin
    Current := File_Root^.Start ;
    Dummy := 0 ;
    while( Current <> nil ) do
    begin
        move( Current^.Text[ 1 ], T[ Dummy ], length( Current^.Text ) ) ;
        Dummy := Dummy + length( Current^.Text ) ;
        T[ Dummy ] := #13 ;
        T[ Dummy + 1 ] := #10 ;
        Dummy := Dummy + 2 ;
        Current := Current^.Next_Line ;                  
    end ;
    T[ Dummy ] := #0 ;
end ; { tEditor.Get_Text }


procedure tEditor.Half_Page ; { Advance one-half page }

var A, Dummy : integer ; { Dummy variables }

begin
    if( Current_Line = nil ) then Exit ; { Done if already at EOB }
    Do_Update ;
    Physical_Text := '' ;

    { Move forward n lines, but not past EOB }
    A := 0 ;
    for Dummy := 1 to Middle_Row do if( Current_Line <> nil ) then
    begin
	Current_Line := Current_Line^.Next_Line ;
	A := A + 1 ;
    end ;
    Set_Cursor_Row( Cursor_Row + A ) ;

    if( ( Cursor_Row > Middle_Row ) and ( File_Root^.Last_Line > Max_Row - 1 ) ) then
    { Need to scroll }
    begin
	Screen_Line := Screen_Line + Cursor_Row - Middle_Row ;
	{ Move down n lines }
	Set_Cursor_Row( Middle_Row ) ; { Fix cursor }
	if( Screen_Line + Max_Row - 1 > File_Root^.Last_Line ) then
	{ Leave EOB at bottom }
	begin
	    Set_Cursor_Row( Cursor_Row + Trunc( Screen_Line + Max_Row -
		2 - File_Root^.Last_Line ) ) ;
	    Screen_Line := File_Root^.Last_Line - ( Max_Row - 2 ) ;
	end ;
	Notify( edn_Scroll ) ;
	Redraw( 1 ) ;
    end ;

    if( Current_Line <> nil ) then
    begin
        Physical_Text := Current_Line^.Text ;
    end ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Adjust ;
end ;


procedure tEditor.Select ; {begin selection}

begin
    if( not Selection ) then
    begin
	Selection := True ;
	Anchor.Line := Screen_Line + Cursor_Row - 1 ;
	Anchor.P_Column := Physical_Column ;
	Anchor.Column := Cursor_Column ;
        Selection_Is_Frozen := False ;
    end else
    begin
	Notify( edn_Selection ) ;
    end ;
end ;


procedure tEditor.Select_Text( Start_Line, Start_Column, End_Line, End_Column : integer ) ;
{ Select specified text }

begin
    if( not Selection ) then
    begin
        if( Start_Line < 1 ) then
        begin
            Start_Line := 1 ;
        end ;
        if( Start_Column < 1 ) then
        begin
            Start_Column := 1 ;
        end ;
        if( End_Line > File_Root^.Last_Line ) then
        begin
            End_Line := File_Root^.Last_Line ;
        end ;
	Selection := True ;
	Anchor.Line := Start_Line ;
	Anchor.P_Column := Start_Column ;
	Anchor.Column := Start_Column ;
        Selection_Is_Frozen := True ;
        Anchor.End_Line := End_Line ;
        if( End_Column < 255 ) then { Avoid overflow }
        begin
            End_Column := End_Column + 1 ;
        end ;
        Anchor.End_P_Column := End_Column ;
        Anchor.End_Column := End_Column ;
    end else
    begin
	Notify( edn_Selection ) ;
    end ;
end ;


procedure TEditor.Freeze_Selection ;

begin
    if( Selection ) then
    begin
        if( Selection_Is_Frozen ) then
        begin
            Notify( edn_Frozen ) ;
        end else
        begin
            Selection_Is_Frozen := True ;
            Anchor.End_Line := Screen_Line + Cursor_Row - 1 ;
            Anchor.End_P_Column := Physical_Column ;
            Anchor.End_Column := Cursor_Column ;
        end ;
    end else
    begin
        Notify( edn_NoSelection ) ;
    end ;
end ;


procedure tEditor.Redraw_Line ; {Redraw current line}

begin
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    if( Quiet_Mode ) then
    begin
	exit ;
    end ;
    Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
    if( Current_Line = nil ) then { At EOB }
    begin
	Logical := End_Of_Buffer_String ;
    end ;
    Text_Out( Screen_Window, Logical, length( Logical ) ) ; {Show new line}
    Erase_EOL( Screen_Window ) ;
    Update_Window( Screen_Window ) ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
    Position_Caret ;
end ;


procedure TEditor.Word_Range( var Starting, Ending : integer ) ;

begin
    { Bounds check... }
    if( Physical_Column > Length( Physical_Text ) ) then { At EOL }
    begin
        Starting := 0 ;
        Ending := 0 ;
	exit ;
    end ;

    { Find start of word }
    Starting := Physical_Column ;
    Ending := Starting ;
    if(
        ( Ending <= length( Physical_Text ) )
        and
	( Physical_Text[ Ending ] <> ' ' )
        and
	( Physical_Text[ Ending ] <> HT )
      ) then { Starting on a non-white space character }
    begin
	{ Skip to next white space character }
	while(
               ( Ending <= Length( Physical_Text ) )
               and
	       ( Physical_Text[ Ending ] <> ' ' ) { not space }
               and
	       ( Physical_Text[ Ending ] <> HT ) { not tab }
             ) do
	begin
	    inc( Ending ) ;
	end ;
    end ;

    { Find end of word }
    if( Starting = Ending ) then { Move at least one character }
    begin
	inc( Ending ) ;
    end ;
    while(
           ( Ending <= Length( Physical_Text ) )
           and
	   ( Physical_Text[ Ending ] = ' ' )
         ) do
    begin
	{ Jump to next non-space character }
	inc( Ending ) ;
    end ;
end ; { TEditor.Word_Range }


procedure tEditor.Delete_Word ; { Delete current word }

var A : integer ;

begin
    if( Physical_Column > Length( Physical_Text ) ) then { At EOL }
    begin
	Cursor_Right ;
	Delete_Char( False ) ;
	exit ;
    end ;

    { Delete word... }
    Word_Range( A, Physical_Column ) ;
    Physical_Text := Copy( Physical_Text, 1, A - 1 ) +
	Copy( Physical_Text, Physical_Column, 255 ) ;
    Flag_Changes ;
    Physical_Column := A ; {Restore physical column}
    Redraw_Line ;
end ;


procedure tEditor.Delete_Current_Line ; {Delete to EOL}

begin
    if( Current_Line = nil ) then {At EOB}
    begin
	Notify( edn_Bottom ) ;
	Exit ;
    end ;
    if ( Current_Line^.Next_Line = nil ) and ( Length( Physical_Text ) > 0 ) and
	( Physical_Column > Length( Physical_Text ) ) then
	{if at end of last line, ignore}
    begin
	Cursor_Right ;
	Exit ;
    end ;

    if( Physical_Column <= length( Physical_Text ) ) then
    begin
        SetLength( Physical_Text, Physical_Column - 1 ) ;
    end ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Flag_Changes ;
    if( Current_Line^.Next_Line = nil ) then {On last line in buffer}
    begin
	if( Physical_Column = 1 ) then {Deleted whole line}
	begin
	    Delete_Char( False ) ; {Delete CRLF}
	    Cursor_Right ; {Go to EOB}
	end ;
    end else
    begin
	Cursor_Right ; {Go to beginning of next line}
	Delete_Char( False ) ; {Delete CRLF}
    end ;
    Redraw_Line ;
end ;


procedure tEditor.Unselect ; { Unselect text }

begin
    if( not Selection ) then Exit ; { No selection in progress }
    Selection := False ;
    Do_Update ;
    Redraw( 1 ) ; { Remove reverse-video }
end ;


procedure tEditor.Uppercase_Word ; { Convert current word to uppercase }

var Starting, Ending : integer ;

begin
    if( Physical_Column > length( Physical_Text ) + 1 ) then
    begin
        exit ; { No current word }
    end ;

    { Find start of word }
    Starting := Physical_Column ;
    while(
           ( Starting > 0 )
           and
           ( Physical_Text[ Starting ] <> ' ' )
           and
           ( Physical_Text[ Starting ] <> HT )
         ) do
    begin
        dec( Starting ) ;
    end ;
    inc( Starting ) ;

    { Find end of word }
    Ending := Starting ;
    while(
           ( Ending <= length( Physical_Text ) + 1 )
           and
           ( Physical_Text[ Ending ] <> ' ' )
           and
           ( Physical_Text[ Ending ] <> HT )
         ) do
    begin
        inc( Ending ) ;
    end ;
    dec( Ending ) ;
    if( Starting >= Ending ) then { No current word }
    begin
        exit ;
    end ;

    { Change word }
    Changes := True ;
    while( Starting < Ending ) do
    begin
        Physical_Text[ Starting ] := upcase( Physical_Text[ Starting ] ) ;
        inc( Starting ) ;
    end ;
    Redraw_Line ;
end ;


procedure tEditor.Word_Backward ;

var A : integer ;
    Character : string ;

begin
    if( Physical_Column = 1 ) then { beginning of line }
    begin
	Cursor_Left ;
    end else { Move by word }
    begin
        dec( Physical_Column ) ;
	A := Physical_Column ;
        { Skip to next white space character }
        while(
               ( Physical_Column > 1 )
               and
               (
                 ( Physical_Column > Length( Physical_Text ) )
                 or
                 ( Physical_Text[ Physical_Column ] = ' ' ) { space }
                 or
                 ( Physical_Text[ Physical_Column ] = HT ) { tab }
               )
             ) do
        begin
            Physical_Column := Physical_Column - 1 ;
	end ;
	if( A = Physical_Column ) then { Move at least one character }
	begin
	    Physical_Column := Physical_Column - 1 ;
	end ;
	while(
               ( Physical_Column > 1 )
               and
               (
                 ( Physical_Column > Length( Physical_Text ) )
                 or
	         ( Physical_Text[ Physical_Column ] <> ' ' )
               )
             ) do
	begin
	    { Jump to next non-space character }
	    Physical_Column := Physical_Column - 1 ;
	end ;
        if( Physical_Column < 1 ) then
        begin
            Physical_Column := 1 ;
        end ;
	Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
	Format( Character ) ;
	Set_Cursor_Column( Length( Logical ) + 1 ) ;
	Logical_Column := Cursor_Column ;
    end ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ;


procedure tEditor.Word_Forward ;

var A : integer ;
    Character : string ;

begin
    if Physical_Column > Length( Physical_Text ) then {end of line}
    begin
        if( ( Options and eds_NoCursorWrap ) <> 0 ) then
        begin
            exit ;
        end ;
	Cursor_Down ;
	Set_Cursor_Column( 1 ) ;
	Logical_Column := 1 ;
	Physical_Column := 1 ;
    end else {Move by word}
    begin
	A := Physical_Column ;
	if ( Physical_Column <= length( Physical_Text ) ) and
	    ( Physical_Text[ Physical_Column ] <> ' ' ) and
	    ( Physical_Text[ Physical_Column ] <> '     ' ) then
	    {Starting on a non-white space character}
	begin
	    {Skip to next white space character}
	    while ( Physical_Column <= Length( Physical_Text ) ) and
		( Physical_Text[ Physical_Column ] <> ' ' ) and {not space}
		( Physical_Text[ Physical_Column ] <> HT ) {not tab} do
	    begin
		Physical_Column := Physical_Column + 1 ;
	    end ;
	end ;
	if A = Physical_Column then {Move at least one character}
	begin
	    Physical_Column := Physical_Column + 1 ;
	end ;
	while ( Physical_Column <= Length( Physical_Text ) ) and
	    ( Physical_Text[ Physical_Column ] = ' ' ) do
	begin
	    {Jump to next non-space character}
	    Physical_Column := Physical_Column + 1 ;
	end ;
	Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
	Format( Character ) ;
	Set_Cursor_Column( Length( Logical ) + 1 ) ;
	Logical_Column := Cursor_Column ;
    end ;
    Handle_Selection( Cursor_Row, Physical_Column, Physical_Text ) ;
end ;


procedure tEditor.Insert_Blank_Line ; { Insert blank line }

var Character : string[ 2 ] ;
    Saved_Row, Saved_Column : integer ;

begin
    { Save current position... }
    Saved_Column := Cursor_Column ;
    Saved_Row := Screen_Line + Cursor_Row - 1 ;

    { Insert the new line... }
    Character := CR ;
    Insert_Chars( Character ) ;

    { Return to previous position }
    Position_Cursor( Saved_Row, Saved_Column ) ;
end ;


procedure tEditor.Change_Case ;

begin
    if( Physical_Column > Length( Physical_Text ) ) then Exit ;
    { Ignore if past end of line or at EOB }
    if ( Physical_Text[ Physical_Column ] >= 'A' ) and
	( Physical_Text[ Physical_Column ] <= 'Z' )
	    then
		Physical_Text[ Physical_Column ] :=
		    Chr( Ord( Physical_Text[ Physical_Column ] ) + 32 )
	    else
		if ( Physical_Text[ Physical_Column ] >= 'a' ) and
		    ( Physical_Text[ Physical_Column ] <= 'z' )
			then
			    Physical_Text[ Physical_Column ] :=
			    Chr( Ord( Physical_Text[ Physical_Column ] ) - 32 )
			else
			    begin
				Cursor_Right ; {Skip non-alpha characters}
				Exit ;
			    end ;
    Position_Caret ;
    Text_Out( Screen_Window, Physical_Text[ Physical_Column ], 1 ) ;
    Update_Window( Screen_Window ) ;
    Flag_Changes ;
    Cursor_Right ;
end ;


procedure tEditor.Bottom ;

var Draw : boolean ;

begin
    Do_Update ;
    Draw := False ;
    if( Screen_Line + Max_Row - 2 < File_Root^.Last_Line ) then
    {EOB not on screen}
    begin
	Screen_Line := File_Root^.Last_Line - ( Max_Row - 2 ) ;
	Set_Cursor_Row( Trunc( File_Root^.Last_Line - Screen_Line ) + 2 ) ;
	Draw := True ;
    end ;
    if( Screen_Column > 1 ) then
    begin
	Last_Screen_Line := 0 ;
	Screen_Column := 1 ;
	Draw := True ;
    end ;
    Set_Cursor_Column( 1 ) ;
    Set_Cursor_Row( Trunc( File_Root^.Last_Line - Screen_Line ) + 2 ) ;
    Position_Caret ;
    Current_Line := nil ;
    Physical_Text := '' ;
    Logical_Text := '' ;
    Logical_Column := 1 ;
    if( Draw ) then
    begin
	Notify( edn_Scroll ) ;
    end ;
    if( Selection or Draw ) then
    begin
	Redraw( 1 ) ;
    end ;
end ;


procedure tEditor.Set_Line( X : integer ; Y : string ) ;

var Line_P, T_Pointer : Line_Pointer ;
    Update_Current : boolean ;

begin
    if( X = -1 ) then { Modify current line }
    begin
        Physical_Text := Y ;
        Changes := True ;
        exit ;
    end ;

    Do_Update ;
    Line_P := Get_Line_Pointer( X ) ;
    if( Line_P <> nil ) then
    begin
        Update_Current := ( Current_Line = Line_P ) ;
	if Maxavail < 256 then {Ran out of memory}
	begin
	    Notify( edn_Memory ) ;
	    exit ;
	end ;
        Any_Changes := True ;
	Getmem( T_Pointer, Length( Y ) + Line_Overhead ) ;
	{Allocate new line}
	T_Pointer^.Last_Line := Line_P^.Last_Line ; {Link in new line}
	T_Pointer^.Next_Line := Line_P^.Next_Line ;
	if T_Pointer^.Last_Line <> nil then {Link previous line to new one}
	    T_Pointer^.Last_Line^.Next_Line := T_Pointer ;
	if T_Pointer^.Next_Line <> nil then {Link next line to new one}
	    T_Pointer^.Next_Line^.Last_Line := T_Pointer ;
	T_Pointer^.Text := Y ; {Store text for line}
	if File_Root^.Ending = Line_P then
            File_Root^.Ending := T_Pointer ;
	if File_Root^.Start = Line_P then
            File_Root^.Start := T_Pointer ;
	Freemem( Line_P, Length( Line_P^.Text ) + Line_Overhead ) ;
	{Delete old line}

        if( Update_Current ) then
        begin
            Current_Line := T_Pointer ;
            Physical_Text := Y ;
        end ;
    end ;
end ; { tEditor.Set_Line }


function tEditor.Something_Selected : boolean ; { True if something selected }

var First_Line, Final_Line : longint ;
    First_Column, Final_Column : integer ;

begin
    Something_Selected := False ;
    if( Selection ) then
    begin
        Get_Selection_Range( First_Line, Final_Line, First_Column, Final_Column ) ;
        if( ( First_Line <> Final_Line ) or ( First_Column <= Final_Column ) ) then
        begin
            Something_Selected := True ;
        end ;
    end ;
end ;


function tEditor.Text_Size : longint ; { Size of all text, in bytes }

var Dummy : Line_Pointer ; { Current line }
    Total : longint ;

begin
    Total := File_Root^.Last_Line * 2 ; { CRLF for all lines }
    Dummy := File_Root^.Start ;
    while( Dummy <> nil ) do
    begin
        Total := Total + length( Dummy^.Text ) ;
        Dummy := Dummy^.Next_Line ;
    end ;
    Text_Size := Total ; { Return result }
end ; { tEditor.Text_Size }


function tEditor.Text_To_Screen_Column( Line, Column : longint ) : longint ;

var This_Line : Line_Pointer ;
    This_Text : string ;

begin
    Do_Update ;

    { Bounds checks... }
    if( Column < 1 ) then
    begin
        Text_To_Screen_Column := 1 ;
        exit ;
    end ;
    if( Line < 1 ) then
    begin
        Line := 1 ;
    end ;
    if( Line > File_Root^.Last_Line ) then
    begin
        Line := File_Root^.Last_Line + 1 ;
    end ;

    { Get line text... }
    This_Line := Get_Line_Pointer( Line ) ;
    if( This_Line = nil ) then
    begin
        This_Text := '' ;
    end else
    begin
        This_Text := This_Line^.Text ;
    end ;

    { Now translate into actual physical position... }
    Format( copy( This_Text, 1, Column - 1 ) ) ;
    Text_To_Screen_Column := length( Logical ) + 1 ;
end ; { tEditor.Text_To_Screen_Column }


procedure tEditor.Top ; { Go to top of buffer }

var Draw : boolean ;

begin
    Do_Update ;
    Set_Cursor_Row( 1 ) ;
    Set_Cursor_Column( 1 ) ;
    Physical_Column := 1 ;
    Draw := False ;
    if Screen_Line > 1 then
    begin
	Screen_Line := 1 ;
	Draw := True ;
    end ;
    if( Screen_Column > 1 ) then
    begin
	Last_Screen_Line := 0 ;
	Screen_Column := 1 ;
	Draw := True ;
    end ;
    Current_Line := File_Root^.Start ;
    Physical_Text := Current_Line^.Text ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Logical_Column := 1 ;
    Position_Caret ;
    if( Draw ) then
    begin
	Notify( edn_Scroll ) ;
    end ;
    if( Selection or Draw ) then
    begin
	Redraw( 1 ) ;
    end ;
end ;


procedure tEditor.Paste( Paste_Buffer : pLine_Buffer ) ;
{Insert text from paste buffer}

Var A, Dummy : integer ; {Dummy variables}
    Character : string ;
    Flag : Boolean ; {True if cursor should remain at EOB}
    F_Pointer, L_Pointer, N_Pointer, T_Pointer : Line_Pointer ;

begin
    if( Paste_Buffer^.Last_Line = 0 ) then
    begin
	Exit ; {Nothing in paste buffer}
    end ;
    if( Paste_Buffer^.Last_Line = 1 ) then {No implicit CRLF}
    begin
	if Length( Physical_Text ) + Length( Paste_Buffer^.Start^.Text ) > 255 then
	begin
	    Notify( edn_Length ) ;
	    Exit ;
	end ;
	Physical_Text := Copy( Physical_Text, 1, Physical_Column - 1 ) +
	    Paste_Buffer^.Start^.Text + Copy( Physical_Text, Physical_Column, 255 ) ;
	Physical_Column := Physical_Column + Length( Paste_Buffer^.Start^.Text ) ;
	Format( Physical_Text ) ;
	Logical_Text := Logical ;
	Character := Copy( Physical_Text, 1, Physical_Column - 1 ) ;
	Format( Character ) ;
	Set_Cursor_Column( Length( Logical ) + 1 ) ;
	Flag_Changes ;
	Move_To( Screen_Window, Cursor_Row, 2 - Screen_Column ) ;
	Text_Out( Screen_Window, Logical_Text, length( Logical_Text ) ) ;
	Update_Window( Screen_Window ) ;
	Position_Caret ;
	Exit ;
    end ;
    Do_Update ;
    T_Pointer := Paste_Buffer^.Start ;
    F_Pointer := nil ;
    L_Pointer := nil ;
    N_Pointer := nil ;
    Flag := False ;
    A := -1 ; {Initialize for count of inserted lines}
    if Current_Line = nil then {At EOB} A := 0 ; {One more line}
    while( T_Pointer <> nil ) do {Make copy of paste buffer}
    begin
	if( Maxavail < 256 + Line_Overhead ) then {Ran out of memory}
	begin
	    Notify( edn_Memory ) ;
	    exit ;
	end ;
	Getmem( N_Pointer, Length( T_Pointer^.Text ) + Line_Overhead ) ;
	{Create next line}
	if( F_Pointer = nil ) then {if first line} F_Pointer := N_Pointer ;
	N_Pointer^.Text := T_Pointer^.Text ; {Copy next line}
	T_Pointer := T_Pointer^.Next_Line ; {Go to next line of past buffer}
	N_Pointer^.Last_Line := L_Pointer ;
	if( L_Pointer <> nil ) then L_Pointer^.Next_Line := N_Pointer ;
	L_Pointer := N_Pointer ; {Save this pointer for later linking}
	A := A + 1 ; {Increment count of new lines}
    end ;
    if( Current_Line = nil ) then { At EOB }
    begin
	if( Length( N_Pointer^.Text ) = 0 ) then { if last line is just CRLF }
	begin
	    A := A - 1 ;
	    L_Pointer := N_Pointer ;
	    N_Pointer := N_Pointer^.Last_Line ;
	    Freemem( L_Pointer, Line_Overhead ) ;
	    Flag := True ;
	end else
	begin
	    Set_Cursor_Row( Cursor_Row - 1 ) ;
	    Flag := False ;
	end ;
    end ;
    N_Pointer^.Next_Line := nil ;
    File_Root^.Last_Line := File_Root^.Last_Line + A ; { Last line in buffer }
    Notify( edn_LineCount ) ;
    Set_Cursor_Row( Cursor_Row + A ) ; {Jump down to bottom of pasted text}
    if( Cursor_Row > Upper_Third_Row ) then if( File_Root^.Last_Line > Max_Row - 1 ) then
    { Must adjust screen }
    begin
	Dummy := Cursor_Row - Upper_Third_Row ; {Number of lines to move up}
	while( Screen_Line + Max_Row - 1 - Dummy > File_Root^.Last_Line ) do
	begin
	    Dummy := Dummy - 1 ; {EOB should be at bottom}
	end ;
	while( Screen_Line + Dummy + Max_Row - 2 > File_Root^.Last_Line ) do
	begin
	    Dummy := Dummy - 1 ;
	end ;
	Set_Cursor_Row( Cursor_Row - Dummy ) ;
	Last_Screen_Line := 0 ; {Force Redraw}
	Screen_Line := Screen_Line + Dummy ;
	Notify( edn_Scroll ) ;
    end ;
    {At this point, F_Pointer is the beginning of the paste buffer copy
     and N_Pointer is the end of the paste buffer copy}
    if File_Root^.Start = nil then {Nothing in main buffer}
    begin
        File_Root^.Start := F_Pointer ;
        File_Root^.Ending := N_Pointer ;
    end else {Something in file}
    begin
        if( Current_Line = nil ) then {At EOB}
        begin
            File_Root^.Ending^.Next_Line := F_Pointer ;
            F_Pointer^.Last_Line := File_Root^.Ending ;
            File_Root^.Ending := N_Pointer ;
            if not Flag then
            begin
                Current_Line := File_Root^.Ending ;
                Physical_Text := Current_Line^.Text ;
                Format( Physical_Text ) ;
                Logical_Text := Logical ;
                Set_Cursor_Column( Length( Logical ) ) ;
                Logical_Column := Cursor_Column ;
                Physical_Column := Length( Physical_Text ) ;
            end ;
        end else
        begin
            if Physical_Column + Length( Paste_Buffer^.Start^.Text ) > 256 then
            begin
                Notify( edn_Length ) ;
                Exit ;
            end ;
            if Length( N_Pointer^.Text ) + Length( Physical_Text ) + 1 -
                Physical_Column > 255 then
            begin
                Notify( edn_Length ) ;
                Exit ;
            end ;
            Character := Copy( Physical_Text, Physical_Column, 255 ) ;
            {Text following cursor position}
            Physical_Text := Copy( Physical_Text, 1, Physical_Column - 1 ) +
                Paste_Buffer^.Start^.Text ; {Recreate first line}
            Flag_Changes ;
            Do_Update ;
            N_Pointer^.Next_Line := Current_Line^.Next_Line ;
            {Save pointer to line after pasted text}
            if( N_Pointer^.Next_Line <> nil ) then
            begin
                N_Pointer^.Next_Line^.Last_Line := N_Pointer ;
            end ;
            Current_Line^.Next_Line := F_Pointer^.Next_Line ;
            Current_Line^.Next_Line^.Last_Line := Current_Line ;
            Freemem( F_Pointer, Length( F_Pointer^.Text ) +
                Line_Overhead ) ;

            if( Current_Line = File_Root^.Ending ) then
            begin
                File_Root^.Ending := N_Pointer ;
            end ;
            Current_Line := N_Pointer ; {Go to line after paste buffer}
            Physical_Text := Current_Line^.Text ; {Get text}
            Format( Physical_Text ) ;
            Logical_Text := Logical ;
            Set_Cursor_Column( Length( Logical_Text ) + 1 ) ;
            Physical_Text := N_Pointer^.Text + Character ;
            {Recreate last line}
            Flag_Changes ;
            Do_Update {Change line}
        end {if Current_Line=nil ELSE}
    end ; {if File_Root^.Start=nil ELSE}
    Position_Caret ;
    Redraw( 1 ) ;
end ; {tEditor.Paste}


constructor tEditor.Init ;

begin
    Initialize ;
end ;


destructor tEditor.Done ;

begin
    { Don't send anymore notifications... }
    Notification_Routine := nil ;
{$IFDEF DELPHI}
    Notification_Routine1 := nil ;
{$ENDIF}

    { Release allocated resources... }
    Clear( False ) ;
    if( Screen_Window <> nil ) then
    begin
        Delete_Window( Screen_Window ) ;
    end ;
end ;


procedure tEditor.Set_Buffer( X : pLine_Buffer ) ;

begin
    Calculate_Rows( Window_Height ) ;
    File_Root := X ;
    Current_Line := File_Root^.Start ;
    if( Current_Line <> nil ) then
    begin
	Physical_Text := Current_Line^.Text ;
    end else
    begin
	Physical_Text := '' ;
    end ;
    Format( Physical_Text ) ;
    Logical_Text := Logical ;
    Last_Screen_Line := 0 ; {Force redraw}
    Cursor_Row := 1 ;
    Cursor_Column := 1 ;
    Screen_Line := 1 ;
    Logical_Column := 1 ;
    Physical_Column := 1 ;
end ;


procedure TEditor.Initialize ;

var Dummy : integer ;

begin
{$IFDEF DELPHI}
    {$HINTS OFF}
    System.Initialize( Logical_Text ) ;
    System.Initialize( Physical_Text ) ;
    {$HINTS ON}
{$ENDIF}
    for Dummy := 0 to 255 do
    begin
	Translation[ Dummy ] := chr( Dummy ) ;
    end ;
    Translation[ 0 ] := chr( 178 ) ;
    Translation[ 8 ] := chr( 178 ) ;
    Translation[ 254 ] := chr( 178 ) ;
{$IFDEF WINDOWS}
    Border := 1 ;
    End_Of_Buffer_String := '' ;
    End_Of_Line_Character := ' ' ;
{$ELSE}
    Border := 0 ;
    End_Of_Buffer_String := '[EOB]' ;
    End_Of_Line_Character := '' ; {Diamond}
{$ENDIF}
    Quiet_Mode := False ;
    Tab_Size := 8 ;
    Any_Changes := False ; {No changesd yet}
    Screen_Line := 1 ; {Line in file that corresponds to screen line 1}
    Last_Screen_Line := 0 ; {Force Redraw on initial display}
    Cursor_Row := 1 ; {Cursor position on screen}
    Cursor_Column := 1 ; {Cursor column on line}
    Physical_Column := 1 ;
    Logical_Column := 1 ;
    Changes := False ;
    Selection := False ;
    Error_Flag := False ;
    Forwards := True ;
    Notification_Routine := nil ;
    Screen_Column := 1 ;
    File_Root := new( pLine_Buffer, Init ) ;
    Current_Line := nil ;
    Screen_Window := nil ;
end ;


{$IFDEF Test}
procedure TEditor.Test_Class ;

begin
    Initialize ;
end ;



procedure Test_Unit ;

var Editor : TEditor ;

begin
    Editor.Test_Class
end ;
{$ENDIF}


end.
