{
        Program Name : _Propert
        Package Name : Subroutine Library - Layer 0
        Purpose      : Dynamic properties definition.
        Institution  : Conroy & Conroy Co.
        Date Written : 14-February-2005
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2005 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TDynamic_Properties is an abstract base class definition for objects
        which provide dynamic object properties.
}

unit _Propert ;

interface

{$I EDEFINES.INC}

uses _Common, { TObject }
     ComInter, { TCommon_COM_Object }
     _GenVal ; { TGeneric_Value }

type TSDynamic_Properties = object( TObject )
                                public { API... }
                                    { Adds a new property, returning the index.  Returns
                                      -1 if a property already exists by that name. }
                                    function Add( const Nam : string ;
                                        Value : PSGeneric_Value ) : longint ; virtual ;

                                    { Deletes all properties. }
                                    procedure Clear ; virtual ;

                                    { Count of defined properties... }
                                    function Count : longint ; virtual ;

                                    { Returns the index of the property with the
                                      specified name.  Returns -1 if the named
                                      property doesn't exist. }
                                    function Index_Of( const Nam : string ) : longint ;
                                        virtual ;

                                    { Retrieves string name of property with
                                      given index.  If index is invalid, returns
                                      a null string. }
                                    function Get_Name( Index : longint ) : string ;
                                        virtual ;

                                    { Gets the value for the specified index.  Returns
                                      nil if the index is invalid. }
                                    function Get_Value( Index : longint ) : PSGeneric_Value ;
                                        virtual ;

                                    { Sets the value for the specified index. }
                                    procedure Set_Value( Index : longint ;
                                        Value : PSGeneric_Value ) ; virtual ;
                            end ;
     PSDynamic_Properties = ^TSDynamic_Properties ;

{$IFDEF DELPHI}
type TDynamic_Properties = class( TDelphi_Object )
                               public { API... }
                                   function Add( const Nam : string ;
                                        Value : TGeneric_Value ) : longint ;
                                        virtual ; abstract ;

                                    procedure Clear ; virtual ; abstract ;

                                    function Count : longint ;
                                        virtual ; abstract ;

                                    function Index_Of( const Nam : string ) : longint ;
                                        virtual ; abstract ;

                                    function Get_Name( Index : longint ) : string ;
                                        virtual ; abstract ;

                                    function Get_Value( Index : longint ) : TGeneric_Value ;
                                        virtual ; abstract ;

                                    procedure Set_Value( Index : longint ;
                                        Value : TGeneric_Value ) ;
                                        virtual ; abstract ;
                           end ;

type TCOM_Dynamic_Properties = class( TCommon_COM_Interface )
                                   public { API... }
                                       function Add( Nam : PChar ;
                                            Value : TCOM_Generic_Value ) : longint ;
                                            virtual ; stdcall ; abstract ;

                                        procedure Clear ;
                                            virtual ; stdcall ; abstract ;

                                        function Count : longint ;
                                            virtual ; stdcall ; abstract ;

                                        function Index_Of( Nam : PChar ) : longint ;
                                            virtual ; stdcall ; abstract ;

                                        function Get_Name( Index : longint ) : PChar ;
                                            virtual ; stdcall ; abstract ;

                                        function Get_Value( Index : longint ) : TCOM_Generic_Value ;
                                            virtual ; stdcall ; abstract ;

                                        procedure Set_Value( Index : longint ;
                                            Value : TCOM_Generic_Value ) ;
                                            virtual ; stdcall ; abstract ;
                               end ;
{$ENDIF}

implementation

{ TSDynamic_Properties methods... }

{ API... }

function TSDynamic_Properties.Add( const Nam : string ;
    Value : PSGeneric_Value ) : longint ;

begin
    Add := -1 ;
    Abstract_Error ;
end ;


procedure TSDynamic_Properties.Clear ;

begin
    Abstract_Error ;
end ;


function TSDynamic_Properties.Count : longint ;

begin
    Count := 0 ;
    Abstract_Error ;
end ;


function TSDynamic_Properties.Index_Of( const Nam : string ) : longint ;

begin
    Index_Of := -1 ;
    Abstract_Error ;
end ;


function TSDynamic_Properties.Get_Name( Index : longint ) : string ;

begin
    Get_Name := '' ;
    Abstract_Error ;
end ;


function TSDynamic_Properties.Get_Value( Index : longint ) : PSGeneric_Value ;

begin
    Get_Value := nil ;
    Abstract_Error ;
end ;


procedure TSDynamic_Properties.Set_Value( Index : longint ;
    Value : PSGeneric_Value ) ;

begin
    Abstract_Error ;
end ;



end.
