/*
              Program Name : BACKUP
              Package Name : LB
              Purpose      : Backup
              Institution  : Conroy & Conroy Co.
              Date Written : 10-Jul-1991
	      Last Modified: 04-Jan-1994
              Written By   : Alan Conroy
              Version      : 2.1A

		TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

	04-Jan-1994	EAC	Allow use of command file.

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

	  This program is a callable backup utility.
*/

#include <dos.h>
#include <dir.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <math.h>
#include <errno.h>
#include <alloc.h>
#include <sys\stat.h>

#include "clib.h"

typedef struct	{
		    char Saveset[ 13 ] ; /* Name of saveset */
		    int Version ; /* Version of saveset */
		    int Volume ; /* Volume of saveset */
		    int Date_Y, Date_M, Date_D ; /* Date of saveset */
		    int Expire_Y, Expire_M, Expire_D ;
		    /* Expiration date */
		    char Filler[ 15 ] ; /* Reserved */
		} Volume_Header ;
typedef struct	{
		    char Compression ; /* Type of compression.  0 = none */
		    char Attributes ; /* DOS File attributes */
		    int Time ; /* Update time */
		    int Date ; /* Update date */
		    long int Size ; /* DOS file size */
		    long int Saveset_Size ; /* Size on saveset */
		    long int Start ;
		    /* Where in DOS file the following data bytes begin */
		    long int Flags ; /* Reserved */
		    char Filename[ 256 ] ; /* Name */
		} File_Header ;

static int Abort ; /* True to abort operation */

void Abort_Backup()
{
    Abort = True ; /* Set abort flag */
} ;

static char *strerr( int X )
{
    static char E[ 256 ] ;

    strcpy( E, strerror( X ) ) ;
    Edit( E, E, 4 ) ;
    return &E ;
} ;

static void ( *Ep )() ;
static void ( *Request )() ;
static File_Header Fh ;
static char *Buffer ;
static char Command[ 256 ] ; /* Input or command line */
static char Command_File[ 256 ] ; /* Command file name */
static int Control ; /* Loop control variable */
static unsigned int Count ; /* Count of bytes to read or write */
static char Current_File[ 256 ] ; /* Name of current file being processed */
static int Dummy ; /* Dummy loop variable */
static int Dummy1 ;
static struct FILE *Fcommand ; /* Command file */
static int hFcommand ; /* Handle to Fcommand */
static int Ignore ; /* True if ignoring creation errors */
static char Image[ 16384 ] ; /* Image of FAT/directory to write */
static int Initialize ; /* True if initializing output device */
static char Input_Spec[ 256 ], Output_Spec[ 256 ] ;
static struct FILE *Input_File, *Output_File ;
static int hInput_File, hOutput_File ; /* Handles */
static int Log ; /* True if logging is on (default) */
static int No_Warnings ;
static int Offset ; /* Offset in buffer */
static union REGS Reg ;
static struct SREGS SReg ;
static int Restore ; /* True if doing a restore operation */
static int Result ;
static int Select ; /* True if /SELECT used */
static char Selection[ 256 ] ; /* Files to restore or list */
static char Temp[ 256 ] ;
static struct FILE *Tt ;
static int hTt ; /* Handle to Tt */
static Volume_Header Vh ;
static int Volume ; /* Current backup disk volume number */


static void Show_Label()
{

struct ffblk S ;
int Dummy ;
char Temp[ 256 ] ;

    if( !Log )
    {
	return ;
    } ;
    Temp[ 0 ] = Command[ 0 ] ;
    Temp[ 1 ] = 0 ;
    strcat( Temp, ":*.*" ) ;
    Dummy = findfirst( Temp, &S, FA_LABEL ) ;
    while( Dummy == 0 )
    {
	if( S.ff_attrib & FA_LABEL )
	{
	    fprintf( Tt, "Volume ID is %s\n", S.ff_name ) ;
	    return ; /* Done if we found the disk label */
	} ;
	Dummy = findnext( &S ) ;
    } ;
} ;


static long int Do_Write( unsigned int Offset, unsigned int Count )
{

int Adjust ; /* True if adjusting amount to output */
struct dfree DStruct ;
long int Number ;
unsigned int Result ;

    getdfree( Command[ 0 ] - 64, &DStruct ) ; /* Room left on output disk */
    if( DStruct.df_avail == 0 ) /* Disk is full */
    {
	return 101 ; /* Disk full error */
    } ;

    Number = DStruct.df_avail ; /* Calculate free space on disk */
    Number = Number * DStruct.df_bsec ;
    Number = Number * DStruct.df_sclus ;

    Adjust = False ;
    if( Number < Count ) /* Not enough room on disk to fit entire block */
    {
	Count = Number ; /* Write only what fits */
	Adjust = True ;
    } ;
    Result = fwrite( &Buffer[ Offset ], 1, Count, Output_File ) ;
    /* Write to saveset */
    if( Result == Count )
    {
	Result = 0 ;
    }
    else
    {
	Result = errno ; /* Get error */
	clearerr( Output_File ) ;
    } ;
    if( Result == 0 && Adjust )
    {
	Result = 101 ; /* Didn't all fit */
    } ;
    return Result ; /* Return result */
} ;

static long int Do_Read( unsigned int *Count )
{

long int Number ;
unsigned int Result, Temp ;

    Number = filelength( hInput_File ) - ftell( Input_File ) ;
    /* Room left in saveset */

    if( Number == 0 ) /* EOF */
    {
	return 0 ;
    } ;

    if( Number < *Count ) /* Not enough room on volume for entire block */
    {
	Temp = Number ; /* Convert to unsigned integer */
	*Count = Temp ; /* Read only what is there */
    } ;
    Result = fread( Buffer, *Count, 1, Input_File ) ; /* Read from saveset */
    if( Result != 1 )
    {
	if( errno != 0 )
	{
	    return errno ;
	} ;
	return 101 ;
    } ;
    return 0 ;
} ;

static void Show_Error( char *S )
{
    if( Ep == NULL )
    {
	fputs( S, Tt ) ;
    }
    else
    {
	Ep( S ) ;
    } ;
} ;


static int A, B, C, Dummy ; /* Dummy variables */
static char Local_Buffer[ 512 ] ; /* Buffer for DISKREAD */
static double Dumm ;
static unsigned int Dummie ;
static char Fat ; /* FAT entry size */
static int Fat_Root ; /* Offset of FAT entry 0 */
static int Fat_Sector ; /* First sector of FAT */
static long int Max_Fat ; /* Maximum cluster number */
static unsigned int Media_Id ; /* Type of media */
static int Flags, Result ; /* Result and flags */
static char S[ 256 ] ;
static int Sector_Size ; /* Size of sectors in bytes */

static unsigned int Next_Fat( unsigned int Cluster )
{

int Offset ;
unsigned int Temp ;

    if( Fat == 12 ) /* 12-bit FAT entries? */
    {
	Offset = Fat_Root + ( Cluster * 3 / 2 ) ; /* Byte offset into FAT */
	Temp = ( unsigned int ) Image[ Offset ] +
	    ( ( unsigned int ) Image[ Offset + 1 ] << 8 ) ;
	if( ( Cluster & 1 ) == 1 ) /* Odd cluster */
	{
	    Temp = Temp / 16 ;
	} ;
	return Temp & 0xFFF ;
    }
    else /* 16 bit entries */
    {
	return ( unsigned int ) Image[ Fat_Root + Cluster * 2 ] +
	    ( ( unsigned int ) Image[ Fat_Root + Cluster * 2 + 1 ] << 8 ) ;
    } ;
} ;

static void Clear_Fat( unsigned int Cluster )
{

int Offset ;
unsigned int Temp ;

    if( Fat == 12 ) /* 12-bit FAT entries? */
    {
	Offset = Fat_Root + ( Cluster * 3 / 2 ) ; /* Byte offset into FAT */
	Temp = ( unsigned int ) Image[ Offset ] +
	    ( ( unsigned int ) Image[ Offset + 1 ] << 8 ) ;
	if( ( Cluster & 1 ) == 1 ) /* Odd cluster */
	{
	    Temp = Temp & 0xF ;
	}
	else
	{
	    Temp = Temp & 0xF000 ;
	} ;
	Image[ Offset ] = Temp & 255 ;
	Image[ Offset + 1 ] = ( Temp >> 8 ) & 255 ;
    }
    else /* 16 bit entries */
    {
	Image[ Fat_Root + Cluster * 2 ] = 0 ;
	Image[ Fat_Root + Cluster * 2 + 1 ] = 0 ;
    } ;
} ;

static void Format( char Drive ) /* Format a floppy */
/* NOTE: It is assumed that Reg contains drive format information from DOS */
{

char Error_Text[ 80 ] ;

    Reg.w.ax = 0x1C00 ;
    Reg.w.dx = Drive - 64 ;
    intdosx( &Reg, &Reg, &SReg ) ; /* Get drive information */
    if( Reg.h.al == 255 ) /* Error */
    {
	strcpy( Error_Text, "Invalid output drive: " ) ;
	strcat( Error_Text, Command ) ;
	strcat( Error_Text, ":" ) ;
	Show_Error( Error_Text ) ;
	exit( 0 ) ;
    } ;
    memset( Image, 0, sizeof( Image ) ) ; /* Zero out image */
    Media_Id = peekb( SReg.ds, Reg.w.bx ) ; /* Get media ID byte */
    if( Reg.w.dx < 4079 )
    {
	Fat = 12 ; /* 12 bit FAT entry size */
    }
    else
    {
	Fat = 16 ; /* 16 bit FAT entry size */
    } ;
    Sector_Size = Reg.w.cx ;
    Max_Fat = Reg.w.dx + 2 ; /* Maximum cluster number */
    Dummy = peekb( 0xF000, 0xFFFE ) ; /* Get machine type */
    if( ( Dummy == 6 || Dummy == 0x94 ) && Media_Id == 0xFA ) /* Rainbow RX50 */
    {
	Fat_Sector = 20 ; /* FAT begins at sector 20, but offset 0 in Image */
	Fat_Root = 0 ;
    }
    else /* Non-rainbow */
    {
	A = Drive - 65 ;
	absread( A, 1, 0, Local_Buffer ) ;
	Fat_Sector = Local_Buffer[ 0x0E ] + Local_Buffer[ 0x0F ] * 256 ;
	/* FAT 1 begins after reserved sectors */
	Fat_Root = Fat_Sector * Sector_Size ;
    } ;
    if( Fat == 12 ) /* Get bytes in first part of FAT */
    {
	Dumm = Max_Fat * 1.5 ;
    }
    else
    {
	Dumm = Max_Fat * 2.0 ;
    } ;
    if( Dumm * 2 + Fat_Root > 16383 ) /* FAT is too big for Image */
    {
	Show_Error( "FAT is too big" ) ;
	exit( 0 ) ;
    } ;
    Dumm = Dumm / Sector_Size ; /* Convert to sectors */
    if( Dumm != floor( Dumm ) )
    {
	Dumm = floor( Dumm ) + 1 ;
    } ;
    Dummy = Dumm ; /* Round and integerize for sector count */
    A = Drive - 65 ; /* Get drive number for diskread/diskwrite */
    absread( A, Dummy * 2, Fat_Sector, &Image[ Fat_Root ] ) ;

    /* Now set-up FAT #1 */
    for( Dummy = 2; Dummy <= Max_Fat; Dummy++ ) /* Check all clusters */
    {
	Dummie = Next_Fat( Dummy ) ; /* Get FAT value for cluster */
        if( Fat == 12 && Dummie == 0xFF7 )
	{
	    Dummie = 0xFFF7 ;
	} ;
	if( Dummie != 0xFFF7 )
	{
	    Clear_Fat( Dummy ) ; /* Zero out non-bad clusters */
	} ;
    } ;

    /* Now set-up FAT #2 */
    Fat_Root = Fat_Root + Dumm * Sector_Size ;
    for( Dummy = 2; Dummy <= Max_Fat; Dummy++ ) /* Check all clusters */
    {
	Dummie = Next_Fat( Dummy ) ; /* Get FAT value for cluster */
	if( Fat == 12 && Dummie == 0xFF7 )
	{
	    Dummie = 0xFFF7 ;
	} ;
	if( Dummie != 0xFFF7 )
	{
	    Clear_Fat( Dummy ) ; /* Zero out non-bad clusters */
	} ;
    } ;
    if( Media_Id == 250 &&
	( peekb( 0xF000, 0xFFFE ) == 6 || peekb( 0xF000, 0xFFFE == 0x94 ) ) )
    /* RX50 on a Rainbow */
    {
	B = 20 ; /* LSN of FAT */
	C = 7 ; /* Count */
    }
    else
    {
	B = 1 ; /* LSN */
	C = 32 ; /* Count */
    } ;
Loop:
    if( abswrite( A, C, B, Image ) == -1 ) /* Error */
    {
	A = 1 ;
	Flags = errno >> 8 ;
	if( ( Flags & 3 ) == 3 )
	{
	    Flags = 9 ;
	}
	else
	{
	    while( ( Flags & 1 ) == 0 ) /* Find first error code */
	    {
		A = A + 1 ;
		Flags = Flags >> 1 ;
	    } ;
	} ;
	switch( A )
	{
	    case 1 : strcpy( S, "Unknown command or invalid drive" ) ;
		     break ;
	    case 2 : strcpy( S, "Bad address mark" ) ;
		     break ;
	    case 3 : strcpy( S, "Sector not found" ) ;
		     break ;
	    case 4 : strcpy( S, "DMA failure" ) ;
		     break ;
	    case 5 : strcpy( S, "Data error (CRC)" ) ;
		     break ;
	    case 6 : strcpy( S, "Controller failure" ) ;
		     break ;
	    case 7 : strcpy( S, "Seek failure" ) ;
		     break ;
	    case 8 : strcpy( S, "Drive does not respond" ) ;
		     break ;
	    case 9 : strcpy( S, "Drive is write locked" ) ;
	} ;
	switch( Dummy & 255 )
	{
	    case  0 : strcat( S, ".  Drive is write locked" ) ;
		      break ;
	    case  1 : strcat( S, ".  Unknown unit" ) ;
		      break ;
	    case  2 : strcat( S, ".  Drive not ready" ) ;
		      break ;
	    case  3 : strcat( S, ".  Unknown command in ABSOLUTE SECTOR WRITE" ) ;
		      break ;
	    case  4 : strcat( S, ".  Data error (CRC)" ) ;
		      break ;
	    case  5 : strcat( S, ".  Bad request structure length" ) ;
		      break ;
	    case  6 : strcat( S, ".  Seek error" ) ;
		      break ;
	    case  7 : strcat( S, ".  Unknown media type" ) ;
		      break ;
	    case  8 : strcat( S, ".  Sector not found" ) ;
		      break ;
	    case  9 : strcat( S, ".  Printer out of paper" ) ;
		      break ;
	    case 10 : strcat( S, ".  Write fault" ) ;
		      break ;
	    case 11 : strcat( S, ".  Read fault" ) ;
		      break ;
	    case 12 : strcat( S, ".  General failure" ) ;
		      break ;
	    case 13 : strcat( S, ".  Error 13 in ABSOLUTE SECTOR WRITE" ) ;
		      break ;
	    case 14 : strcat( S, ".  Error 14 in ABSOLUTE SECTOR WRITE" ) ;
		      break ;
	    case 15 : strcat( S, ".  Invalid disk change" ) ;
	} ;
	Show_Error( S ) ;
	if( Request == NULL )
	{
	    fputs(
		"Initialization error.  Correct and press any key to continue",
		Tt ) ;
	}
	else
	{
	    Request( "Initialization error", 0 ) ;
	} ;
	strcpy( Command, InKey() ) ; /* Wait for keypress */
	goto Loop ;
    } ;
    sleep( 2 ) ;
    /* This is settle-down time for DOS to recognize free space on disk */
} ;


static void Write_Header() /* Write volume header to disk */
{

char S[ 256 ] ;
struct dosdate_t T ;

    Control = 1 ; /* Execute loop at least once */
    while( Control != 0 )
    {
	if( Request == NULL )
	{
	    fprintf( Tt,
		"Place saveset volume %i into drive %s: and press RETURN\n",
		Vh.Volume, Command ) ;
	    Clear_Type_Ahead() ;
	    gets( Temp ) ;
	}
	else
	{
	    memset( S, 0, 256 ) ;
	    sprintf( S, "Place saveset volume %i into drive %s:",
		Vh.Volume, Command ) ;
	    Request( S, 1 ) ;
	} ;
	if( Initialize )
	{
	    Format( Command[ 0 ] ) ; /* Format disk */
	} ;
	Show_Label() ;
	strcpy( S, Command ) ;
	strcat( S, ":" ) ;
	strcat( S, Output_Spec ) ;
	hOutput_File = open( S, O_CREAT | O_TRUNC | O_RDWR | O_BINARY,
	    S_IREAD | S_IWRITE ) ;
	if( hOutput_File >= 0 )
	{
	    Control = 0 ;
	    Output_File = fdopen( hOutput_File, "w+b" ) ;
	}
	else
	{
	    Control = errno ;
	} ;
	if( Control == 0 ) /* File was opened OK */
	{
	    _dos_getdate( &T ) ;
	    Vh.Date_Y = T.year ;
	    Vh.Date_M = T.month ;
	    Vh.Date_D = T.day ;
	    if( fwrite( &Vh, sizeof( Vh ), 1, Output_File ) == 1 )
	    {
		Control = 0 ;
	    }
	    else
	    {
		Control = errno ;
	    } ;
	} ;
	if( Control != 0 )
	{
	    sprintf( S,
		"%s while writing volume header to %s:.  Please use another disk.",
		strerr( Control ), Command ) ;
	    Show_Error( S ) ;
	} ;
    } ;
} ;

static void Do_Next_Volume() /* Handle next volume for output */
{
    fclose( Output_File ) ;
    close( hOutput_File ) ;
    Vh.Volume = Vh.Volume + 1 ; /* Next volume */
    Write_Header() ;
    if( Log )
    {
	printf( "\nStarting volume %i    ", Vh.Volume ) ;
	Show_Label() ;
	printf( "\n\n" ) ;
    } ;
} ;


static void Get_Next_Volume() /* Handle next volume for input */
{

int Dummy, Dummy1 ;
char S[ 256 ], Work[ 256 ] ; /* Work variable */

    close( hInput_File ) ;
    fclose( Input_File ) ;
    Dummy = Vh.Volume + 1 ;
    sleep( 2 ) ; /* Necessary for SMARTDRV to notice change of disks */
Loop:
    if( Request == NULL )
    {
	fprintf( Tt, "Place volume %i in drive %c: and press RETURN",
	    Dummy, Command[ 0 ] ) ;
        Clear_Type_Ahead() ;
	gets( Temp ) ;
    }
    else
    {
	sprintf( S, "Place volume %i in drive %c:", Dummy, Command[ 0 ] ) ;
	Request( S, 1 ) ;
    } ;
    if( Abort )
    {
	return ;
    } ;
    strcpy( Work, Input_Spec ) ;
    if( strstr( Work, ":" ) == NULL )
    {
	strcpy( S, Command ) ;
	strcat( S, ":" ) ;
	strcat( S, Work ) ;
	strcpy( Work, S ) ;
    } ;
    hInput_File = open( Work, O_RDONLY | O_BINARY ) ;
    if( hInput_File >= 0 )
    {
	Dummy1 = 0 ;
	Input_File = fdopen( hInput_File, "rb" ) ;
    }
    else
    {
	Dummy1 = errno ;
    } ;
    if( Dummy1 != 0 )
    {
	sprintf( S, "%s opening %s", strerr( Dummy1 ), Input_Spec ) ;
	Show_Error( S ) ;
	if( Dummy == 1 ) /* First volume */
	{
	    Abort = True ;
	} ;
        if( Abort )
	{
	    return ;
	} ;
	goto Loop ;
    } ;
    if( filelength( hInput_File ) < sizeof( Vh ) )
    {
	Show_Error( "Invalid saveset format" ) ;
	if( Dummy == 1 ) /* First volume */
	{
	    Abort = True ;
	} ;
        if( Abort )
	{
	    return ;
	} ;
	goto Loop ;
    } ;
    fread( &Vh, 1, sizeof( Vh ), Input_File ) ; /* Get volume header */
    if( Vh.Date_D < 1 || Vh.Date_D > 31 || Vh.Date_M < 1 || Vh.Date_M > 12 )
    {
	Show_Error( "Invalid saveset format" ) ;
	if( Dummy == 1 ) /* First volume */
	{
	    Abort = True ;
	} ;
        if( Abort )
	{
	    return ;
	} ;
	goto Loop ;
    } ;
    if( Vh.Version != 0 )
    {
	Show_Error(
	  "Saveset was created by a later version of XBACKUP - cannot read saveset" ) ;
	if( Dummy == 1 ) /* First volume */
	{
	    Abort = True ;
	} ;
	if( Abort )
	{
	    return ;
	} ;
	goto Loop ;
    } ;
    if( strcmp( strc( Vh.Saveset ), Right( Work, 3 ) ) != 0 && No_Warnings == False )
    {
	memset( S, 0, 256 ) ;
	sprintf( S,
	    "WARNING: Internal saveset name of %s does not match external name",
	    strc( Vh.Saveset ) ) ;
	Show_Error( S ) ;
    } ;
    if( Abort )
    {
	return ;
    } ;
    if( Vh.Volume != Dummy && No_Warnings == False )
    {
	Show_Error( "WARNING: Not the next volume of the saveset" ) ;
    } ;
    if( Log )
    {
	printf( "Volume %i\n", Vh.Volume ) ;
	Show_Label() ;
    } ;
} ;


static void Write_File_Header() /* Write file header to saveset */
{

struct diskfree_t D ;
int Control ; /* Loop control variable */
char S[ 256 ] ;

    Control = 1 ;
    while( Control != 0 )
    {
	if( fwrite( &Fh, 23 + Fh.Filename[ 0 ], 1, Output_File ) == 1 )
	{
	    Control = 0 ;
	}
	else
	{
	    Control = errno ;
	} ;
	if( Control != 0 ) /* Error */
	{
	    if( Control == 101 ) /* Disk full */
	    {
		_dos_getdiskfree( Command[ 0 ] - 64, &D ) ;
		if( D.avail_clusters != 0 )
		{
		    sprintf( S,
			"Disk write error while writing to %s:.  Skipping to next volume.\n",
			Command ) ;
		    Show_Error( S ) ;
		} ;
	    }
	    else
	    {
		sprintf( S,
		    "%s while writing to %s:.  Skipping to next volume.",
		    strerr( Control ), Command ) ;
		Show_Error( S ) ;
	    } ;
	    Do_Next_Volume() ;
	} ;
    } ;
} ;


static void Do_List() /* List a saveset */
{

int Printed ; /* True if printed name of current file */
char S[ 256 ], Temp[ 256 ] ;

    Abort = False ;
    if( strlen( Input_Spec ) == 2 )
    {
	strcat( Input_Spec, "BACKUP." ) ;
    } ;
    hInput_File = open( Input_Spec, O_RDONLY | O_BINARY ) ;
    if( hInput_File >= 0 )
    {
	Dummy = 0 ;
	Input_File = fdopen( hInput_File, "rb" ) ;
    }
    else
    {
	Dummy = errno ;
    } ;
    if( Dummy != 0 )
    {
	sprintf( S, "%s while opening saveset %s",
	    strerr( Dummy ), Input_Spec ) ;
	Show_Error( S ) ;
	return ;
    } ;
    if( filelength( hInput_File ) < sizeof( Vh ) )
    {
	Show_Error( "Invalid saveset format" ) ;
	return ;
    } ;
    printf( "\nList of saveset %s\n\n", Input_Spec ) ;
    fread( &Vh, 1, sizeof( Vh ), Input_File ) ; /* Get volume header */
    if( Vh.Date_D < 1 || Vh.Date_D > 31 || Vh.Date_M < 1 || Vh.Date_M > 12 )
    {
	Show_Error( "Invalid saveset format" ) ;
	return ;
    } ;
    if( Vh.Version != 0 )
    {
	Show_Error( "Saveset was created by a later version of XBACKUP - cannot read saveset" ) ;
	return ;
    } ;
    if( strcmp( strc( Vh.Saveset ), Right( Input_Spec, 3 ) ) != 0 && No_Warnings == False )
    {
	sprintf( S,
	    "WARNING: Internal saveset name of %s does not match external name",
	    strc( Vh.Saveset ) ) ;
	Show_Error( S ) ;
    } ;
    if( Vh.Volume != 1 && No_Warnings == False )
    {
	Show_Error( "WARNING: Not beginning with first volume of saveset" ) ;
    } ;
    printf( "Volume %i\n", Vh.Volume ) ;
    Printed = False ; /* No filenames printed yet */
Loop:
    Result = fread( &Fh, 1, 23, Input_File ) ; /* Get first part of file header */
    if( Result < 23 ) /* Need next volume */
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ;
    if( Fh.Filename[ 0 ] == 0 )
    {
	return ; /* End of saveset */
    } ;

    /* Read in the file name */
    Result = fread( &Fh.Filename[ 1 ], 1, Fh.Filename[ 0 ], Input_File ) ;
    if( Result != Fh.Filename[ 0 ] )
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ;
    strcpy( Temp, strc( Fh.Filename ) ) ;
    if( Wc_Match( Selection, Temp ) )
    {
	if( !Printed )
	{
	    printf( "    %s  %li\n", Temp, Fh.Size ) ;
	    /* Show name and size */
	} ;
	Printed = True ;
    } ;
    if( ftell( Input_File ) + Fh.Size - Fh.Start > filelength( hInput_File ) )
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ;
    Printed = False ;
    fseek( Input_File, ftell( Input_File ) + Fh.Size - Fh.Start, SEEK_SET ) ;
    /* Jump to next file */
    goto Loop ;
} ;


static int Create_Dir( char *Passed_X )
{

int Dummy ;
char Temp[ 256 ], Work[ 256 ] ; /* Work variable */
char X[ 256 ] ;

    if( Passed_X == NULL )
    {
	X[ 0 ] = 0 ;
    }
    else
    {
	strcpy( X, Passed_X ) ;
    } ;

    Dummy = strlen( X ) - 1 ;
    while( X[ Dummy ] != '\\' && Dummy > 0 )
    {
	Dummy = Dummy - 1 ;
    } ;
    strcpy( X, Left( X, Dummy ) ) ; /* Strip filename, etc. */
    if( strlen( X ) < 4 ) /* Only root directory */
    {
	return True ; /* OK */
    } ;

    Dummy = Instr( 1, X, "\\" ) ;
    strcpy( Work, Left( X, Dummy ) ) ; /* Nothing so far */
    while( strlen( Work ) < strlen( X ) ) /* Until full path exists */
    {
	Dummy = strlen( Work ) + 2 ;
	while( X[ Dummy - 1 ] != '\\' && Dummy <= strlen( X ) )
	{
            Dummy = Dummy + 1 ; /* Find end of next directory in path */
	} ;
	strcpy( Work, Left( X, Dummy - 1 ) ) ; /* Get partial path */
	Dummy = mkdir( Work ) ; /* Create the directory */
	if ( Dummy == -1 && errno == ENOENT )
        /* Error (other than already exists) */
	{
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "%s while creating directory path: %s",
		strerr( Dummy ), Work ) ;
	    Show_Error( Temp ) ;
	    return False ; /* Failure */
	} ;
	strcat( Work, "\\" ) ;
    } ;
    return True ; /* OK */
} ;


static void Qualify_Name( char *X ) /* Fully qualify a filename */
{

int Dummy ;

    if( X[ strlen( X ) - 1 ] == '\\' ) /* No filename at all */
    {
	strcat( X, "*.*" ) ;
    }
    else
    {
	if( Extension_Pos( X ) == 0 ) /* No extension */
	{
	    strcat( X, ".*" ) ; /* Default extension */
	} ;
    } ;
} ; /* Qualify_Name */


static void Do_Restore() /* Restore files from a saveset */
{

unsigned int Count ; /* Number of bytes to read/write */
File_Header Fh_Temp ; /* Temporary file header */
long int Files_Matched ; /* Number of matches found to restore */
struct FILE *Output_File ; /* File being created */
int HOutput_File ;
char Temp[ 256 ], Output_Name[ 256 ] ; /* Name of output file */
struct dfree DF ;

    /* Validate output drive */
    if( Switch_Present( Output_Spec ) > 0 )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "Invalid switch: %s", Output_Spec ) ;
	Show_Error( Temp ) ;
	return ;
    } ;
    Files_Matched = 0 ; /* No files thus far */
    if( Instr( 1, Output_Spec, ":" ) == 0 )
    {
	getcurdir( 0, Current_File ) ;
	Temp[ 0 ] = Current_File[ 0 ] ;
	Temp[ 1 ] = 0 ;
	strcat( Temp, ":" ) ;
	strcat( Temp, Output_Spec ) ;
	strcpy( Output_Spec, Temp ) ;
    }
    else
    {
	if( Instr( 1, Output_Spec, ":" ) != 2 )
	{
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "Invalid output drive: ", Output_Spec ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;
    } ;
    Qualify_Name( Output_Spec ) ; /* Fully qualify name */

    /* Validate input drive(s) */
    if( Command[ 0 ] < 'A' || Command[ 0 ] > 'Z' )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "Invalid input drive: ", Command ) ;
	Show_Error( Temp ) ;
	return ;
    } ;

    getdfree( Output_Spec[ 0 ] - 64, &DF ) ;
    if( DF.df_sclus == - 1 )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "Invalid output drive: %c:", Output_Spec[ 0 ] ) ;
	Show_Error( Temp ) ;
	exit ;
    } ;

    getdfree( Command[ 0 ] - 64, &DF ) ;
    if( DF.df_sclus == - 1 )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "Invalid input drive: %s:", Command ) ;
	Show_Error( Temp ) ;
	exit ;
    } ;

    if( Log )
    {
	fprintf( Tt, "Beginning restore of %s:%s to %s on %s at %s\n", Command,
            Input_Spec, Output_Spec, Date(), Time() ) ;
    } ;
    if( strlen( Input_Spec ) == 2 )
    {
	strcat( Input_Spec, "*.*" ) ;
    } ;

    if( Output_Spec[ 0 ] == Command[ 0 ] ) /* From/to same disk */
    {
	if( Request == NULL )
	{
	    printf( "Place source disk into drive %s: and press RETURN\n",
		Command ) ;
	    Clear_Type_Ahead() ;
	    gets( Temp ) ;
	}
	else
	{
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "Place source disk into drive %s:", Command ) ;
	    Request( Temp, 1 ) ;
	} ;
    } ;
    Abort = False ;
    if( strlen( Input_Spec ) == 2 )
    {
	strcat( Input_Spec, "BACKUP." ) ;
    } ;
    strcpy( Temp, Command ) ;
    strcat( Temp, ":" ) ;
    strcat( Temp, Input_Spec ) ;
    hInput_File = open( Temp, O_RDONLY | O_BINARY ) ;
    if( hInput_File == -1 )
    {
	Input_File = NULL ;
    }
    else
    {
	Input_File = fdopen( hInput_File, "rb" ) ;
    } ;
    if( Input_File == NULL )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "%s while opening saveset %s:%s", strerr( errno ),
	    Command, Input_Spec ) ;
	Show_Error( Temp ) ;
	return ;
    } ;
    if( filelength( hInput_File ) < sizeof( Vh ) )
    {
	Show_Error( "Invalid saveset format" ) ;
	return ;
    } ;
    fread( &Vh, 1, sizeof( Vh ), Input_File ) ; /* Get volume header */
    if( Vh.Date_D < 1 || Vh.Date_D > 31 || Vh.Date_M < 1 || Vh.Date_M > 12 )
    {
	Show_Error( "Invalid saveset format" ) ;
	return ;
    } ;
    if( Vh.Version != 0 )
    {
	Show_Error(
	    "Saveset was created by a later version of XBACKUP - cannot read saveset" ) ;
	return ;
    } ;
    if( strcmp( strc( Vh.Saveset ), Input_Spec ) != 0 && No_Warnings == False )
    {
	memset( Temp, 0, 256 ) ;
	sprintf( Temp, "WARNING: Internal saveset name of %s does not match external name",
	    strc( Vh.Saveset ) ) ;
	Show_Error( Temp ) ;
    } ;
    if( Vh.Volume != 1 && No_Warnings == False )
    {
	Show_Error( "WARNING: Not beginning with first volume of saveset" ) ;
    } ;
    if( Log )
    {
	printf( "Volume %i\n", Vh.Volume ) ;
    } ;


    /* Code to actually do restore follows: */
Next_Selection:
    if( strlen( Command_File ) > 0 )
    {
	if( fgets( Selection, 256, Fcommand ) == NULL )
	{
	    return ;
	} ;
	Edit( Selection, Selection, 4 ) ;
    } ;

Loop:
    if( ftell( Input_File ) + 23 > filelength( hInput_File ) ) /* EOF */
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ; /* if ftell( Input_File ) + 23 > filelength( hInput_File ) */

    Result = fread( &Fh, 1, 23, Input_File ) ;
    /* Get first part of file header */
    if( Result < 23 ) /* Need next volume */
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ; /* if( Result < 23 ) */
    if( Fh.Filename[ 0 ] == 0 )
    {
	if( Files_Matched == 0 )
	{
	    Show_Error( "No files found" ) ;
	} ;
	return ; /* End of saveset */
    } ;

    if( ftell( Input_File ) + Fh.Filename[ 0 ] > filelength( hInput_File ) )
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ;

    Result = fread( &Fh.Filename[ 1 ], 1, Fh.Filename[ 0 ], Input_File ) ;
    /* Read name */
    if( Result < Fh.Filename[ 0 ] )
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ; /* if( Result < FH.Filename[ 0 ] ) */

    strcpy( Temp, strc( Fh.Filename ) ) ;
    if( Wc_Match( Selection, Temp ) && Fh.Start == 0 )
    /* If a selected file and starting at beginning of file */
    {
	Files_Matched = Files_Matched + 1 ; /* Count matches */
	strcpy( Output_Name, Wc_Xlate( Selection, Temp, Output_Spec ) ) ;
	if( Log )
	{
	    puts( Output_Name ) ; /* Show name */
	} ;
	if( !Create_Dir( Output_Name ) )
	{
	    return ; /* Create any needed sub-directories, exit if error */
	} ;

	hOutput_File = open( Output_Name, O_CREAT | O_TRUNC | O_RDWR | O_BINARY,
	    S_IREAD | S_IWRITE ) ;
	if( hOutput_File == -1 )
	{
	    Output_File = NULL ;
	}
	else
	{
	    Output_File = fdopen( hOutput_File, "w+b" ) ;
	} ;
	if( Output_File == NULL )
	{
	    if( Ignore ) /* Ignoring creation errors */
	    {
		if( Log )
		{
		    printf( "%s while creating file %s - skipping file\n",
			strerr( errno ), Output_Name ) ;
		} ;
		goto Skip_File ;
	    } ;
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "%s while creating file %s", strerr( errno ),
		Output_Name ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;

	while( filelength( hOutput_File ) < Fh.Size )
	/* Until entire file restored */
	{
	    if( Fh.Size - filelength( hOutput_File ) > 65520 )
	    {
		Count = 65520 ;
	    }
	    else
	    {
		Count = Fh.Size - filelength( hOutput_File ) ;
	    } ;
	    Dummy = Do_Read( &Count ) ; /* Read into buffer */
	    if( Dummy != 0 )
	    {
		memset( Temp, 0, 256 ) ;
		sprintf( Temp, "%s while reading saveset", strerr( Dummy ) ) ;
		Show_Error( Temp ) ;
		return ;
	    } ; /* if( Dummy != 0 ) */

	    if( Count > 0 ) /* If anything was read in from saveset */
	    {
		if( fwrite( Buffer, 1, Count, Output_File ) != Count )
		/* Write to output */
		{
		    memset( Temp, 0, 256 ) ;
		    sprintf( Temp, "%s while writing file %s", strerr( errno ),
			Output_Name ) ;
		    Show_Error( Temp ) ;
		    return ;
		} ;
	    } ; /* if( Count > 0 ) */

	    if( ftell( Input_File ) == filelength( hInput_File ) )
	    /* Need next volume */
	    {
		Fh_Temp = Fh ;
		Fh_Temp.Filename[ 0 ] = Fh.Filename[ 0 ] + 1 ;
		Fh_Temp.Filename[ Fh_Temp.Filename[ 0 ] ] = 'A' ;
		strcpy( Temp, strc( Fh_Temp.Filename ) ) ;
		while( ( strcmp( Temp, strc( Fh.Filename ) ) != 0 ) ||
			( Fh_Temp.Size != Fh.Size ) )
		/* We must do this to prevent synchronization problems (ie
		   a matching saveset name/volume number, but not really
		   from the same saveset). */
		{
		    Dummy = Vh.Volume ;
		    Get_Next_Volume() ;
		    if( Abort )
		    {
			return ; /* Exit on error */
		    } ;
		    Result = fread( &Fh_Temp, 1, 23, Input_File ) ;
		    /* Get first part of file header */
		    if( Result == 23 ) /* File header was read */
		    {
			if( ftell( Input_File ) + Fh_Temp.Filename[ 0 ] <=
			    filelength( hInput_File ) )
			{
			    Result = fread( &Fh_Temp.Filename[ 1 ], 1,
				Fh_Temp.Filename[ 0 ], Input_File ) ;
			    /* Read file name */
			    if( Result == Fh_Temp.Filename[ 0 ] )
			    /* File name was read */
			    {
				strcpy( Temp, strc( Fh_Temp.Filename ) ) ;
				if( ( strcmp( Temp,
				    strc( Fh.Filename ) ) != 0 ) ||
				    ( Fh_Temp.Size != Fh.Size ) )
				{
				    Vh.Volume = Dummy ;
				    Show_Error(
					"Not the next volume of the saveset" ) ;
				}
				else
				{
				    Fh = Fh_Temp ;
				} ;
			    }
			    else
			    {
				Fh_Temp.Filename[ 0 ] =
				    Fh.Filename[ 0 ] + 1 ;
				Fh_Temp.Filename[ Fh_Temp.Filename[ 0 ] ] = 'A' ;
			    } ;
			} ; /* if... */
		    } ; /* if... */
		    strcpy( Temp, strc( Fh_Temp.Filename ) ) ;
		} ; /* while... */
	    } ; /* if( ftell( Input_File ) = filelength( hInput_File ) ) */
	} ; /* while( filelength( hOutput_File ) < FH.Size ) */

	_dos_setftime( hOutput_File, Fh.Date, Fh.Time ) ; /* Set file date/time */
	fclose( Output_File ) ;
	close( hOutput_File ) ; /* Done restoring file */
	_dos_setfileattr( Output_Name, Fh.Attributes ) ;
	/* Set file attributes */
	if ( strstr( Selection, "*" ) == NULL &&
			strstr( Selection, "..." ) == NULL &&
			strstr( Selection, "?" ) == NULL ) /* No wildcards */
	{
	    if( strlen( Command_File ) == 0 ) /* No command file */
	    {
		return ;
	    }
	    else
	    {
		goto Next_Selection ;
	    } ;
	} ;
	goto Loop ;
    } ;

Skip_File:
    if( ftell( Input_File ) + Fh.Size - Fh.Start > filelength( hInput_File ) )
    {
	Get_Next_Volume() ;
	if( Abort )
	{
	    return ; /* Exit on error */
	} ;
	goto Loop ; /* Try again */
    } ;
    fseek( Input_File, ftell( Input_File ) + Fh.Size - Fh.Start, SEEK_SET ) ;
    /* Jump to next file */
    goto Loop ;
} ;


static char *Do_Lookup( char *Passed_X )
{
/* Do a wildcard lookup, but ignore certain files */

int D ;
char Filename[ 256 ], Result[ 256 ] ;
char X[ 256 ] ;
static char Buffer[ 256 ] ;

    strcpy( X, Passed_X ) ;
Loop:
    strcpy( Result, Lookup( X ) ) ; /* Lookup first/next file */
    X[ 0 ] = 0 ; /* In case we loop back */
    if( strlen( Result ) == 0 ) /* Quick exit if no more matches */
    {
	Buffer[ 0 ] = 0 ;
	return ;
    } ;
    D = strlen( Result ) ;
    while( D > 0 && Result[ D - 1 ] != '\\' )
    {
	D = D - 1 ;
    } ;
    strcpy( Filename, &Result[ D ] ) ;
    if( strcmp( Filename, "." ) == 0 || strcmp( Filename, ".." ) == 0 )
    {
	goto Loop ; /* Skip . and .. */
    } ;

    strcpy( Buffer, Result ) ;
    return Buffer ;
} ; /* Do_Lookup */


static void End_Saveset() /* End a saveset */
{
    Fh.Filename[ 0 ] = 0 ; /* This signifies the EOS */
    Write_File_Header() ;
} ; /* End_Saveset */


static void Validate_Input() /* Validate input drive */
{

char Temp[ 256 ] ;
struct dfree DF ;

    if( strstr( Input_Spec, ":" ) == NULL )
    {
	getcurdir( 0, Current_File ) ;
	Temp[ 0 ] = Current_File[ 1 ] ;
	Temp[ 1 ] = 0 ;
	strcat( Temp, ":" ) ;
	strcat( Temp, Input_Spec ) ;
	strcpy( Input_Spec, Temp ) ;
    }
    else
    {
	if( Instr( 1, Input_Spec, ":" ) != 2 )
	{
	    strcpy( Temp, "Invalid input drive: " ) ;
	    strcat( Temp, Input_Spec ) ;
	    Show_Error( Temp ) ;
	    Abort = True ;
	    return ;
	} ;
    } ;
    if( strstr( Input_Spec, "\\" ) == 0 )
    {
	strcpy( Temp, Left( Input_Spec, 2 ) ) ;
	strcat( Temp, "\\...\\" ) ;
	strcat( Temp, &Input_Spec[ 2 ] ) ;
	strcpy( Input_Spec, Temp ) ;
    } ;
    Qualify_Name( Input_Spec ) ;
    getdfree( Input_Spec[ 0 ] - 64, &DF ) ;
    if( DF.df_sclus == - 1 )
    {
	strcpy( Temp, "Invalid input drive: " ) ;
	strcat( Temp, Left( Input_Spec, 1 ) ) ;
	strcat( Temp, ":" ) ;
	Show_Error( Temp ) ;
	Abort = True ;
	return ;
    } ;
    if( strlen( Input_Spec ) == 2 )
    {
	strcat( Input_Spec, "*.*" ) ;
    } ;
} ; /* Validate_Input */


static void Do_Backup() /* Perform backup */
{

char Temp[ 256 ] ;
struct dfree DF ;

    if( Select ) /* /SELECT is only allowed on restores and lists */
    {
	Show_Error( "/SELECT is only allowed on restores and listings" ) ;
	return ;
    } ;
    if( Switch_Present( Input_Spec ) > 0 )
    {
	strcpy( Temp, "Invalid switch: " ) ;
	strcat( Temp, Input_Spec ) ;
	Show_Error( Temp ) ;
	return ;
    } ;

    /* Validate output drive(s) */
    if( Command[ 0 ] < 'A' || Command[ 0 ] > 'Z' )
    {
	strcpy( Temp, "Invalid output drive: " ) ;
	strcat( Temp, Command ) ;
	strcat( Temp, ":" ) ;
	Show_Error( Temp ) ;
	return ;
    } ;

    getdfree( Command[ 0 ] - 64, &DF ) ;
    if( DF.df_sclus == - 1 )
    {
	strcpy( Temp, "Invalid output drive: " ) ;
	strcat( Temp, Command ) ;
	strcat( Temp, ":" ) ;
	Show_Error( Temp ) ;
	return ;
    } ;

    Reg.w.ax = 0x1C00 ;
    Reg.w.dx = Command[ 0 ] - 64 ;
    intdosx( &Reg, &Reg, &SReg ) ; /* Get drive information */
    if( peekb( SReg.ds, Reg.w.bx ) == 0xF8 ) /* Fixed drive */
    {
	Show_Error( "Cannot initialize a non-removable disk" ) ;
	return ;
    } ;

    strpas( Vh.Saveset, Output_Spec ) ; /* Saveset name */
    Vh.Version = 0 ; /* Version 0 of saveset format */
    Vh.Volume = 1 ; /* First volume */
    Vh.Expire_Y = 9999 ; /* Expiration date */
    Vh.Expire_M = 1 ;
    Vh.Expire_D = 1 ;
    Write_Header() ; /* Write header to disk */

    if( strlen( Command_File ) == 0 )
    {
	Validate_Input ();
	if( Abort )
	{
	    return ;
	} ;
    } ;

    if( Log )
    {
	printf( "Beginning backup of %s to %s: on %s at %s\n", Input_Spec,
	    Command, Output_Spec, Date(), Time() ) ;
    } ;

    if( Input_Spec[ 0 ] == Command[ 0 ] ) /* From/to same disk */
    {
	if( Request == NULL )
	{
	    fprintf( Tt, "Place source disk into drive %s: and presss RETURN\n",
		    Command ) ;
	    Clear_Type_Ahead() ;
	    gets( Temp ) ;
	}
	else
	{
	    strcpy( Temp, "Place source disk into drive " ) ;
	    strcat( Temp, Command ) ;
	    strcat( Temp, ":" ) ;
	    Request( Temp, 1 ) ;
	} ;
    } ;

Next_Selection:
    if( strlen( Command_File ) > 0 )
    {
	if( fgets( Input_Spec, 256, Fcommand ) == NULL )
	{
	    End_Saveset() ;
	    return ; /* Done */
	} ;
	Edit( Input_Spec, Input_Spec, 4 ) ;
	Validate_Input() ;
	if( Abort )
	{
	    return ;
	} ;
    } ;

    strcpy( Current_File, Do_Lookup( Input_Spec ) ) ; /* Find first file */
    while( strlen( Current_File ) != 0 ) /* Until no more files to process */
    {
	_dos_getfileattr( Current_File, &Dummy1 ) ; /* DOS File attributes */
	if( Dummy1 & _A_SUBDIR ) /* Skip directories */
	{
	    goto Skip_File ;
	} ;
	if( Log )
	{
	    puts( Current_File ) ; /* Show file being backed up */
	} ;
	Fh.Attributes = Dummy1 ;
	hInput_File = open( Current_File, O_RDONLY | O_BINARY ) ;
	if( hInput_File == -1 )
	{
	    Input_File = NULL ;
	}
	else
	{
	    Input_File = fdopen( hInput_File, "rb" ) ;
	} ;
	if( Input_File == NULL )
	{
	    if( Ignore )
	    {
		if( Log )
		{
		    printf( "%s while opening file %s - skipping file\n",
			strerr( Dummy ), Current_File ) ;
		} ;
		goto Skip_File ;
	    } ;
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "%s while opening file %s", strerr( errno ),
		Current_File ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;

	Fh.Compression = 0 ; /* Compression not supported yet */
	_dos_getftime( hInput_File, &Fh.Date, &Fh.Time ) ;
	Fh.Size = filelength( hInput_File ) ; /* Get DOS file size */
	Fh.Saveset_Size = Fh.Size ; /* Same as DOS with no compression */
	Fh.Flags = 0 ; /* Not used yet */
	strpas( Fh.Filename, Current_File ) ;
	Fh.Start = 0 ; /* Offset 0 */
	if( Input_Spec[ 0 ] == Command[ 0 ] ) /* From/to same disk */
	{
	    if( Request == NULL )
	    {
		fprintf( Tt, "Place saveset volume %i into drive %s: and press RETURN\n",
			Vh.Volume, Command ) ;
	        Clear_Type_Ahead() ;
	        gets( Temp ) ;
	    }
	    else
	    {
		memset( Temp, 0, 256 ) ;
		sprintf( Temp, "Place saveset volume %i into drive %s:",
			Vh.Volume, Command ) ;
		Request( Temp, 1 ) ;
	    } ;
	} ;
	Write_File_Header() ;

	if( Input_Spec[ 0 ] == Command[ 0 ] ) /* From/to same disk */
	{
	    if( Request == NULL )
	    {
		fprintf( Tt, "Place source disk into drive %s: and press RETURN\n",
			Command ) ;
	        Clear_Type_Ahead() ;
	        gets( Temp ) ;
	    }
	    else
	    {
		memset( Temp, 0, 256 ) ;
		sprintf( Temp, "Place source disk into drive %s:", Command ) ;
		Request( Temp, 1 ) ;
	    } ;
	} ;
	while( ftell( Input_File ) < filelength( hInput_File ) )
	{
	    Count = fread( Buffer, 1, 65520, Input_File ) ;
	    /* Read as much as we can */
	    if( Input_Spec[ 0 ] == Command[ 0 ] ) /* From/to same disk */
	    {
		if( Request == NULL )
		{
		    fprintf( Tt,
			"Place saveset volume %i into drive %s: and press RETURN\n",
			Vh.Volume, Command ) ;
		    Clear_Type_Ahead() ;
		    gets( Temp ) ;
		}
		else
		{
		    memset( Temp, 0, 256 ) ;
		    sprintf( Temp, "Place saveset volume %i into drive %s:",
			Vh.Volume, Command ) ;
		    Request( Temp, 1 ) ;
		} ;
	    } ;
	    Result = 0 ;
	    Offset = 0 ;
	    while( Result < Count )
	    {
		Result = ftell( Output_File ) ;
		Control = Do_Write( Offset, Count ) ; /* Write to saveset */
		if( Control == 0 )
		{
		    Result = Count ;
		}
		else /* Error */
		{
		    if( Control == 101 ) /* Disk full */
		    {
			getdfree( Command[ 0 ] - 64, &DF ) ;
			if( DF.df_avail != 0 )
			{
			    memset( Temp, 0, 256 ) ;
			    sprintf( Temp,
				"Disk write error while writing to %s:.  Skipping to next volume.",
				Command ) ;
			    Show_Error( Temp ) ;
			} ;
		    }
		    else
		    {
			memset( Temp, 0, 256 ) ;
			sprintf( Temp,
			    "%s while writing to %s:.  Skipping to next volume.",
			    strerr( Control ), Command ) ;
			Show_Error( Temp ) ;
		    } ;
		    Result = ftell( Output_File ) - Result ;
		    /* Bytes actually written */
		    Do_Next_Volume() ;
		    Fh.Start = ftell( Input_File ) - Count + Result ;
		    Offset = Result ;
		    Count = Count - Result ; /* Number of bytes left to transfer */
		    Write_File_Header() ;
		    Result = 0 ; /* Force loop to keep executing */
		} ; /* if( Control == 0 ) */
	    } ; /* while( Result < Count ) */
	} ; /* while( ftell( Input_File ) < filelength( hInput_File ) ) */
	fclose( Input_File ) ; /* Done with file */
	close( hInput_File ) ;

Skip_File :
	strcpy( Current_File, Do_Lookup( "" ) ) ; /* Find next file */
    } ; /* while( strlen( Current_File ) <> 0 ) */
    if( strlen( Command_File ) > 0 )
    {
	goto Next_Selection ;
    } ;
    End_Saveset() ;
} ;


/* Backup interface entry point */
void Backup_Backup( char *Passed_Command0, void ( *E_P )(), void ( *R_P )() )
{

char Command0[ 256 ] ;

    if( Passed_Command0 == NULL )
    {
	Command0[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Command0, Passed_Command0 ) ;
    } ;
    Buffer = malloc( 65520 ) ; /* Allocate memory for buffer */
    if( Buffer == NULL )
    {
	puts( "Not enough memory to allocate work buffer" ) ;
	return ;
    } ;
    Ep = E_P ;
    Request = R_P ;
    Tt = fopen( "CON", "a+" ) ;

    /* Parse switches */

    if( Parse_Switch( "NOW|ARNINGS", "", Command0, Temp ) == 1 )
    {
	if( strlen( Temp ) > 0 ) /* Parameter to switch */
	{
	    Show_Error( "Illegal parameter on /NOWARNINGS" ) ;
	    return ;
	} ;
	No_Warnings = True ; /* No warnings */
    }
    else
    {
	No_Warnings = False ;
    } ;

    if( Parse_Switch( "I|GNORE", "", Command0, Temp ) == 1 )
    {
	if( strlen( Temp ) > 0 ) /* Parameter to switch */
	{
	    Show_Error( "Illegal parameter on /IGNORE" ) ;
	    return ;
	} ;
	Ignore = True ; /* No log */
    }
    else
    {
	Ignore = False ;
    } ;

puts( Command0 ) ;
    if( Parse_Switch( "NOLO|G", "LO|G", Command0, Temp ) == 1 )
    {
	if( strlen( Temp ) > 0 ) /* Parameter to switch */
	{
	    Show_Error( "Illegal parameter on /LOG or /NOLOG" ) ;
	    return ;
	} ;
	Log = False ; /* No log */
    }
    else
    {
	Log = True ;
    } ;

    if( Parse_Switch( "R|ESTORE", "", Command0, Temp ) == 1 )
    {
	if( strlen( Temp ) > 0 ) /* Parameter to switch */
	{
	    Show_Error( "Illegal parameter on /RESTORE" ) ;
	    return ;
	} ;
	Restore = True ; /* Restore from a saveset */
    }
    else
    {
	Restore = False ;
    } ;

    Command_File[ 0 ] = 0 ;
    Dummy = Parse_Switch( "C|OMMAND", "", Command0, Temp ) ;
    if( Dummy == 1 ) /* /COMMAND */
    {
	if( strlen( Temp ) == 0 ) /* No parameter */
	{
	    Show_Error( "Missing parameter for /COMMAND" ) ;
	    return ;
	} ;
	strcpy( Command_File, Temp ) ;
	hFcommand = open( Temp, O_RDONLY | O_TEXT ) ;
	if( hFcommand == -1 )
	{
	    Fcommand = NULL ;
	}
	else
	{
	    Fcommand = fopen( Temp, "rt" ) ;
	} ;
	if( Fcommand == NULL )
	{
	    memset( Temp, 0, 256 ) ;
	    sprintf( Temp, "%s while opening saveset command file",
		strerr( errno ) ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;
    } ;

    Dummy = Parse_Switch( "NOI|NITIALIZE", "I|NITIALIZE", Command0, Temp ) ;
    if( Dummy == 1 ) /* /NOINITIALIZE */
    {
	if( strlen( Temp ) > 0 ) /* Parameter to switch */
	{
	    Show_Error( "Illegal parameter on /INITIALIZE or /NOINITIALIZE" ) ;
	    return ;
	} ;
	Initialize = False ; /* No log */
    }
    else
    {
	if( Restore )
	    {
		if( Dummy == 2 ) /* /INITIALIZE */
		{
		    Show_Error( "Incompatible switches" ) ;
		    return ;
		} ;
		Initialize = False ;
	    }
	    else
	    {
		Initialize = True ;
	    } ;
    } ;

    Edit( Command0, Command0, 8 | 128 ) ; /* No leading/trailing spaces */

    strcpy( Input_Spec, Parse_Parameter( " ", Command0 ) ) ;
    /* Input specification is the first parameter */

    strcpy( Selection, "*:\\...\\*.*" ) ;
    /* Default to restoring/listing everything */
    if( Parse_Switch( "S|ELECT", "", Input_Spec, Temp ) == 1 )
    {
	if( strlen( Command_File ) > 0 )
	{
	    Show_Error( "Incompatible switches" ) ;
	    return ;
	} ;
	if( strlen( Temp ) > 0 )
	{
	    strcpy( Selection, Temp ) ;
	} ;
	Select = True ; /* Indicate that the switch was used */
	if( strstr( Selection, ":" ) == NULL )
	{
	    strcpy( Temp, "*:" ) ;
	    strcat( Temp, Selection ) ;
	    strcpy( Selection, Temp ) ;
	    /* Default to all devices */
	} ;
	if( strstr( Selection, "\\" ) == NULL )
	{
	/* Default to all directories */
	    strcpy( Temp, Left( Selection, 2 ) ) ;
	    strcat( Temp, "\\" ) ;
	    strcat( Temp, &Selection[ 2 ] ) ;
	    strcpy( Selection, Temp ) ;
	} ;
	Qualify_Name( Selection ) ;
    }
    else
    {
	Select = False ; /* No selection specified */
    } ;
    Edit( Command0, Command0, 32 ) ;
    Edit( Input_Spec, Input_Spec, 32 ) ;
    if( Restore )
    {
	strcpy( Output_Spec, Command0 ) ;
	Command0[ 0 ] = Input_Spec[ 0 ] ;
	Command0[ 1 ] = 0 ;
	strcpy( Input_Spec, &Input_Spec[ 2 ] ) ;
	if( strlen( Input_Spec ) == 0 )
	{
	    strcpy( Input_Spec, "BACKUP." ) ;
	} ;
	/* Input_Spec now contains the saveset name */
	if( strlen( Input_Spec ) > 12 )
	{
	    strcpy( Temp, "Invalid saveset name: " ) ;
	    strcat( Temp, Input_Spec ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;
    } ;

    if( strlen( Command0 ) == 0 ) /* Only one parameter */
    {
	if( Parse_Switch( "LI|ST", "", Input_Spec, Temp ) == 1 )
	{
	    if( Ignore || Restore || ( !Log ) || strlen( Command_File ) > 0 )
	    /* Cannot do /LIST and /RESTORE or /LIST and /NOLOG or /LIST and
	     /IGNORE or /LIST and /COMMAND */
	    {
		Show_Error( "Incompatible switches" ) ;
		return ;
	    } ;
	    if( strlen( Temp ) > 0 ) /* Parameter to switch */
	    {
		Show_Error( "Illegal parameter on /LIST" ) ;
	    }
	    else
	    {
		Do_List() ; /* List a saveset */
	    } ;
	    return ;
	} ;
	Show_Error( "Missing output device" ) ;
	return ;
    } ;
    if( !Restore )
    {
	strcpy( Output_Spec, Command0 ) ;
	/* Output specification is the second parameter */

	/* Parse output specification */

	if( Output_Spec[ 1 ] != ':' )
	{
	    strcpy( Temp, "Invalid output drive: " ) ;
	    strcat( Temp, Output_Spec ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;
	Command0[ 0 ] = Output_Spec[ 0 ] ;
	Command0[ 1 ] = 0 ;
	strcpy( Output_Spec, &Output_Spec[ 2 ] ) ;
	if( strlen( Output_Spec ) == 0 )
	{
	    strcpy( Output_Spec, "BACKUP." ) ;
	    /* Output_Spec now contains the saveset name */
	} ;
        if( Extension_Pos( Output_Spec ) == 0 )
	{
           strcat( Output_Spec, "." ) ;
	} ;
	if( strlen( Output_Spec ) > 12 )
	{
	    strcpy( Temp, "Invalid saveset name: " ) ;
	    strcat( Temp, Output_Spec ) ;
	    Show_Error( Temp ) ;
	    return ;
	} ;
    } ;
    if( Switch_Present( Output_Spec ) > 0 )
    {
	strcpy( Temp, "Invalid switch: " ) ;
	strcat( Temp, Output_Spec ) ;
	Show_Error( Temp ) ;
	return ;
    } ;
    strcpy( Command, Command0 ) ;

    if( Restore )
    {
	Do_Restore() ;
    }
    else
    {
	Do_Backup() ;
    } ;

    /* Now end up */
    fclose( Input_File ) ;
    close( hInput_File ) ;
    fclose( Output_File ) ;
    close( hOutput_File ) ;
    free( Buffer ) ; /* Deallocate memory for buffer */
} ;
