.RADIX  10
.8086

;              Program Name : CVT
;              Package Name : LB
;              Purpose      : Conversion routines
;              Institution  : Conroy & Conroy Company.
;              Date Written : 04-Apr-86
;              Written By   : Alan Conroy
;              Version      : 9.2-10
;
;	 Copyright (C) 1986 by Alan Conroy.  Released to the public domain.
;
;	 TO THE GLORY OF GOD THROUGH JESUS CHRIST
;
;        *********************************************************
;        *                                                       *
;        *        M O D I F I C A T I O N   H I S T O R Y        *
;        *                                                       *
;        *     DATE      BY             REASON                   *
;        *                                                       *
;        *********************************************************
;
;        *********************************************************
;        *                                                       *
;        *           P R O G R A M   P U R P O S E               *
;        *                                                       *
;        *********************************************************
;
;          These routines perform various and sundry conversions.
;

; BIN2OCTR
;
; This routine converts a binary value in a register to ASCII octal.
; Data passed is AX=Value to convert, DS:BX = Location of output string.
; Data returned is DS:BX = Byte right after last byte in output string/
;

BIN2OCTR:
        PUSH  AX    ; Save registers
        PUSH  CX
        PUSH  DX
        PUSH  SI
        PUSHF
        MOV   SI,0  ; Digits flag (no nonzero digits yet)
        MOV   CL,15 ; Number of times to shift
        MOV   DX,AX ; Save value
CVT6:   MOV   AX,DX ; Get value
        SHR   AX,CL ; Shift value
        AND   AL,7  ; Get one octal digit
        CMP   AL,0  ; See if zero
        JNE   CVT8  ; Nope
        CMP   SI,0  ; Digits flag set?
        JE    CVT9  ; No
CVT8:   INC   SI    ; Set digits flag
        ADD   AL,48 ; Convert to ASCII
        MOV   [BX],AL ; Store digit
        INC   BX    ; Increment offset
CVT9:   CMP   CL,0  ; Done yet?
        JZ    CVT7  ; Yup
        SUB   CL,3  ; Next digit
        JMP   CVT6
CVT7:   CMP   SI,0  ; Digit flag set?
        JNE   CVT10 ; Yup
        MOV   AL,48 ; Otherwise set to zero
        MOV   [BX],AL
        INC   BX
CVT10:  POPF        ; Restore register
        POP   SI
        POP   DX
        POP   CX
        POP   AX
        RET

;
; DEC2BINR
;
; This routine converts an ASCII integer decimal value into binary form.
; Data passed is BX=Location of string, CL=length of string.
; Data returned is AX=Binary form of integer, CH=0 if number is OK, <>0 if bad.
; CL is undefined.

DEC2BINR:
        PUSH  SI ; Save registers
        PUSH  DI
        PUSH  DX
        PUSHF
        MOV   DI, 10 ; Value of ten
        MOV   SI, 1 ; Offset flag
        MOV   DH, 0 ; Clear negation flag
        MOV   DL, 0 ; Clear digit flag
        MOV   AX, 0 ; Zero accumulator
        CMP   CL, 0 ; Zero length?
        JE    CVT11 ; Yes - all done
        MOV   CH, [BX+SI] ; Get first character
        CMP   CH, '-' ; Is it a minus sign?
        JNE   CVT13 ; Nope
        INC   SI    ; Increment offset
        DEC   CL    ; One less character to examine
        INC   DH    ; Set negation flag
CVT13:  CMP   CL, 0 ; Zero length?
        JE    CVT11 ; Yes - all done
        DEC   CL    ; One less doorbell to answer...
        MOV   CH, [BX+SI] ; Get next character
        INC   SI    ; Increment offset
        CMP   CH, 48 ; Less than a "0"?
        JL    CVT14 ; Yes - error
        SUB   CH, 48 ; Convert to binary value
        CMP   CH, 9 ; Greater than a "9"
        JG    CVT14 ; yes - error
        CMP   CH, 0 ; Is it a zero?
        JNE   CVT15 ; Nope
        CMP   DL, 0 ; Digit flag set?
        JE    CVT13 ; Nope, so ignore zero
CVT15:  INC   DL    ; Set digit flag
        CMP   AX, 3276 ; See if multiplication would result in overflow
        JG    CVT14 ; Yes - error
        PUSH  DX ; Save DX
        MUL   DI ; Multiply by ten
        POP   DX ; Restore DX
        PUSH  CX ; Save CX
        MOV   CL, CH
        MOV   CH, 0
        ADD   AX, CX ; Add in next digit
        POP   CX ; Restore CX
        TEST  AX, 8000H ; See if overflow to high bit
        JNZ   CVT14 ; Yes - error
        JMP   CVT13 ; Loop
CVT14:  MOV   CH, 255 ; Error!
        JMP   CVT12 ; End
CVT11:  MOV   CH, 0 ; Number is OK
        CMP   DH, 0 ; Negation flag?
        JE    CVT12 ; Nope
        NEG   AX    ; Yes, so negate return value
CVT12:  POPF     ; Restore registers
        POP   DX
        POP   DI
        POP   SI
        RET      ; All done

;
; BIN2DECR
;
; This routine converts a binary value in a register to ASCII decimal.
; Data passed is AX=Value to convert, DS:BX = Location of output string.
; Data returned is DS:BX = Byte right after last byte in output string.
;

BIN2DECR:
        PUSH  AX ; Save registers
        PUSH  CX
        PUSH  DX
        PUSH  DI
        PUSH  SI
        PUSHF
        TEST  AX,8000H  ; See if sign bit is set
        JZ    CVT1      ; Not set
        PUSH  AX
        MOV   AL,45
        MOV   [BX], AL  ; Display sign bit
        POP   AX
        INC   BX        ; Increment output byte count
        NEG   AX        ; Convert to a positive value
CVT1:   MOV   CX,10000D ; Load register with highest power of ten
        MOV   DI,0
CVT2:   PUSH  AX        ; Save AX
        MOV   DX,0
        DIV   CX        ; Divide AX by the value
        CMP   AX,0      ; See if digit is zero
        JNE   CVT3      ; If not zero
        MOV   AX,DI     ; Load flags
        CMP   AX,0      ; See if any non-zero digits yet
        JE    CVT4      ; Nope
        MOV   AX,0      ; Load zero back into AX
CVT3:   ADD   AL, 48D   ; Add to 48 to get ASCII value of digit
        MOV   [BX], AL  ; Store digit
        INC   BX
        INC   DI        ; Set flag to say non-zero digit encountered
        SUB   AL,48D    ; Get original value
        MUL   CX        ; Multiply by divisor
        MOV   SI,AX     ; Save result
        POP   AX        ; Get original number
        SUB   AX,SI     ; Get new quotient
        PUSH  AX        ; Save new quotient
CVT4:   MOV   AX,CX     ; Move divisor into AX
        MOV   DX,0
        MOV   CX,10
        DIV   CX        ; Divide by ten
        MOV   CX,AX
        CMP   AX,0      ; See if done
        POP   AX        ; Restore AX
        JNE   CVT2      ; Not done
        MOV   AX,DI     ; Any digits?
        CMP   AX,0
        JNE   CVT5      ; Yes
        MOV   AL,48
        MOV   [BX],AL   ; Output a 0
        INC   BX
CVT5:   POPF            ; Restore registers
        POP   SI
        POP   DI
        POP   DX
        POP   CX
        POP   AX
        RET             ; All done

	END
