/*
              Program Name : Filnam
              Package Name : LB
              Purpose      : File name translation
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 2.0A

              Copyright (C) 1987-1988 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This function searches the current set of paths for a file, returning
        the name of the file (including path specification, if found).
*/

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <io.h>
#include <ctype.h>

#include "files.h"

static int Nested ;
static char Buffer[ 256 ] ;

static void Find_File( char *X )
{
/* Parse X and find file.  Return actual name in Buffer. */

long A, AA ;
char B[ 256 ], Env[ 256 ], Path[ 256 ], Work[ 256 ] ;
char *Temp ;
int F ;

    /* First, see if we are nested more than eight (8) deep.  If so, we
     automatically fail.  This prevents any infinite loop due to recursive
     symbol definitions. */
    Nested = Nested + 1 ; /* Nested one more level */
    if( Nested == 8 ) /* Nested too deep */
    {
	Buffer[ 0 ] = 0 ; /* Not found */
    } ;

    /* Now try to find the file */

    while( strlen( X ) > 0 ) /* Check all items in list */
    {
	/* Get next item in list */

	A = 0 ;
	strcpy( B, X ) ;
	while( X[ A ] != 0 && X[ A ] != ';' )
	{
	    A = A + 1 ;
	} ;
	B[ A ] = 0 ; /* Terminate list item */
	if( X[ A ] == 0 ) /* No semicolon found */
	{
	    X[ 0 ] = 0 ; /* What remains */
	}
	else
	{
	    strcpy( X, &X[ A + 1 ] ) ; /* What remains */
	} ;

	/* Check for literal device name */

	if( B[ 0 ] == '_' ) /* Use literal device */
	{
	    if( ( F = open( &B[ 1 ], O_RDONLY ) ) != -1 )
	    {
		close( F ) ;
		strcpy( Buffer, &B[ 1 ] ) ;
		return ;
	    } ;
	} ; /* if ( B[ 0 ] == '_' ) */

	/* Handle case where there is no device/logical specified */

	A = 0 ;
	while( B[ A ] != 0 && B[ A ] != ':' )
	{
	    A = A + 1 ;
	} ;
	if( B[ A ] == 0 ) /* No device/logical */
	{
	    if( ( F = open( &B, O_RDONLY ) ) != -1 )
	    {
		close( F ) ;
		strcpy( Buffer, &B ) ;
		return ;
	    } ;
	    continue ;
	} ; /* if ( B[ A ] == 0 ) */

	/* Try opening with symbol as a logical */

	Path[ 0 ] = 0 ;
	strcpy( Work, B ) ;
	Work[ A ] = 0 ; /* Get device/logical name */
	AA = 0 ;
	while( Work[ AA ] != 0 )
	{
	    Work[ AA ] = toupper( Work[ AA ] ) ;
	    AA = AA + 1 ;
	} ;
	Temp = getenv( Work ) ; /* Get device/logical translation */
	if( Temp == NULL )
	{
	    Env[ 0 ] = 0 ; /* Not found */
	}
	else
	{
	    strcpy( Env, Temp ) ;
	} ;
	strcpy( Work, &B[ A + 1 ] ) ; /* Get only file name/extension */
	while( strlen( Env ) > 0 ) /* Try all specifications */
	{
	    A = 0 ;
	    while( Env[ A ] != 0 && Env[ A ] != ';' )
	    {
		A = A + 1 ;
	    } ;
	    strcpy( Path, Env ) ;
	    Path[ A ] = 0 ; /* Get next specification */
	    if( Env[ A ] == 0 )
	    {
		Env[ 0 ] = 0 ; /* What remains is nothing */
	    }
	    else
	    {
		strcpy( Env, &Env[ A + 1 ] ) ; /* What remains */
	    } ;
	    if( strlen( Path ) > 0 )
	    {
		if( Path[ strlen( Path ) - 1 ] == ':' ) /* Nested logical */
		{
		    strcat( Path, Work ) ;
		    Find_File( Path ) ;
		    if( strlen( Buffer ) != 0 ) /* Found it */
		    {
			return ;
		    } ;
		    goto Skip ;
		} ;
		if( Path[ strlen( Path ) - 1 ] != '\\' )
		{
		    strcat( Path, "\\" ) ; /* Must end with backslash */
		} ;
	    } ;
	    strcat( Path, Work ) ;
	    if( ( F = open( &Path, O_RDONLY ) ) != -1 )
	    {
		close( F ) ;
		strcpy( Buffer, Path ) ;
		return ;
	    } ;
	} ; /* while( strlen( Env ) > 0 ) */

Skip:
        /* Now try device as a real device name */

	if( ( F = open( &B, O_RDONLY ) ) != -1 )
	{
	    close( F ) ;
	    strcpy( Buffer, B ) ;
	    return ;
	} ;
    } ;
    Buffer[ 0 ] = 0 ; /* File was not found */
} ; /* Find_File */


char *Filnam( char *X )
{
/* Find a file given a file name.  Name can contain logicals (symbols), and
 search lists (;).  If no device/directory is specified and file is not
 found with name given, it is searched for in PATH:.  Literal devices are
 specified with underscores (_).  If a logical translation fails, the
 "logical" is tried as if a literal device name. */

int Dummy ;
char S[ 256 ], Work[ 256 ] ;

    /* Pass 1 : Try as passed */
    strcpy( S, X ) ;
    while( strlen( S ) > 0 ) /* Try each listed item */
    {
	Dummy = 0 ;
	while( S[ Dummy ] != 0 && S[ Dummy ] != ';' )
	{
	    Dummy = Dummy + 1 ;
	} ;
	strcpy( Work, S ) ;
	Work[ Dummy ] = 0 ; /* Get next specification */
	if( S[ Dummy ] == 0 )
	{
	    S[ 0 ] = 0 ;
	}
	else
	{
	    strcpy( S, &S[ Dummy + 1 ] ) ;
	} ;
	Nested = 0 ;
	Find_File( Work ) ;
	if( Buffer[ 0 ] != 0 ) /* Found file as passed */
	{
	    return &Buffer ;
	} ;
    } ;


    /* Pass 2 : Try it with PATH: */

    strcpy( S, X ) ;
    while( strlen( S ) > 0 ) /* Try each listed item */
    {
	Dummy = 0 ;
	while( S[ Dummy ] != 0 && S[ Dummy ] != ';' )
	{
	    Dummy = Dummy + 1 ;
	} ;
	strcpy( Work, S ) ;
	Work[ Dummy ] = 0 ; /* Get next specification */
	if( S[ Dummy ] == 0 )
	{
	    S[ 0 ] = 0 ;
	}
	else
	{
	    strcpy( S, &S[ Dummy + 1 ] ) ;
	} ;
	Dummy = 0 ;
	while( Work[ Dummy ] != ':' && Work[ Dummy ] != '\\' && Work[ Dummy ] != 0 )
	{
	    Dummy = Dummy + 1 ;
	} ;
	if( Work[ Dummy ] == 0 ) /* No device/directory */
	{
	    Nested = 0 ;
	    strcpy( Buffer, "PATH:" ) ;
	    strcat( Buffer, Work ) ;
	    strcpy( Work, Buffer ) ;
	    Find_File( Work ) ;
	    if( Buffer[ 0 ] != 0 ) /* Found file */
	    {
		return &Buffer ;
	    } ; /* if( Buffer[ 0 ] != 0 ) */
	} ; /* if( Work[ Dummy ] == 0 ) */
    } ; /* while( strlen( S ) > 0 ) */
    return NULL ; /* Not found at all */
} ; /* Filnam */

