.RADIX  10
.8086

;              Program Name : INPUT
;              Package Name : LB
;              Purpose      : Keyboard input routine
;              Institution  : Conroy & Conroy Company
;              Date Written : 04-Apr-86
;              Written By   : Alan Conroy
;              Version      : 9.2-10
;
;	 Copyright ( C) 1986 by Alan Conroy.  Released to the public domain.
;
;        *********************************************************
;        *                                                       *
;        *        M O D I F I C A T I O N   H I S T O R Y        *
;        *                                                       *
;        *     DATE      BY             REASON                   *
;        *                                                       *
;        *********************************************************
;
;        *********************************************************
;        *                                                       *
;        *           P R O G R A M   P U R P O S E               *
;        *                                                       *
;        *********************************************************
;
;          These routines do keyboard input.  This is designed to be used as
;	 an include file.  The IO.ASM file must have been included prior to the
;	 inclusion of this file.
;
;
; INPUT
;
; This routine takes input from the keyboard.  No registers are changed upon
; return from this routine.  The input is placed into the InBuff and the length
; in InBLen.
;

INPUT:  PUSH  AX ; Save registers
        PUSH  BX
        PUSH  CX
        PUSH  DX
        PUSHF

In1:    ; Do input and initial special character checks
        MOV   AH,08H
        INT   21H     ; Get byte of input
        AND   AL,127  ; Strip high bit
        CMP   AL,0    ; Is it a null?
        JE    In1     ; If so, ignore it
        PUSH  AX      ; Save input
        CMP   AL,_LF  ; Is it a LF?
        JE    In2     ; yes
        CMP   AL,_VT  ; Is it a VT?
        JE    In2     ; yes
        CMP   AL,_FF  ; Is it a FF?
        JE    In2     ; yes
        CMP   AL,_CR  ; Is it a CR?
        JE    In2     ; yes
        CMP   AL,_DEL  ; Is it a delete?
        JE    Inc     ; yes - do delete stuff
        CMP   AL,21   ; Is character a control-U?
        JE    Ina     ; Yup - skip input buffer length check
        CMP   AL,18   ; Is it a control-R?
        JE    Ina     ; Yup - skip input buffer length check

        ; Special handling for escapes
        CMP   AL,_ESC ; See if an escape
        JNE   Inb     ; Nope
        MOV   AH,02H
        MOV   DL,'$'
        INT   21H     ; Output dollar sign
        POP   AX      ; Refresh AX
        PUSH  AX
        JMP   In3

Inb:    ; Check length of input buffer (maximum=255 bytes without delimiter)
        MOV   BX, OFFSET InBLen
        MOV   AL,[BX] ; Get length of input so far
        CMP   AL,255  ; Is length = 255?
        JNE   Ina     ; Nope
        MOV   AH,02H
        MOV   DL,_BEL
        INT   21H     ; Output bell
        POP   AX      ; Clear stack
        JMP   In1     ; Go back for more input

Ina:    ; Special stuff for control characters
        POP   AX      ; refresh AX
        PUSH  AX
        CMP   AL,31   ; See if a control character
        JG    In2     ; Nope
        CMP   AL,_BEL ; Is is a bell?
        JE    In2     ; If so, ignore it
        CMP   AL,_BS  ; Is it a backspace?
        JE    In2     ; If so, ignore it
        CMP   AL,_HT  ; Is it a tab?
        JE    In2     ; If so, ignore it
        MOV   AH,02H
        MOV   DL,'^'
        INT   21H     ; Output circumflex
        POP   AX      ; Refresh AX from stack
        PUSH  AX
        ADD   AL,64   ; Convert to a letter

In2:    ; Display input character
        ; Also special stuff for control-R
        MOV   AH,02H
        MOV   DL,AL
        INT   21H     ; Output character
        POP   AX      ; Refresh AX
        PUSH  AX
        CMP   AL,18   ; Is it a control-R?
        JNE   Inf     ; Nope
        POP   AX      ; Clear stack
        CALL  CTRLR   ; handle it otherwise
        JMP   In1     ; Go back for more input

Inf:    ; Control-U handler
        CMP   AL,21   ; See if a control-U
        JNE   In3     ; Nope
        POP   AX      ; Clear stack
        MOV   BX, OFFSET InBLen
        MOV   AL,0
        MOV   [BX],AL ; Zero out input buffer
        CALL  CRLF
        JMP   In1     ; Go back for input

Inc:    ; Delete handler
        POP   AX      ; Clear stack
        MOV   BX, OFFSET InBLen
        MOV   AL,[BX] ; Get input buffer length
        CMP   AL,0    ; Is there any input to delete?
        JNE   Ind     ; Yes
        JMP   In1     ; If not, go back for input
Ind:    MOV   AH,0
        PUSH  AX      ; Save length
        DEC   AL      ; Reduce length of input buffer by one character
        MOV   [BX],AL
        CALL  In7     ; Erase character on screen
        POP   SI
        MOV   BX, OFFSET InBuff
        MOV   AL,[BX+SI] ; Look at deleted character
        CMP   AL,32   ; Is it a control character?
        JL    Ine     ; Yes
        JMP   In1     ; If not, we is done
Ine:    CMP   AL,_BEL ; A bell?
        JE    Ing     ; Yes
        CMP   AL,_BS  ; A backspace?
        JE    Ing     ; Yes
        CMP   AL,_HT  ; A tab?
        JE    Ing     ; Yes
        CALL  In7     ; Erase caret on screen
Ing:    JMP   In1     ; Go back for more input

In3:    ; Store character in input buffer
        MOV   BX, OFFSET InBLen
        MOV   AL,[BX]
        MOV   AH,0
        INC   AL      ; Increment buffer length
        MOV   SI,AX   ; Get input buffer length (offset into buffer)
        MOV   BX, OFFSET InBLen
        MOV   [BX],AL ; and update length of input buffer
        MOV   BX, OFFSET InBuff
        POP   AX      ; Restore character
        PUSH  AX
        MOV   [BX+SI],AL ; Store character in input buffer
        POP   AX      ; Restore character and clear stack
        CMP   AL,27   ; Is it a ESC?
        JE    In6 ; yes
        CMP   AL,26 ; Is it a ^Z?
        JE    In6 ; yes
        CMP   AL,_LF ; Is it a LF?
        JE    In4 ; yes
        CMP   AL,_VT ; Is it a VT?
        JE    In4 ; yes
        CMP   AL,_FF ; Is it a FF?
        JE    In4 ; yes
        CMP   AL,_CR ; Is it a CR?
        JE    In4 ; yes
        JMP   In1 ; no - go back for next character
In4:    CMP   AL,_CR ; Is it a CR?
        JNE   In5 ; no
        MOV   AH,02H
        MOV   DL,_LF
        INT   21H ; Output LF
        JMP   In6
In5:    MOV   AH,02H
        MOV   DL,_CR
        INT   21H ; Output CR
In6:    POPF  ; Restore registers
        POP   DX
        POP   CX
        POP   BX
        POP   AX
        RET ; All done

In7:    ; Delete last character on screen
        MOV   AH,02H
        MOV   DL,_BS
        INT   21H     ; Output a backspace
        MOV   AH,02H
        MOV   DL,_Space
        INT   21H     ; Output a space
        MOV   AH,02H
        MOV   DL,_BS
        INT   21H     ; Output a backspace
        RET

CTRLR:  ; Handle control-R
        CALL  CRLF
        MOV   BX, OFFSET InBLen ; Get length of input
        MOV   AL,[BX]
        CMP   AL,0    ; Any input at all?
        JG    In8     ; Yes
        RET           ; otherwise we're done

In8:    MOV   SI,0    ; Start at beginning of input buffer

In9:    ; Beginning of loop
        INC   SI      ; Increment counter
        MOV   BX, OFFSET InBLen ; Get length of input buffer
        MOV   AL, [BX]
        MOV   AH,0
        CMP   SI, AX  ; Are there still characters to process?
        JLE   In10    ; Yes
        RET           ; We're done

In10:   MOV   BX, OFFSET InBuff ; Get next character
        MOV   AL,[BX+SI]
        PUSH  AX      ; Save character
        CMP   AL,_LF  ; Is it a LF?
        JE    In12    ; yes
        CMP   AL,_VT  ; Is it a VT?
        JE    In12    ; yes
        CMP   AL,_FF  ; Is it a FF?
        JE    In12    ; yes
        CMP   AL,_CR  ; Is it a CR?
        JE    In12    ; yes
        CMP   AL,31   ; See if a control character
        JG    In12    ; Nope
        CMP   AL,_BEL ; Is is a bell?
        JE    In12    ; If so, ignore it
        CMP   AL,_BS  ; Is it a backspace?
        JE    In12    ; If so, ignore it
        CMP   AL,_HT  ; Is it a tab?
        JE    In12    ; If so, ignore it
        CMP   AL,_ESC ; See if an escape
        JNE   In11    ; Nope
        PUSH  SI      ; Save counter
        MOV   AH,02H
        MOV   DL,'$'
        INT   21H     ; Output dollar sign
        POP   SI      ; Restore counter
        POP   AX      ; Clear stack
        JMP   In9     ; Go back for next character

In11:   ; Special stuff for control characters
        PUSH  SI      ; Save counter
        MOV   AH,02H
        MOV   DL,'^'
        INT   21H     ; Output circumflex
        POP   SI      ; Restore counter
        POP   AX      ; Refresh AX
        PUSH  AX
        ADD   AL,64   ; Convert to a letter

In12:   ; Display character
        PUSH  SI ; Save counter
        MOV   AH,02H
        MOV   DL,AL
        INT   21H     ; Output character
        POP   SI      ; Restore counter
        POP   AX      ; Refresh AX
        PUSH  AX
        CMP   AL,26 ; Is it a ^Z?
        JNE   In15  ; No
        POP   AX    ; Clear stack
        JMP   In9   ; Go back for more
In15:   CMP   AL,_LF ; Is it a LF?
        JE    In13  ; yes
        CMP   AL,_VT ; Is it a VT?
        JE    In13  ; yes
        CMP   AL,_FF ; Is it a FF?
        JE    In13  ; yes
        CMP   AL,_CR ; Is it a CR?
        JE    In13  ; yes
        POP   AX    ; Clear stack
        JMP   In9   ; no - go back for next character
In13:   CMP   AL,_CR ; Is it a CR?
        JNE   In14 ; no
        PUSH  SI ; Save counter
        MOV   AH,02H
        MOV   DL,_LF
        INT   21H ; Output LF
        POP   SI ; Restore counter
        POP   AX ; Clear stack
        JMP   In9 ; Go back for next character

In14:   PUSH  SI ; Save counter
        MOV   AH,02H
        MOV   DL,_CR
        INT   21H ; Output CR
        POP   SI ; Restore counter
        POP   AX ; Clear stack
        JMP   In9 ; Go for next character

;
; CLEARIN
;
; This routine clears the input buffer
;

CLEARIN:
        PUSH  BX ; Save registers
        PUSHF
        MOV   BX, OFFSET InBLen
        MOV   BYTE PTR [BX], 0 ; Set length to zero
        POPF     ; Restore registers
        POP   BX
        RET ; All done

	END
