/*
              Program Name : Wildcard
              Package Name : LB
              Purpose      : Filename wildcard lookup
              Institution  : Conroy & Conroy
              Date Written : 31-Mar-89
              Written By   : Alan Conroy
              Version      : 2.1

	      Copyright (C) 1989 by Alan Conroy.  Released to the public domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *     DATE      BY             REASON                   *
        *                                                       *
        *********************************************************

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This function provides for Extended DOS Wildcard lookups.
*/

#include <stdlib.h>
#include <dir.h>
#include <dos.h>
#include <string.h>

#include "edit.h"
#include "clib.h"
#include "ustrings.h"
#include "misc.h"
#include "instr.h"
#include "wildcard.h"

struct ffblk File_Stuff ;
static char File_Name[ 256 ] ;
static char File_Type[ 256 ] ;
static char Path_Name[ 256 ] ;
static char Device_Name[ 4 ] ;
static char Directories[ 34 ][ 14 ] ;
static char Directory_Name[ 256 ] ;
static int Directory_Wildcard[ 34 ] ;
static short int Wildcard_Level ; /* Starting level of ... specification */
static char File_List[ 256 ] = { 0 } ; /* List of remaining paths to search */
static int A, B, Dummy, Dummy1 ; /* Dummy variables */
static struct ffblk Dir ;
static char Direc[ 256 ], Temp[ 256 ] ;
static char Filename[ 256 ] ;


static char *Find_Next_Dir( char *Passed_Path, char *Passed_Direct )
/* Find next directory after Direct in Path, and return name.  Return null if
 no more directories after Direct in Path. */
{

char Buffer[ 128 ], Temp[ 256 ] ;
char Path[ 256 ], Direct[ 256 ] ;

    strcpy( Path, Passed_Path ) ;
    strcpy( Direct, Passed_Direct ) ;
    strcpy( Temp, Device_Name ) ;
    strcat( Temp, Path ) ;
    strcat( Temp, "*.*" ) ;
    if( findfirst( Temp, &Dir, FA_DIREC ) ) /* Failure */
    {
	Buffer[ 0 ] = 0 ;
	return Buffer ;
    } ;

    /* Start at beginning of path */
    if( Direct[ strlen( Direct ) - 1 ] == '\\' )
    {
	strcpy( Direct, Left( Direct, strlen( Direct ) - 1 ) ) ;
    } ;
    if( strcmp( Direct, "*" ) == 0 || strcmp( Direct, "*.*" ) == 0 )
    {
	Direct[ 0 ] = 0 ; /* Get first directory */
    } ;
    if( Direct[ 0 ] != 0 )
    {
	while( strcmp( Direct, Dir.ff_name ) != 0 )
	{
	    if( findnext( &Dir ) ) /* Find this directory */
	    {
		Buffer[ 0 ] = 0 ;
		return Buffer ;
		/* Failure - Sub-directory must have been deleted */
	    } ;
	} ;
	if( findnext( &Dir ) ) /* Now find next directory */
	{
	    Buffer[ 0 ] = 0 ;
	    return Buffer ; /* Failure - No sub-directories */
	} ;
    } ;

    while( Dir.ff_name[ 0 ] == '.' || ( Dir.ff_attrib & 0x10 ) != 0x10 )
    /* Ignore ., .., and non-directories */
    {
	if( findnext( &Dir ) )
	{
	    Buffer[ 0 ] = 0 ;
	    return Buffer ; /* Failure - no sub-directories */
	} ;
    } ;
    strcpy( Buffer, Dir.ff_name ) ;
    strcat( Buffer, "\\" ) ;
    return Buffer ;
} ;

static char *Do_Lookup( char *Path_Name ) /* Return Filename */
{

int A, B, Dummy, Dummy1 ;
char P[ 128 ] ;
int Wildcard_Failure ;
char Buffer[ 256 ], Temp[ 256 ] ;
int Result ;

Try_Again:
    if( strlen( Path_Name ) > 0 ) /* First time */
    {
	A = 0 ;
        for( Dummy = 1; Dummy <= 33; Dummy++ ) /* For all directory levels */
	{
	    if( strcmp( Directories[ Dummy ], "*\\" ) == 0
		|| strcmp( Directories[ Dummy ], "*.*\\" ) == 0 )
	    {
	        Directory_Wildcard[ Dummy ] = True ;
		if( A == 0 )
		{
		    A = Dummy ; /* First wildcard */
		} ;
	    }
	    else
	    {
		Directory_Wildcard[ Dummy ] = False ;
	    } ;
	} ;
        Direc[ 0 ] = 0 ;
	Dummy = 1 ;
        /* On first file, replace directory wildcards with first directory name */
	while( Dummy <= 33 )
	{
	  if( Directory_Wildcard[ Dummy ] )
	  {
	      if( strcmp( Directories[ Dummy ], "*\\" ) != 0 &&
		strcmp( Directories[ Dummy ], "*.*\\" ) != 0 )
	      /* Already did a wildcard substitution on this directory */
	      {
		strcpy( Directories[ Dummy ],
			Find_Next_Dir( Direc, Directories[ Dummy ] ) ) ;
		if( strlen( Directories[ Dummy ] ) == 0 ) /* Lookup failure */
		{
		    if( A == Dummy ) /* First directory */
		    {
			Buffer[ 0 ] = 0 ;
			return Buffer ;
		    } ;
		    /* Backup to previous wildcard directory to try again */
		    while( Dummy > 0 && Directory_Wildcard[ Dummy ] == False )
		    {
			Dummy = Dummy - 1 ;
		    } ;
		    Dummy = Dummy - 1 ;
		    /* Decrement Dummy, since end of loop increments Dummy, so
		     that the next loop begins on this directory. */
		} ;
	      }
	      else /* Need first lookup for this wildcard directory */
	      {
		strcpy( Directories[ Dummy ], Find_Next_Dir( Direc, "" ) ) ;
		if( strlen( Directories[ Dummy ] ) == 0 ) /* Lookup failure */
		{
		    if( A == Dummy ) /* First directory */
		    {
			Buffer[ 0 ] = 0 ;
			return Buffer ;
		    } ;
		    /* Backup to previous wildcard directory to try again */
		    while( Dummy > 0 && Directory_Wildcard[ Dummy ] == False )
		    {
			Dummy = Dummy - 1 ;
		    } ;
		    Dummy = Dummy - 1 ;
		    /* Decrement Dummy, since end of loop increments Dummy, so
		     that the next loop begins on this directory. */
		}
		else
		{
		    strcat( Direc, Directories[ Dummy ] ) ;
		} ;
	      } ;
	  }
	  else
	  {
	    strcat( Direc, Directories[ Dummy ] ) ;
	  } ;
	  Dummy = Dummy + 1 ;
	} ; /* while */
	strcpy( Temp, Device_Name ) ;
	strcat( Temp, Direc ) ;
	strcat( Temp, File_Name ) ;
	strcat( Temp, File_Type ) ;
	Result = findfirst( Temp, &File_Stuff,
	    FA_RDONLY | FA_HIDDEN | FA_SYSTEM | FA_LABEL | FA_DIREC | FA_ARCH ) ;
    }
    else
    {
        Direc[ 0 ] = 0 ; /* Build valid path in case this works */
        for( Dummy = 1; Dummy <= 33; Dummy++ )
	{
	    strcat( Direc, Directories[ Dummy ] ) ;
	} ;
	Result = findnext( &File_Stuff ) ;
    } ; /* if( strlen( Path_Name ) > 0 ) */


    if( Result != 0 ) /* Lookup error */
    {
	/* At this point, we need to check for the next wildcard to change to
	 (if any) */

	Wildcard_Failure = False ;
        A = 33 ; /* Start with last directory in path */

Backup:
	if( A < 1 )
	{
	    Buffer[ 0 ] = 0 ; /* Done */
	    return Buffer ;
	} ;

	/* Find last used directory in list */
        Dummy1 = 0 ; /* Dummy1 will be the last used directory */
	for( Dummy = A; Dummy >= 1; Dummy-- )
	{
	    if( strlen( Directories[ Dummy ] ) > 0 )
	    {
		if( Dummy1 == 0 )
		{
		    Dummy1 = Dummy ;
		} ;
	    } ;
	} ;

        /* Build path up to current directory */
        Direc[ 0 ] = 0 ;
	for( Dummy = 1; Dummy < Dummy1; Dummy++ )
	{
	    strcat( Direc, Directories[ Dummy ] ) ;
	} ;

	/* These next lines are needed for "\*\...\" constructs */
	Dummy = Wildcard_Level - 1 ;
	if( Wildcard_Failure )
	{
	    Dummy = Dummy + 1 ;
	} ;

	/* If "..." lookup failure, then look for wildcard directories */
	if( Dummy1 >= Dummy ) /* a "..." match needed */
	{
	    while( strlen( Directories[ Dummy1 + 1 ] ) == 0 )
	    {
		strcpy( Temp, Direc ) ;
		strcat( Temp, Directories[ Dummy1 ] ) ;
		strcpy( Directories[ Dummy1 + 1 ], Find_Next_Dir( Temp, "" ) ) ;
		/* Find first sub-directory in current path */
Next_Sub_Directory:
		if( strlen( Directories[ Dummy1 + 1 ] ) == 0 )
		/* No subdirectories */
		{
		    if( Dummy1 < Wildcard_Level ) /* No "..." matches */
		    {
			A = Wildcard_Level - 1 ;
			Wildcard_Failure = True ;
			goto Backup ;
		    } ;

		    strcpy( Directories[ Dummy1 ],
			Find_Next_Dir( Direc, Directories[ Dummy1 ] ) ) ;
			/* Find next directory after current one in path */
		    Dummy1 = Dummy1 - 1 ; /* Backup one level */

		    /* Build path up to new level */
		    Direc[ 0 ] = 0 ;
		    for( Dummy = 1; Dummy < Dummy1; Dummy++ )
		    {
			strcat( Direc, Directories[ Dummy ] ) ;
		    } ;
		    if( strlen( Directories[ Dummy1 + 1 ] ) == 0 )
		    /* That was the last subdir */
		    {
			goto Next_Sub_Directory ;
		    } ;
		} ; /* if */
	    } ; /* while */
	}
	else
	{
Next_Directory:
	    /* Find previous wildcard directory */
	    Dummy1 = 0 ;
	    for( Dummy = A; Dummy >= 1; Dummy-- )
	    {
		if( Directory_Wildcard[ Dummy ] )
		{
		    if( Dummy1 == 0 )
		    {
			Dummy1 = Dummy ;
		    } ;
		} ;
	    } ;
	    if( Dummy1 == 0 )
	    {
		Buffer[ 0 ] = 0 ; /* Failure */
		return Buffer ;
	    } ;

	    /* Build path up to current wildcard */
	    Direc[ 0 ] = 0 ;
	    for( Dummy = 1; Dummy < Dummy1; Dummy++ )
	    {
		strcat( Direc, Directories[ Dummy ] ) ;
	    } ;

	    /* Find next directory after current one */
	    strcpy( Directories[ Dummy1 ],
		Find_Next_Dir( Direc, Directories[ Dummy1 ] ) ) ;
	    if( strlen( Directories[ Dummy1 ] ) == 0 ) /* No more directories */
	    {
		A = Dummy1 - 1 ;
		goto Next_Directory ;
	    } ;

	    /* Fill rest of wildcards in path beyond this directory */
	    strcat( Direc, Directories[ Dummy1 ] ) ;
	    for( B = Dummy1 + 1; B <= 32; B++ )
	    {
		if( Directory_Wildcard[ B ] )
		{
		    strcpy( Directories[ B ], Find_Next_Dir( Direc, "" ) ) ;
		    /* Find first directory in new path up to this point */
	            if( strlen( Directories[ B ] ) == 0 )
		    /* No sub-directory found */
		    {
			A = Dummy1 ;
			goto Next_Directory ;
		    } ;
		} ;
		strcat( Direc, Directories[ B ] ) ;
	    } ; /* for */
	} ;

	Direc[ 0 ] = 0 ;
	for( A = 1; A <= 32; A++ )
	{
	    if( strlen( Directories[ A ] ) != 0 )
	    {
		strcat( Direc, Directories[ A ] ) ; /* Build path */
	    } ;
	} ;

	/* Try next file */
	strcpy( Temp, Device_Name ) ;
	strcat( Temp, Direc ) ;
	strcat( Temp, File_Name ) ;
	strcat( Temp, File_Type ) ;
	if( findfirst( Temp, &File_Stuff,
	    FA_RDONLY | FA_HIDDEN | FA_SYSTEM | FA_LABEL | FA_DIREC | FA_ARCH ) )
	{
	    goto Try_Again ; /* No files on this path */
	} ;
    } ;
    strcpy( Buffer, Device_Name ) ;
    strcat( Buffer, Direc ) ;
    strcat( Buffer, File_Stuff.ff_name ) ;
    return Buffer ;
} ;


static char *Lookup_File( char *Passed_Filename ) /* Do a lookup */
{

char Buffer[ 256 ], Filename[ 256 ] ;
char Save_Directories[ 34 ][ 14 ] ;

    if( Passed_Filename == NULL )
    {
	Filename[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Filename, Passed_Filename ) ;
    } ;

    if( strlen( Filename ) > 0 ) /* First time */
    {
        Edit( Filename, Filename, 32 ) ;
        if( Filename[ strlen( Filename ) - 1 ] == '\\' )
	{
	    strcat( Filename, "*.*" ) ;
	} ;
        Wildcard_Level = 34 ; /* No "..." */
        for( Dummy = 1; Dummy <= 33; Dummy++ )
	{
	    Directories[ Dummy ][ 0 ] = 0 ;
	} ;
        _splitpath( Filename, Device_Name, Directory_Name, File_Name,
	    File_Type ) ;
        if( strlen( File_Name ) == 0 && strcmp( File_Type, "." ) == 0 )
	{
	    strcpy( File_Name, "*" ) ; /* Convert filename of . to *.  */
	} ;
        Dummy1 = 1 ;
        Dummy = 1 ;
        A = 0 ;
        while( Dummy1 <= strlen( Directory_Name ) )
        {
            while( Directory_Name[ Dummy1 - 1 ] != '\\' &&
                Dummy1 < strlen( Directory_Name ) )
	    {
		Dummy1 = Dummy1 + 1 ; /* Find next backslash */
	    } ;
            A = A + 1 ;
            if( Wildcard_Level < A ) /* Illegal format in directory spec */
            {
                Buffer[ 0 ] = 0 ;
		return Buffer ;
            } ;
            if( Dummy == Dummy1 ) /* Single character directory name */
	    {
		Directories[ A ][ 0 ] = Directory_Name[ Dummy - 1 ] ;
		Directories[ A ][ 1 ] = 0 ;
	    }
	    else
	    {
		strcpy( Directories[ A ],
		    Mid( Directory_Name, Dummy, Dummy1 - Dummy + 1 ) ) ;
	    } ;
            B = Instr( 1, Directories[ A ], "..." ) ;
            if( B > 0 )
            {
                if( strcmp( Directories[ A ], "...\\" ) == 0 )
                {
                    Directories[ A ][ 0 ] = 0 ;
                    Wildcard_Level = A ;
                }
		else
                {
		    strcpy( Directories[ A ],
			Mid( Directories[ A ], 1, B - 1 ) ) ;
		    strcat( Directories[ A ], "\\" ) ;
                    Wildcard_Level = A + 1 ;
                } ;
                A = A + 1 ;
            } ;
            Dummy1 = Dummy1 + 1 ;
            Dummy = Dummy1 ;
        } ; /* while( Dummy1 <= strlen( Directory_Name ) ) */

        if( strcmp( Device_Name, "*:" ) == 0 ) /* Wildcard device */
	{
	    Device_Name[ 0 ] = 'A' ;
	    Directory_Wildcard[ 0 ] = True ;
	    for( Dummy = 1; Dummy <= 33; Dummy++ )
	    {
		strcpy( Save_Directories[ Dummy], Directories[ Dummy ] ) ;
	    } ;
	}
	else
	{
	    Directory_Wildcard[ 0 ] = False ;
	} ;
	strcpy( Filename, Device_Name ) ;
	strcat( Filename, Directory_Name ) ;
	strcat( Filename, File_Name ) ;
	strcat( Filename, File_Type ) ;
    } ; /* if( strlen( Filename ) > 0 ) */

Next_Device:
    strcpy( Temp, Do_Lookup( Filename ) ) ;
    if( strlen( Temp ) == 0 )
    {
	if( Directory_Wildcard[ 0 ] )
	/* No files found for current specification and a wildcard device */
	{
            while( strlen( Temp ) == 0 ) /* Check next devices */
            {
                if( Device_Name[ 0 ] == 'Z' ) /* End of devices */
                {
                    Buffer[ 0 ] = 0 ;
		    return Buffer ;
                } ;
                Device_Name[ 0 ] = Device_Name[ 0 ] + 1 ;
                strcpy( Filename, Device_Name ) ;
		strcat( Filename, Directory_Name ) ;
		strcat( Filename, File_Name ) ;
		strcat( Filename, File_Type ) ;
		for( Dummy = 1; Dummy <= 33; Dummy++ )
		{
		    strcpy( Directories[ Dummy ], Save_Directories[ Dummy ] ) ;
		} ;
                goto Next_Device ;
            } ;
	} ;
    } ;
    if( ( File_Stuff.ff_attrib & 0x08 ) == 0x08 ) /* Label */
    {
	if( strlen( Temp ) > 0 )
	{
	    Filename[ 0 ] = 0 ; /* Go to next file */
	    goto Next_Device ; /* Ignore labels */
	} ;
    } ;
    return Temp ; /* Return value */
} ;

char *Lookup( char *Passed_Filename ) /* Do a lookup */
{

int A ;
char Filename[ 256 ], Temp[ 256 ] ;

    if( Passed_Filename == NULL )
    {
	Filename[ 0 ] = 0 ;
    }
    else
    {
	strcpy( Filename, Passed_Filename ) ;
    } ;

    if( strlen( Filename ) > 0 ) /* Starting new lookup */
    {
	strcpy( File_List, Filename ) ;
	Temp[ 0 ] = 0 ; /* Assume failure of previous lookup */
    }
    else
    {
	strcpy( Temp, Lookup_File( "" ) ) ; /* Lookup next matching file */
    } ;
    while( strlen( Temp ) == 0 && strlen( File_List ) > 0 )
    /* While lookup failures */
    {
	A = Instr( 1, File_List, "," ) ;
	/* Find end of next file specification */
	if( A == 0 )
	{
	    A = strlen( File_List ) + 1 ;
	} ;
	strcpy( Temp, Left( File_List, A - 1 ) ) ; /* Get file specification */
	strcpy( File_List, &File_List[ A ] ) ; /* Strip from list */
	strcpy( Temp, Lookup_File( Temp ) ) ; /* Do the lookup */
    } ;
    return Temp ; /* Return result */
} ;
